// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <stdlib.h>

#include "colorfilter.h"
#include "image_init.h"
#include "../shared/error.h"
#include "../shared/mqttpub.h"

void
colfilt_planaryuv (scwstate_t * state, int hshift, int vshift,
		   unsigned char **plane, int *linesize)
{
  scwhistimg_t *imgp = &state->histimg[state->curhist];
  int *dest = imgp->img;
  int x, y;
  unsigned char *in0line;
  unsigned char *in1line;
  unsigned char *in2line;
  int *outline;
  int filt_cb, filt_cr;

  filt_cb = state->colorfilter_cb_autohd >> 8;
  filt_cr = state->colorfilter_cr_autohd >> 8;

  for (y = imgp->area_ymin; y < imgp->area_ymax; y++)
    {
      in0line = plane[0] + y * linesize[0];
      in1line = plane[1] + (y >> vshift) * linesize[1];
      in2line = plane[2] + (y >> vshift) * linesize[2];
      outline = dest + y * state->width;

      for (x = imgp->area_xmin; x < imgp->area_xmax; x++)
	{
	  int yy, cb, cr;

	  yy = in0line[x];
	  cb = in1line[x >> hshift];
	  cr = in2line[x >> hshift];

	  // filter value = distance to wanted cb,cr
	  // and *4 for extra selectivity
	  yy = yy - ((abs (cb - filt_cb) + abs (cr - filt_cr)) << 2);

	  // yy often <0 now, and max 255
	  // and destimage is *256
	  if (yy < 0)
	    outline[x] = 0;
	  else
	    outline[x] = yy << 8;
	}
    }
}

// Ffmpeg integration considers this a header file and demands a prototype.
void cbcradjust (scwstate_t * state, int r, int g, int b, int yy,
		 int cb, int cr);
// yy,cb,cr are used for calculation; r,g,b, *only* for debug report.
void
cbcradjust (scwstate_t * state, int r, int g, int b, int yy, int cb, int cr)
{
  int autor = 0, autog = 0, autob = 0;
  int maxy, calcy;

  // Slowly adjust the filter color.
  // Note that this runs for most blinks.

  state->colorfilter_cb_autohd =
    (31 * state->colorfilter_cb_autohd + (cb << 8)) / 32;
  state->colorfilter_cr_autohd =
    (31 * state->colorfilter_cr_autohd + (cr << 8)) / 32;

  if (state->colorfilter_cb_autohd <
      state->colorfilter_cb_sethd - state->colorfilter_autobound_hd)
    state->colorfilter_cb_autohd =
      state->colorfilter_cb_sethd - state->colorfilter_autobound_hd;

  if (state->colorfilter_cb_autohd >
      state->colorfilter_cb_sethd + state->colorfilter_autobound_hd)
    state->colorfilter_cb_autohd =
      state->colorfilter_cb_sethd + state->colorfilter_autobound_hd;

  if (state->colorfilter_cr_autohd <
      state->colorfilter_cr_sethd - state->colorfilter_autobound_hd)
    state->colorfilter_cr_autohd =
      state->colorfilter_cr_sethd - state->colorfilter_autobound_hd;

  if (state->colorfilter_cr_autohd >
      state->colorfilter_cr_sethd + state->colorfilter_autobound_hd)
    state->colorfilter_cr_autohd =
      state->colorfilter_cr_sethd + state->colorfilter_autobound_hd;

  // Coarse max for Y to get all RGB under 255 (namely max ~240)
  // e.g. full Blue can only have Y = ~10 to still fit in RGB.
  maxy = 235;			// official max for YCbCr BT.601
  calcy = (240 - (408 * (state->colorfilter_cr_autohd >> 8) / 256
		  - 223)) * 256 / 298;
  if (calcy < maxy)
    maxy = calcy;
  calcy = (240 - (-100 * (state->colorfilter_cb_autohd >> 8) / 256
		  - 208 * (state->colorfilter_cr_autohd >> 8) / 256
		  + 136)) * 256 / 298;
  if (calcy < maxy)
    maxy = calcy;
  calcy = (240 - (516 * (state->colorfilter_cb_autohd >> 8) / 256
		  - 277)) * 256 / 298;
  if (calcy < maxy)
    maxy = calcy;

  ycbcr2rgb (maxy, state->colorfilter_cb_autohd >> 8,
	     state->colorfilter_cr_autohd >> 8, &autor, &autog, &autob);

  if (0)
    {
      MDBG
	("this rgb %d %d %d ycbcr %d %d %d  avg rgb %d %d %d ycbcr %d %d %d",
	 r, g, b, yy, cb, cr, autor, autog, autob, maxy,
	 state->colorfilter_cb_autohd >> 8,
	 state->colorfilter_cr_autohd >> 8);
    }

  // Send a report once every while
  if (state->colorfilter_adjustreport <= 0)
    {
      mqttpubjson (state->mosq, state->topicprefix, "cam", state->camname,
		   "status", NULL,
		   json_pack ("{s: {s:{s:i,s:i,s:i}, s:{s:i,s:i,s:i}} }",
			      "blinkcolor",
			      "rgb",
			      "r", autor,
			      "g", autog,
			      "b", autob,
			      "ycbcr",
			      "y", maxy,
			      "cb", state->colorfilter_cb_autohd >> 8,
			      "cr", state->colorfilter_cr_autohd >> 8));
      state->colorfilter_adjustreport = 50;
    }
}

void
filteradjust_planaryuv (scwstate_t * state, int hshift, int vshift,
			unsigned char **plane, int *linesize)
{
  int x, y;
  unsigned char *in0line;
  unsigned char *in1line;
  unsigned char *in2line;
  int yy, cb, cr, r = 0, g = 0, b = 0;
  int yy2;

  x = state->brightestx;
  y = state->brightesty;
  if (x < 0 || y < 0 || x >= state->width || y >= state->height)
    return;

  in0line = plane[0] + y * linesize[0];
  in1line = plane[1] + (y >> vshift) * linesize[1];
  in2line = plane[2] + (y >> vshift) * linesize[2];

  yy = in0line[x];
  cb = in1line[x >> hshift];
  cr = in2line[x >> hshift];

  // Provide r,g,b for debugprint; clip yy to prevent >300 RGB values.
  if (yy > 180)
    yy2 = 180;
  else
    yy2 = yy;
  ycbcr2rgb (yy2, cb, cr, &r, &g, &b);

  // This uses yy,cb,cr for calculation; and r,g,b only for debugprint.
  cbcradjust (state, r, g, b, yy, cb, cr);
}


static inline void
int_rgb2ycbcr (int r, int g, int b, int *y, int *cb, int *cr)
{
  // Fast integer version, further see image_init.c
  // Values may go somewhat beyond bounds.

  *y = 16 + ((66 * r + 129 * g + 25 * b) >> 8);

  *cb = 128 + ((-38 * r - 74 * g + 112 * b) >> 8);

  *cr = 128 + ((112 * r - 94 * g - 18 * b) >> 8);
}

// We use (Y)CbCr because that is mostly independent of lightness
// (sum of r/g/b weighing is zero)
void
colfilt_packedrgb (scwstate_t * state, unsigned char *inimg)
{
  scwhistimg_t *imgp = &state->histimg[state->curhist];
  int *dest = imgp->img;
  int x, y;
  unsigned char *in0line;
  int *outline;
  int filt_cb, filt_cr;

  filt_cb = state->colorfilter_cb_autohd >> 8;
  filt_cr = state->colorfilter_cr_autohd >> 8;

  for (y = imgp->area_ymin; y < imgp->area_ymax; y++)
    {
      in0line = inimg + y * state->width * 3;
      outline = dest + y * state->width;

      for (x = imgp->area_xmin; x < imgp->area_xmax; x++)
	{
	  int r, g, b, yy = 0, cb = 0, cr = 0;

	  r = in0line[x * 3];
	  g = in0line[x * 3 + 1];
	  b = in0line[x * 3 + 2];
	  int_rgb2ycbcr (r, g, b, &yy, &cb, &cr);

	  // filter value = distance to wanted cb,cr
	  // and *4 for extra selectivity
	  yy = yy - ((abs (cb - filt_cb) + abs (cr - filt_cr)) << 2);

	  // yy often <0 now, and max 255
	  // and destimage is *256
	  if (yy < 0)
	    outline[x] = 0;
	  else
	    outline[x] = yy << 8;
	}
    }
}

void
filteradjust_packedrgb (scwstate_t * state, unsigned char *inimg)
{
  int x, y;
  int r, g, b, yy = 0, cb = 0, cr = 0;

  x = state->brightestx;
  y = state->brightesty;
  if (x < 0 || y < 0 || x >= state->width || y >= state->height)
    return;

  r = inimg[(y * state->width + x) * 3];
  g = inimg[(y * state->width + x) * 3 + 1];
  b = inimg[(y * state->width + x) * 3 + 2];

  int_rgb2ycbcr (r, g, b, &yy, &cb, &cr);

  // This uses yy,cb,cr for calculation; and r,g,b only for debugprint.
  cbcradjust (state, r, g, b, yy, cb, cr);
}
