// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "configfile.h"
#include "../shared/error.h"


// This does all state init that depends on config.
// Returns:
// 0 = okay
// -1 = error, message already printed.
int
parse_configfile (scwstate_t * state)
{
  json_t *jroot;
  json_error_t jerror;
  int retval = 0;
  const char *cc;


  // Caller should make sure that *state is set to all-zero as far
  // as we use it, i.e. NULL for alloc'able strings.

  // Useful defaults
  state->svr_hostname = "127.0.0.1";
  state->svr_port = 1883;
  state->topicprefix = "scwips";
  state->imgpersec_min = 15;
  state->detect_wait_max = 8;


  jroot = json_load_file (state->configfilename, 0, &jerror);
  if (jroot == NULL)
    {
      STDERR ("config error on line %d: %s", jerror.line, jerror.text);
      return -1;
    }

  // Note that jansson is perfectly okay handling NULL pointers

  cc = json_string_value (json_object_get (jroot, "name"));
  if (cc != NULL)
    state->camname = strdup (cc);
  else
    {
      STDERR ("no cam name specified in configfile");
      state->camname = "";
      retval = -1;
    }

  state->width_config = json_integer_value (json_object_get (jroot, "pixw"));
  state->height_config = json_integer_value (json_object_get (jroot, "pixh"));
  if (state->width_config <= 1 || state->height_config <= 1)
    {
      STDERR ("no cam width/height specified in configfile");
      retval = -1;
    }

  state->jcamtype = json_incref (json_object_get (jroot, "type"));
  if (state->jcamtype == NULL)
    {
      STDERR ("no cam type specified in configfile");
      retval = -1;
    }

  state->jcamdistortion = json_incref (json_object_get (jroot, "distortion"));
  if (state->jcamdistortion == NULL)
    {
      STDERR ("no cam distortion specified in configfile");
      retval = -1;
    }

  {
    json_t *jmqtt = json_object_get (jroot, "mqtt");
    json_t *jport;

    cc = json_string_value (json_object_get (jmqtt, "host"));
    if (cc != NULL)
      state->svr_hostname = strdup (cc);

    jport = json_object_get (jmqtt, "port");
    if (json_is_integer (jport))
      state->svr_port = json_integer_value (jport);

    cc = json_string_value (json_object_get (jmqtt, "topicprefix"));
    if (cc != NULL)
      state->topicprefix = strdup (cc);
  }

  cc = json_string_value (json_object_get (jroot, "shutdowncmd_touch"));
  if (cc != NULL)
    state->shutdown_touchfile = strdup (cc);

  do
    {
      json_t *jcol = json_object_get (jroot, "colorfilter");
      if (jcol == NULL)
	break;

      state->colorfilter_r = json_integer_value (json_object_get (jcol, "r"));
      state->colorfilter_g = json_integer_value (json_object_get (jcol, "g"));
      state->colorfilter_b = json_integer_value (json_object_get (jcol, "b"));
    }
  while (0);

  {
    json_t *jblink = json_object_get (jroot, "blinkthres_min");
    if (json_is_integer (jblink))
      state->blinkthres_min = json_integer_value (jblink);
  }

  {
    json_t *jfocus = json_object_get (jroot, "focusarea_min");
    if (json_is_integer (jfocus))
      state->focusarea_min = json_integer_value (jfocus);
  }

  {
    json_t *jfps = json_object_get (jroot, "fps_min");
    if (json_is_integer (jfps))
      state->imgpersec_min = json_integer_value (jfps);
  }


  // Additional sanity checks

  if (state->svr_hostname == NULL || state->svr_hostname[0] == '\0')
    {
      STDERR ("no mqtt hostname specified");
      retval = -1;
    }
  if (state->svr_port < 1 || state->svr_port > 65535)
    {
      STDERR ("invalid mqtt port specified: %d", state->svr_port);
      retval = -1;
    }
  if (state->topicprefix == NULL || state->topicprefix[0] == '\0')
    {
      STDERR ("no mqtt topic prefix specified");
      retval = -1;
    }
  if (strlen (state->topicprefix) + strlen (state->camname) > TOPICLEN - 200)
    {
      STDERR ("mqtt topicprefix+camname too long, increase TOPICLEN");
      retval = -1;
    }
  if (state->camname[0] == '.' || strchr (state->camname, '/') != NULL)
    {
      STDERR ("camera name may not start with dot and may not"
	      " contain slash: %s", state->camname);
      retval = -1;
    }
  if (!strncmp (state->camname, "Refpt_", 6))
    {
      STDERR ("camera name may not start with Refpt_ : %s", state->camname);
      retval = -1;
    }


  // Prepare mqtt topics

  if (retval == 0)
    {
      // topicprefix etc. checked for sensible length above
      snprintf (state->broadcmd_topic, TOPICLEN, "%s/cmd",
		state->topicprefix);
      snprintf (state->allcamcmd_topic, TOPICLEN, "%s/cam/cmd",
		state->topicprefix);
      snprintf (state->camcmd_topic, TOPICLEN, "%s/cam/%s/cmd",
		state->topicprefix, state->camname);
      snprintf (state->calcpix_topic, TOPICLEN, "%s/cam/%s/calcpix",
		state->topicprefix, state->camname);

    }

  // Further state init see image_init()

  json_decref (jroot);
  return retval;
}
