// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "detect.h"
#include "mqtt.h"
#include "../shared/error.h"

// prev = 0, -1, -2; only underflow supported
#define HISTWRAP(prev)  (state->curhist + (prev)) >= 0 ?	\
			 (state->curhist + (prev)) :		\
			 (state->curhist + (prev) + HISTIMGS)


// Once rows have been excluded, column sums will get different too.
// Do full recalculation from detection results.
// Ffmpeg integration considers this a header file and demands a prototype.
void
recalc_weights_from_detectimg (scwstate_t * state, int *detect_img,
			       int *weight_cols, int *weight_rows,
			       int xmin, int xmax, int ymin, int ymax);
void
recalc_weights_from_detectimg (scwstate_t * state, int *detect_img,
			       int *weight_cols, int *weight_rows,
			       int xmin, int xmax, int ymin, int ymax)
{
  int x, y;

  memset (weight_cols, 0, state->width * sizeof (*(weight_cols)));
  memset (weight_rows, 0, state->height * sizeof (*(weight_rows)));

  for (y = ymin; y < ymax; y++)
    {
      int *detectpix = &detect_img[y * state->width + xmin];

      for (x = xmin; x < xmax; x++, detectpix++)
	if (*detectpix > 0)
	  {
	    // Add to sum arrays
	    weight_cols[x] += *detectpix;
	    weight_rows[y] += *detectpix;
	  }
    }
}

// Detect largest hole in weight array, keep only biggest side.
// Updates min/max.
// Returns: 1 if anything changed.
// Ffmpeg integration considers this a header file and demands a prototype.
int
remove_array_hole (scwstate_t * state, int *weight_array, int *min, int *max);
int
remove_array_hole (scwstate_t * state, int *weight_array, int *min, int *max)
{
  int i;
  int thresweight;
  int totalused;
  int thisblank_start, thisblank_len;
  int longestblank_start, longestblank_len;

  // Find the max weight from the array, use it to set threshold.
  thresweight = 0;
  for (i = *min; i < *max; i++)
    if (weight_array[i] > thresweight)
      thresweight = weight_array[i];
  thresweight /= 16;		// zero is fine

  // Then find the longest blank, may also be at begin or end.
  // Also count the total number of used entries.

  totalused = 0;
  thisblank_start = -1;
  thisblank_len = 0;
  longestblank_start = -1;
  longestblank_len = 0;

  for (i = *min; i < *max; i++)
    if (weight_array[i] > thresweight)
      {
	totalused++;

	// finalize previous blank
	if (thisblank_len > longestblank_len)
	  {
	    longestblank_start = thisblank_start;
	    longestblank_len = thisblank_len;
	  }
	thisblank_len = 0;
      }
    else
      {
	if (thisblank_len <= 0)
	  thisblank_start = i;
	thisblank_len++;
      }

  // finalize last blank
  if (thisblank_len > longestblank_len)
    {
      longestblank_start = thisblank_start;
      longestblank_len = thisblank_len;
    }

  if (totalused > 1 && longestblank_len > 10 &&
      longestblank_len > 3 * totalused)
    {
      long weightleft, weightright;

      weightleft = 0;
      for (i = *min; i < longestblank_start; i++)
	if (weight_array[i] > thresweight)
	  weightleft += weight_array[i];

      weightright = 0;
      for (i = longestblank_start + longestblank_len; i < *max; i++)
	if (weight_array[i] > thresweight)
	  weightright += weight_array[i];

      noMDBG ("from %d - %d hole %d - %d weight left %ld right %ld",
	      *min, *max,
	      longestblank_start, longestblank_start + longestblank_len,
	      weightleft, weightright);

      if (weightleft > weightright)
	{
	  *max = longestblank_start;	// max is exclusive
	  return 1;
	}
      else
	{
	  *min = longestblank_start + longestblank_len;
	  return 1;
	}
    }
  return 0;
}

// Reflections are often visible far away from original blink.
// For columns (first) and rows (second), see if there are any big holes;
// then for the largest hole, choose the largest useful side. Repeats until
// no large holes left.
// This will often also remove start and end blanks.
// Updates xmin/xmax/ymin/ymax.
// Ffmpeg integration considers this a header file and demands a prototype.
void
remove_reflections (scwstate_t * state, int *detect_img, int *weight_cols,
		    int *weight_rows, int *xmin, int *xmax, int *ymin,
		    int *ymax);
void
remove_reflections (scwstate_t * state, int *detect_img, int *weight_cols,
		    int *weight_rows, int *xmin, int *xmax, int *ymin,
		    int *ymax)
{
  int done = 0;

  while (!done)
    {
      done = 1;

      // Columns first: reflection is often cased by vertical surfaces
      // so horizontal hole is most obvious.
      if (remove_array_hole (state, weight_cols, xmin, xmax))
	{
	  noMDBG ("removed hole in x, now %d - %d", *xmin, *xmax);

	  recalc_weights_from_detectimg (state, detect_img,
					 weight_cols, weight_rows,
					 *xmin, *xmax, *ymin, *ymax);
	  done = 0;
	}

      if (remove_array_hole (state, weight_rows, ymin, ymax))
	{
	  noMDBG ("removed hole in y, now %d - %d", *ymin, *ymax);

	  recalc_weights_from_detectimg (state, detect_img,
					 weight_cols, weight_rows,
					 *xmin, *xmax, *ymin, *ymax);
	  done = 0;
	}
    }
}


// Returns weighted mean; <0. = nothing at all, then other params unchanged.
// Sets weightsum; changes min/max to boundaries of really-used entries.
// Ffmpeg integration considers this a header file and demands a prototype.
double
find_array_weightedmean (scwstate_t * state, int *weight_array, int *min,
			 int *max, long *weightsum);
double
find_array_weightedmean (scwstate_t * state, int *weight_array, int *min,
			 int *max, long *weightsum)
{
  int i;
  int thresweight;
  long weightedsum, weights;
  int used_min, used_max;

  // Find the max weight from the array, use it to set threshold.
  thresweight = 0;
  for (i = *min; i < *max; i++)
    if (weight_array[i] > thresweight)
      thresweight = weight_array[i];

  if (thresweight <= 0)
    return -1.;

  thresweight /= 4;

  // Aggregate over the array, only using "significant" entries.
  weightedsum = 0;
  weights = 0;
  used_min = -1;
  used_max = -1;
  for (i = *min; i < *max; i++)
    if (weight_array[i] > thresweight)
      {
	weightedsum += weight_array[i] * i;
	weights += weight_array[i];

	if (used_min < 0)
	  used_min = i;

	used_max = i;
      }
  used_max++;			// max is exclusive

  if (0)			// debug
    {
      char s[100000];
      int slen = 0;
      int r;

      r = sprintf (s + slen, "filt %d wsum %ld w %ld minmax %d %d",
		   thresweight, weightedsum, weights, used_min, used_max);
      slen += r;

      for (i = *min; i < *max; i++)
	if (weight_array[i] > 0)
	  {
	    r = sprintf (s + slen, " [%d]=%d", i, weight_array[i]);
	    slen += r;
	  }

      MDBG ("%s", s);
    }

  if (weights <= 0)
    return -1.;

  *min = used_min;
  *max = used_max;
  *weightsum = weights;
  return ((double) weightedsum) / ((double) weights);
}



// For edge (rise or fall), find weighted-average pixel location.
// Ffmpeg integration considers this a header file and demands a prototype.
void
find_mean_rowcol (scwstate_t * state, scwedge_t * edge, int *weight_cols,
		  int *weight_rows, int xmin, int xmax, int ymin, int ymax);
void
find_mean_rowcol (scwstate_t * state, scwedge_t * edge, int *weight_cols,
		  int *weight_rows, int xmin, int xmax, int ymin, int ymax)
{
  long weightsumx, weightsumy;
  int used_xmin, used_xmax, used_ymin, used_ymax;

  used_xmin = xmin;
  used_xmax = xmax;
  weightsumx = 0;
  edge->blinkx_d =
    find_array_weightedmean (state, weight_cols, &used_xmin, &used_xmax,
			     &weightsumx);
  if (edge->blinkx_d < 0.)
    return;

  used_ymin = ymin;
  used_ymax = ymax;
  weightsumy = 0;
  edge->blinky_d =
    find_array_weightedmean (state, weight_rows, &used_ymin, &used_ymax,
			     &weightsumy);
  if (edge->blinky_d < 0.)
    return;

  if (weightsumx <= 0 || weightsumy <= 0)
    return;			// leave edge->weightsum at zero (memset'ed)

  // Since different columns and rows may have been selected,
  // weight sums may be different too. Store average.
  edge->weightsum = (weightsumx + weightsumy) / 2;

  edge->blinkw = used_xmax - used_xmin;
  edge->blinkh = used_ymax - used_ymin;
}

// Find xmin/xmax/ymin/ymax subset that is covered by all history images.
// Returns:
// 1 = useful area to use
// 0 = useless: no overlapping area anywhere
static inline int
findarea (scwstate_t * state, int *xmin, int *xmax, int *ymin, int *ymax)
{
  int i;

  *xmin = 0;
  *xmax = state->width;
  *ymin = 0;
  *ymax = state->height;

  for (i = 0; i < HISTIMGS; i++)
    {
      if (state->histimg[i].area_xmin > *xmin)
	*xmin = state->histimg[i].area_xmin;
      if (state->histimg[i].area_xmax < *xmax)
	*xmax = state->histimg[i].area_xmax;
      if (state->histimg[i].area_ymin > *ymin)
	*ymin = state->histimg[i].area_ymin;
      if (state->histimg[i].area_ymax < *ymax)
	*ymax = state->histimg[i].area_ymax;
    }

  if (*xmax <= *xmin || *ymax <= *ymin)
    return 0;
  // xmin starts at 0 and can only go higher,
  // xmax starts at width and can only go lower,
  // so if values ever get out-of-picture then this check will always trigger.

  return 1;
}

// Ffmpeg integration considers this a header file and demands a prototype.
void detect_edges (scwstate_t * state);
void
detect_edges (scwstate_t * state)
{
  int xmin, xmax, ymin, ymax;
  int *curhistimg, *minus1histimg, *minus2histimg;
  int blinkthres, maxval, minval;
  int x, y;
  scwedge_t *rise;
  scwedge_t *fall;
  int weight;

  // New color-filtered image is in current histimg slot.

  // ==> SEE Blinkdetect.txt

  curhistimg = state->histimg[state->curhist].img;
  minus1histimg = state->histimg[HISTWRAP (-1)].img;
  minus2histimg = state->histimg[HISTWRAP (-2)].img;
  blinkthres = state->blinkthres_cur_hd >> 8;

  rise = &state->histimg[state->curhist].rise;
  fall = &state->histimg[state->curhist].fall;
  memset (rise, 0, sizeof (*rise));
  memset (fall, 0, sizeof (*fall));
  rise->brightestx = -1;
  rise->brightesty = -1;
  fall->brightestx = -1;
  fall->brightesty = -1;

  if (!findarea (state, &xmin, &xmax, &ymin, &ymax))
    // No area to check.
    // Areas are grown in updates_afterimage(), hopefully fixed soon enough.
    return;

  memset (state->weight_cols_rise, 0,
	  state->width * sizeof (*(state->weight_cols_rise)));
  memset (state->weight_rows_rise, 0,
	  state->height * sizeof (*(state->weight_rows_rise)));
  memset (state->weight_cols_fall, 0,
	  state->width * sizeof (*(state->weight_cols_fall)));
  memset (state->weight_rows_fall, 0,
	  state->height * sizeof (*(state->weight_rows_fall)));

  for (y = ymin; y < ymax; y++)
    {
      int *curpix = &curhistimg[y * state->width + xmin];
      int *minus1pix = &minus1histimg[y * state->width + xmin];
      int *minus2pix = &minus2histimg[y * state->width + xmin];

      // Rise/fall images are always stored, but only used when
      // removing reflections.
      int *detectrise_pix = &state->detectrise_img[y * state->width + xmin];
      int *detectfall_pix = &state->detectfall_img[y * state->width + xmin];

      for (x = xmin; x < xmax;
	   x++, curpix++, minus1pix++, minus2pix++, detectrise_pix++,
	   detectfall_pix++)
	{
	  *detectrise_pix = 0;	// default for early "continue"
	  *detectfall_pix = 0;

	  // We need maxval and minval for both rise and fall.
	  // But maxval can easily filter *any edge* candidates.
	  maxval = *curpix;
	  if (*minus1pix > maxval)
	    maxval = *minus1pix;
	  if (*minus2pix > maxval)
	    maxval = *minus2pix;

	  if (maxval < blinkthres)
	    // can't be any edge (e.g. not the right color), hopefully often
	    continue;

	  minval = *curpix;
	  if (*minus1pix < minval)
	    minval = *minus1pix;
	  if (*minus2pix < minval)
	    minval = *minus2pix;

	  if (maxval - minval < blinkthres)
	    // can't be an edge, too little difference
	    continue;

	  // Now the difference is above threshold, so it must be
	  // at least one edge (and can even be rise+fall/fall+rise).

	  // Weight is difference between max an min,
	  // and subtract thres since will always be larger.
	  weight = (maxval - minval) - blinkthres;

	  // Weighted sums may get quite large:
	  // weight up to 64k, x/y up to 1920(4k..)
	  // and possibly quite a large number of pixels.
	  // So reduce weight some more, shouldn't loose too much precision.
	  weight = weight >> 8;

	  // Now weight might be shifted out entirely, then ignore.

	  // Rising edge on current or previous frame
	  if (weight > 0 &&
	      (*minus1pix < *curpix - blinkthres ||
	       *minus2pix < *curpix - blinkthres ||
	       *minus2pix < *minus1pix - blinkthres))
	    {
	      // Store into detection image
	      *detectrise_pix = weight;

	      // Add to sum arrays
	      state->weight_cols_rise[x] += weight;
	      state->weight_rows_rise[y] += weight;

	      if (maxval - minval > rise->brightestblink)
		{
		  rise->brightestblink = maxval - minval;

		  // Brighestx/y for auto-finetuning color from current frame.
		  // So pixel must still be active in current frame.
		  if (*minus1pix < *curpix - blinkthres ||
		      *minus2pix < *curpix - blinkthres)
		    {
		      rise->brightestx = x;
		      rise->brightesty = y;
		    }
		}
	    }

	  // Falling edge on current or previous frame
	  if (weight > 0 &&
	      (*minus1pix > *curpix + blinkthres ||
	       *minus2pix > *curpix + blinkthres ||
	       *minus2pix > *minus1pix + blinkthres))
	    {
	      // Store into detection image
	      *detectfall_pix = weight;

	      // Add to sum arrays
	      state->weight_cols_fall[x] += weight;
	      state->weight_rows_fall[y] += weight;

	      if (maxval - minval > fall->brightestblink)
		{
		  fall->brightestblink = maxval - minval;

		  // Brighestx/y for auto-finetuning color from current frame.
		  // Never applies for falling edge: pixel not active any more.
		}
	    }
	}
    }

  // Now the detected weights are gathered per-column and per-row.

  // Rise
  {
    int adj_xmin = xmin;
    int adj_xmax = xmax;
    int adj_ymin = ymin;
    int adj_ymax = ymax;

    remove_reflections (state, state->detectrise_img,
			state->weight_cols_rise, state->weight_rows_rise,
			&adj_xmin, &adj_xmax, &adj_ymin, &adj_ymax);

    find_mean_rowcol (state, rise, state->weight_cols_rise,
		      state->weight_rows_rise, xmin, xmax, ymin, ymax);
  }

  // Fall
  {
    int adj_xmin = xmin;
    int adj_xmax = xmax;
    int adj_ymin = ymin;
    int adj_ymax = ymax;

    remove_reflections (state, state->detectfall_img,
			state->weight_cols_fall, state->weight_rows_fall,
			&adj_xmin, &adj_xmax, &adj_ymin, &adj_ymax);

    find_mean_rowcol (state, fall, state->weight_cols_fall,
		      state->weight_rows_fall, xmin, xmax, ymin, ymax);
  }


  MDBG ("rise x/y/wsum %d %d %d w/h %d %d bri/x/y %d %d %d  "
	"fall x/y/wsum %d %d %d w/h %d %d bri/x/y %d %d %d  ",
	(int) rise->blinkx_d, (int) rise->blinky_d, rise->weightsum,
	rise->blinkw, rise->blinkh,
	rise->brightestblink, rise->brightestx, rise->brightesty,
	(int) fall->blinkx_d, (int) fall->blinky_d, fall->weightsum,
	fall->blinkw, fall->blinkh,
	fall->brightestblink, fall->brightestx, fall->brightesty);
}


void
detect_blink (scwstate_t * state)
{
  scwedge_t *rise;
  scwedge_t *fall;
  scwedge_t *prevrise;
  scwedge_t *prevfall;
  scwedge_t *reportedge;
  scwedge_t *filtadjustedge;
  const char *reportedge_txt;
  const char *filtadjustedge_txt;
  int blinkx, blinky;
  int toofaraway;
  int minmargin = state->focusarea_min;


  // These are "return values", set to default for early returns.
  state->brightestx = -1;
  state->brightesty = -1;

  detect_edges (state);

  rise = &state->histimg[state->curhist].rise;
  fall = &state->histimg[state->curhist].fall;
  prevrise = &state->histimg[HISTWRAP (-1)].rise;
  prevfall = &state->histimg[HISTWRAP (-1)].fall;

  state->detect_wait_since++;
  reportedge = NULL;
  reportedge_txt = "none";
  filtadjustedge = NULL;
  filtadjustedge_txt = "none";

  if (state->detect_state == DSTATE_WAIT)
    {
      // Previous frame was nothing interesting.

      // Weightsum always >=0, so if this matches than also >0.
      if (rise->weightsum > fall->weightsum)
	state->detect_state = DSTATE_FIRSTRISE;
      else if (fall->weightsum > rise->weightsum)
	state->detect_state = DSTATE_FIRSTFALL;
    }

  else if (state->detect_state == DSTATE_FIRSTRISE)
    {
      // Previous frame was first rising edge frame.
      // In most cases, this frame will have strictly more pixels,
      // but may have less weight if some min/max shifted out of
      // the window already.

      // But we only react to edges that are big enough.
      if (rise->weightsum >= state->weightsum_slowfilt / 4 ||
	  prevrise->weightsum >= state->weightsum_slowfilt / 4)
	{
	  // We have an edge to possibly report below.

	  if (rise->weightsum >= prevrise->weightsum)
	    {
	      reportedge = rise;
	      reportedge_txt = "rise";
	      filtadjustedge = rise;
	      filtadjustedge_txt = "rise";
	    }
	  else
	    {
	      reportedge = prevrise;
	      reportedge_txt = "prevrise";

	      // Filter color adjustment can only look up brightestx/y
	      // in current frame.
	      // Even though current frame is "worse" blink, we can still
	      // use it for color finetuning under strict conditions to
	      // make sure it's not really too bad.
	      if (rise->weightsum > 1 &&
		  rise->weightsum > 3 * prevrise->weightsum / 4 &&
		  rise->brightestblink > 3 * prevrise->brightestblink / 4)
		{
		  // Okay, can use for color finetuning
		  filtadjustedge = rise;
		  filtadjustedge_txt = "rise";
		}
	    }

	  // Consider what to do next.
	  if (fall->weightsum > rise->weightsum / 4)
	    // This frame is already quite much fall
	    state->detect_state = DSTATE_FIRSTFALL;
	  else
	    state->detect_state = DSTATE_IGNORERISE;
	}
      else
	{
	  // Edge is not big enough.

	  // Consider what to do next.
	  if (rise->weightsum > fall->weightsum)
	    // Handle as first again, but if the next frame is empty
	    // then won't report because still too small.
	    state->detect_state = DSTATE_FIRSTRISE;
	  else if (fall->weightsum > rise->weightsum)
	    state->detect_state = DSTATE_FIRSTFALL;
	  else
	    state->detect_state = DSTATE_WAIT;
	}

    }
  else if (state->detect_state == DSTATE_IGNORERISE)
    {
      // Previous frame reported rise,
      // this frame probably gets some rise re-detection.

      if (fall->weightsum > 0)
	state->detect_state = DSTATE_FIRSTFALL;
      else
	state->detect_state = DSTATE_WAIT;
    }

  else if (state->detect_state == DSTATE_FIRSTFALL)
    {
      // Previous frame was first falling edge frame.
      // In most cases, this frame will have strictly more pixels,
      // but may have less weight if some min/max shifted out of
      // the window already.

      // But we only react to edges that are big enough.
      if (fall->weightsum >= state->weightsum_slowfilt / 4 ||
	  prevfall->weightsum >= state->weightsum_slowfilt / 4)
	{
	  // We have an edge to possibly report below.

	  if (fall->weightsum >= prevfall->weightsum)
	    {
	      reportedge = fall;
	      reportedge_txt = "fall";

	      // Filter color can't be finetuned on fall, no color any more.
	    }
	  else
	    {
	      reportedge = prevfall;
	      reportedge_txt = "prevfall";

	      // Filter color can't be finetuned on fall, no color any more.
	    }

	  // Consider what to do next.
	  if (rise->weightsum > fall->weightsum / 4)
	    // This frame is already quite much rise
	    state->detect_state = DSTATE_FIRSTRISE;
	  else
	    state->detect_state = DSTATE_IGNOREFALL;
	}
      else
	{
	  // Edge is not big enough.

	  // Consider what to do next.
	  if (rise->weightsum > fall->weightsum)
	    // Handle as first again, but if the next frame is empty
	    // then won't report because still too small.
	    state->detect_state = DSTATE_FIRSTRISE;
	  else if (fall->weightsum > rise->weightsum)
	    state->detect_state = DSTATE_FIRSTFALL;
	  else
	    state->detect_state = DSTATE_WAIT;
	}

    }
  else if (state->detect_state == DSTATE_IGNOREFALL)
    {
      // Previous frame reported fall,
      // this frame probably gets some fall re-detection.

      if (rise->weightsum > 0)
	state->detect_state = DSTATE_FIRSTRISE;
      else
	state->detect_state = DSTATE_WAIT;
    }

  else
    // Illegal state
    state->detect_state = DSTATE_WAIT;

  MDBG
    ("blinkthres %d mincfg %d weightslowfilt %d since %d newst %d report %s filtadj %s",
     state->blinkthres_cur_hd >> 8, state->blinkthres_min,
     state->weightsum_slowfilt, state->detect_wait_since,
     state->detect_state, reportedge_txt, filtadjustedge_txt);


  blinkx = 0;
  blinky = 0;
  toofaraway = 0;
  if (reportedge != NULL)
    {
      // Integer values for our own calc.
      // Tests above guarantee that reportedge always has weightsum > 0.
      blinkx = reportedge->blinkx_d;
      blinky = reportedge->blinky_d;

      // Also, before shifting our center, check that new result
      // is close enough to "previous".
      if (abs (blinkx - state->ourarea_x) > state->focusarea_min / 2 ||
	  abs (blinky - state->ourarea_y) > state->focusarea_min / 2)
	toofaraway = 1;
    }


  if (reportedge != NULL)
    {
      int target, diff, maxdiff, shrink;

      // There is an edge, and it is significant considering slowfilt.
      // It may still not be reported for various reasons,
      // but adjust various filters already.

      // Regulate blinkthres.
      // There is an edge, so brightestblink will be set.
      // This happens only on blinks, so adjust quickly: at least
      // 4~5 times faster than down-adjust below.

      target = 800 * reportedge->brightestblink / 1024;

      state->blinkthres_cur_hd =
	(31 * state->blinkthres_cur_hd + (target << 8)) / 32;

      if (state->blinkthres_cur_hd < (state->blinkthres_min << 8))
	state->blinkthres_cur_hd = (state->blinkthres_min << 8);


      // Slowly adjust weightsum filter.
      // This is not "HD": values are easily >1000 so enough resolution,
      // and prevent possible overflow.

      if (reportedge->weightsum >= state->weightsum_slowfilt)
	{
	  // more weight than before
	  state->weightsum_slowfilt =
	    (7 * state->weightsum_slowfilt + reportedge->weightsum) / 8;
	}
      else
	{
	  // less weight than before
	  state->weightsum_slowfilt =
	    (7 * state->weightsum_slowfilt + reportedge->weightsum) / 8;
	}


      // Move our focusarea towards blink.
      // (Note: also moved on backcalcpix reception, see mqtt.c)

      maxdiff = state->focusarea_min / 4;
      if (maxdiff < 1)
	maxdiff = 1;

      diff = blinkx - state->ourarea_x;
      if (diff < -maxdiff)
	diff = -maxdiff;
      else if (diff > maxdiff)
	diff = maxdiff;
      state->ourarea_x += diff;

      diff = blinky - state->ourarea_y;
      if (diff < -maxdiff)
	diff = -maxdiff;
      else if (diff > maxdiff)
	diff = maxdiff;
      state->ourarea_y += diff;


      // Also quickly shrink our focusarea.
      // Note: always *enlarged* in updates_afterimage().

      shrink = state->maxdimension / 32;
      if (shrink < 6)
	shrink = 6;
      // Since updates_afterimage() always enlarges for every frame,
      // shrinking here should be at least 7~8 times quicker.

      // Keep margins around entire blink, and at least two times
      // the blink size (for really large blinks: 0.5*blinkw/h will
      // already be occupied by the blink itself).
      minmargin = state->focusarea_min + reportedge->blinkw;
      if (state->focusarea_min + reportedge->blinkh > minmargin)
	minmargin = state->focusarea_min + reportedge->blinkh;
      if (reportedge->blinkw * 2 > minmargin)
	minmargin = reportedge->blinkw * 2;
      if (reportedge->blinkh * 2 > minmargin)
	minmargin = reportedge->blinkh * 2;

      if (state->ourarea_margin > minmargin)
	{
	  state->ourarea_margin -= shrink;
	  if (state->ourarea_margin < minmargin)
	    state->ourarea_margin = minmargin;
	}
      //else: our margin is too small already,
      // i.e we have a particularly large blink now.
      // Do not jump-enlarge but leave that to updates_afterimage().

      // Make sure the detected blink is still within the area:
      // shift to inside of focusarea_min/8 from margin edges.

      if (state->ourarea_x - state->ourarea_margin + state->focusarea_min / 8
	  > blinkx)
	state->ourarea_x =
	  blinkx + state->ourarea_margin - state->focusarea_min / 8;
      if (state->ourarea_x + state->ourarea_margin -
	  state->focusarea_min / 8 < blinkx)
	state->ourarea_x =
	  blinkx - state->ourarea_margin + state->focusarea_min / 8;

      if (state->ourarea_y - state->ourarea_margin + state->focusarea_min / 8
	  > blinky)
	state->ourarea_y =
	  blinky + state->ourarea_margin - state->focusarea_min / 8;
      if (state->ourarea_y + state->ourarea_margin -
	  state->focusarea_min / 8 < blinky)
	state->ourarea_y =
	  blinky - state->ourarea_margin + state->focusarea_min / 8;

      // Since the backcalc margin is chosen whenever it is smaller,
      // make sure it never gets smaller than required for large blinks.
      if (state->backcalcarea_margin < minmargin)
	state->backcalcarea_margin = minmargin;
    }
  else
    {
      // No (significant) edge in this frame: adjust slowly downwards.

      // Regulate blinkthres to minimum value from config.

      // CAREFUL!: normal thres/value is up to 64k,
      // _hd is << 8
      // so *255 will reach 32bit limit!
      // (possibly set _hd to <<4 instead)
      state->blinkthres_cur_hd =
	(127 * state->blinkthres_cur_hd + (state->blinkthres_min << 8)) / 128;


      // Regulate weightsum filter to zero.

      state->weightsum_slowfilt = (31 * state->weightsum_slowfilt) / 32;
    }


  // Some reasons to not report blink.

  if (reportedge == NULL)
    // No blink at all
    return;

  if (state->detect_wait_since > state->detect_wait_max)
    {
      // There is an edge, but the previous one was too long ago.
      // Hopefully the next edge comes sooner.
      MDBG ("noreport: waitsince %d waitmax %d", state->detect_wait_since,
	    state->detect_wait_max);
      state->detect_wait_since = 0;
      return;
    }
  state->detect_wait_since = 0;

  if (toofaraway)
    {
      // Too far from our center; hopefully next time our center
      // has shifted somewhat further towards it.
      MDBG ("noreport: toofaraway");
      return;
    }

  // Already checked against state->weightsum_slowfilt / 4

  if (reportedge->weightsum > state->weightsum_slowfilt * 4)
    {
      // Really much more weight compared to previous, allow filter to
      // grow towards it.
      MDBG ("noreport: weightsum %d slow %d", reportedge->weightsum,
	    state->weightsum_slowfilt);
      return;
    }

  if (state->ourarea_margin > minmargin * 2
      && state->backcalcarea_margin > minmargin * 2)
    {
      // Area still too big, wait until shrunk somewhat.
      // (I.e. okay if either one is small enough)
      // Note that this is after our shrinking above,
      // so we can be quite selective here.
      MDBG ("noreport: ourmarg %d backmarg %d minmarg %d",
	    state->ourarea_margin, state->backcalcarea_margin, minmargin);
      return;
    }


  // Okay, everything is really truly right now.

  state->blinkspersec_cnt++;
  state->fixlost_timeout = 0;

  if (filtadjustedge != NULL)
    {
      state->brightestx = filtadjustedge->brightestx;
      state->brightesty = filtadjustedge->brightesty;
    }

  mqttpubjson (state->mosq, state->topicprefix, "cam", state->camname,
	       "pix", NULL,
	       json_pack ("{s:f, s:f, s:i, s:i}",
			  "pixx", reportedge->blinkx_d,
			  "pixy", reportedge->blinky_d,
			  "pixw", reportedge->blinkw,
			  "pixh", reportedge->blinkh));
}
