// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <stdio.h>
#include <stdlib.h>

#include "image_init.h"
#include "mqtt.h"
#include "../shared/error.h"


// Ffmpeg integration considers this a header file and demands a prototype.
void stdin_image_init (scwstate_t * state);
void
stdin_image_init (scwstate_t * state)
{
  if (state->recvimg_formatbase == IMGF_RGB)
    {
      state->recvimg_size = state->width * state->height * 3;
    }
  else if (state->recvimg_formatbase == IMGF_YUV)
    {
      // https://en.wikipedia.org/wiki/Chroma_subsampling
      // and ffmpeg libavutil/pixdesc.c list of .log2_chroma_w/h

      switch (state->recvimg_formatdetail)
	{
	case IMGF_410p:
	  state->recvimg_yuv_hshift = 2;
	  state->recvimg_yuv_vshift = 1;
	  // Ffmpeg has 2,2; wikipedia contradicts itself (4x2 vs 4x4 base).
	  // But we can just make up another IMGF_ if we ever encounter it.
	  break;
	case IMGF_411p:
	  state->recvimg_yuv_hshift = 2;
	  state->recvimg_yuv_vshift = 0;
	  break;
	case IMGF_420p:
	  state->recvimg_yuv_hshift = 1;
	  state->recvimg_yuv_vshift = 1;
	  break;
	case IMGF_422p:
	  state->recvimg_yuv_hshift = 1;
	  state->recvimg_yuv_vshift = 0;
	  break;
	case IMGF_440p:
	  state->recvimg_yuv_hshift = 0;
	  state->recvimg_yuv_vshift = 1;
	  break;
	case IMGF_444p:
	  state->recvimg_yuv_hshift = 0;
	  state->recvimg_yuv_vshift = 0;
	  break;
	default:
	  STDERR ("Formatdetail %d not supported",
		  state->recvimg_formatdetail);
	  exit (1);
	}

      // For every luma pixel, the x>>hshift and y>>vshift are expected
      // to point to the correct color pixel.
      // That means width=5 with shift=2 will need to have 2 color pixels
      // (just like with width=8).
      state->recvimg_yuv_cbcrwidth =
	CEIL_RSHIFT (state->width, state->recvimg_yuv_hshift);
      state->recvimg_yuv_cbcrheight =
	CEIL_RSHIFT (state->height, state->recvimg_yuv_vshift);

      state->recvimg_size = state->width * state->height +
	state->recvimg_yuv_cbcrwidth * state->recvimg_yuv_cbcrheight * 2;
    }
  else
    {
      STDERR ("Format not supported");
      exit (1);
    }

  if (state->recvimg_size < state->headbuf_size)
    {
      // entire image fits in headbuf, possibly part of next image too...
      STDERR ("Input image too small, algorithm fails");
      exit (1);
    }

  state->recvimg = malloc (state->recvimg_size);
  if (state->recvimg == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  if (state->recvimg_formatbase == IMGF_YUV)
    {
      state->recvimg_yuv_plane[0] = state->recvimg;
      state->recvimg_yuv_plane[1] = state->recvimg
	+ (state->width * state->height);
      state->recvimg_yuv_plane[2] = state->recvimg
	+ (state->width * state->height
	   + state->recvimg_yuv_cbcrwidth * state->recvimg_yuv_cbcrheight);
    }
}

void
image_init (scwstate_t * state)
{
  int i;
  int yy = 0, cb = 0, cr = 0;

  if (state->img_init_flag)
    {
      STDERR ("Not intended to be called multiple times");
      exit (1);
    }

  if (state->width != state->width_config &&
      state->height != state->height_config)
    {
      STDERR ("Configured width/height %d x %d does not match input %d x %d. "
	      "Cameras are known to apply weird scaling, please calibrate at intended size.",
	      state->width_config, state->height_config, state->width,
	      state->height);
      exit (1);
    }

  state->maxdimension = state->width;
  if (state->height > state->maxdimension)
    state->maxdimension = state->height;

  if (state->headbuf_size > 0)
    stdin_image_init (state);	// state->recvimg*

  for (i = 0; i < HISTIMGS; i++)
    {
      state->histimg[i].img =
	calloc (state->width * state->height,
		sizeof (*(state->histimg[i].img)));
      if (state->histimg[i].img == NULL)
	{
	  STDERR ("calloc failed");
	  exit (1);
	}
    }
  state->curhist = 0;

  state->detectrise_img =
    malloc (state->width * state->height * sizeof (*(state->detectrise_img)));
  state->detectfall_img =
    malloc (state->width * state->height * sizeof (*(state->detectfall_img)));
  if (state->detectrise_img == NULL || state->detectfall_img == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  // Pubimg largest possible, i.e. RGB 3*8bits, plus some header
  state->pubimg = malloc (2000 + state->width * state->height * 3);
  if (state->pubimg == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  state->weight_cols_rise =
    calloc (state->width, sizeof (*(state->weight_cols_rise)));
  state->weight_rows_rise =
    calloc (state->height, sizeof (*(state->weight_rows_rise)));
  state->weight_cols_fall =
    calloc (state->width, sizeof (*(state->weight_cols_fall)));
  state->weight_rows_fall =
    calloc (state->height, sizeof (*(state->weight_rows_fall)));
  if (state->weight_cols_rise == NULL ||
      state->weight_rows_rise == NULL ||
      state->weight_cols_fall == NULL || state->weight_rows_fall == NULL)
    {
      STDERR ("calloc failed");
      exit (1);
    }


  state->ourarea_x = state->width / 2;
  state->ourarea_y = state->height / 2;
  state->ourarea_margin = state->focusarea_min;

  state->backcalcarea_x = state->width / 2;
  state->backcalcarea_y = state->height / 2;
  state->backcalcarea_margin = 2 * state->focusarea_min;	// begin larger

  state->blinkthres_cur_hd = state->blinkthres_min << 8;
  state->weightsum_slowfilt = 0;

  rgb2ycbcr (state->colorfilter_r, state->colorfilter_g,
	     state->colorfilter_b, &yy, &cb, &cr);
  state->colorfilter_cb_sethd = cb << 8;
  state->colorfilter_cr_sethd = cr << 8;
  state->colorfilter_cb_autohd = state->colorfilter_cb_sethd;
  state->colorfilter_cr_autohd = state->colorfilter_cr_sethd;
  state->colorfilter_autobound_hd = 20 << 8;

  if (0)			// debug
    {
      int r2 = 0;
      int g2 = 0;
      int b2 = 0;

      ycbcr2rgb (yy, cb, cr, &r2, &g2, &b2);
      STDERR ("colfilt r/g/b %d %d %d y/cb/cr %d %d %d backr/g/b %d %d %d",
	      state->colorfilter_r, state->colorfilter_g,
	      state->colorfilter_b, yy, cb, cr, r2, g2, b2);
    }

  state->img_init_flag = 1;

  mqtt_sendcaminfo (state);
}
