// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include <mosquitto.h>
#include <sys/time.h>
#include <sys/types.h>

#include "mqtt.h"
#include "../shared/error.h"
#include "../shared/mqttpub.h"


void
mqtt_sendcaminfo (scwstate_t * state)
{
  if (!state->img_init_flag)
    // Nothing to announce yet
    return;

  if (state->connstate != CS_CONNECTED)
    // Nobody is listening; will be done in mqttconnect_callback()
    return;

  mqttpubjson (state->mosq, state->topicprefix, "cam", state->camname, "info",
	       NULL,
	       json_pack ("{s:s,s:O,s:O,s:i,s:i}",
			  "name", state->camname,
			  "type", state->jcamtype,
			  "distortion", state->jcamdistortion,
			  "pixw", state->width, "pixh", state->height));
}


// For the usual packed RGB bytes.
// Note: width/height are specifiable but should fit within state->pubimg!
void
pubimg_packedrgb (scwstate_t * state, const char *imgtype, int width,
		  int height, unsigned char *img)
{
  char topic[TOPICLEN + 5];
  int headerbytes;

  if (!state->img_init_flag)
    // Pubimg not yet allocd
    return;

  if (strlen (state->topicprefix) + strlen (state->camname) + strlen (imgtype)
      > TOPICLEN - 500)
    {
      MDBG ("imgpub topic too long, increase TOPICLEN");
      return;
    }
  snprintf (topic, TOPICLEN, "%s/cam/%s/img/%s/pnm", state->topicprefix,
	    state->camname, imgtype);

  headerbytes =
    snprintf ((char *) (state->pubimg), 100, "P6\n%d %d\n255\n",
	      width, height);
  if (headerbytes < 5 || headerbytes > 90)
    {
      MDBG ("Can't make proper imgpublish header");
      return;
    }

  // pubimg is allocated for RGB bytes

  memcpy (state->pubimg + headerbytes, img, width * height * 3);

  // Note: this is binary image data, don't view this topic on tty...
  mosquitto_publish (state->mosq, NULL, topic,
		     headerbytes + width * height * 3, state->pubimg, 0, 0);
}

// For integer RGB values as used in histimg[].
// Note: width/height are specifiable but should fit within state->pubimg!
void
pubimg_intgrey (scwstate_t * state, const char *imgtype, int width,
		int height, int *img)
{
  char topic[TOPICLEN + 5];
  int headerbytes;
  int *imgend;
  unsigned char *destpix;

  if (!state->img_init_flag)
    // Pubimg not yet allocd
    return;

  if (strlen (state->topicprefix) + strlen (state->camname) + strlen (imgtype)
      > TOPICLEN - 500)
    {
      MDBG ("imgpub topic too long, increase TOPICLEN");
      return;
    }
  snprintf (topic, TOPICLEN, "%s/cam/%s/img/%s/pnm", state->topicprefix,
	    state->camname, imgtype);

  headerbytes =
    snprintf ((char *) (state->pubimg), 100, "P5\n%d %d\n255\n",
	      width, height);
  if (headerbytes < 5 || headerbytes > 90)
    {
      MDBG ("Can't make proper imgpublish header");
      return;
    }

  // pubimg is allocated for RGB bytes, so will always fit for greybytes.

  imgend = img + (width * height);
  destpix = state->pubimg + headerbytes;
  for (; img < imgend; img++, destpix++)
    if (*img < 0)
      *destpix = 0;
    else if (*img > 65535)
      *destpix = 255;
    else
      *destpix = (*img) >> 8;

  // Note: this is binary image data, don't view this topic on tty...
  mosquitto_publish (state->mosq, NULL, topic,
		     headerbytes + width * height, state->pubimg, 0, 0);
}

// For YUV in 3 separate planes
// Catering for ffmpeg that has separate linesize for all planes
// plane[0..2] = start of Y/Cb/Cr
// linesize[0..2] = increase per line of Y/Cb/Cr
// Note: width/height are specifiable but should fit within state->pubimg!
void
pubimg_planaryuv (scwstate_t * state, const char *imgtype, int width,
		  int height, int hshift, int vshift, unsigned char **plane,
		  int *linesize)
{
  char topic[TOPICLEN + 5];
  int headerbytes;
  int cbcrwidth, cbcrheight;
  unsigned char *inptr;
  unsigned char *destptr;
  int f1, f2, f3;
  int i;

  if (!state->img_init_flag)
    // Pubimg not yet allocd
    return;

  if (strlen (state->topicprefix) + strlen (state->camname) + strlen (imgtype)
      > TOPICLEN - 500)
    {
      MDBG ("imgpub topic too long, increase TOPICLEN");
      return;
    }
  snprintf (topic, TOPICLEN, "%s/cam/%s/img/%s/yuv4mpeg",
	    state->topicprefix, state->camname, imgtype);

  // This works except for possibly 410.
  // Could also have user pass formatstring?
  f1 = 4;
  f2 = f1 >> hshift;
  f3 = f2;
  if (vshift >= 1)
    f3 = 0;

  headerbytes =
    snprintf ((char *) (state->pubimg), 200,
	      "YUV4MPEG2 W%d H%d F30:1 Ip A0:0 C%d%d%d XYSCSS=%d%d%d\nFRAME\n",
	      width, height, f1, f2, f3, f1, f2, f3);
  if (headerbytes < 5 || headerbytes > 90)
    {
      MDBG ("Can't make proper imgpublish header");
      return;
    }

  state->pubimg[headerbytes] = '\0';

  // pubimg is allocated for RGB bytes, so will always fit even for YUV444

  cbcrwidth = CEIL_RSHIFT (width, hshift);
  cbcrheight = CEIL_RSHIFT (height, vshift);

  destptr = state->pubimg + headerbytes;

  inptr = plane[0];
  for (i = 0; i < height; i++)
    {
      memcpy (destptr, inptr, width);
      inptr += linesize[0];
      destptr += width;
    }

  inptr = plane[1];
  for (i = 0; i < cbcrheight; i++)
    {
      memcpy (destptr, inptr, cbcrwidth);
      inptr += linesize[1];
      destptr += cbcrwidth;
    }

  inptr = plane[2];
  for (i = 0; i < cbcrheight; i++)
    {
      memcpy (destptr, inptr, cbcrwidth);
      inptr += linesize[2];
      destptr += cbcrwidth;
    }

  // Note: this is binary image data, don't view this topic on tty...
  mosquitto_publish (state->mosq, NULL, topic,
		     headerbytes + width * height +
		     cbcrwidth * cbcrheight * 2, state->pubimg, 0, 0);
}




/* This connect callback is called when CONNACK is received, i.e. when
   the broker has actively allowed us access.
*/
static void
mqttconnect_callback (struct mosquitto *mosq, void *usrobj, int result)
{
  scwstate_t *state = usrobj;

  if (result != 0)
    {
      // Access actively refused.
      // Socket will be closed by library via
      // mosquitto_loop_read() -> _mosquitto_loop_rc_handle()
      return;
    }

  state->connstate = CS_CONNECTED;

  // Send subscriptions
  mosquitto_subscribe (state->mosq, NULL, state->broadcmd_topic, 0);
  mosquitto_subscribe (state->mosq, NULL, state->allcamcmd_topic, 0);
  mosquitto_subscribe (state->mosq, NULL, state->camcmd_topic, 0);
  mosquitto_subscribe (state->mosq, NULL, state->calcpix_topic, 0);

  // Announce ourself (if info is already available)
  mqtt_sendcaminfo (state);

  // Do NOT send initial "lost":"fix", but keep the "lost":"connection"
  // (last-will) as last published state in case that camera has stopped
  // producing a stream.
  // If there are frames coming, the normal fixlost_timeout frame counter
  // will overflow soon enough, and "lost":"fix" will be posted anyway.
}

/* This disconnect callback is called whenever socket is (already) closed
   for whatever reason.
*/
static void
mqttdisconnect_callback (struct mosquitto *mosq, void *usrobj, int reason)
{
  scwstate_t *state = usrobj;

  // We are disconnected already, can't send any message.
  // But if we wanted disconnection, we already sent "shutdown".
  // And if we got disconnected without our request, the broker will
  // already have sent our Last Will = position-lost message.

  if (state->connstate == CS_EOFDISCONNECTING
      || state->connstate == CS_EOFDISCONNECTED)
    {
      state->connstate = CS_EOFDISCONNECTED;
    }
  else
    {
      state->connstate = CS_UNCONNECTED;
      state->connect_nexttime = time (NULL) + 5;
    }
}

/* This message callback is called whenever a message is received; may be 
   called _multiple_ times from 1 mosquitto_loop_read() call.
*/
static void
mqttmessage_callback (struct mosquitto *mosq, void *usrobj,
		      const struct mosquitto_message *message)
{
  scwstate_t *state = usrobj;
  json_t *jroot;
  json_error_t jerror;

  // We are non-threaded, and all messages can always be handled immediately
  // when they arrive, no queueing needed. So we just handle the message
  // right here inside the callback.

  // We don't want topicless.
  if (message->topic[0] == '\0')
    return;

  // We don't want too little or too much.
  if (message->payloadlen < 0 || message->payloadlen > 1000)
    return;

  // All incoming messages are json
  jroot = json_loadb (message->payload, message->payloadlen, 0, &jerror);
  if (jroot == NULL)
    {
      MDBG ("incoming message json decode failure at bytepos %d: %s",
	    jerror.position, jerror.text);
      return;
    }

  do
    {
      // Easy "break" for decref & return

      if (!strcmp (message->topic, state->broadcmd_topic) ||
	  !strcmp (message->topic, state->allcamcmd_topic) ||
	  !strcmp (message->topic, state->camcmd_topic))
	{
	  const char *cmd;

	  cmd = json_string_value (json_object_get (jroot, "cmd"));
	  if (cmd == NULL)
	    {
	      MDBG ("no cmd in message");
	      break;
	    }

	  if (!strcmp (cmd, "postinfo"))
	    {
	      mqtt_sendcaminfo (state);
	      break;
	    }

	  if (!strcmp (cmd, "postimg"))
	    {
	      const char *img;

	      img = json_string_value (json_object_get (jroot, "img"));
	      if (img == NULL)
		{
		  MDBG ("postimg cmd without specifier");
		  break;
		}

	      if (!strcmp (img, "orig"))
		state->pubimg_orig_flag = 1;
	      if (!strcmp (img, "colfilt"))
		state->pubimg_colfilt_flag = 1;

	      break;
	    }

	  if (!strcmp (cmd, "shutdown"))
	    {
	      // Report connection-lost (same as last-will)
	      mqttprintf (state->mosq, state->topicprefix, "cam",
			  state->camname, "pix", NULL,
			  "{\"lost\": \"connection\"}");

	      if (state->shutdown_touchfile != NULL)
		{
		  FILE *f;
		  f = fopen (state->shutdown_touchfile, "w");
		  if (f != NULL)
		    fclose (f);
		}

	      // Initiate mqtt disconnect, will lead to shutdown after disconnected
	      // by broker, or after timeout; handled in mainloop().

	      if (state->connstate == CS_CONNECTED)
		{
		  state->connstate = CS_EOFDISCONNECTING;
		  state->connect_nexttime = time (NULL) + 5;
		  mosquitto_disconnect (state->mosq);	// just queues DISCONNECT
		}
	      else if (state->connstate != CS_EOFDISCONNECTING
		       && state->connstate != CS_EOFDISCONNECTED)
		{
		  // Not connected at all, so sending DISCONNECT would be
		  // useless.
		  // (This should not happen in practice, since we are handling
		  // the shutdown command that arrived via mqtt.)
		  state->connstate = CS_EOFDISCONNECTED;
		  // This will cause loop break immediately in next round.
		}
	      break;
	    }

	  break;
	}

      if (!strcmp (message->topic, state->calcpix_topic)
	  && state->img_init_flag)
	{
	  int x, y, diff, mindiff;
	  int res;

	  res = json_unpack (jroot, "{s:i, s:i}", "pixx", &x, "pixy", &y);

	  if (res == 0)
	    {
	      // x/y may be totally out-of-range
	      if (x < 0)
		x = 0;
	      if (x > state->width - 1)
		x = state->width - 1;
	      if (y < 0)
		y = 0;
	      if (y > state->height - 1)
		y = state->height - 1;

	      state->backcalcarea_x = x;
	      state->backcalcarea_y = y;

	      // Do not immediately set backcalcarea_margin to focusarea_min:
	      // on large blinks, focusarea should leave enough margin.
	      // This is done at end of each detect() when blink is found,
	      // also immediately enlarges backcalcarea_margin if needed.
	      // So just shrink here, at rate of ~16 times growth rate for each
	      // frame in process_afterimage().
	      diff = state->maxdimension / 16;
	      state->backcalcarea_margin -= diff;
	      if (state->backcalcarea_margin < state->focusarea_min)
		state->backcalcarea_margin = state->focusarea_min;

	      // Assume backcalc is correct; slowly move our area center
	      // to backcalc position.
	      // Note that this is slow enough to keep history within margins.
	      mindiff = state->focusarea_min / 8;
	      if (mindiff < 1)
		mindiff = 1;

	      diff = state->backcalcarea_x - state->ourarea_x;
	      if (diff < -mindiff)
		diff = -mindiff;
	      else if (diff > mindiff)
		diff = mindiff;
	      state->ourarea_x += diff;

	      diff = state->backcalcarea_y - state->ourarea_y;
	      if (diff < -mindiff)
		diff = -mindiff;
	      else if (diff > mindiff)
		diff = mindiff;
	      state->ourarea_y += diff;
	    }
	  else
	    {
	      // "lost":"fix" etc, immediately forget and
	      // only consider our own area.
	      state->backcalcarea_x = state->width / 2;
	      state->backcalcarea_y = state->height / 2;
	      state->backcalcarea_margin = state->maxdimension;
	    }

	  break;
	}

    }
  while (0);

  json_decref (jroot);
  // *message is free()'d by the library after this callback returns.
}


/* Spends at most usectime in select(), processes any result.
   usectime must be >=0 and <1000000.
   Returns:
   1 = stdin readable (if requested)
   0 = okay, nothing else to do
   -1 = our disconnect was honoured and finished: terminate program
   -2 = fatal error: terminate program (select(), <fps)
*/
int
mqtt_selectround (scwstate_t * state, int usectime, int readstdin)
{
  time_t now;
  fd_set readfds, writefds;
  int mosqsock, maxfd;
  struct timeval tv;
  int selectres;
  int retval = 0;

  now = time (NULL);

  // Server has disconnected after we asked
  if (state->connstate == CS_EOFDISCONNECTED)
    return -1;

  // Server did not answer after we asked DISCONNECT
  // (Test: suspend broker with Ctrl-Z just before DISCONNECT.)
  // Don't wait longer, forcibly close socket with _destroy().
  if (state->connstate == CS_EOFDISCONNECTING
      && now >= state->connect_nexttime)
    return -1;

  // If disconnected without our asking, attempt to reconnect periodically
  // (=also immediately on startup).
  if (state->connstate == CS_UNCONNECTED && now >= state->connect_nexttime)
    {
      state->connstate = CS_CONNECTING;

      if (mosquitto_connect
	  (state->mosq, state->svr_hostname, state->svr_port,
	   10) != MOSQ_ERR_SUCCESS)
	{
	  // Failed, try again after some time
	  now = time (NULL);	// DNS/connect may have taken time
	  state->connstate = CS_UNCONNECTED;
	  state->connect_nexttime = now + 5;
	}

      // mosquitto_connect() returns after TCP connect() finished, with
      // CONNECT packet queued to be sent by next mosquitto_loop_write().
      // DNS resolving and connect() may be protected by alarm() by caller.

      // We keep ->connstate in CS_CONNECTING until CONNACK received
      // via mqttconnect_callback().

      // CONNACK is waited for up to keepalive time, checked in
      // _mosquitto_check_keepalive() via mosquitto_loop_misc().
      // (state->mosq->state is not cs_connected yet; _no_ pings are sent.)
      // Waiting-for-CONNACK is easily testable: Ctrl-S on broker.
    }

  FD_ZERO (&readfds);
  FD_ZERO (&writefds);
  maxfd = 0;

  if (readstdin)
    FD_SET (0, &readfds);
  // already maxfd = 0;

  mosqsock = mosquitto_socket (state->mosq);
  if (mosqsock != -1)
    {
      FD_SET (mosqsock, &readfds);
      if (mosquitto_want_write (state->mosq) == true)
	FD_SET (mosqsock, &writefds);
      maxfd = mosqsock;
    }

  tv.tv_sec = 0;
  tv.tv_usec = usectime;

  selectres = select (maxfd + 1, &readfds, &writefds, NULL, &tv);
  // We could handle errno==EINTR, but shouldn't occur in practice.
  if (selectres == -1)
    return -2;			// Stop on any error

  if (readstdin && FD_ISSET (0, &readfds))
    retval = 1;

  if (mosqsock != -1 && FD_ISSET (mosqsock, &readfds))
    {
      mosquitto_loop_read (state->mosq, 1);
      // Most callbacks are called from here, including
      // mqttmessage_callback() which handles any incoming messages
      // (possibly multiple, if piled up in tcp buffer),
      // and starts new foreground communication if possible.

      // Don't mind exitcode; socket already disconnected in case of
      // trouble, then also mqttdisconnect_callback() was called.
    }

  if (mosqsock != -1 && FD_ISSET (mosqsock, &writefds))
    {
      mosquitto_loop_write (state->mosq, 1);
    }

  // Schedule keepalives and check them
  mosquitto_loop_misc (state->mosq);


  // Handle timeouts etc.

  now = time (NULL);		// select() may have waited up to 1 sec

  if (state->imgpersec_time == 0)
    {
      // startup situation (state cleared)
      state->imgpersec_time = now + 3;	// allow 3-second startup
      state->imgpersec_cnt = 10000;	// pre-load for first seconds
    }
  else if (now > state->imgpersec_time)
    {
      if (state->imgpersec_cnt < state->imgpersec_min)
	{
	  STDERR ("less than %d fps, webcam broken? CPU too slow?",
		  state->imgpersec_min);
	  return -2;
	}

      if (state->imgpersec_cnt < 10000)
	mqttpubjson (state->mosq, state->topicprefix, "cam", state->camname,
		     "status", NULL,
		     json_pack ("{s:i}", "fps", state->imgpersec_cnt));

      state->imgpersec_time = now;
      state->imgpersec_cnt = 0;
    }

  if (state->blinkspersec_time == 0)
    {
      // startup situation (state cleared)
      state->blinkspersec_time = now;
      state->blinkspersec_cnt = 10000;	// pre-load for first seconds
    }
  else if ((now & 0x03) == 0 && now > state->blinkspersec_time)
    {
      // every 4 seconds

      if (state->blinkspersec_cnt < 10000)
	mqttpubjson (state->mosq, state->topicprefix, "cam", state->camname,
		     "status", NULL,
		     json_pack ("{s:f, s:i}",
				"blinkspersec",
				(double) state->blinkspersec_cnt /
				(double) (now -
					  state->blinkspersec_time) / 2.,
				"blinkthres", state->blinkthres_cur_hd >> 8));
      // Note that we count both edges, i.e. blink freq is half of that.

      state->blinkspersec_cnt = 0;
      state->blinkspersec_time = now;
    }


  return retval;
}


// Returns 0=okay !0=error
int
mqtt_init (scwstate_t * state)
{
  char clientid[MOSQ_MQTT_ID_MAX_LENGTH + 5];
  char hostname[TOPICLEN + 5];
  int res;

  // Note that broker kicks any connected client with equal clientid.
  // Clientid is max 23 chars.
  if (gethostname (hostname, TOPICLEN) != 0)
    hostname[0] = '\0';
  res =
    snprintf (clientid, MOSQ_MQTT_ID_MAX_LENGTH, "%d/%s", getpid (),
	      hostname);
  // Fake resultvalue handling to make gcc stop complaining about the
  // intended truncation, which will actually always have res LARGER than 1.
  if (res < 1)
    exit (1);

  mosquitto_lib_init ();	// cannot return errors

  state->mosq = mosquitto_new (clientid, true, state);
  if (state->mosq == NULL)
    {
      STDERR ("mosquitto_new failed: %s", strerror (errno));
      return 1;
    }

  mosquitto_connect_callback_set (state->mosq, mqttconnect_callback);
  mosquitto_disconnect_callback_set (state->mosq, mqttdisconnect_callback);
  mosquitto_message_callback_set (state->mosq, mqttmessage_callback);

  // Set Last Will
  if (strlen (state->topicprefix) + strlen (state->camname) + 20 < TOPICLEN)
    {
      char topic[TOPICLEN];
      const char *payload = "{\"lost\": \"connection\"}";

      snprintf (topic, TOPICLEN, "%s/cam/%s/pix", state->topicprefix,
		state->camname);
      mosquitto_will_set (state->mosq, topic, strlen (payload), payload, 0,
			  0);
    }

  return 0;			// all okay
}

void
mqtt_stop (scwstate_t * state)
{
  mosquitto_destroy (state->mosq);
  // This is the _only_ way an mqtt client can forcibly close the socket.
}
