// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "process_image.h"
#include "mqtt.h"
#include "colorfilter.h"
#include "detect.h"
#include "../shared/error.h"


// Ffmpeg integration considers this a header file and demands a prototype.
void updates_afterimage (scwstate_t * state);
void
updates_afterimage (scwstate_t * state)
{
  int i;

  state->imgpersec_cnt++;

  if (state->colorfilter_adjustreport > 0)
    state->colorfilter_adjustreport--;
  else
    state->colorfilter_adjustreport = 0;

  // Grow area margins
  // 10sec @25fps to max borders
  // (when in center, will already cover entire image at 5sec)
  i = state->maxdimension / 250;
  if (i < 1)
    i = 1;

  state->ourarea_margin += i;
  if (state->ourarea_margin > state->maxdimension)
    state->ourarea_margin = state->maxdimension;

  state->backcalcarea_margin += i;
  if (state->backcalcarea_margin > state->maxdimension)
    state->backcalcarea_margin = state->maxdimension;

  MDBG ("ourx/y/m %d %d %d backcalcx/y/m %d %d %d",
	state->ourarea_x, state->ourarea_y, state->ourarea_margin,
	state->backcalcarea_x, state->backcalcarea_y,
	state->backcalcarea_margin);


  // Report lost position if no blink seen for too long
  state->fixlost_timeout++;
  if (state->fixlost_timeout == state->detect_wait_max * 2)
    {
      mqttprintf (state->mosq, state->topicprefix, "cam", state->camname,
		  "pix", NULL, "{\"lost\": \"fix\"}");
      // let counter overflow, no problem to send "lost" again then
    }
}

// Ffmpeg integration considers this a header file and demands a prototype.
void prepare_histslot (scwstate_t * state);
void
prepare_histslot (scwstate_t * state)
{
  scwhistimg_t *imgp;

  state->curhist++;
  if (state->curhist >= HISTIMGS)
    state->curhist = 0;

  imgp = &state->histimg[state->curhist];

  // Find area of interest,
  // backcalc has precedence if smaller.
  if (state->backcalcarea_margin < state->ourarea_margin)
    {
      imgp->area_xmin = state->backcalcarea_x - state->backcalcarea_margin;
      imgp->area_xmax = state->backcalcarea_x + state->backcalcarea_margin;
      imgp->area_ymin = state->backcalcarea_y - state->backcalcarea_margin;
      imgp->area_ymax = state->backcalcarea_y + state->backcalcarea_margin;
    }
  else
    {
      imgp->area_xmin = state->ourarea_x - state->ourarea_margin;
      imgp->area_xmax = state->ourarea_x + state->ourarea_margin;
      imgp->area_ymin = state->ourarea_y - state->ourarea_margin;
      imgp->area_ymax = state->ourarea_y + state->ourarea_margin;
    }

  if (imgp->area_xmin < 0)
    imgp->area_xmin = 0;
  if (imgp->area_xmin >= state->width)
    imgp->area_xmin = state->width - 1;;
  if (imgp->area_xmax < 0)
    imgp->area_xmax = 0;
  if (imgp->area_xmax >= state->width)
    imgp->area_xmax = state->width - 1;;

  if (imgp->area_ymin < 0)
    imgp->area_ymin = 0;
  if (imgp->area_ymin >= state->height)
    imgp->area_ymin = state->height - 1;;
  if (imgp->area_ymax < 0)
    imgp->area_ymax = 0;
  if (imgp->area_ymax >= state->height)
    imgp->area_ymax = state->height - 1;;
}



void
process_stdin_image (scwstate_t * state)
{
  // First publish original image if requested
  if (state->pubimg_orig_flag)
    {
      if (state->recvimg_formatbase == IMGF_RGB)
	{
	  pubimg_packedrgb (state, "orig", state->width,
			    state->height, state->recvimg);
	}
      else if (state->recvimg_formatbase == IMGF_YUV)
	{
	  int linesize[3];
	  linesize[0] = state->width;
	  linesize[1] = state->recvimg_yuv_cbcrwidth;
	  linesize[2] = state->recvimg_yuv_cbcrwidth;

	  pubimg_planaryuv (state, "orig", state->width, state->height,
			    state->recvimg_yuv_hshift,
			    state->recvimg_yuv_vshift,
			    state->recvimg_yuv_plane, linesize);
	}

      state->pubimg_orig_flag = 0;
    }

  prepare_histslot (state);

  // Color filter depends on input format

  if (state->recvimg_formatbase == IMGF_RGB)
    {
      // Note: This converts to YUV (YCbCr) for filtering,
      // so don't use an RGB stream when you have YUV already!
      colfilt_packedrgb (state, state->recvimg);
    }
  else if (state->recvimg_formatbase == IMGF_YUV)
    {
      int linesize[3];
      linesize[0] = state->width;
      linesize[1] = state->recvimg_yuv_cbcrwidth;
      linesize[2] = state->recvimg_yuv_cbcrwidth;

      colfilt_planaryuv (state, state->recvimg_yuv_hshift,
			 state->recvimg_yuv_vshift,
			 state->recvimg_yuv_plane, linesize);
    }

  // Publish color-filtered image if requested
  if (state->pubimg_colfilt_flag)
    {
      pubimg_intgrey (state, "colfilt", state->width, state->height,
		      state->histimg[state->curhist].img);
      state->pubimg_colfilt_flag = 0;
    }

  detect_blink (state);

  updates_afterimage (state);

  if (state->brightestx >= 0)
    {
      if (state->recvimg_formatbase == IMGF_RGB)
	{
	  filteradjust_packedrgb (state, state->recvimg);
	}
      else if (state->recvimg_formatbase == IMGF_YUV)
	{
	  int linesize[3];
	  linesize[0] = state->width;
	  linesize[1] = state->recvimg_yuv_cbcrwidth;
	  linesize[2] = state->recvimg_yuv_cbcrwidth;

	  filteradjust_planaryuv (state, state->recvimg_yuv_hshift,
				  state->recvimg_yuv_vshift,
				  state->recvimg_yuv_plane, linesize);
	}
    }
}
