// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "stdin.h"
#include "image_init.h"
#include "mqtt.h"
#include "../shared/error.h"


// Helper functions for PPM (P6) header parsing according to specs,
// i.e. any-whitespace-separated

static inline char *
find_firstnonwhite (char *cp, char *lastp)
{
  while (cp < lastp)
    {
      if (*cp != ' ' && *cp != '\t' && *cp != '\n' && *cp != '\r')
	return cp;
      cp++;
    }
  // Can only be here if cp >= lastp
  STDERR ("Input header too large or wrong input format.");
  return NULL;
}

static inline char *
find_firstwhite (char *cp, char *lastp)
{
  while (cp < lastp)
    {
      if (*cp == ' ' || *cp == '\t' || *cp == '\n' || *cp == '\r')
	return cp;
      cp++;
    }
  // Can only be here if cp >= lastp
  STDERR ("Input header too large or wrong input format.");
  return NULL;
}

static inline char *
find_firstlineend (char *cp, char *lastp)
{
  while (cp < lastp)
    {
      if (*cp == '\n' || *cp == '\r')
	return cp;
      cp++;
    }
  // Can only be here if cp >= lastp
  STDERR ("Input header too large or wrong input format.");
  return NULL;
}

static inline char *
find_firstnonwhite_skipcomments (char *cp, char *lastp)
{
  while (1)
    {
      cp = find_firstnonwhite (cp, lastp);
      if (cp == NULL)
	return NULL;
      if (*cp == '#')
	{
	  // This is a comment: skip entire line.
	  cp = find_firstlineend (cp, lastp);
	  if (cp == NULL)
	    return NULL;
	  cp++;
	  // Now cp points to first char of next line,
	  // try again from there.
	}
      else
	return cp;
    }
}


// Returns consumed header bytes
// or <0 = error
int
ppmheaderparse (scwstate_t * state)
{
  char *startp;
  char *endp;
  char *lastp;
  char *headerend;
  int width, height, maxval;

  lastp = (char *) (state->headbuf + state->inheadbuf);

  startp = (char *) state->headbuf;
  endp = find_firstwhite (startp, lastp);
  if (endp == NULL)
    return -1;
  *endp = '\0';

  if (strcmp (startp, "P6"))
    {
      STDERR ("Only P6 image format is supported");
      return -1;
    }

  startp = find_firstnonwhite_skipcomments (endp + 1, lastp);
  if (startp == NULL)
    return -1;
  endp = find_firstwhite (startp, lastp);
  if (endp == NULL)
    return -1;
  *endp = '\0';

  width = atoi (startp);
  if (width < 10 || width > 10000)
    {
      STDERR ("Input width too small or too large");
      return -1;
    }

  startp = find_firstnonwhite_skipcomments (endp + 1, lastp);
  if (startp == NULL)
    return -1;
  endp = find_firstwhite (startp, lastp);
  if (endp == NULL)
    return -1;
  *endp = '\0';

  height = atoi (startp);
  if (height < 10 || height > 10000)
    {
      STDERR ("Input height too small or too large");
      return -1;
    }

  startp = find_firstnonwhite_skipcomments (endp + 1, lastp);
  if (startp == NULL)
    return -1;
  headerend = find_firstlineend (startp, lastp);
  if (headerend == NULL)
    return -1;
  // P6 must assume that no comments follow maxval line,
  // no way to differentiate between comment and binary image content...

  endp = find_firstwhite (startp, lastp);
  if (endp == NULL)
    return -1;
  *endp = '\0';			// Possibly overwrites headerend = \n, no problem

  maxval = atoi (startp);
  if (maxval != 255)
    {
      STDERR ("Only maxval 255 is supported");
      return -1;
    }

  if (!state->img_init_flag)
    {
      state->recvimg_formatbase = IMGF_RGB;
      state->recvimg_formatdetail = IMGF_RGB;
      state->width = width;
      state->height = height;
    }
  else
    {
      if (width != state->width || height != state->height)
	{
	  STDERR
	    ("Changing input image size not supported (%d x %d -> %d x %d)",
	     state->width, state->height, width, height);
	  return -1;
	}
    }

  // headerend points to final newline char
  return (headerend - (char *) state->headbuf) + 1;
}

// Returns consumed header bytes
// or <0 = error
int
yuvheaderparse (scwstate_t * state)
{
  // File header like:
  //   YUV4MPEG2 W640 H480 F30:1 Ip A0:0 C422 XYSCSS=422 <more> \n
  // Frame header like:
  //   FRAME <more> \n
  //
  // All fields separated by single space; always one line.
  // See: http://linux.die.net/man/5/yuv4mpeg

  char *cp, *line2p;
  int width, height;

  if (!memcmp (state->headbuf, "YUV4MPEG2", 9))
    {
      // This is the first buffer, which must have file header first.
      line2p = memchr (state->headbuf, '\n', state->inheadbuf);
      if (line2p == NULL)
	{
	  STDERR ("no newline after file header");
	  return -1;
	}
      *line2p = '\0';
      line2p++;

      width = -1;
      height = -1;
      state->recvimg_formatdetail = IMGF_NONE;

      for (cp = strtok ((char *) state->headbuf, " "); cp;
	   cp = strtok (NULL, " "))
	{
	  if (*cp == 'W')
	    width = atoi (cp + 1);
	  else if (*cp == 'H')
	    height = atoi (cp + 1);
	  else if (*cp == 'I')
	    {
	      if (*(cp + 1) != 'p')
		{
		  STDERR ("Only progressive-scan supported (Ip)");
		  return -1;
		}
	    }
	  else if (*cp == 'C')
	    {
	      if (!strcasecmp (cp + 1, "411"))
		state->recvimg_formatdetail = IMGF_411p;
	      else if (!strcasecmp (cp + 1, "420"))
		state->recvimg_formatdetail = IMGF_420p;
	      else if (!strcasecmp (cp + 1, "420paldv"))
		state->recvimg_formatdetail = IMGF_420p;
	      else if (!strcasecmp (cp + 1, "420mpeg2"))
		state->recvimg_formatdetail = IMGF_420p;
	      else if (!strcasecmp (cp + 1, "420jpeg"))
		state->recvimg_formatdetail = IMGF_420p;
	      else if (!strcasecmp (cp + 1, "422"))
		state->recvimg_formatdetail = IMGF_422p;
	      else if (!strcasecmp (cp + 1, "444"))
		state->recvimg_formatdetail = IMGF_444p;
	      else
		{
		  STDERR ("Unsupported color format %s", cp + 1);
		  return -1;
		}
	    }
	}
      if (width <= 0 || height <= 0)
	{
	  STDERR ("No width or height in input stream");
	  return -1;
	}

      if (state->recvimg_formatdetail == IMGF_NONE)
	{
	  STDERR ("No color format in input stream");
	  return -1;
	}

      if (!state->img_init_flag)
	{
	  state->recvimg_formatbase = IMGF_YUV;
	  state->width = width;
	  state->height = height;
	}
      else
	{
	  if (width != state->width || height != state->height)
	    {
	      STDERR
		("Changing input image size not supported (%d x %d -> %d x %d)",
		 state->width, state->height, width, height);
	      return -1;
	    }
	}

      // Also skip frame header, below.
    }
  else
    {
      // 2nd or later frame: only frame header
      line2p = (char *) state->headbuf;
    }

  // Frame header is easy, just skip it.
  if (memcmp (line2p, "FRAME", 5))
    {
      STDERR ("no FRAME header");
      return -1;
    }
  cp = memchr (line2p, '\n',
	       state->inheadbuf - (line2p - (char *) state->headbuf));
  if (cp == NULL)
    {
      STDERR ("No newline following FRAME header");
      return -1;
    }

  // cp points to newline char following FRAME header
  return (cp - (char *) state->headbuf) + 1;
}

// Returns consumed header bytes
// or <0 = error
int
headerparse (scwstate_t * state)
{
  if (!memcmp (state->headbuf, "P6", 2))
    return ppmheaderparse (state);
  else if (!memcmp (state->headbuf, "YUV4MPEG2", 9))
    return yuvheaderparse (state);
  else if (!memcmp (state->headbuf, "FRAME", 5))
    return yuvheaderparse (state);
  else
    {
      STDERR ("Unrecognized input video - use PPM P6 or YUV4MPEG2 format");
      return -1;
    }
}


// Read stdin and report completed image
//
// Note: We do image_init() as soon as needed, since that will
// allocate state->recvimg that we need to copy remaining non-header
// bytes to.
// Returns:
// 1 = complete image in state->recvimg, caller can process it.
// 0 = not yet complete, nothing to do.
// <0 = error or EOF, message printed; program should be terminated.
int
read_stdin (scwstate_t * state, int fd)
{
  int bytestogo;
  int bytesread;

  // Every frame starts with a variable-length header,
  // then the *fixed*-length image (namely uncompressed RGB or YUV).
  // We start reading the full headbuf; and anything that's left
  // will be transferred to the recvimg.
  // Then we read the full recvimg, which will never have anything left.

  if (state->readinghead)
    {
      int headerused;

      bytestogo = state->headbuf_size - state->inheadbuf;
      if (bytestogo <= 0)
	{
	  STDERR ("Shouldn't happen: bytestogo <= 0");
	  return -1;
	}
      bytesread = read (fd, state->headbuf + state->inheadbuf, bytestogo);
      if (bytesread <= 0)
	{
	  // EOF or error
	  STDERR ("EOF or input error during header");
	  return -1;
	}
      state->inheadbuf += bytesread;
      if (state->inheadbuf < state->headbuf_size)
	// We must read some more next time
	return 0;

      // Header should be complete now
      headerused = headerparse (state);
      if (headerused < 0)
	return headerused;

      // First header should be enough to know image params
      if (!state->img_init_flag)
	{
	  image_init (state);
	  // Allocates state->recvimg that we need to copy non-header bytes to.
	}

      // Move remaining data from headbuf to recvimg.
      state->inrecvimg = state->inheadbuf - headerused;
      memcpy (state->recvimg, state->headbuf + headerused, state->inrecvimg);

      // Next time continue with image reads.
      state->readinghead = 0;

      // We have read, so need another select before trying to read more.
      return 0;
    }

  // else: reading image data

  bytestogo = state->recvimg_size - state->inrecvimg;
  if (bytestogo <= 0)
    {
      STDERR ("Shouldn't happen: bytestogo <= 0");
      return -1;
    }
  bytesread = read (fd, state->recvimg + state->inrecvimg, bytestogo);
  if (bytesread <= 0)
    {
      // EOF or error
      STDERR ("EOF or input error during image");
      return -1;
    }
  state->inrecvimg += bytesread;
  if (state->inrecvimg < state->recvimg_size)
    // We must read some more next time
    return 0;

  // Image is complete

  if (state->inrecvimg != state->recvimg_size)
    {
      STDERR ("Algorithm error: inrecvimg %d != recvimg_size %d",
	      state->inrecvimg, state->recvimg_size);
      return -1;
    }

  // Start with new image next time
  state->inheadbuf = 0;
  state->readinghead = 1;

  // And caller can process it
  // (namely by calling process_stdin_image() below)
  return 1;
}


void
stdin_init (scwstate_t * state)
{
  state->headbuf_size = 1024;
  state->headbuf = malloc (state->headbuf_size);
  if (state->headbuf == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  state->inheadbuf = 0;
  state->readinghead = 1;
}
