// USAGE: camcalibplanes < input.json
//
// Calculates camera parameters from lists of many known real-world points
// in flat surfaces (planes) mapped to camera pixels.
//
// Input format: see examples.
//
// See ./opencv-findcorners/ for "easy" cases,
// see ./hugin/ for "manual" cases.

// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#define PROGNAME  "camcalibplanes"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <jansson.h>

#include "../shared/error.h"
#include "../camfuncs/camfuncs.h"
#include "../camfuncs/camhelpers.h"
#include "../cmpfit/mpfit.h"

#define FIT_ACCEPTSUMSQ  0.050

#define FITDBG(...)		/*nothing */
//#define FITDBG(format, args...)  printf(format "\n", ##args)


const char *progname = PROGNAME;


typedef struct
{
  char *name;
  int camtype;			// index to camtype[]
  double *typeparam;		// malloc'd for specified type
  int type_fitparamidx;		// where typeparam is in param array
  int camdistortion;		// index to camdistortion[]
  double *distortionparam;	// malloc'd for specified distortion
  int dist_fitparamidx;		// where distortionparam is in param array

  int w;			// camera #pixels
  int h;
  double xscale;		// calculate pixel to -1..+1 coords
  double xoffset;
  double yscale;
  double yoffset;
} cam_t;

// parse_cam_json requires cam_t to be defined
#define CAMERR_RETURN(args...)  do { STDERR(args); exit(1); } while(0)
#include "../shared/parse_cam_json.h"
#undef CAMERR_RETURN

typedef struct
{
  double posrot[6];		// position + rotation: x,y,z,a,b,g
  int fitparamidx;		// where x,y,z,a,b,g are in param array
  double focallength;		// guess for this position, later avg'd
} campos_t;

// We support multiple different grids
typedef struct
{
  double ix;			// interval between points (meters)
  double iy;
} grid_t;

// All observations in one list (okay, database table if you prefer)
typedef struct
{
  int gridnum;
  int camposnum;
  double gridn;
  double gridm;
  double pixx;			// -1..+1 calculated
  double pixy;
  int farthestcorner;
} gridpix_t;


// We support multiple grids (e.g big and small), multiple camera positions,
// and very many gridpoint observation pixels.
grid_t *grid = NULL;		// mallocd
int grid_allocd = 0;
int grid_used = 0;

gridpix_t *gridpix = NULL;	// mallocd
int gridpix_allocd = 0;
int gridpix_used = 0;

int campos_used = 0;

double *pix2pixdist = NULL;	// mallocd

camtype_func_t *camfunc_basic;
distortion_func_t *distfunc_none;
int camfunc_basic_idx;
int distfunc_none_idx;

int
new_grid ()
{
  if (grid_used >= grid_allocd)
    {
      grid_allocd = grid_used * 2;
      if (grid_allocd < 10)
	grid_allocd = 10;
      grid = realloc (grid, sizeof (*grid) * grid_allocd);
      if (grid == NULL)
	{
	  STDERR ("realloc failed");
	  exit (1);
	}
    }

  memset (&grid[grid_used], 0, sizeof (grid[grid_used]));

  grid_used++;
  return grid_used - 1;
}

int
new_gridpix ()
{
  if (gridpix_used >= gridpix_allocd)
    {
      gridpix_allocd = gridpix_used * 2;
      if (gridpix_allocd < 10)
	gridpix_allocd = 10;
      gridpix = realloc (gridpix, sizeof (*gridpix) * gridpix_allocd);
      if (gridpix == NULL)
	{
	  STDERR ("realloc failed");
	  exit (1);
	}
    }

  memset (&gridpix[gridpix_used], 0, sizeof (gridpix[gridpix_used]));

  gridpix_used++;
  return gridpix_used - 1;
}

inline int
param_nonzero (double d)
{
  return (fabs (d) > 0.0000001);
}


// Adapted from testmpfit.c
void
mp_printresult (double *x, mp_result * result)
{
  int i;

  if ((x == 0) || (result == 0))
    return;
  printf ("  CHI-SQUARE = %f    (%d DOF)\n",
	  result->bestnorm, result->nfunc - result->nfree);
  printf ("   eqv resid = %f\n", sqrt (result->bestnorm / result->nfunc));
  // This is not the same as average residual, since residuals are weighted
  // by themselves, i.e. larger residuals weigh more in the sum-of-squares.
  // When residual values are really spread out, there will be relatively more
  // large residuals with large weights, so the difference will get more
  // pronounced.

  printf ("  params     = %d\n", result->npar);
  printf ("   - free    = %d\n", result->nfree);
  printf ("   - pegged  = %d\n", result->npegged);
  printf ("  iterations = %d\n", result->niter);
  printf ("  func evals = %d\n", result->nfev);
  for (i = 0; i < result->npar; i++)
    {
      printf ("  P[%d] = %f +/- %f\n", i, x[i], result->xerror[i]);
    }
  if (result->resid != NULL)
    {
      double sum = 0.;
      double abssum = 0.;
      double sqsum = 0.;
      double absmax = 0.;
      for (i = 0; i < result->nfunc; i++)
	{
	  sum += result->resid[i];
	  abssum += fabs (result->resid[i]);
	  sqsum += result->resid[i] * result->resid[i];
	  if (fabs (result->resid[i]) > absmax)
	    absmax = fabs (result->resid[i]);
	  printf ("  resid[%d] = %f\n", i, result->resid[i]);
	}
      printf ("  resid avg = %f   avg-of-abs = %f   sum-of-squares = %f\n",
	      sum / result->nfunc, abssum / result->nfunc, sqsum);
      printf ("  resid max-of-abs = %f\n", absmax);
    }
}


// Returns objpos[] = grid point position x,y,z
// We imagine grid pasted to wall with y=0; n to right=x+ and m to bottom=z-.
inline void
getgridxyz (int gridnum, double n, double m, double *objpos)
{
  objpos[0] = n * grid[gridnum].ix;
  objpos[1] = 0;
  objpos[2] = -m * grid[gridnum].iy;
}


// Special version: Camtype+params/dist+params specified separately.
// Distortion not out-of-range-linearized, just as is, since we are
// fitting the distortion parameters.
// 
// double camposrot[6] = camera position+rotation x/y/z/a/b/g
// double objpos[3] = object position x/y/z
static inline void
calc_campix2 (cam_t * cam_p, double *camposrot,
	      camtype_func_t * camfunc,
	      double *camtypeparam,
	      distortion_func_t * distfunc,
	      double *camdistparam,
	      double *objpos, double *campixx, double *campixy)
{
  double rotx, roty, rotz;
  double officialx, officialy;

  FITDBG ("  position obj x %f y %f z %f", objpos[0], objpos[1], objpos[2]);

  FITDBG ("  position cam %s x %f y %f z %f", cam_p->name,
	  camposrot[0], camposrot[1], camposrot[2]);

  // now object moves to objx-camx, objy-camy, objz-camz
  FITDBG ("  translated objx %f y %f z %f",
	  objpos[0] - camposrot[0],
	  objpos[1] - camposrot[1], objpos[2] - camposrot[2]);

  FITDBG ("  rotation cam %s a %f b %f g %f", cam_p->name,
	  camposrot[3], camposrot[4], camposrot[5]);

  rotate_3d (objpos[0] - camposrot[0],
	     objpos[1] - camposrot[1],
	     objpos[2] - camposrot[2],
	     camposrot[3], camposrot[4], camposrot[5], &rotx, &roty, &rotz);

  FITDBG ("  rotated objx %f y %f z %f", rotx, roty, rotz);

  FITDBG ("  camera typeparam %f %f %f", camtypeparam[0],
	  camtypeparam[1], camtypeparam[2]);

  camfunc (rotx, roty, rotz, &officialx, &officialy, camtypeparam);

  FITDBG ("  camera official x %f y %f", officialx, officialy);

  FITDBG ("  camera distortionparam %f %f %f",
	  camdistparam[0], camdistparam[1], camdistparam[2]);

  distfunc (officialx, officialy, campixx, campixy, camdistparam);

  FITDBG ("  camera distorted x %f y %f", *campixx, *campixy);
}



// Fit functions work like this:
//
// Caller has guessed param[nparam], and this fit function must calculate
// how correct those guesses are.
// It does so by assuming the guesses are fully correct, and calculates what
// the results would have been of the nmeas measurements that were done.
// Of course, those calculated results will be different from the measured
// results, and these differences must be stored in residual[nmeas].
// Caller will then use these residuals to adjust the param guesses.
//
// *private can be used to pass any data that is needed, such as the measured
// results to compare against.
//
// (derivvec[])[], if non-NULL, can be used to store derivatives if calculable.
//
// Returns 0 = success.


typedef struct
{
  cam_t *cam;
  campos_t *campos;
  int *gridpoint_indexes;
} private_firstguess_t;

// Specialized edition for limited number of grid points, e.g. just 4 points
// Result is a reasonably accurate guess for camera position/rotation
// that will be further refined in the complete optimization.
int
gridpix_firstguess_fitfunc (int nmeas, int nparam, double *param,
			    double *residual, double **derivvec,
			    void *privatevoid)
{
  private_firstguess_t *private = privatevoid;
  cam_t *cam = private->cam;
  campos_t *campos = private->campos;

  // List of grid point indexes that we must do now.
  // Note: doubled! even = pixx, odd = pixy
  // E.g. { 5, 5, 16, 16, 27, 27, 42, 42 }   length = nmeas
  int *gp_idx = private->gridpoint_indexes;

  int i;

  // Always check to avoid really nasty surprises.
  for (i = 0; i < nmeas; i += 2)
    if (gp_idx[i] != gp_idx[i + 1])
      {
	STDERR ("FATAL: called with unequal even/odd "
		"gp_idx[%d]=%d != gp_idx[%d+1]=%d, "
		"which is not currently handled", i, gp_idx[i],
		i, gp_idx[i + 1]);
	exit (1);
      }

  for (i = 0; i < nmeas; i += 2)
    {
      int j = gp_idx[i];	// j = gridpix[] index
      double objpos[3];
      int p_pos, p_cam, p_dist;
      double campixx, campixy;

      getgridxyz (gridpix[j].gridnum,
		  gridpix[j].gridn, gridpix[j].gridm, objpos);

      FITDBG ("  %s for meas %d gp_idx %d = grid %f %f = x %f y %f z %f",
	      __FUNCTION__, i, j, gridpix[j].gridn, gridpix[j].gridm,
	      objpos[0], objpos[1], objpos[2]);

      // Camera pos/rot x,y,z,a,b,g are being guessed.
      // This is different for each camposnum -- but we handle one only.
      p_pos = campos[gridpix[j].camposnum].fitparamidx;	// always 0

      p_cam = cam->type_fitparamidx;	// always 6, length 1 (camfunc_basic)
      p_dist = cam->dist_fitparamidx;	// always 0, length 0 (distfunc_none)

      calc_campix2 (cam, &param[p_pos],
		    camfunc_basic,
		    &param[p_cam],
		    distfunc_none,
		    &param[p_dist], objpos, &campixx, &campixy);

      residual[i] = gridpix[j].pixx - campixx;
      residual[i + 1] = gridpix[j].pixy - campixy;

      FITDBG ("  wanted x %f y %f  resid x %f y %f",
	      gridpix[j].pixx, gridpix[j].pixy, residual[i], residual[i + 1]);
    }

  return 0;
}


typedef struct
{
  cam_t *cam;
  campos_t *campos;
  int funcevals;
} private_fitfunc_t;

int
gridpix_fitfunc (int nmeas, int nparam, double *param, double *residual,
		 double **derivvec, void *privatevoid)
{
  private_fitfunc_t *private = privatevoid;
  cam_t *cam = private->cam;
  campos_t *campos = private->campos;
  int i;

  private->funcevals++;
  if (private->funcevals % 1000 == 0)
    {
      printf (".");
      fflush (NULL);
    }

  // Check that nmeas is always same, i.e. that we never get called with
  // a meas subset; that's required for even/odd split to keep working.
  FITDBG ("  %s for nmeas %d nparam %d", __FUNCTION__, nmeas, nparam);
  // -> okay, but do check always to avoid really nasty surprises.
  if (nmeas != 2 * gridpix_used)
    {
      STDERR ("FATAL: called with nmeas = %d instead of expected %d, "
	      "which is not currently handled", nmeas, 2 * gridpix_used);
      exit (1);
    }

  for (i = 0; i < nmeas; i += 2)
    {
      int j = i >> 1;		// j = gridpix[] index
      double objpos[3];
      int p_pos, p_cam, p_dist;
      double campixx, campixy;

      getgridxyz (gridpix[j].gridnum,
		  gridpix[j].gridn, gridpix[j].gridm, objpos);

      FITDBG ("  %s for meas %d = grid %f %f = x %f y %f z %f",
	      __FUNCTION__, i, gridpix[j].gridn, gridpix[j].gridm,
	      objpos[0], objpos[1], objpos[2]);

      // Camera pos/rot x,y,z,a,b,g are being refined.
      // This is different for each camposnum.
      p_pos = campos[gridpix[j].camposnum].fitparamidx;

      p_cam = cam->type_fitparamidx;
      p_dist = cam->dist_fitparamidx;

      calc_campix2 (cam, &param[p_pos],
		    camtype[cam->camtype].func,
		    &param[p_cam],
		    camdistortion[cam->camdistortion].func,
		    &param[p_dist], objpos, &campixx, &campixy);

      residual[i] = gridpix[j].pixx - campixx;
      residual[i + 1] = gridpix[j].pixy - campixy;

      FITDBG ("  wanted x %f y %f  resid x %f y %f",
	      gridpix[j].pixx, gridpix[j].pixy, residual[i], residual[i + 1]);
    }

  return 0;
}


// Find the best grid-corner pixels to use in first fit attempt,
// usually the four real outside corners if available.
//
// This is useful for cases where chessboard may not be complete.
// Method: calc all pixel-to-pixel distances for wanted camera position
// select the two points with largest distance, then add some more points
// with largest distance sum to already-selected.
//
// Clears all gridpix[i].farthestcorner that don't belong to wanted campos.
void
select_farthest_corners (int camposi)
{
  int goodgridnum = -1;
  int i, j, n;
  int largesti, largestj;
  double largestdist, distsum;

  for (i = 0; i < gridpix_used; i++)
    gridpix[i].farthestcorner = 0;

  // Calc all distances for this campos only; we won't use !=camposi.
  // We also won't use i==j.
  // At the same time, keep track of largest distance = farthest corners.
  largestdist = -1.;
  largesti = -1;
  largestj = -1;
  for (i = 0; i < gridpix_used; i++)
    {
      if (gridpix[i].camposnum != camposi)
	continue;

      // Check same gridnum while we're at it.
      if (goodgridnum < 0)
	goodgridnum = gridpix[i].gridnum;	// lock onto it
      else if (gridpix[i].gridnum != goodgridnum)
	{
	  STDERR ("FATAL: can't have same camera position %d "
		  "seeing multiple grids %d and %d",
		  camposi, goodgridnum, gridpix[i].gridnum);
	  exit (1);
	}

      for (j = 0; j < gridpix_used; j++)
	{
	  if (j == i || gridpix[j].camposnum != camposi)
	    continue;

	  double d = hypot (gridpix[i].pixx - gridpix[j].pixx,
			    gridpix[i].pixy - gridpix[j].pixy);

	  pix2pixdist[i * gridpix_used + j] = d;

	  if (d > largestdist)
	    {
	      largestdist = d;
	      largesti = i;
	      largestj = j;
	    }
	}
    }

  if (largesti < 0 || largestj < 0 || largestdist < 0.)
    {
      STDERR ("campos %d has less than two gridpix, useless", camposi);
      exit (1);
    }
  gridpix[largesti].farthestcorner = 1;
  gridpix[largestj].farthestcorner = 1;
  FITDBG
    ("for firstguess campos %d farthestcorner gridpix[%d] n %.1f m %.1f pixx %f y %f distance %f",
     camposi, largesti, gridpix[largesti].gridn, gridpix[largesti].gridm,
     gridpix[largesti].pixx, gridpix[largesti].pixy, largestdist);
  FITDBG
    ("for firstguess campos %d farthestcorner gridpix[%d] n %.1f m %.1f pixx %f y %f distance %f",
     camposi, largestj, gridpix[largestj].gridn, gridpix[largestj].gridm,
     gridpix[largestj].pixx, gridpix[largestj].pixy, largestdist);

  // Okay, we have selected the two farthest pixels.
  // Now add some more pixels with largest total distance. This reliably
  // selects the two other farthest corners, but no more useful points
  // after that.
  //
  // Could also add points by selecting greatest minimum distance:
  // https://se.mathworks.com/matlabcentral/answers/42622-how-can-i-choose-a-subset-of-k-points-the-farthest-apart
  // https://stackoverflow.com/questions/48925086/choosing-subset-of-farthest-points-in-given-set-of-points
  // (Note that this will *not* select the farthest corners in many cases.)
  // But four points is perfectly fine for firstguess, so just leave at that.
  for (n = 0; n < 2; n++)
    {
      largestdist = -1.;
      largesti = -1;

      // See if i can be added
      for (i = 0; i < gridpix_used; i++)
	{
	  if (gridpix[i].farthestcorner || gridpix[i].camposnum != camposi)
	    continue;

	  distsum = 0.;

	  for (j = 0; j < gridpix_used; j++)
	    if (gridpix[j].farthestcorner)
	      // given that [i] is for camposi (just checked above),
	      // and [j] is a farthestcorner which must also be camposi,
	      // pix2pixdist will always have been calculated above.
	      distsum += pix2pixdist[i * gridpix_used + j];

	  if (distsum > largestdist)
	    {
	      largestdist = distsum;
	      largesti = i;
	    }
	}

      if (largesti < 0 || largestdist < 0.)
	{
	  STDERR ("campos %d has less than %d gridpix, not handled", camposi,
		  2 + 1 + n);
	  exit (1);
	}
      gridpix[largesti].farthestcorner = 1;
      FITDBG
	("for firstguess campos %d farthestcorner gridpix[%d] n %.1f m %.1f pixx %f y %f distance %f",
	 camposi, largesti, gridpix[largesti].gridn, gridpix[largesti].gridm,
	 gridpix[largesti].pixx, gridpix[largesti].pixy, largestdist);
    }
}


// First easy guess for each camera position/rotation relative to its grid.
void
calc_campositions_firstguess (cam_t * cam, campos_t * campos,
			      double focallength)
{
  int camposi;
  int nmeas;
  int *gp_idx;
  double *resid;
  double param[10];		// only 7 params (x,y,z,a,b,g, focallength)
  double param_error[10];
  mp_par param_settings[10];
  mp_result result;
  mp_config mpconfig;

  // Allocate full set even if only a few will be used
  gp_idx = malloc (sizeof (int) * gridpix_used);
  resid = malloc (sizeof (double) * gridpix_used);

  if (pix2pixdist == NULL || gp_idx == NULL || resid == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  for (camposi = 0; camposi < campos_used; camposi++)
    {
      int i;
      int fitres;
      private_firstguess_t private;

      printf ("Camera position %d calculating first guess...\n", camposi);

      // Find which gridpix items are most useful for this campos,
      // usually the four outer points only.
      select_farthest_corners (camposi);

      nmeas = 0;
      for (i = 0; i < gridpix_used; i++)
	if (gridpix[i].farthestcorner)
	  {
	    gp_idx[nmeas] = i;	// even = x coord
	    nmeas++;
	    gp_idx[nmeas] = i;	// odd = y coord
	    nmeas++;

	    //thisgrid = gridpix[i].gridnum;
	  }

      memset (&param_settings, 0, sizeof (param_settings));

      campos[camposi].fitparamidx = 0;

      // Starting guesses x,y,z,a,b,g
      // Grid is pasted to wall at 0,0,0 and extending to x+ and z-;
      // put camera somewhere before it.
      param[0] = 0.10;
      param[1] = -1.0;
      param[2] = -0.10;
      param[3] = 0.001;
      param[4] = 0.001;
      param[5] = 0.001;

      // We want the negative-Y coord, stay facing the grid
      param_settings[1].limited[1] = 1;	// upper bound active
      param_settings[1].limits[1] = -0.001;	// upper bound value

      param_settings[0].step = 0.001;
      param_settings[0].advlimited = 1;
      param_settings[0].advlimit = 0.2;

      param_settings[1].step = 0.001;
      param_settings[1].advlimited = 1;
      param_settings[1].advlimit = 0.2;

      param_settings[2].step = 0.001;
      param_settings[2].advlimited = 1;
      param_settings[2].advlimit = 0.2;

      // wraps into -period/2 .. period/2
      param_settings[3].periodic = 1;
      param_settings[3].period = 2. * M_PI;
      param_settings[3].advlimited = 1;
      param_settings[3].advlimit = 2.;
      param_settings[3].step = 0.001;

      param_settings[4].periodic = 1;
      param_settings[4].period = 2. * M_PI;
      param_settings[4].advlimited = 1;
      param_settings[4].advlimit = 2.;
      param_settings[4].step = 0.001;

      param_settings[5].periodic = 1;
      param_settings[5].period = 2. * M_PI;
      param_settings[5].advlimited = 1;
      param_settings[5].advlimit = 2.;
      param_settings[5].step = 0.001;

      // Firstguess_fitfunc always uses camfunc_basic
      cam->type_fitparamidx = 6;

      // Only focal length, simple starting guess, and keep fixed for now
      param[6] = focallength;
      param_settings[6].fixed = 1;

      // Firstguess_fitfunc always uses distfunc_none
      cam->dist_fitparamidx = 0;

      memset (&mpconfig, 0, sizeof (mpconfig));
      mpconfig.maxiter = 2000;	// usually *good* result needs <200

      memset (&result, 0, sizeof (result));
      result.xerror = param_error;
      result.resid = resid;

      private.cam = cam;
      private.campos = campos;
      private.gridpoint_indexes = gp_idx;

      fitres =
	mpfit (gridpix_firstguess_fitfunc, nmeas, 7, param, param_settings,
	       &mpconfig, (void *) &private, &result);
      printf ("\n  mpfit status = %d\n", fitres);
      mp_printresult (param, &result);

#if 0
      // Below goes all wrong for fisheye lenses;
      // and above result is very useful already anyway.
      //
      // Also note that we only have 4 points here, useless to
      // try the real camera model: will be totally underdetermined.

      printf ("Camera position %d refining focal length...\n", camposi);

      // Second round unfixes focal length.
      param_settings[6].fixed = 0;

      memset (&result, 0, sizeof (result));
      result.xerror = param_error;
      result.resid = resid;

      memset (&mpconfig, 0, sizeof (mpconfig));
      mpconfig.maxiter = 2000;	// usually *good* result needs <200

      fitres = mpfit (gridpix_firstguess_fitfunc, nmeas, 7,
		      param, param_settings, &mpconfig, (void *) &private,
		      &result);
      printf ("\n  mpfit status = %d\n", fitres);
      mp_printresult (param, &result);
#endif

      campos[camposi].posrot[0] = param[0];
      campos[camposi].posrot[1] = param[1];
      campos[camposi].posrot[2] = param[2];
      campos[camposi].posrot[3] = param[3];
      campos[camposi].posrot[4] = param[4];
      campos[camposi].posrot[5] = param[5];

      // The focal length guess is for one position only,
      // these will be averaged to one starting guess for full-param calc.
      campos[camposi].focallength = param[6];

      printf
	("Camera at [%d] (%6.3f, %6.3f, %6.3f) rotation (%6.3f, %6.3f, %6.3f)\n",
	 camposi, campos[camposi].posrot[0], campos[camposi].posrot[1],
	 campos[camposi].posrot[2], campos[camposi].posrot[3],
	 campos[camposi].posrot[4], campos[camposi].posrot[5]);
      printf ("Guessed focal length %f\n", campos[camposi].focallength);
    }

  free (gp_idx);
  free (resid);
}


// Find all camera + distortion params, refine each camera position/rotation.
// Returns sum of squares, or -1 if fit failed.
double
calc_fullparams (cam_t * cam, campos_t * campos)
{
  int nmeas;
  int nparam;
  double *resid;
  double *param;
  double *param_error;
  mp_par *param_settings;
  mp_result result;
  mp_config mpconfig;
  int i;
  int fitres;
  private_fitfunc_t private;

  // We use all gridpoint observations, pixx/pixy separately.
  nmeas = 2 * gridpix_used;

  nparam = 0;

  cam->type_fitparamidx = nparam;
  nparam += camtype[cam->camtype].nparam;

  cam->dist_fitparamidx = nparam;
  nparam += camdistortion[cam->camdistortion].nparam;

  for (i = 0; i < campos_used; i++)
    {
      campos[i].fitparamidx = nparam;
      nparam += 6;
    }

  // Now we can allocate the arrays
  resid = calloc (nmeas, sizeof (double));
  param = calloc (nparam, sizeof (double));
  param_error = calloc (nparam, sizeof (double));
  param_settings = calloc (nparam, sizeof (mp_par));

  if (resid == NULL || param == NULL || param_error == NULL
      || param_settings == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  // Initialize starting guesses for type params
  {
    int p = cam->type_fitparamidx;
    int n = camtype[cam->camtype].nparam;

    camtype[cam->camtype].initfunc (param + p, param_settings + p);

    // Copy better starting guesses if non-zero provided in input,
    // and also fixate for now.
    // (At least focaldistance.)
    for (i = 0; i < n; i++)
      {
	if (param_nonzero (cam->typeparam[i]))
	  {
	    param[p + i] = cam->typeparam[i];
	    param_settings[p + i].fixed = 1;
	  }
      }

    for (i = 0; i < n; i++)
      printf ("Camera type starting guess param[%d] = %f\n", i, param[p + i]);
  }

  // Initialize starting guesses for distortion params
  {
    int p = cam->dist_fitparamidx;
    int n = camdistortion[cam->camdistortion].nparam;

    camdistortion[cam->camdistortion].initfunc (param + p,
						param_settings + p);

    // Copy better starting guesses if non-zero provided in input,
    // and also fixate for now.
    for (i = 0; i < n; i++)
      {
	if (param_nonzero (cam->distortionparam[i]))
	  {
	    param[p + i] = cam->distortionparam[i];
	    param_settings[p + i].fixed = 1;
	  }
      }

    for (i = 0; i < n; i++)
      printf ("Camera distortion starting guess param[%d] = %f\n",
	      i, param[p + i]);
  }

  // Initialize starting guesses for all camera positions from earlier results
  for (i = 0; i < campos_used; i++)
    {
      int p = campos[i].fitparamidx;

      // Copy from first guesses
      param[p] = campos[i].posrot[0];
      param[p + 1] = campos[i].posrot[1];
      param[p + 2] = campos[i].posrot[2];
      param[p + 3] = campos[i].posrot[3];
      param[p + 4] = campos[i].posrot[4];
      param[p + 5] = campos[i].posrot[5];

      // negative-Y coord
      param_settings[p + 1].limited[1] = 1;	// upper bound active
      param_settings[p + 1].limits[1] = -0.001;	// upper bound value

      param_settings[p].step = 0.001;
      param_settings[p].advlimited = 1;
      param_settings[p].advlimit = 0.2;

      param_settings[p + 1].step = 0.001;
      param_settings[p + 1].advlimited = 1;
      param_settings[p + 1].advlimit = 0.2;

      param_settings[p + 2].step = 0.001;
      param_settings[p + 2].advlimited = 1;
      param_settings[p + 2].advlimit = 0.2;

      // wraps into -period/2 .. period/2
      param_settings[p + 3].periodic = 1;
      param_settings[p + 3].period = 2. * M_PI;
      param_settings[p + 3].advlimited = 1;
      param_settings[p + 3].advlimit = 2.;
      param_settings[p + 3].step = 0.001;

      param_settings[p + 4].periodic = 1;
      param_settings[p + 4].period = 2. * M_PI;
      param_settings[p + 4].advlimited = 1;
      param_settings[p + 4].advlimit = 2.;
      param_settings[p + 4].step = 0.001;

      param_settings[p + 5].periodic = 1;
      param_settings[p + 5].period = 2. * M_PI;
      param_settings[p + 5].advlimited = 1;
      param_settings[p + 5].advlimit = 2.;
      param_settings[p + 5].step = 0.001;
    }

  memset (&mpconfig, 0, sizeof (mpconfig));
  mpconfig.maxiter = 2000;	// usually *good* result needs <200

  memset (&result, 0, sizeof (result));
  result.xerror = param_error;
  result.resid = resid;

  private.cam = cam;
  private.campos = campos;
  private.funcevals = 0;

  printf ("Calculating with fixed parameters\n"
	  "  max %d iterations, approx %d func evals...\n",
	  mpconfig.maxiter, mpconfig.maxiter * (nparam + 1));

  fitres =
    mpfit (gridpix_fitfunc, nmeas, nparam, param, param_settings, &mpconfig,
	   (void *) &private, &result);
  printf ("\n  mpfit status = %d\n", fitres);
  mp_printresult (param, &result);

  if (fitres >= 1 && fitres <= 4)
    {
      for (i = 0; i < nparam; i++)
	param_settings[i].fixed = 0;

      printf ("Calculating further with unfixed parameters\n"
	      "  max %d iterations, approx %d func evals...\n",
	      mpconfig.maxiter, mpconfig.maxiter * (nparam + 1));

      fitres =
	mpfit (gridpix_fitfunc, nmeas, nparam, param, param_settings,
	       &mpconfig, (void *) &private, &result);
      printf ("\n  mpfit status = %d\n", fitres);
      mp_printresult (param, &result);
    }

  // Copy result back to arrays
  {
    int p, n, i;

    p = cam->type_fitparamidx;
    n = camtype[cam->camtype].nparam;
    for (i = 0; i < n; i++)
      {
	cam->typeparam[i] = param[p + i];
      }

    p = cam->dist_fitparamidx;
    n = camdistortion[cam->camdistortion].nparam;
    for (i = 0; i < n; i++)
      {
	cam->distortionparam[i] = param[p + i];
      }

    for (i = 0; i < campos_used; i++)
      {
	p = campos[i].fitparamidx;

	campos[i].posrot[0] = param[p];
	campos[i].posrot[1] = param[p + 1];
	campos[i].posrot[2] = param[p + 2];
	campos[i].posrot[3] = param[p + 3];
	campos[i].posrot[4] = param[p + 4];
	campos[i].posrot[5] = param[p + 5];
      }
  }

  if (fitres < 1 || fitres > 4)
    {
      printf ("Fit failed\n");
      result.bestnorm = -1;
    }

  free (resid);
  free (param);
  free (param_error);
  free (param_settings);

  return result.bestnorm;
}

void
print_resultjson (cam_t * cam, campos_t * campos)
{
  int n, i;

  printf
    ("\n"
     "Camera positions are calculated as if grids were hanging on wall at y=0\n"
     "  (x+/z- plane) and camera on tripod before it (negative y position).\n"
     "  Rotation is like tripod: pan (compass), tilt (up/down), oblique (landscape/\n"
     "  portrait), given in radians\n");

  for (i = 0; i < campos_used; i++)
    {
      printf
	("Camera at [%d] (%6.3f, %6.3f, %6.3f) rotation (%6.3f, %6.3f, %6.3f)\n",
	 i, campos[i].posrot[0], campos[i].posrot[1], campos[i].posrot[2],
	 campos[i].posrot[3], campos[i].posrot[4], campos[i].posrot[5]);
    }

  printf ("\nResult seems to be acceptable.\n"
	  "Please verify that camera positions above are approximately correct.\n\n"
	  "In camera configuration, use:\n\n");

  printf ("\"pixw\": %d,\n\"pixh\": %d,\n", cam->w, cam->h);


  // Print compact form that jansson can't do easily
  printf ("\"type\": {\n  \"name\": \"%s\",\n  \"param\": [",
	  camtype[cam->camtype].name);

  n = camtype[cam->camtype].nparam;
  for (i = 0; i < n; i++)
    printf ("%f%s", cam->typeparam[i], i < (n - 1) ? ", " : "");
  printf ("]\n},\n");


  // Print compact form that jansson can't do easily
  printf ("\"distortion\": {\n  \"name\": \"%s\",\n  \"param\": [",
	  camdistortion[cam->camdistortion].name);

  n = camdistortion[cam->camdistortion].nparam;
  for (i = 0; i < n; i++)
    printf ("%f%s", cam->distortionparam[i], i < (n - 1) ? ", " : "");
  printf ("]\n},\n\n");
}


inline void
parse_point (json_t * jpoint, cam_t * cam, int curgrid, int curcampos,
	     int fix_righttoleft, int jsonpointidx)
{
  int curgridpix;
  json_t *jrealh, *jrealv, *jpixx, *jpixy;
  double realh, realv, pixx, pixy;

  if (!json_is_object (jpoint))
    return;			// Probably comment, etc.

  curgridpix = new_gridpix ();
  gridpix[curgridpix].gridnum = curgrid;
  gridpix[curgridpix].camposnum = curcampos;

  jrealh = json_object_get (jpoint, "realh");
  jrealv = json_object_get (jpoint, "realv");
  jpixx = json_object_get (jpoint, "pixx");
  jpixy = json_object_get (jpoint, "pixy");

  if (!json_is_number (jrealh) || !json_is_number (jrealv) ||
      !json_is_number (jpixx) || !json_is_number (jpixy))
    {
      STDERR
	("Grid %d picture %d point (arrayitem) %d not specified completely",
	 curgrid, curcampos, jsonpointidx);
      exit (1);
    }

  realh = json_number_value (jrealh);
  realv = json_number_value (jrealv);
  pixx = json_number_value (jpixx);
  pixy = json_number_value (jpixy);

  if (pixx < 0. || pixx >= (double) cam->w ||
      pixy < 0. || pixy >= (double) cam->h)
    {
      STDERR
	("Grid %d picture %d point (arrayitem) %d pixel (%f, %f) out of camera image (%d x %d)",
	 curgrid, curcampos, jsonpointidx, pixx, pixy, cam->w, cam->h);
      exit (1);
    }

  if (fix_righttoleft)
    gridpix[curgridpix].gridn = -realh;
  else
    gridpix[curgridpix].gridn = realh;

  gridpix[curgridpix].gridm = realv;

  gridpix[curgridpix].pixx = (pixx - cam->xoffset) * cam->xscale;

  gridpix[curgridpix].pixy = (pixy - cam->yoffset) * cam->yscale;
}

void
parse_picture (json_t * jpic, cam_t * cam, int curgrid)
{
  int curcampos;
  int fix_righttoleft;
  json_t *jpoints;
  int i;

  jpoints = json_object_get (jpic, "points");

  if (!json_is_array (jpoints))
    return;			// Probably comment, etc.

  // New picture means new camera position
  curcampos = campos_used;
  campos_used++;

  fix_righttoleft =
    json_boolean_value (json_object_get (jpic, "fix-righttoleft"));
  // Note: upside-down pattern does NOT need any fix, works fine,
  // will calculate proper camera rotation.

  for (i = 0; i < json_array_size (jpoints); i++)
    parse_point (json_array_get (jpoints, i), cam, curgrid, curcampos,
		 fix_righttoleft, i);
}

void
parse_grid (json_t * jgrid, cam_t * cam)
{
  int curgrid;
  double ix, iy, dx, dy;

  if (!json_is_object (jgrid))
    return;			// Probably comment, etc.

  curgrid = new_grid ();

  ix =
    json_number_value (json_object_get
		       (json_object_get
			(json_object_get (jgrid, "scale"),
			 "horizontal"), "meters"));
  iy =
    json_number_value (json_object_get
		       (json_object_get
			(json_object_get (jgrid, "scale"),
			 "vertical"), "meters"));
  dx =
    json_number_value (json_object_get
		       (json_object_get
			(json_object_get (jgrid, "scale"),
			 "horizontal"), "divideby"));
  dy =
    json_number_value (json_object_get
		       (json_object_get
			(json_object_get (jgrid, "scale"),
			 "vertical"), "divideby"));

  if (ix < 0.001 || iy < 0.001 || dx < 0.001 || dy < 0.001)
    {
      STDERR ("Grid scale not fully specified; use 1.0 for unity scale");
      exit (1);
    }

  grid[curgrid].ix = ix / dx;
  grid[curgrid].iy = iy / dy;

  {
    json_t *jpics = json_object_get (jgrid, "pictures");
    int i;

    for (i = 0; i < json_array_size (jpics); i++)
      parse_picture (json_array_get (jpics, i), cam, curgrid);
  }
}

void
parse_stdin (cam_t * cam)
{
  json_t *jroot;
  json_error_t jerror;

  jroot = json_loadfd (0, 0, &jerror);
  if (jroot == NULL)
    {
      STDERR ("input error on line %d: %s", jerror.line, jerror.text);
      exit (1);
    }

  // Note that jansson is perfectly okay handling NULL pointers

  parse_cam_json (jroot, cam, 0);
  cam->name = "<noname>";	// always only one cam, fake name

  {
    json_t *jgrids = json_object_get (jroot, "grids");
    int i;

    for (i = 0; i < json_array_size (jgrids); i++)
      parse_grid (json_array_get (jgrids, i), cam);
  }

#if 0
  {
    int i;
    for (i = 0; i < gridpix_used; i++)
      printf ("  gridpix[%d]=grid %d campos %d n %f m %f x %f y %f\n",
	      i, gridpix[i].gridnum, gridpix[i].camposnum,
	      gridpix[i].gridn, gridpix[i].gridm,
	      gridpix[i].pixx, gridpix[i].pixy);
  }
#endif

  if (gridpix_used < 4)
    {
      STDERR ("Too few points specified, cannot calculate");
      exit (1);
    }
}

void
find_cambasic_distnone ()
{
  int i, found;

  found = 0;
  for (i = 0; camtype[i].name != NULL; i++)
    if (!strcmp (camtype[i].name, "basic"))
      {
	camfunc_basic = camtype[i].func;
	camfunc_basic_idx = i;
	found = 1;
	break;
      }
  if (!found)
    {
      STDERR ("Cameratype basic is not available");
      exit (1);
    }
  if (camtype[i].nparam != 1)
    {
      STDERR ("Cameratype basic does not have 1 parameter, please fix code");
      exit (1);
    }

  found = 0;
  for (i = 0; camdistortion[i].name != NULL; i++)
    if (!strcmp (camdistortion[i].name, "none"))
      {
	distfunc_none = camdistortion[i].func;
	distfunc_none_idx = i;
	found = 1;
	break;
      }
  if (!found)
    {
      STDERR ("Cameradistortion none is not available");
      exit (1);
    }
  if (camdistortion[i].nparam != 0)
    {
      STDERR
	("Cameradistortion none has non-zero number of parameters, please fix code");
      exit (1);
    }
}


int
main (void)
{
  // We have several "working copies" of cam info and camera positions
  cam_t cam_config;
  cam_t cam_typetry;
  cam_t cam_typebest;
  cam_t cam_disttry;
  cam_t cam_distbest;
  campos_t *campos_firstguess;	// mallocd
  campos_t *campos_typetry;
  campos_t *campos_typebest;
  campos_t *campos_disttry;
  campos_t *campos_distbest;
  double leastsquares_best;
  double focallength;
  int trycamtype;
  int trydistortion;

  find_cambasic_distnone ();

  cam_config.camtype = -1;
  cam_config.camdistortion = -1;

  parse_stdin (&cam_config);

  pix2pixdist = malloc (sizeof (double) * gridpix_used * gridpix_used);
  campos_firstguess = calloc (campos_used, sizeof (campos_t));
  campos_typetry = calloc (campos_used, sizeof (campos_t));
  campos_typebest = calloc (campos_used, sizeof (campos_t));
  campos_disttry = calloc (campos_used, sizeof (campos_t));
  campos_distbest = calloc (campos_used, sizeof (campos_t));
  if (pix2pixdist == NULL || campos_firstguess == NULL
      || campos_typetry == NULL || campos_typebest == NULL
      || campos_disttry == NULL || campos_distbest == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  focallength = 1.;

  // If type is specified, use focal length
  if (cam_config.camtype >= 0)
    {
      if (param_nonzero (cam_config.typeparam[0]))
	focallength = cam_config.typeparam[0];
    }

  printf ("Calculating first guesses for camera positions...\n");

  calc_campositions_firstguess (&cam_config, campos_firstguess, focallength);

  // Take average focallength from all position-firstguesses
  // TODO: check if median is better
  {
    int i;
    focallength = 0;
    for (i = 0; i < campos_used; i++)
      focallength += campos_firstguess[i].focallength;
    focallength = focallength / (double) campos_used;
  }


  // Attempt all camera types without distortion, remember best fitting.

  cam_typebest.typeparam = NULL;
  leastsquares_best = 1000. * FIT_ACCEPTSUMSQ;
  for (trycamtype = 0; camtype[trycamtype].name != NULL; trycamtype++)
    {
      double thisleastsquares;

      // If type is specified, then attempt only that one
      if (cam_config.camtype >= 0 && trycamtype != cam_config.camtype)
	continue;

      cam_typetry = cam_config;
      // also copies *pointers* to typeparam, distortionparam

      cam_typetry.camtype = trycamtype;
      cam_typetry.camdistortion = distfunc_none_idx;

      cam_typetry.typeparam =
	calloc (camtype[cam_typetry.camtype].nparam, sizeof (double));
      cam_typetry.typeparam[0] = focallength;

      memcpy (campos_typetry, campos_firstguess,
	      campos_used * sizeof (campos_t));

      // If type is specified, copy nonzero params
      if (cam_config.camtype >= 0
	  && cam_typetry.camtype == cam_config.camtype)
	{
	  int i;
	  for (i = 0; i < camtype[cam_config.camtype].nparam; i++)
	    {
	      if (param_nonzero (cam_config.typeparam[i]))
		cam_typetry.typeparam[i] = cam_config.typeparam[i];
	    }
	}

      printf ("Attempting camera type %s without distortion...\n",
	      camtype[cam_typetry.camtype].name);

      thisleastsquares = calc_fullparams (&cam_typetry, campos_typetry);

      if (thisleastsquares > 0
	  && thisleastsquares < 10. * FIT_ACCEPTSUMSQ
	  && thisleastsquares < leastsquares_best)
	{
	  leastsquares_best = thisleastsquares;

	  if (cam_typebest.typeparam != NULL)
	    free (cam_typebest.typeparam);

	  cam_typebest = cam_typetry;
	  // also copies *pointers* to typeparam, distortionparam

	  memcpy (campos_typebest, campos_typetry,
		  campos_used * sizeof (campos_t));

	  printf
	    ("Camera type %s has best result until now.\n",
	     camtype[cam_typebest.camtype].name);
	}
      else
	{
	  free (cam_typetry.typeparam);
	}
    }

  if (leastsquares_best >= 10. * FIT_ACCEPTSUMSQ)
    {
      printf ("Sorry, no camera type seems to match properly.");
      exit (1);
    }


  // For best type, attempt all distortions, remember best fitting.

  cam_distbest.typeparam = NULL;
  cam_distbest.distortionparam = NULL;
  leastsquares_best = 1000. * FIT_ACCEPTSUMSQ;
  for (trydistortion = 0; camdistortion[trydistortion].name != NULL;
       trydistortion++)
    {
      double thisleastsquares;

      // If distortion is specified, then attempt only that one
      if (cam_config.camdistortion >= 0
	  && trydistortion != cam_config.camdistortion)
	continue;

      cam_disttry = cam_typebest;
      // also copies *pointers* to typeparam, distortionparam

      cam_disttry.camdistortion = trydistortion;

      cam_disttry.typeparam = calloc (camtype[cam_disttry.camtype].nparam,
				      sizeof (double));
      memcpy (cam_disttry.typeparam, cam_typebest.typeparam,
	      camtype[cam_disttry.camtype].nparam * sizeof (double));

      cam_disttry.distortionparam =
	calloc (camdistortion[cam_disttry.camdistortion].nparam,
		sizeof (double));

      memcpy (campos_disttry, campos_typebest,
	      campos_used * sizeof (campos_t));

      // If distortion is specified, copy nonzero params
      if (cam_config.camdistortion >= 0
	  && cam_disttry.camdistortion == cam_config.camdistortion)
	{
	  int i;
	  for (i = 0; i < camdistortion[cam_config.camdistortion].nparam; i++)
	    {
	      if (param_nonzero (cam_config.distortionparam[i]))
		cam_disttry.distortionparam[i] =
		  cam_config.distortionparam[i];
	    }
	}

      printf ("Attempting camera type %s with distortion %s...\n",
	      camtype[cam_disttry.camtype].name,
	      camdistortion[cam_disttry.camdistortion].name);

      thisleastsquares = calc_fullparams (&cam_disttry, campos_disttry);

      if (thisleastsquares > 0
	  && thisleastsquares < FIT_ACCEPTSUMSQ
	  && thisleastsquares < leastsquares_best)
	{
	  leastsquares_best = thisleastsquares;

	  if (cam_distbest.typeparam != NULL)
	    free (cam_distbest.typeparam);
	  if (cam_distbest.distortionparam != NULL)
	    free (cam_distbest.distortionparam);

	  cam_distbest = cam_disttry;
	  // also copies *pointers* to typeparam, distortionparam

	  memcpy (campos_distbest, campos_disttry,
		  campos_used * sizeof (campos_t));

	  printf
	    ("Camera type %s with distortion %s has best result until now.\n",
	     camtype[cam_distbest.camtype].name,
	     camdistortion[cam_distbest.camdistortion].name);
	}
      else
	{
	  free (cam_disttry.typeparam);
	  free (cam_disttry.distortionparam);
	}
    }

  if (leastsquares_best >= FIT_ACCEPTSUMSQ)
    {
      printf ("Sorry, no distortion seems to match properly.");
      exit (1);
    }

  print_resultjson (&cam_distbest, campos_distbest);

  free (pix2pixdist);
  return 0;
}
