// Inspired by example at:
// https://github.com/sturkmen72/opencv_samples/blob/master/findChessboardCorners.cpp
// Also see:
// https://stackoverflow.com/questions/17665912/findchessboardcorners-fails-for-calibration-image

// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#define PROGNAME  "opencv-findcorners"

#include <opencv2/opencv.hpp>
#include <stdio.h>

using namespace cv;
using namespace std;

void
print_header (char *fname, int w, int h)
{
  printf ("      {\n"
	  "        \"#\": \"%s %d %d %s\",\n"
	  "        \"points\":\n        [\n", PROGNAME, w, h, fname);
}

void
print_footer ()
{
  printf ("          \"# (additional item to allow comma above)\"\n"
	  "        ]\n      },\n");
}

int
main (int argc, char **argv)
{
  if (argc != 4)
    {
      fprintf (stderr,
	       "\n"
	       "USAGE: %s <width> <height> <imagefile> > <output.txt>\n"
	       "\n"
	       "Where <width> and <height> are number of _inner_ corners of chessboard.\n"
	       "\n"
	       "Output format is suitable for camcalibplanes *IF* correct number\n"
	       "of corners are found; else numbering will be totally wrong and needs to be\n"
	       "adjusted manually.\n\n",
		PROGNAME);
      return 1;
    }


  int w = atoi (argv[1]);
  int h = atoi (argv[2]);
  Mat img = imread (argv[3]);

  Size patternsize (w, h);	// interior number of corners
  vector < Point2f > corners;	// will be filled by the detected corners

  //imshow("image",img);
  //moveWindow("image",40,40);

  Mat gray;
  cvtColor (img, gray, COLOR_BGR2GRAY);	// greyscale of source image

  bool patternfound = findChessboardCorners (gray, patternsize, corners,
					     0 + CALIB_CB_ADAPTIVE_THRESH
					     //+ CALIB_CB_NORMALIZE_IMAGE
					     //+ CALIB_CB_FAST_CHECK
    );


  // Draw before the subpixel refinement
  if (0)
    {
      Mat dispimg2 = img.clone ();
      drawChessboardCorners (dispimg2, patternsize, Mat (corners),
			     patternfound);
      const char *winname = "result before subpixel";
      namedWindow (winname, WINDOW_NORMAL);
      resizeWindow (winname, dispimg2.cols, dispimg2.rows);
      imshow (winname, dispimg2);
    }


  // Subpixel refinement, each corner separately.
  // Detect suitable refinement area based on closest-neighbor distance.
  vector < Point2f > corners_subp;
  for (int i = 0; i < corners.size (); i++)
    {
      double myx = corners[i].x;
      double myy = corners[i].y;
      double mindistance = 50.;

      // Check all other corners and find the closest
      for (int j = 0; j < corners.size (); j++)
	if (j != i)
	  {
	    double otherx = corners[j].x;
	    double othery = corners[j].y;
	    double dist = hypot (myx - otherx, myy - othery);
	    if (dist < mindistance)
	      mindistance = dist;
	  }

      mindistance /= 4.;
      // cornerSubPix will check mindistance to all sides,
      // i.e. a 2*mindistance+1 horizontal and vertical square.

      vector < Point2f > onecorner;
      onecorner.push_back (corners[i]);

      if (mindistance >= 3)
	cornerSubPix (gray, onecorner, Size (mindistance, mindistance),
		      Size (-1, -1),
		      TermCriteria (CV_TERMCRIT_EPS + CV_TERMCRIT_ITER,
				    30, 0.1));

      corners_subp.push_back (onecorner[0]);

      if (0)
	fprintf (stderr,
		 "# i %d myx %.2f myy %.2f mindist %.2f newx %.2f newy %.2f\n",
		 i, myx, myy, mindistance,
		 (double) onecorner[0].x, (double) onecorner[0].y);
    }


  // Output list in correct format
  if (corners_subp.size () == w * h)
    {
      print_header (argv[3], w, h);

      int x = 0, y = 0;
      for (int i = 0; i < corners_subp.size (); i++)
	{
	  printf
	    ("          {\"realh\": %2d, \"realv\": %2d, \"pixx\": %10.5f, \"pixy\": %10.5f},\n",
	     x, y, (double) corners_subp[i].x, (double) corners_subp[i].y);

	  x++;
	  if (x >= w)
	    {
	      x -= w;
	      y++;
	    }
	}

      print_footer ();
    }
  else
    {
      print_header (argv[3], w, h);

      printf ("\n"
	      "          \"# SORRY! Found %d corners instead of %d\",\n"
	      "          \"# Cannot properly order them, please order manually.\",\n"
	      "          \"# Note that camcalibplanes will work fine with only a subset of grid points.\",\n"
	      "\n", corners_subp.size (), w * h);

      for (int i = 0; i < corners_subp.size (); i++)
	printf
	  ("          {\"realh\":  , \"realv\":  , \"pixx\": %10.5f, \"pixy\": %10.5f},\n",
	   (double) corners_subp[i].x, (double) corners_subp[i].y);

      print_footer ();
    }


  // Draw result
  Mat dispimg = img.clone ();
  drawChessboardCorners (dispimg, patternsize, Mat (corners_subp),
			 patternfound);
  const char *winname = "result after subpixel";
  namedWindow (winname, WINDOW_NORMAL);
  resizeWindow (winname, dispimg.cols, dispimg.rows);
  imshow (winname, dispimg);

  // Make sure stdout is printed before stderr below
  fflush(NULL);

  fprintf (stderr,
	   "# Enlarge window to take a closer look.\n"
	   "# Close window to exit (or press escape or 'q' in window).\n");
  while (getWindowProperty(winname, WND_PROP_AUTOSIZE) >= -0.1)
    {
      int k = waitKey (100);
      //fprintf (stderr, "# key %d\n", k);
      // Shift adds 65536
      if (k == 'q' || k == 65536 + 'Q' || k == 27)
	break;
    }

  destroyAllWindows ();
  return 0;
}
