// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#ifndef CAMHELPERS_H
#define CAMHELPERS_H

#include "camhelpers.h"
#include <math.h>


// These are inline functions for speed; must be static so that they
// can be included "anonymously" multiple times in combined .o files


// theta is the angle respective to optical axis = positive z-axis
// phi is rotation in image plane from positive x-axis
static inline void
obj2angles (double objx, double objy, double objz, double *theta, double *phi)
{
  double r;

  r = sqrt (objx * objx + objy * objy + objz * objz);	// always positive and >=objz
  if (r != 0.)
    {
      *theta = acos (objz / r);
      // positive objz: objz/r is always >=0 and <=1
      // negative objz: objz/r is always <=0 and >=-1
      // acos() is well-defined acos(1)=0 -- acos(0)=Pi/2 -- acos(-1)=Pi
      // i.e. totally neat half-circle.

      *phi = atan2 (objy, objx);
      // atan2() has proper wrapping so that sin/cos will work,
      // incl. y/0 and 0/0.
    }
  else
    {
      *theta = 0.;
      *phi = 0.;
    }
}

static inline void
angles2img (double r, double phi, double *camx, double *camy)
{
  *camx = r * cos (phi);
  *camy = r * sin (phi);
}

static inline void
rotate_3d_officialUNUSED (double origx, double origy, double origz,
			  double alpha, double beta, double gamma,
			  double *newx, double *newy, double *newz)
{
  // http://planning.cs.uiuc.edu/node102.html

  *newx = origx * cos (alpha) * cos (beta)
    + origy * (cos (alpha) * sin (beta) * sin (gamma) -
	       sin (alpha) * cos (gamma)) +
    origz * (cos (alpha) * sin (beta) * cos (gamma) +
	     sin (alpha) * sin (gamma));
  *newy =
    origx * sin (alpha) * cos (beta) +
    origy * (sin (alpha) * sin (beta) * sin (gamma) +
	     cos (alpha) * cos (gamma)) +
    origz * (sin (alpha) * sin (beta) * cos (gamma) -
	     cos (alpha) * sin (gamma));
  *newz =
    -origx * sin (beta) + origy * cos (beta) * sin (gamma) +
    origz * cos (beta) * cos (gamma);
}

static inline void
rotate_3d (double origx, double origy, double origz, double alpha,
	   double beta, double gamma, double *newx, double *newy,
	   double *newz)
{
  // Tripod variant:
  // Start horizontal-north
  // First alpha = north - west - east - south
  // Then beta = up - flat - down
  // Finally gamma = twist landscape - portrait
  // See doc/formulas/rotation-3d/

  // Direct cut-paste from Maple sheet:
  *newx =
    (cos (gamma) * cos (alpha) +
     sin (gamma) * sin (beta) * sin (alpha)) * origx +
    (-cos (gamma) * sin (alpha) +
     sin (gamma) * sin (beta) * cos (alpha)) * origy -
    sin (gamma) * cos (beta) * origz;

  *newy =
    (-sin (gamma) * cos (alpha) +
     cos (gamma) * sin (beta) * sin (alpha)) * origx +
    (sin (gamma) * sin (alpha) +
     cos (gamma) * sin (beta) * cos (alpha)) * origy -
    cos (gamma) * cos (beta) * origz;

  *newz =
    cos (beta) * sin (alpha) * origx + cos (beta) * cos (alpha) * origy +
    sin (beta) * origz;
}


// Special tan with linearization <=-Pi/2 and >=Pi/2
static inline double
tanlin (double x)
{
  static int prepd = 0;
  static double minbound, maxbound, minadd, maxadd;
  static double fact = 10.;

  if (!prepd)
    {
      minbound = -(M_PI / 2. - 0.1);
      maxbound = (M_PI / 2. - 0.1);
      minadd = tan (minbound) - fact * minbound;
      maxadd = tan (maxbound) - fact * maxbound;
      prepd = 1;
    }

  if (x < minbound)
    return minadd + fact * x;
  if (x > maxbound)
    return maxadd + fact * x;
  return tan (x);
}


// Special sin with linearization <=-Pi/2 and >=Pi/2
static inline double
sinlin (double x)
{
  static int prepd = 0;
  static double minbound, maxbound, minadd, maxadd;
  // fact = 1 in this case

  if (!prepd)
    {
      minbound = -(M_PI / 2.);
      maxbound = (M_PI / 2.);
      minadd = sin (minbound) - minbound;	// -1 - -Pi/2
      maxadd = sin (maxbound) - maxbound;	// 1 - Pi/2
      prepd = 1;
    }

  if (x < minbound)
    return minadd + x;
  if (x > maxbound)
    return maxadd + x;
  return sin (x);
}


#endif /* CAMHELPERS_H */
