// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "camfuncs.h"
#include <math.h>
#include "camhelpers.h"

// DECL: camtype_func_t camtype_fisheye;
// DECL: camtype_initfunc_t camtype_fisheye_init;
// REG_CAMTYPE: {"fisheye", 4, camtype_fisheye, camtype_fisheye_init}

void
camtype_fisheye (double objx, double objy, double objz, double *camx,
		 double *camy, double *params)
{
  // This is generic fisheye, where objects get somewhat
  // _compressed_ in the direction of the border as they go near it
  // (i.e. "border pushes objects back").
  //
  // Specifically: orthographical f*sin(th) via equisolidangle f*2*sin(th/2)
  // up to eqidistant f*th which is f*big*sin(th/big).
  // Together that becomes f*b*sin(th/b) where f and b are fit params.

  // This is nonlinear in objx and objy, so we necessarily have to go
  // via angles. So no possibility of different focus for x and y here,
  // but can still do any stretching via distortion.
  // Also, this means that center shift becomes important at this point:
  // we still consider z=0 the true optical axis, and shift the sensor
  // as last step.

  // NOTE: param[0] is base focus, fixed to 1. during calib.
  double f = params[0];
  double b = params[1];
  //double b_inv = param[..]
  // Separate 1/b would save float div, but can't do that while fitting b
  // since that needs to move together with 1/b.
  // Note that we have lots of trig funcs anyway, so probably no big win.
  double shiftx = params[2];
  double shifty = params[3];
  double theta, phi, r;

  obj2angles (objx, objy, objz, &theta, &phi);
  r = f * b * sinlin (theta / b);
  angles2img (r, phi, camx, camy);
  *camx += shiftx;
  *camy += shifty;
}

void
camtype_fisheye_init (double *param, mp_par * param_settings)
{
  // NOTE: param[0] is base focus, fixed to 1. during calib.
  param[0] = 1.;
  param[1] = 1.5;
  param[2] = 0.001;
  param[3] = 0.001;

  // We want positive focusdist (and not zero)
  param_settings[0].limited[0] = 1;	// lower bound active
  param_settings[0].limits[0] = 0.001;	// lower bound value
  param_settings[0].step = 0.001;
  param_settings[0].advlimited = 1;	// limit advance per iteration
  param_settings[0].advlimit = 2.;	// advance limit value

  // b negative works but just the same as positive;
  // b=0..1 works but gives weird unphysical behaviour
  param_settings[1].limited[0] = 1;	// lower bound active
  param_settings[1].limits[0] = 0.001;	// lower bound value
  param_settings[1].step = 0.001;
  param_settings[1].advlimited = 1;	// limit advance per iteration
  param_settings[1].advlimit = 0.2;	// advance limit value

  // centershift shouldn't be too much
  param_settings[2].step = 0.001;
  param_settings[2].advlimited = 1;	// limit advance per iteration
  param_settings[2].advlimit = 0.1;	// advance limit value
  param_settings[3].step = 0.001;
  param_settings[3].advlimited = 1;	// limit advance per iteration
  param_settings[3].advlimit = 0.1;	// advance limit value

  // Note: don't forget adjusting #params in REG_CAMTYPE above
}
