// USAGE: campixtest < stdin
// Simple test tool: given camera type, position and orientation,
// and given object position, print camera pixel that object will appear on.
// (Just stripped version of camcalibplanes)

// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#define PROGNAME  "campixtest"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <jansson.h>

#include "../shared/error.h"
#include "../camfuncs/camfuncs.h"
#include "../camfuncs/camhelpers.h"

//#define FITDBG(args...)  STDERR(args)
#define FITDBG(args...)		/* nothing */


const char *progname = PROGNAME;


typedef struct
{
  char *name;
  int camtype;			// index to camtype[]
  double *typeparam;		// malloc'd for specified type
  int camdistortion;		// index to camdistortion[]
  double *distortionparam;	// malloc'd for specified distortion

  int w;			// camera #pixels
  int h;
  double xscale;		// calculate pixel to -1..+1 coords
  double xoffset;
  double yscale;
  double yoffset;
} cam_t;

// parse_cam_json requires cam_t to be defined
#define CAMERR_RETURN(args...)  do { STDERR(args); exit(1); } while(0)
#include "../shared/parse_cam_json.h"
#undef CAMERR_RETURN

// We only have one cam
cam_t cam[2];
int camnum = 0;


// Special version: Camtype+params/dist+params specified separately.
// Distortion not out-of-range-linearized, just as-is.
// 
// double camposrot[6] = camera position+rotation x/y/z/a/b/g
// double objpos[3] = object position x/y/z
static inline void
calc_campix2 (cam_t * cam_p, double *camposrot,
	      camtype_func_t * camfunc,
	      double *camtypeparam,
	      distortion_func_t * distfunc,
	      double *camdistparam,
	      double *objpos, double *campixx, double *campixy)
{
  double rotx, roty, rotz;
  double officialx, officialy;

  FITDBG ("  position obj x %f y %f z %f", objpos[0], objpos[1], objpos[2]);

  FITDBG ("  position cam %s x %f y %f z %f", cam_p->name,
	  camposrot[0], camposrot[1], camposrot[2]);

  // now object moves to objx-camx, objy-camy, objz-camz
  FITDBG ("  translated objx %f y %f z %f",
	  objpos[0] - camposrot[0],
	  objpos[1] - camposrot[1], objpos[2] - camposrot[2]);

  FITDBG ("  rotation cam %s a %f b %f g %f", cam_p->name,
	  camposrot[3], camposrot[4], camposrot[5]);

  rotate_3d (objpos[0] - camposrot[0],
	     objpos[1] - camposrot[1],
	     objpos[2] - camposrot[2],
	     camposrot[3], camposrot[4], camposrot[5], &rotx, &roty, &rotz);

  FITDBG ("  rotated objx %f y %f z %f", rotx, roty, rotz);

  FITDBG ("  camera typeparam %f %f %f", camtypeparam[0],
	  camtypeparam[1], camtypeparam[2]);

  camfunc (rotx, roty, rotz, &officialx, &officialy, camtypeparam);

  FITDBG ("  camera official x %f y %f", officialx, officialy);

  FITDBG ("  camera distortionparam %f %f %f",
	  camdistparam[0], camdistparam[1], camdistparam[2]);

  distfunc (officialx, officialy, campixx, campixy, camdistparam);

  FITDBG ("  camera distorted x %f y %f", *campixx, *campixy);
}


void
parse_stdin ()
{
  json_t *jroot;
  json_error_t jerror;
  double camposrot[6];		// position + rotation: x,y,z,a,b,g
  double objpos[3];		// position: x,y,z

  jroot = json_loadfd (0, 0, &jerror);
  if (jroot == NULL)
    {
      STDERR ("input error on line %d: %s", jerror.line, jerror.text);
      exit (1);
    }

  // Note that jansson is perfectly okay handling NULL pointers

  parse_cam_json (jroot, &cam[camnum], 1);
  cam[camnum].name = "<noname>";	// always only one cam, fake name

  {
    json_t *jcampos = json_object_get (jroot, "campos");
    if (jcampos == NULL)
      {
	STDERR ("No camera position specified in configfile");
	exit (1);
      }

    camposrot[0] = json_number_value (json_object_get (jcampos, "x"));
    camposrot[1] = json_number_value (json_object_get (jcampos, "y"));
    camposrot[2] = json_number_value (json_object_get (jcampos, "z"));
    camposrot[3] = json_number_value (json_object_get (jcampos, "a"));
    camposrot[4] = json_number_value (json_object_get (jcampos, "b"));
    camposrot[5] = json_number_value (json_object_get (jcampos, "g"));

    printf
      ("Camera (%6.3f, %6.3f, %6.3f) rotation (%6.3f, %6.3f, %6.3f)\n",
       camposrot[0], camposrot[1], camposrot[2],
       camposrot[3], camposrot[4], camposrot[5]);
  }

  {
    json_t *jpoints = json_object_get (jroot, "points");
    int i;

    for (i = 0; i < json_array_size (jpoints); i++)
      {
	json_t *jpoint;
	double campixx = 0;
	double campixy = 0;

	jpoint = json_array_get (jpoints, i);
	if (!json_is_object (jpoint))
	  continue;

	objpos[0] = json_number_value (json_object_get (jpoint, "x"));
	objpos[1] = json_number_value (json_object_get (jpoint, "y"));
	objpos[2] = json_number_value (json_object_get (jpoint, "z"));

	calc_campix2 (&cam[camnum], camposrot,
		      camtype[cam[camnum].camtype].func,
		      cam[camnum].typeparam,
		      camdistortion[cam[camnum].camdistortion].func,
		      cam[camnum].distortionparam, objpos, &campixx,
		      &campixy);

	printf
	  ("Object (%6.3f, %6.3f, %6.3f) is at camera pixel: (%4ld, %4ld)\n",
	   objpos[0], objpos[1], objpos[2],
	   lround ((campixx / cam[camnum].xscale) + cam[camnum].xoffset),
	   lround ((campixy / cam[camnum].yscale) + cam[camnum].yoffset));
      }

  }

  json_decref (jroot);
}


int
main (void)
{
  parse_stdin ();

  return 0;
}
