// USAGE: camplacingvis < stdin
// Given multiple camera types and positions, visualize the spatial result
// of expected reachability.

// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#define PROGNAME  "camplacingvis"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <jansson.h>

typedef int bool;		// for MathGL 2.4.4
#include <mgl2/mgl_cf.h>

#include "../shared/error.h"
#include "../camfuncs/camfuncs.h"
#include "../camfuncs/camhelpers.h"

//#define FITDBG(args...)  STDERR(args)
#define FITDBG(args...)
//#define MDBG(args...)  STDERR(args)
#define MDBG(args...)


const char *progname = PROGNAME;


typedef struct
{
  char *name;
  int camtype;			// index to camtype[]
  double *typeparam;		// malloc'd for specified type
  int camdistortion;		// index to camdistortion[]
  double *distortionparam;	// malloc'd for specified distortion
  double dist_lin_radius;	// distortion linearized beyond this limit

  int w;			// camera #pixels
  int h;
  double xscale;		// calculate pixel to -1..+1 coords
  double xoffset;
  double yscale;
  double yoffset;
} cam_t;

// parse_cam_json requires cam_t to be defined
#define CAMERR_RETURN(args...)  do { STDERR(args); exit(1); } while(0)
#include "../shared/parse_cam_json.h"
#undef CAMERR_RETURN

typedef struct
{
  char *imgfilename;
  int imgw;
  int imgh;

  double area_xmin;
  double area_xmax;
  double area_ymin;
  double area_ymax;
  double area_xintv;
  double area_yintv;
  double area_z;
  int area_xnum;
  int area_ynum;
  int area_allocd;
  double *area;

  double blinker_sizex;
  double blinker_sizey;
  double blinker_sizez;
  double blinker_minpix;

  // We keep overwriting the same cam
  cam_t cam[2];
  double camposrot[6];		// position + rotation: x,y,z,a,b,g

} state_t;

int camnum = 0;


#define cam_obj_t cam_t

// vvvvv----- UNCHANGED FROM senecalc.c -----vvvvv
void
dist_lin_precalc (cam_obj_t * cam_p)
{
  distortion_func_t *distfunc;
  double xin, yin;
  double xout, yout;
  double prev_xin, prev_yin;
  double prev_xout, prev_yout;
  double xhighend, xlowend, yhighend, ylowend;
  double min;
  int first;

  // We want to end up with a "limiting" radius from the image center.
  // Points outside the radius are then calculated as on-radius value +
  // linear outside-of-radius extension --> distortion_wrapper().

  // We currently detect the radius from the minimum of the pos/neg x/y axes,
  // but that could be improved with diagonals if really needed.
  // Procedure: walk the distortion function from the center,
  // see where it starts moving pixels *inwards* from the previous.
  // That's the point where distortion stops being useful,
  // and where fitting troubles begin.
  // Also, extremely high slopes are detected in the same way.

  // Most cams will be less than 1:2.35, so can keep walk limited.
  // Take coarse steps at ~100 pix per half-width.

  distfunc = camdistortion[cam_p->camdistortion].func;

  prev_xin = 0.;
  prev_xout = 0.;
  prev_xout = 0.;
  prev_yout = 0.;
  first = 1;
  yin = 0.;
  for (xin = 0.; xin < 3.; xin += 0.01)
    {
      distfunc (xin, yin, &xout, &yout, cam_p->distortionparam);
      if (first)
	first = 0;
      else
	{
	  if (xout <= prev_xout)
	    break;		// stepping back
	  else if (xout >= prev_xout + 0.1)
	    break;		// 10x speed-up slope
	}
      prev_xin = xin;
      prev_xout = xout;
    }
  // Now either at problem point, or at end of walk;
  // use linear extrapolation from this point onwards.
  xhighend = prev_xin;

  prev_xin = 0.;
  prev_xout = 0.;
  prev_xout = 0.;
  prev_yout = 0.;
  first = 1;
  yin = 0.;
  for (xin = 0.; xin > -3.; xin -= 0.01)
    {
      distfunc (xin, yin, &xout, &yout, cam_p->distortionparam);
      if (first)
	first = 0;
      else
	{
	  if (xout >= prev_xout)
	    break;		// stepping back
	  else if (xout <= prev_xout - 0.1)
	    break;		// 10x speed-up slope
	}
      prev_xin = xin;
      prev_xout = xout;
    }
  // Now either at problem point, or at end of walk;
  // use linear extrapolation from this point onwards.
  xlowend = prev_xin;

  prev_xin = 0.;
  prev_xout = 0.;
  prev_xout = 0.;
  prev_yout = 0.;
  first = 1;
  xin = 0.;
  for (yin = 0.; yin < 3.; yin += 0.01)
    {
      distfunc (xin, yin, &xout, &yout, cam_p->distortionparam);
      if (first)
	first = 0;
      else
	{
	  if (yout <= prev_yout)
	    break;		// stepping back
	  else if (yout >= prev_yout + 0.1)
	    break;		// 10x speed-up slope
	}
      prev_yin = yin;
      prev_yout = yout;
    }
  // Now either at problem point, or at end of walk;
  // use linear extrapolation from this point onwards.
  yhighend = prev_yin;

  prev_xin = 0.;
  prev_xout = 0.;
  prev_xout = 0.;
  prev_yout = 0.;
  first = 1;
  xin = 0.;
  for (yin = 0.; yin > -3.; yin -= 0.01)
    {
      distfunc (xin, yin, &xout, &yout, cam_p->distortionparam);
      if (first)
	first = 0;
      else
	{
	  if (yout >= prev_yout)
	    break;		// stepping back
	  else if (yout <= prev_yout - 0.1)
	    break;		// 10x speed-up slope
	}
      prev_yin = yin;
      prev_yout = yout;
    }
  // Now either at problem point, or at end of walk;
  // use linear extrapolation from this point onwards.
  ylowend = prev_yin;

  min = xhighend;
  if (-xlowend < min)
    min = -xlowend;
  if (yhighend < min)
    min = yhighend;
  if (-ylowend < min)
    min = -ylowend;

  cam_p->dist_lin_radius = min;

  MDBG ("Cam %s distortion useful limits x %f %f y %f %f  radius %f",
	cam_p->name, xhighend, xlowend, yhighend, ylowend,
	cam_p->dist_lin_radius);
}

inline void
distortion_wrapper (cam_obj_t * cam_p, double officialx, double officialy,
		    double *distortedx, double *distortedy)
{
  distortion_func_t *distfunc;
  double r;
  double xnew, ynew, xadd, yadd;

  // We clip x/y to within the useful radius,
  // calculate the distortion result for that point,
  // and then linearly extend beyond that result.

  xnew = officialx;
  ynew = officialy;
  xadd = 0.;
  yadd = 0.;

  r = hypot (officialx, officialy);
  if (r > cam_p->dist_lin_radius)
    {
      double scale = cam_p->dist_lin_radius / r;
      xnew = officialx * scale;
      ynew = officialy * scale;
      xadd = officialx - xnew;
      yadd = officialy - ynew;
    }

  //MDBG ("distwrap clipped input to %f %f +linear %f %f",
  //      xnew, ynew, xadd, yadd);

  distfunc = camdistortion[cam_p->camdistortion].func;
  distfunc (xnew, ynew, distortedx, distortedy, cam_p->distortionparam);

  *distortedx += xadd;
  *distortedx += yadd;
}

// ^^^^^----- UNCHANGED FROM senecalc.c -----^^^^^


// Special version: Camtype+params/dist+params specified separately.
// 
// double camposrot[6] = camera position+rotation x/y/z/a/b/g
// double objpos[3] = object position x/y/z
static inline void
calc_campix2 (cam_t * cam_p, double *camposrot,
	      camtype_func_t * camfunc,
	      double *camtypeparam,
	      distortion_func_t * distfunc,
	      double *camdistparam,
	      double *objpos, double *campixx, double *campixy)
{
  double rotx, roty, rotz;
  double officialx, officialy;

  FITDBG ("  position obj x %f y %f z %f", objpos[0], objpos[1], objpos[2]);

  FITDBG ("  position cam %s x %f y %f z %f", cam_p->name,
	  camposrot[0], camposrot[1], camposrot[2]);

  // now object moves to objx-camx, objy-camy, objz-camz
  FITDBG ("  translated objx %f y %f z %f",
	  objpos[0] - camposrot[0],
	  objpos[1] - camposrot[1], objpos[2] - camposrot[2]);

  FITDBG ("  rotation cam %s a %f b %f g %f", cam_p->name,
	  camposrot[3], camposrot[4], camposrot[5]);

  rotate_3d (objpos[0] - camposrot[0],
	     objpos[1] - camposrot[1],
	     objpos[2] - camposrot[2],
	     camposrot[3], camposrot[4], camposrot[5], &rotx, &roty, &rotz);

  FITDBG ("  rotated objx %f y %f z %f", rotx, roty, rotz);

  FITDBG ("  camera typeparam %f %f %f", camtypeparam[0],
	  camtypeparam[1], camtypeparam[2]);

  camfunc (rotx, roty, rotz, &officialx, &officialy, camtypeparam);

  FITDBG ("  camera official x %f y %f", officialx, officialy);

  FITDBG ("  camera distortionparam %f %f %f",
	  camdistparam[0], camdistparam[1], camdistparam[2]);

  distortion_wrapper (cam_p, officialx, officialy, campixx, campixy);

  FITDBG ("  camera distorted x %f y %f", *campixx, *campixy);
}


void
alloc_area (state_t * state)
{
  if (state->area_allocd)
    return;

  state->area_xnum =
    ceil ((state->area_xmax - state->area_xmin) / state->area_xintv) + 0.1;
  state->area_ynum =
    ceil ((state->area_ymax - state->area_ymin) / state->area_xintv) + 0.1;

  state->area = calloc (state->area_xnum * state->area_ynum, sizeof (double));
  if (state->area == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  state->area_allocd = 1;
}

// returns: 1 = okay, 0 = error
static inline int
calc_blinker (state_t * state, double blinkx, double blinky, double blinkz,
	      double *blink_xmin, double *blink_xmax, double *blink_ymin,
	      double *blink_ymax)
{
  double campixx = -1, campixy = -1;
  double objpos[3];
  double backscalepixx, backscalepixy;

  objpos[0] = blinkx;
  objpos[1] = blinky;
  objpos[2] = blinkz;

  calc_campix2 (&state->cam[camnum], state->camposrot,
		camtype[state->cam[camnum].camtype].func,
		state->cam[camnum].typeparam,
		camdistortion[state->cam[camnum].camdistortion].func,
		state->cam[camnum].distortionparam,
		objpos, &campixx, &campixy);

  backscalepixx = (campixx / state->cam[camnum].xscale) +
    state->cam[camnum].xoffset;
  backscalepixy = (campixy / state->cam[camnum].yscale) +
    state->cam[camnum].yoffset;
  FITDBG ("  camera pixel x %f y %f", backscalepixx, backscalepixy);

  if (backscalepixx < 0 || backscalepixx >= state->cam[camnum].w)
    return 0;
  if (backscalepixy < 0 || backscalepixy >= state->cam[camnum].h)
    return 0;
  if (*blink_xmin > *blink_xmax)
    {
      // first round
      *blink_xmin = backscalepixx;
      *blink_xmax = backscalepixx;
      *blink_ymin = backscalepixy;
      *blink_ymax = backscalepixy;
    }
  else
    {
      if (backscalepixx < *blink_xmin)
	*blink_xmin = backscalepixx;
      if (backscalepixx > *blink_xmax)
	*blink_xmax = backscalepixx;
      if (backscalepixy < *blink_ymin)
	*blink_ymin = backscalepixy;
      if (backscalepixy > *blink_ymax)
	*blink_ymax = backscalepixy;
    }
  return 1;
}

void
walk_area (state_t * state)
{
  int ix, iy;

  for (iy = 0; iy < state->area_ynum; iy++)
    {
      double centery = state->area_ymin + ((double) iy) * state->area_yintv;
      for (ix = 0; ix < state->area_xnum; ix++)
	{
	  double centerx =
	    state->area_xmin + ((double) ix) * state->area_xintv;

	  double blink_xmin = 1., blink_xmax = -1.;
	  double blink_ymin = 0., blink_ymax = 0.;
	  double bx, by, bz;
	  double blinkarea;
	  int stillokay;

	  stillokay = 1;
	  for (bx = -0.5; bx < 0.6 && stillokay; bx += 1.)
	    for (by = -0.5; by < 0.6 && stillokay; by += 1.)
	      for (bz = -0.5; bz < 0.6 && stillokay; bz += 1.)
		stillokay = calc_blinker (state,
					  centerx + bx * state->blinker_sizex,
					  centery + by * state->blinker_sizey,
					  state->area_z +
					  bz * state->blinker_sizez,
					  &blink_xmin, &blink_xmax,
					  &blink_ymin, &blink_ymax);

#if 0
	  STDERR ("xy %f %f okay %d bl %f %f %f %f",
		  centerx, centery, stillokay, blink_xmin, blink_xmax,
		  blink_ymin, blink_ymax);
#endif

	  if (!stillokay || blink_xmin > blink_xmax)
	    continue;

	  blinkarea = (blink_xmax - blink_xmin) * (blink_ymax - blink_ymin);
	  if (blinkarea >= state->blinker_minpix)
	    state->area[iy * state->area_xnum + ix]++;
	}
    }
}

void
dump_area (state_t * state)
{
  int ix, iy;

  printf ("\n");
  for (iy = state->area_ynum - 1; iy >= 0; iy--)
    {
      for (ix = 0; ix < state->area_xnum; ix++)
	{
	  printf (" %3.1f", state->area[iy * state->area_xnum + ix]);
	}
      printf ("\n");
    }
  printf ("\n");
}


void
add_cam (state_t * state)
{
  alloc_area (state);
  dist_lin_precalc (&state->cam[camnum]);
  walk_area (state);
  //dump_area (state);
}


void
make_graph (state_t * state)
{
  HMDT datx_minushalf_1extra;
  HMDT daty_minushalf_1extra;
  HMDT datz_1extra;
  HMDT datc_1extra;
  mreal *mgldataptr_x;
  mreal *mgldataptr_y;
  mreal *mgldataptr_z;
  mreal *mgldataptr_c;
  mreal zmin;
  mreal zmax;
  HMGL gr;
  double zminround, zmaxround;
  int i, ix, iy;

  // Tiles are plotted left-bottom at given coordinate, right-top at +1,+1,
  // so we shift down/left half a tile.
  // Top row + right column are never plotted, seems inevitable,
  // so we just add another one.

  datx_minushalf_1extra = mgl_create_data_size (state->area_xnum + 1, 1, 1);
  mgldataptr_x = mgl_data_data (datx_minushalf_1extra);
  for (i = 0; i < state->area_xnum + 1; i++)
    mgldataptr_x[i] = state->area_xmin +
      (((double) i) - 0.5) * state->area_xintv;

  daty_minushalf_1extra = mgl_create_data_size (state->area_ynum + 1, 1, 1);
  mgldataptr_y = mgl_data_data (daty_minushalf_1extra);
  for (i = 0; i < state->area_ynum + 1; i++)
    mgldataptr_y[i] = state->area_ymin +
      (((double) i) - 0.5) * state->area_yintv;

  datz_1extra = mgl_create_data_size (state->area_xnum + 1,
				      state->area_ynum + 1, 1);
  datc_1extra = mgl_create_data_size (state->area_xnum + 1,
				      state->area_ynum + 1, 1);
  mgldataptr_z = mgl_data_data (datz_1extra);
  mgldataptr_c = mgl_data_data (datc_1extra);

  for (iy = 0; iy < state->area_ynum + 1; iy++)
    for (ix = 0; ix < state->area_xnum + 1; ix++)
      {
	double d;
	d = NAN;
	if (iy < state->area_ynum && ix < state->area_xnum)
	  {
	    d = state->area[iy * state->area_xnum + ix];
	    if (d < 0.5)
	      d = NAN;
	  }
	mgldataptr_z[iy * (state->area_xnum + 1) + ix] = d;
	mgldataptr_c[iy * (state->area_xnum + 1) + ix] = d;
      }

  gr = mgl_create_graph (state->imgw, state->imgh);

  mgl_inplot (gr, -.22, 1.13, -.22, 1.22);

  mgl_set_range_val (gr, 'x',
		     state->area_xmin + -1.1 * state->area_xintv,
		     state->area_xmin +
		     (state->area_xnum + 0.1) * state->area_xintv);
  mgl_set_range_val (gr, 'y', state->area_ymin + -1.1 * state->area_yintv,
		     state->area_ymin +
		     (state->area_ynum + 0.1) * state->area_yintv);

  zmin = mgl_data_min (datz_1extra);
  zmax = mgl_data_max (datz_1extra);
  //zminround = floor (zmin - 0.01);
  //zmaxround = ceil (zmax + 0.01);
  zminround = zmin - 0.01;
  zmaxround = zmax + 0.01;

  mgl_set_range_val (gr, 'z', zminround, zmaxround);
  mgl_set_range_val (gr, 'c', zminround, zmaxround);

  mgl_aspect (gr, state->imgh * (state->area_xnum + 2),
	      state->imgw * (state->area_ynum + 2), state->imgw);

  // axes/grid position
  mgl_set_origin (gr, 0, 0, zmax + 0.01);

  mgl_set_ticks (gr, 'x', 1., 0, 0);
  mgl_set_ticks (gr, 'y', 1., 0, 0);
  mgl_set_ticks (gr, 'z', 1., 0, 0);
  mgl_set_ticks (gr, 'c', 1., 0, 0);

  mgl_load_font (gr, "heros", "/usr/share/mathgl/fonts");

#define COLORSCHEME "UbcyqR"

  mgl_tile_xy (gr, datx_minushalf_1extra, daty_minushalf_1extra, datz_1extra,
	       COLORSCHEME, "");

  mgl_set_font_size (gr, 1.5);
  mgl_colorbar (gr, COLORSCHEME ">I");

  mgl_set_font_size (gr, 1.5);
  mgl_axis (gr, "xyU", "k1-", "");
  mgl_axis_grid (gr, "xy", "W0-", "");

  mgl_write_frame (gr, state->imgfilename, "");

  mgl_delete_graph (gr);
}


void
parse_camposition (state_t * state, json_t * jroot)
{
  int have_rotation = 0;

  if (!json_is_object (jroot))
    return;			// Probably comment, etc.

  {
    json_t *jcampos = json_object_get (jroot, "position");
    if (jcampos == NULL)
      return;			// Probably comment, etc.

    state->camposrot[0] = json_number_value (json_object_get (jcampos, "x"));
    state->camposrot[1] = json_number_value (json_object_get (jcampos, "y"));
    state->camposrot[2] = json_number_value (json_object_get (jcampos, "z"));

    if (json_object_get (jcampos, "a") != NULL ||
	json_object_get (jcampos, "b") != NULL ||
	json_object_get (jcampos, "g") != NULL)
      {
	state->camposrot[3] =
	  json_number_value (json_object_get (jcampos, "a"));
	state->camposrot[4] =
	  json_number_value (json_object_get (jcampos, "b"));
	state->camposrot[5] =
	  json_number_value (json_object_get (jcampos, "g"));
	have_rotation = 1;
      }
  }

  {
    json_t *jlookat = json_object_get (jroot, "lookingat");
    if (jlookat != NULL)
      {
	double lx, ly, lz, hyp;
	lx = json_number_value (json_object_get (jlookat, "x"));
	ly = json_number_value (json_object_get (jlookat, "y"));
	lz = json_number_value (json_object_get (jlookat, "z"));

	// Calculate back to rotation angles
	hyp = hypot (lx - state->camposrot[0], ly - state->camposrot[1]);
	state->camposrot[3] =
	  atan2 (lx - state->camposrot[0], ly - state->camposrot[1]);
	state->camposrot[4] = atan2 (lz - state->camposrot[2], hyp);
	state->camposrot[5] = 0;
	have_rotation = 1;
      }
  }

  if (!have_rotation)
    {
      STDERR ("No camera rotation or lookingat specified");
      exit (1);
    }

  add_cam (state);
}

void
parse_camera (state_t * state, json_t * jroot)
{
  cam_t *cam = state->cam;

  if (!json_is_object (jroot))
    return;			// Probably comment, etc.

  parse_cam_json (jroot, &cam[camnum], 1);
  cam[camnum].name = "<noname>";	// always only one cam, fake name

  {
    state->blinker_minpix =
      json_number_value (json_object_get (jroot, "blinkdetectpixels"));

    if (state->blinker_minpix < 0.1)
      {
	STDERR ("No or illegal blinkdetectpixels specified");
	exit (1);
      }
  }

  {
    json_t *jpositions = json_object_get (jroot, "positions");
    int i;

    for (i = 0; i < json_array_size (jpositions); i++)
      parse_camposition (state, json_array_get (jpositions, i));
  }
}

void
parse_stdin (state_t * state)
{
  json_t *jroot;
  json_error_t jerror;

  jroot = json_loadfd (0, 0, &jerror);
  if (jroot == NULL)
    {
      STDERR ("input error on line %d: %s", jerror.line, jerror.text);
      exit (1);
    }

  // Note that jansson is perfectly okay handling NULL pointers

  {
    const char *cc;

    cc =
      json_string_value (json_object_get
			 (json_object_get (jroot, "output"), "file"));
    if (cc != NULL)
      state->imgfilename = strdup (cc);
    else
      {
	STDERR ("No outputfile specified");
	exit (1);
      }

    state->imgw =
      json_integer_value (json_object_get
			  (json_object_get (jroot, "output"), "pixw"));
    state->imgh =
      json_integer_value (json_object_get
			  (json_object_get (jroot, "output"), "pixh"));
    if (state->imgw <= 1 || state->imgh <= 1)
      {
	STDERR ("No output size specified");
	exit (1);
      }
  }

  {
    state->area_xmin =
      json_number_value (json_object_get
			 (json_object_get (jroot, "area"), "xmin"));
    state->area_xmax =
      json_number_value (json_object_get
			 (json_object_get (jroot, "area"), "xmax"));
    state->area_ymin =
      json_number_value (json_object_get
			 (json_object_get (jroot, "area"), "ymin"));
    state->area_ymax =
      json_number_value (json_object_get
			 (json_object_get (jroot, "area"), "ymax"));
    state->area_xintv =
      json_number_value (json_object_get
			 (json_object_get (jroot, "area"), "xinterval"));
    state->area_yintv =
      json_number_value (json_object_get
			 (json_object_get (jroot, "area"), "yinterval"));
    state->area_z =
      json_number_value (json_object_get
			 (json_object_get (jroot, "area"), "height"));

    if (state->area_xintv < 0.001 || state->area_yintv < 0.001)
      {
	STDERR ("No or illegal area x/y interval specified");
	exit (1);
      }

    if (state->area_xmax < state->area_xmin + state->area_xintv + 0.001 ||
	state->area_ymax < state->area_ymin + state->area_yintv + 0.001)
      {
	STDERR ("No or illegal area x/y min/max specified");
	exit (1);
      }
  }

  {
    state->blinker_sizex =
      json_number_value (json_object_get
			 (json_object_get (jroot, "blinker"), "sizex"));
    state->blinker_sizey =
      json_number_value (json_object_get
			 (json_object_get (jroot, "blinker"), "sizey"));
    state->blinker_sizez =
      json_number_value (json_object_get
			 (json_object_get (jroot, "blinker"), "sizez"));

    if (state->blinker_sizex < 0.001 || state->blinker_sizey < 0.001
	|| state->blinker_sizez < 0.001)
      {
	STDERR ("No or illegal blinker size specified");
	exit (1);
      }
  }

  {
    json_t *jcameras = json_object_get (jroot, "cameras");
    int i;

    for (i = 0; i < json_array_size (jcameras); i++)
      parse_camera (state, json_array_get (jcameras, i));
  }

  json_decref (jroot);
}


int
main (void)
{
  state_t state;

  memset (&state, 0, sizeof (state));

  parse_stdin (&state);

  //dump_area (&state);
  make_graph (&state);

  return 0;
}
