// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <stdio.h>
#include <stdlib.h>
#include <jansson.h>

#include "configfile.h"
#include "../shared/error.h"


// This does all state init that depends on config.
// Returns:
// 0 = okay
// -1 = error, message(s) already printed.
int
parse_configfile (state_t * state)
{
  json_error_t jerror;
  int retval = 0;

  state->jconfig = json_load_file (state->configfile, 0, &jerror);
  if (state->jconfig == NULL)
    {
      STDERR ("config error on line %d: %s", jerror.line, jerror.text);
      return -1;
    }

  // Note that jansson is perfectly okay handling NULL pointers

  state->name = json_string_value (json_object_get (state->jconfig, "name"));
  if (state->name == NULL || state->name[0] == '\0')
    {
      STDERR ("no name specified in configfile");
      retval = -1;
    }

  state->data_topic =
    json_string_value (json_object_get (state->jconfig, "datatopic"));
  if (state->data_topic == NULL || state->data_topic[0] == '\0')
    {
      STDERR ("no data topic specified in configfile");
      retval = -1;
    }

  state->do_interpolate =
    json_is_true (json_object_get (state->jconfig, "interpolate"));

  {
    int i;
    int axisnum = 0;
    json_t *jaxes;

    jaxes = json_object_get (state->jconfig, "axes");
    if (!json_is_array (jaxes))
      {
	STDERR ("no axes array in configfile");
	retval = -1;
      }

    for (i = 0; i < json_array_size (jaxes); i++)
      {
	json_t *jaxis;
	const char *label;
	double intv;

	jaxis = json_array_get (jaxes, i);
	if (!json_is_object (jaxis))
	  continue;		// Possibly comment, etc.

	label = json_string_value (json_object_get (jaxis, "label"));
	if (label == NULL || label[0] == '\0')
	  continue;		// Possibly comment, etc.

	if (axisnum >= 2)
	  {
	    STDERR ("more than 2 axes specified in configfile");
	    retval = -1;
	    break;
	  }

	state->jaxes[axisnum] = jaxis;

	intv = json_number_value (json_object_get (jaxis, "interval"));
	if (intv <= 0.0000001)
	  {
	    STDERR ("no interval for axis %d specified in configfile",
		    axisnum);
	    retval = -1;
	  }
	else if (axisnum == 0)
	  state->xbinsize = intv;
	else if (axisnum == 1)
	  state->ybinsize = intv;

	axisnum++;
      }

    if (axisnum != 2)
      {
	STDERR ("no 2 axes specified in configfile");
	retval = -1;
      }
  }

  {
    int i;
    int planenum = 0;
    json_t *jplanes;

    jplanes = json_object_get (state->jconfig, "planes");
    if (!json_is_array (jplanes))
      {
	STDERR ("no planes array in configfile");
	retval = -1;
      }

    // Allocate max, may possibly get less due to comments
    state->jplanes =
      malloc (json_array_size (jplanes) * sizeof (*state->jplanes));
    if (state->jplanes == NULL)
      {
	STDERR ("malloc failed");
	exit (1);
      }

    for (i = 0; i < json_array_size (jplanes); i++)
      {
	json_t *jplane;
	const char *label;

	jplane = json_array_get (jplanes, i);
	if (!json_is_object (jplane))
	  continue;		// Possibly comment, etc.

	label = json_string_value (json_object_get (jplane, "label"));
	if (label == NULL || label[0] == '\0')
	  continue;		// Possibly comment, etc.

	state->jplanes[planenum] = jplane;

	planenum++;
      }

    state->jplanes_num = planenum;
    if (planenum < 1)
      {
	STDERR ("no data planes specified in configfile");
	retval = -1;
      }
  }

  state->jconditions = json_object_get (state->jconfig, "conditions");
  if (!json_is_array (state->jconditions))
    {
      state->jconditions = NULL;	// Empty is allowed
    }

  state->jrenders = json_object_get (state->jconfig, "renders");
  if (!json_is_array (state->jrenders))
    {
      STDERR ("no renders array in configfile");
      retval = -1;
    }

  return retval;
}
