// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <string.h>

#include "data.h"
#include "../shared/error.h"

void
init_data (state_t * state, int xbin, int ybin)
{
  int x, y, v;

  //       bin=2 in center
  //           |
  //  0  1 2 3 4 5 6 7 8   data[] index   size=9
  // -2 -1 0 1 2 3 4 5 6   bin coord      zerooffset=size/2-bin=9/2-2=2(.5)

  // Startup size
  state->data_xsize = 20;
  state->data_ysize = 20;

  // Set xbin/ybin in center
  state->data_xzerooffset = (state->data_xsize / 2) - xbin;
  state->data_yzerooffset = (state->data_ysize / 2) - ybin;

  state->data = malloc ((state->data_xsize
			 * state->data_ysize
			 * state->data_valuesperbin) * sizeof (*state->data));
  if (state->data == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  for (y = 0; y < state->data_ysize; y++)
    for (x = 0; x < state->data_xsize; x++)
      for (v = 0; v < state->data_valuesperbin; v++)
	state->data[(y * state->data_xsize + x)
		    * state->data_valuesperbin + v] = NAN;

  // Initialize in the expectation that data will get stored here.
  state->data_xused_min = xbin;
  state->data_xused_max = xbin;
  state->data_yused_min = ybin;
  state->data_yused_max = ybin;


#ifdef INTERP_DATA_KEEP_ALLOCD
  if (state->do_interpolate)
    {
      state->interp_data = malloc ((state->data_xsize
				    * state->data_ysize
				    * state->interp_data_valuesperbin)
				   * sizeof (*state->interp_data));
      if (state->interp_data == NULL)
	{
	  STDERR ("malloc failed");
	  exit (1);
	}
    }
#endif
}

void
grow_data (state_t * state, int xbin, int ybin)
{
  int new_xused_min, new_xused_max, new_yused_min, new_yused_max;
  int new_xsize, new_ysize, new_xzerooffset, new_yzerooffset;
  int copy_old_xfirstidx, copy_old_xlastidx;
  int copy_new_xfirstidx, copy_new_xlastidx;
  int copy_old_yfirstidx, copy_old_ylastidx;
  int copy_new_yfirstidx, copy_new_ylastidx;
  int spillover;
  int x, y, v;
  double *new_data;

  if (state->data == NULL)
    {
      init_data (state, xbin, ybin);
      return;
    }

  // Find out what x/y range will be used
  new_xused_min = state->data_xused_min;
  new_xused_max = state->data_xused_max;
  new_yused_min = state->data_yused_min;
  new_yused_max = state->data_yused_max;
  if (xbin < new_xused_min)
    new_xused_min = xbin;
  if (xbin > new_xused_max)
    new_xused_max = xbin;
  if (ybin < new_yused_min)
    new_yused_min = ybin;
  if (ybin > new_yused_max)
    new_yused_max = ybin;

  // Grow exponentially based on actually used data
  new_xsize = (new_xused_max - new_xused_min + 1) * 2;
  new_ysize = (new_yused_max - new_yused_min + 1) * 2;
  if (new_xsize < 10)
    new_xsize = 10;
  if (new_ysize < 10)
    new_ysize = 10;

  //         1 2 3 4         bin coord used: min=1 max=4 used=max-min+1=4-1+1=4
  //  0  1 2 3 4 5 6 7 8 9   new data[] index   size=10
  // -2 -1 0 1 2 3 4 5 6 7   new bin coord      zerooffset=(size-used)/2-min=2

  // Place used data in center, allows equal growth to both sides.
  //
  // Note: this may shift previously allocated but unused data out of scope.
  // E.g. default 20x20 grid started from (0,0) and filled up to (5,5);
  // new datapoint (20,0) will make 40x10 grid and loose y=-10..-5 coords.

  new_xzerooffset = (new_xsize - (new_xused_max - new_xused_min + 1)) / 2
    - new_xused_min;
  new_yzerooffset = (new_ysize - (new_yused_max - new_yused_min + 1)) / 2
    - new_yused_min;

  //         0 1 2 3         old data[] index   size=4
  //         1 2 3 4         old bin coord      zerooffset=-1
  //
  //  0  1 2 3 4 5 6 7 8 9   new data[] index   size=10
  // -2 -1 0 1 2 3 4 5 6 7   new bin coord      zerooffset=2

  // First assume everything will fit (all copy indexes are *inclusive*)
  copy_old_xfirstidx = 0;
  copy_old_xlastidx = state->data_xsize - 1;
  copy_new_xfirstidx = new_xzerooffset - state->data_xzerooffset;
  copy_new_xlastidx = copy_new_xfirstidx + state->data_xsize - 1;

  copy_old_yfirstidx = 0;
  copy_old_ylastidx = state->data_ysize - 1;
  copy_new_yfirstidx = new_yzerooffset - state->data_yzerooffset;
  copy_new_ylastidx = copy_new_yfirstidx + state->data_ysize - 1;

  // Then correct in case it won't fit anyway
  spillover = -copy_new_xfirstidx;	// too much on low side
  if (spillover > 0)
    {
      copy_old_xfirstidx += spillover;
      copy_new_xfirstidx += spillover;
    }
  spillover = copy_new_xlastidx - (new_xsize - 1);	// too much on high side
  if (spillover > 0)
    {
      copy_old_xlastidx -= spillover;
      copy_new_xlastidx -= spillover;
    }

  spillover = -copy_new_yfirstidx;	// too much on low side
  if (spillover > 0)
    {
      copy_old_yfirstidx += spillover;
      copy_new_yfirstidx += spillover;
    }
  spillover = copy_new_ylastidx - (new_ysize - 1);	// too much on high side
  if (spillover > 0)
    {
      copy_old_ylastidx -= spillover;
      copy_new_ylastidx -= spillover;
    }

  // Sanity checks that should never trigger if algo is correct
  if ((copy_new_xlastidx - copy_new_xfirstidx !=
       copy_old_xlastidx - copy_old_xfirstidx) ||
      (copy_new_ylastidx - copy_new_yfirstidx !=
       copy_old_ylastidx - copy_old_yfirstidx) ||
      (copy_new_xlastidx < copy_new_xfirstidx ||
       copy_new_ylastidx < copy_new_yfirstidx))
    {
      STDERR ("illegal index result, please fix calculation");
      exit (1);
    }

  new_data = malloc ((new_xsize * new_ysize
		      * state->data_valuesperbin) * sizeof (*new_data));
  if (new_data == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  // y first part all NAN
  for (y = 0; y < copy_new_yfirstidx; y++)
    for (x = 0; x < new_xsize; x++)
      for (v = 0; v < state->data_valuesperbin; v++)
	new_data[(y * new_xsize + x) * state->data_valuesperbin + v] = NAN;

  // y middle part contains old data
  for (y = copy_new_yfirstidx; y <= copy_new_ylastidx; y++)
    {
      // x first part all NAN
      for (x = 0; x < copy_new_xfirstidx; x++)
	for (v = 0; v < state->data_valuesperbin; v++)
	  new_data[(y * new_xsize + x) * state->data_valuesperbin + v] = NAN;

      // x middle part contains old data: copy
      memcpy (&new_data[(y * new_xsize
			 + copy_new_xfirstidx) * state->data_valuesperbin],
	      &state->data[((y - copy_new_yfirstidx + copy_old_yfirstidx)
			    * state->data_xsize
			    + copy_old_xfirstidx) * state->data_valuesperbin],
	      (copy_new_xlastidx - copy_new_xfirstidx + 1)
	      * state->data_valuesperbin * sizeof (*new_data));

      // x last part all NAN
      for (x = copy_new_xlastidx + 1; x < new_xsize; x++)
	for (v = 0; v < state->data_valuesperbin; v++)
	  new_data[(y * new_xsize + x) * state->data_valuesperbin + v] = NAN;
    }

  // y last part all NAN
  for (y = copy_new_ylastidx + 1; y < new_ysize; y++)
    for (x = 0; x < new_xsize; x++)
      for (v = 0; v < state->data_valuesperbin; v++)
	new_data[(y * new_xsize + x) * state->data_valuesperbin + v] = NAN;

  // Swap in new data
  free (state->data);
  state->data = new_data;
  state->data_xsize = new_xsize;
  state->data_ysize = new_ysize;
  state->data_xzerooffset = new_xzerooffset;
  state->data_yzerooffset = new_yzerooffset;


#ifdef INTERP_DATA_KEEP_ALLOCD
  if (state->do_interpolate)
    {
      // Interpolate-data is hard to grow, just fully recalculate.
      free (state->interp_data);
      state->interp_data = malloc ((state->data_xsize
				    * state->data_ysize
				    * state->interp_data_valuesperbin)
				   * sizeof (*state->interp_data));
      if (state->interp_data == NULL)
	{
	  STDERR ("malloc failed");
	  exit (1);
	}
    }
#endif
}

inline void
storedata (state_t * state, int xbin, int ybin, int valueidx, double value)
{
  int xmin, xmax, ymin, ymax;

  //  0  1 2 3 4 5 6  data[] index     size=7  zerooffset=2
  // -2 -1 0 1 2 3 4  bin coord        min=-2  max=7-2-1=4

  xmin = -state->data_xzerooffset;
  xmax = state->data_xsize - state->data_xzerooffset - 1;
  ymin = -state->data_yzerooffset;
  ymax = state->data_ysize - state->data_yzerooffset - 1;

  // Zero-init will get xmin=0, xmax=-1, so first datapoint will neatly trigger.

  if (xbin < xmin || xbin > xmax || ybin < ymin || ybin > ymax)
    grow_data (state, xbin, ybin);

  state->data[((ybin + state->data_yzerooffset) * state->data_xsize
	       + (xbin + state->data_xzerooffset))
	      * state->data_valuesperbin + valueidx] = value;

  if (xbin < state->data_xused_min)
    state->data_xused_min = xbin;
  if (xbin > state->data_xused_max)
    state->data_xused_max = xbin;
  if (ybin < state->data_yused_min)
    state->data_yused_min = ybin;
  if (ybin > state->data_yused_max)
    state->data_yused_max = ybin;
}

inline void
storedata_n (state_t * state, int xbin, int ybin, int startidx, int numvals,
	     double *values)
{
  int xmin, xmax, ymin, ymax;

  //  0  1 2 3 4 5 6  data[] index     size=7  zerooffset=2
  // -2 -1 0 1 2 3 4  bin coord        min=-2  max=7-2-1=4

  xmin = -state->data_xzerooffset;
  xmax = state->data_xsize - state->data_xzerooffset - 1;
  ymin = -state->data_yzerooffset;
  ymax = state->data_ysize - state->data_yzerooffset - 1;

  // Zero-init will get xmin=0, xmax=-1, so first datapoint will neatly trigger.

  if (xbin < xmin || xbin > xmax || ybin < ymin || ybin > ymax)
    grow_data (state, xbin, ybin);

  memcpy (&state->data[((ybin + state->data_yzerooffset) * state->data_xsize
			+ (xbin + state->data_xzerooffset))
		       * state->data_valuesperbin + startidx],
	  values, numvals * sizeof (*values));

  if (xbin < state->data_xused_min)
    state->data_xused_min = xbin;
  if (xbin > state->data_xused_max)
    state->data_xused_max = xbin;
  if (ybin < state->data_yused_min)
    state->data_yused_min = ybin;
  if (ybin > state->data_yused_max)
    state->data_yused_max = ybin;
}

// getdata() -> static inline in data.h
