// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <string.h>
#include "interpolate.h"
#include "data.h"

#define DIST_WEIGHT(d)  (1/(d))
//#define DIST_WEIGHT(d)  (1/(d*d))

// Weight if point has data itself, will be same for all directions,
// so value does not matter.
// But is checked for forced inclusion in no-extrapolation areas,
// so must be much higher than weightsum will ever become.
#define SAME_WEIGHT  1000.


static inline void
extend_into_singlepoint (state_t * state, int direction, int x, int y,
			 int *lastdata_x, int *lastdata_y, int prevrow_x,
			 int prevrow_y)
{
  int setx, sety;
  double setweight;

  if (!isnan (state->data[(y * state->data_xsize + x)
			  * state->data_valuesperbin + 0]))
    // only need look at first plane; NAN should be coupled always.
    {
      *lastdata_x = x;
      *lastdata_y = y;
      setx = x;
      sety = y;
      setweight = SAME_WEIGHT;
    }
  else
    {
      // Extend from previous column on current row, or from
      // previous row, whichever is closest.

      double currentrow_dist = -1.;
      double prevrow_dist = -1.;
      int prevx = -1;
      int prevy = -1;

      if (*lastdata_x >= 0)
	currentrow_dist = hypot ((x - *lastdata_x) * state->xbinsize,
				 (y - *lastdata_y) * state->ybinsize);

      if (prevrow_x >= 0 && prevrow_x < state->data_xsize &&
	  prevrow_y >= 0 && prevrow_y < state->data_ysize)
	{
	  prevx = state->interp_data[(prevrow_y * state->data_xsize
				      + prevrow_x)
				     * state->interp_data_valuesperbin
				     + direction].fromx;
	  prevy = state->interp_data[(prevrow_y * state->data_xsize
				      + prevrow_x)
				     * state->interp_data_valuesperbin
				     + direction].fromy;

	  if (prevx >= 0)
	    prevrow_dist = hypot ((x - prevx) * state->xbinsize,
				  (y - prevy) * state->ybinsize);
	}

      if (currentrow_dist >= 0. && prevrow_dist >= 0.)
	{
	  if (currentrow_dist <= prevrow_dist)
	    {
	      // Current row continues
	      setx = *lastdata_x;
	      sety = *lastdata_y;
	      setweight = DIST_WEIGHT (currentrow_dist);
	    }
	  else
	    {
	      // Previous takes over on this row
	      *lastdata_x = prevx;
	      *lastdata_y = prevy;
	      setx = prevx;
	      sety = prevy;
	      setweight = DIST_WEIGHT (prevrow_dist);
	    }
	}
      else if (currentrow_dist >= 0.)
	{
	  // Current row continues
	  setx = *lastdata_x;
	  sety = *lastdata_y;
	  setweight = DIST_WEIGHT (currentrow_dist);
	}
      else if (prevrow_dist >= 0.)
	{
	  // Previous takes over on this row
	  *lastdata_x = prevx;
	  *lastdata_y = prevy;
	  setx = prevx;
	  sety = prevy;
	  setweight = DIST_WEIGHT (prevrow_dist);
	}
      else
	{
	  // No data seen yet.
	  setx = -1;
	  sety = -1;
	  setweight = 0.;
	}
    }

  state->interp_data[(y * state->data_xsize + x)
		     * state->interp_data_valuesperbin
		     + direction].fromx = setx;
  state->interp_data[(y * state->data_xsize + x)
		     * state->interp_data_valuesperbin
		     + direction].fromy = sety;
  state->interp_data[(y * state->data_xsize + x)
		     * state->interp_data_valuesperbin
		     + direction].weight = setweight;
}

// WARNING: The quarters routine is left for documentation purposes,
// but probably won't work well since everything else assumes to be
// handling eighths.
void
quarters_extend_full (state_t * state)
{
  int x, y;
  int lastdata_x, lastdata_y;
  int direction;

  direction = 0;		// x+ y+
  for (y = 0; y < state->data_ysize; y++)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (x = 0; x < state->data_xsize; x++)
	{
	  int prevrow_x = x;
	  int prevrow_y = y - 1;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 1;		// y+ x+
  for (x = 0; x < state->data_xsize; x++)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (y = 0; y < state->data_ysize; y++)
	{
	  int prevrow_x = x - 1;
	  int prevrow_y = y;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 2;		// x- y+
  for (y = 0; y < state->data_ysize; y++)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (x = state->data_xsize - 1; x >= 0; x--)
	{
	  int prevrow_x = x;
	  int prevrow_y = y - 1;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 3;		// y+ x-
  for (x = state->data_xsize - 1; x >= 0; x--)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (y = 0; y < state->data_ysize; y++)
	{
	  int prevrow_x = x + 1;
	  int prevrow_y = y;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 4;		// x+ y-
  for (y = state->data_ysize - 1; y >= 0; y--)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (x = 0; x < state->data_xsize; x++)
	{
	  int prevrow_x = x;
	  int prevrow_y = y + 1;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 5;		// y- x+
  for (x = 0; x < state->data_xsize; x++)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (y = state->data_ysize - 1; y >= 0; y--)
	{
	  int prevrow_x = x - 1;
	  int prevrow_y = y;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 6;		// x- y-
  for (y = state->data_ysize - 1; y >= 0; y--)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (x = state->data_xsize - 1; x >= 0; x--)
	{
	  int prevrow_x = x;
	  int prevrow_y = y + 1;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 7;		// y- x-
  for (x = state->data_xsize - 1; x >= 0; x--)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (y = state->data_ysize - 1; y >= 0; y--)
	{
	  int prevrow_x = x + 1;
	  int prevrow_y = y;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }
}

void
eighths_extend_full (state_t * state)
{
  int x, y;
  int lastdata_x, lastdata_y;
  int direction;

  // NOTE: direction sequence is important for extrapolate check
  // ("no data in >=4 *subsequent* eighths")

  direction = 0;		// x+ y+
  for (y = 0; y < state->data_ysize; y++)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (x = 0; x < state->data_xsize; x++)
	{
	  int prevrow_x = x - 1;
	  int prevrow_y = y - 1;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 1;		// y+ x+
  for (x = 0; x < state->data_xsize; x++)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (y = 0; y < state->data_ysize; y++)
	{
	  int prevrow_x = x - 1;
	  int prevrow_y = y - 1;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 2;		// y+ x-
  for (x = state->data_xsize - 1; x >= 0; x--)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (y = 0; y < state->data_ysize; y++)
	{
	  int prevrow_x = x + 1;
	  int prevrow_y = y - 1;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 3;		// x- y+
  for (y = 0; y < state->data_ysize; y++)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (x = state->data_xsize - 1; x >= 0; x--)
	{
	  int prevrow_x = x + 1;
	  int prevrow_y = y - 1;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 4;		// x- y-
  for (y = state->data_ysize - 1; y >= 0; y--)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (x = state->data_xsize - 1; x >= 0; x--)
	{
	  int prevrow_x = x + 1;
	  int prevrow_y = y + 1;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 5;		// y- x-
  for (x = state->data_xsize - 1; x >= 0; x--)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (y = state->data_ysize - 1; y >= 0; y--)
	{
	  int prevrow_x = x + 1;
	  int prevrow_y = y + 1;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 6;		// y- x+
  for (x = 0; x < state->data_xsize; x++)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (y = state->data_ysize - 1; y >= 0; y--)
	{
	  int prevrow_x = x - 1;
	  int prevrow_y = y + 1;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }

  direction = 7;		// x+ y-
  for (y = state->data_ysize - 1; y >= 0; y--)
    {
      lastdata_x = -1;
      lastdata_y = -1;
      for (x = 0; x < state->data_xsize; x++)
	{
	  int prevrow_x = x - 1;
	  int prevrow_y = y + 1;

	  extend_into_singlepoint (state, direction, x, y, &lastdata_x,
				   &lastdata_y, prevrow_x, prevrow_y);
	}
    }
}

void
copy_one_direction_to_data (state_t * state)
{
  int x, y;
  int direction;
  int showselect;

  // Note that debug==0 won't work, always add something.
  direction = state->interpolate_debug & 0x07;
  showselect = (state->interpolate_debug & 0xf0) >> 4;

  for (y = 0; y < state->data_ysize; y++)
    for (x = 0; x < state->data_xsize; x++)
      {
	double newz;
	int fromx = state->interp_data[(y * state->data_xsize + x)
				       * state->interp_data_valuesperbin
				       + direction].fromx;
	int fromy = state->interp_data[(y * state->data_xsize + x)
				       * state->interp_data_valuesperbin
				       + direction].fromy;
	double weight = state->interp_data[(y * state->data_xsize + x)
					   * state->interp_data_valuesperbin
					   + direction].weight;
	(void) weight;		// silence gcc complaints "unused"

	if (fromx < 0)
	  {
	    state->data[(y * state->data_xsize + x)
			* state->data_valuesperbin
			+ state->jplanes_num] = NAN;
	    continue;
	  }

	if (showselect == 2)
	  newz = fromx;
	else if (showselect == 3)
	  newz = fromy;
	else if (showselect == 4)
	  newz = weight;
	else
	  {
	    // includes showselect == 1
	    newz = state->data[(fromy * state->data_xsize + fromx)
			       * state->data_valuesperbin + 0];
	  }

	state->data[(y * state->data_xsize + x)
		    * state->data_valuesperbin + state->jplanes_num] = newz;
      }
}

inline void
interpolate_single (state_t * state, int x, int y)
{
  int i, j;
  double sum_weight;
  double sum_plane[state->jplanes_num];
  int plane;
  int num_included;
  int epcheck_firstvalid = -1;	// extrapolation checks
  int epcheck_lastvalid = -1;
  int epcheck_okay = 1;

  memset (sum_plane, 0, sizeof (sum_plane));
  sum_weight = 0;
  num_included = 0;

  for (i = 0; i < state->interp_data_valuesperbin; i++)
    {
      int already_included;
      int fromx = state->interp_data[(y * state->data_xsize + x)
				     * state->interp_data_valuesperbin
				     + i].fromx;
      int fromy = state->interp_data[(y * state->data_xsize + x)
				     * state->interp_data_valuesperbin
				     + i].fromy;
      double weight = state->interp_data[(y * state->data_xsize + x)
					 * state->interp_data_valuesperbin
					 + i].weight;
      if (fromx < 0)
	{
	  // No point in this direction

	  if (i >= epcheck_lastvalid + 4)
	    {
	      //   0 1 2 3 4 5 6 7
	      //   valid|invalid|valid
	      //       ^lastvalid             6  >=  2+4=6
	      //               ^match
	      // This is the 4th subsequent direction with no data:
	      // we must be extrapolating. Also holds starting at -1.
	      epcheck_okay = 0;
	    }

	  continue;
	}

      if (epcheck_firstvalid < 0)
	epcheck_firstvalid = i;
      epcheck_lastvalid = i;

      // skip if already included before
      already_included = 0;
      for (j = 0; j < i; j++)
	{
	  if (state->interp_data[(y * state->data_xsize + x)
				 * state->interp_data_valuesperbin
				 + j].fromx == fromx &&
	      state->interp_data[(y * state->data_xsize + x)
				 * state->interp_data_valuesperbin
				 + j].fromy == fromy)
	    {
	      already_included = 1;
	      break;
	    }
	}
      if (already_included)
	continue;

      for (plane = 0; plane < state->jplanes_num; plane++)
	sum_plane[plane] +=
	  state->data[(fromy * state->data_xsize + fromx)
		      * state->data_valuesperbin + plane] * weight;
      sum_weight += weight;
      num_included++;
    }

  if (epcheck_firstvalid + state->interp_data_valuesperbin
      > epcheck_lastvalid + 4)
    {
      //       __________________
      //      / / /          \ \ \ v--firstvalid+8
      //     0 1 2 3 4 5 6 7 [8 9 10]
      // invalid| valid |invalid
      //         ^firstvalid              2+8=10  >  5+4=9
      //               ^lastvalid
      // 4th subsequent direction with no data via wraparound:
      // we must be extrapolating.
      epcheck_okay = 0;
    }

#if 0
  // Don't extrapolate when less than 3 samples available
  if (num_included < 3)
    epcheck_okay = 0;
#endif

  // But always include point itself
  if (sum_weight > (SAME_WEIGHT - 0.1))
    epcheck_okay = 1;

  for (plane = 0; plane < state->jplanes_num; plane++)
    {
      if (epcheck_okay)
	state->data[(y * state->data_xsize + x)
		    * state->data_valuesperbin
		    + state->jplanes_num + plane] =
	  sum_plane[plane] / sum_weight;
      else
	state->data[(y * state->data_xsize + x)
		    * state->data_valuesperbin
		    + state->jplanes_num + plane] = NAN;
    }
}

void
interpolate_to_data (state_t * state)
{
  int x, y;

  if (state->interpolate_debug > 0)
    {
      copy_one_direction_to_data (state);
      return;
    }

  for (y = 0; y < state->data_ysize; y++)
    for (x = 0; x < state->data_xsize; x++)
      interpolate_single (state, x, y);
}
