// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <jansson.h>
#include <mosquitto.h>
#include <math.h>
#include <sys/time.h>
#include <string.h>
#include <stdint.h>

#include "mqttpub.h"
#include "interpolate.h"
#include "../shared/mqttpub.h"
#include "../shared/error.h"


void
mqtt_sendinfo (state_t * state)
{
  if (state->data == NULL)
    // Nothing to announce yet
    return;

  if (state->connstate != CS_CONNECTED)
    // Nobody is listening; will be done in mqttconnect_callback()
    return;

  mqttpubjson (state->mosq, state->topicprefix, "collect", state->name,
	       "info", NULL, json_pack ("{s:s}", "name", state->name));
}

void
mqtt_send_bindata (state_t * state)
{
  char topic[TOPICLEN + 5];
  json_t *jsend;
  json_t *jsend_axes;
  json_t *jsend_planes;
  char *jsend_str;
  int jsend_str_len;
  unsigned char *senddata;
  int xitems, yitems;
  int i;
  uint32_t magicheader = 0x0000DA7A;
  uint32_t json_header_len;
  int header_padding;
  double magic_double_7654321 = 765.3421;
  double magic_double_NAN = NAN;
  int outpos;
  int xstartbin, ystartbin;
  int x, y;

  if (state->data == NULL)
    // Nothing to announce yet
    return;


  if (state->do_interpolate)
    {
#ifndef INTERP_DATA_KEEP_ALLOCD
      state->interp_data = malloc ((state->data_xsize
				    * state->data_ysize
				    * state->interp_data_valuesperbin)
				   * sizeof (*state->interp_data));
      if (state->interp_data == NULL)
	{
	  STDERR ("malloc failed");
	  exit (1);
	}
#endif

      //quarters_extend_full(state);
      eighths_extend_full (state);

      interpolate_to_data (state);

#ifndef INTERP_DATA_KEEP_ALLOCD
      free (state->interp_data);
#endif
    }


  xitems = state->data_xused_max - state->data_xused_min + 1;
  yitems = state->data_yused_max - state->data_yused_min + 1;

  jsend_axes = json_array ();

  {
    json_t *jcopy = json_copy (state->jaxes[0]);
    json_t *jadd = json_pack ("{s:i, s:f}",
			      "items", xitems,
			      "firstitem",
			      state->data_xused_min * state->xbinsize);
    json_object_update (jcopy, jadd);
    json_decref (jadd);
    json_array_append_new (jsend_axes, jcopy);
  }

  {
    json_t *jcopy = json_copy (state->jaxes[1]);
    json_t *jadd = json_pack ("{s:i, s:f}",
			      "items", yitems,
			      "firstitem",
			      state->data_yused_min * state->ybinsize);
    json_object_update (jcopy, jadd);
    json_decref (jadd);
    json_array_append_new (jsend_axes, jcopy);
  }

  jsend_planes = json_array ();

  for (i = 0; i < state->jplanes_num; i++)
    {
      json_t *jcopy = json_copy (state->jplanes[i]);

      json_t *jadd = json_pack ("{s:s}", "itemformat", "double-x86");
      json_object_update (jcopy, jadd);
      json_decref (jadd);

      if (!strcmp
	  (json_string_value (json_object_get (jcopy, "label")),
	   "autogenerated:time"))
	{
	  struct timeval tv;
	  double dnow;
	  json_t *jadd;

	  gettimeofday (&tv, NULL);
	  dnow = (double) tv.tv_sec;
	  dnow += 0.000001 * (double) tv.tv_usec;

	  jadd = json_pack ("{s:f, s:f}",
			    "autogenerated:time-last", state->lasttime_d,
			    "autogenerated:time-now", dnow);
	  json_object_update (jcopy, jadd);
	  json_decref (jadd);
	}

      json_array_append_new (jsend_planes, jcopy);
    }

  jsend = json_pack ("{s:s, s:o, s:o, s:O, s:{s:f, s:f}}",
		     "format", "xy,data:interleaved",
		     "axes", jsend_axes,
		     "planes", jsend_planes,
		     "renders", state->jrenders,
		     "last-received",
		     "x", state->lastrecv_x, "y", state->lastrecv_y);
  jsend_str = json_dumps (jsend, JSON_INDENT (2));
  json_decref (jsend);

  if (jsend_str == NULL)
    return;

  jsend_str_len = strlen (jsend_str);
  json_header_len = ((jsend_str_len / sizeof (double)) + 1) * sizeof (double);
  header_padding = json_header_len - jsend_str_len;

  senddata = malloc (8
		     + json_header_len
		     + sizeof (double) * 2
		     + sizeof (double) * xitems * yitems * state->jplanes_num
		     + 1000);
  if (senddata == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  outpos = 0;
  memcpy (senddata + outpos, &magicheader, 4);
  outpos += 4;
  memcpy (senddata + outpos, &json_header_len, 4);
  outpos += 4;
  memcpy (senddata + outpos, jsend_str, jsend_str_len);
  outpos += jsend_str_len;
  memset (senddata + outpos, '\n', header_padding);
  outpos += header_padding;

  if (outpos != 8 + json_header_len)
    {
      STDERR ("calculation error, please fix code");
      exit (1);
    }

  memcpy (senddata + outpos, &magic_double_7654321, sizeof (double));
  outpos += sizeof (double);
  memcpy (senddata + outpos, &magic_double_NAN, sizeof (double));
  outpos += sizeof (double);


  // Assume used_min/max are always kept within allocated region.
  if (sizeof (double) != sizeof (*state->data))
    {
      STDERR ("data is not double type, please fix code");
      exit (1);
    }
  xstartbin = state->data_xused_min + state->data_xzerooffset;
  ystartbin = state->data_yused_min + state->data_yzerooffset;
  for (y = 0; y < yitems; y++)
    {
      if (state->do_interpolate)
	{
	  // Interpolated data is kept as additional interleaved values
	  for (x = 0; x < xitems; x++)
	    {
	      memcpy (senddata + outpos,
		      &state->data[((ystartbin + y) * state->data_xsize
				    + xstartbin +
				    x) * state->data_valuesperbin +
				   state->jplanes_num],
		      state->jplanes_num * sizeof (double));
	      outpos += state->jplanes_num * sizeof (double);
	    }
	}
      else
	{
	  // Linear copy for entire row (as far as useful)
	  memcpy (senddata + outpos,
		  &state->data[((ystartbin + y) * state->data_xsize
				+ xstartbin) * state->data_valuesperbin],
		  xitems * state->jplanes_num * sizeof (double));
	  outpos += xitems * state->jplanes_num * sizeof (double);
	}
    }

  if (outpos != 8 + json_header_len
      + (2 + xitems * yitems * state->jplanes_num) * sizeof (double))
    {
      STDERR ("calculation error, please fix code");
      exit (1);
    }


  i =
    snprintf (topic, TOPICLEN, "%s/collect/%s/bindata", state->topicprefix,
	      state->name);
  if (i >= TOPICLEN)
    {
      STDERR ("too long topic, increase TOPICLEN");
      exit (1);
    }

  mosquitto_publish (state->mosq, NULL, topic, outpos, senddata, 0, 0);

  free (senddata);
  free (jsend_str);
}
