// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <string.h>
#include <sys/time.h>
#include <math.h>

#include "receive_data.h"
#include "data.h"
#include "mqttpub.h"

#define LTYPE_AXIS      10
#define LTYPE_PLANE     11
#define LTYPE_CONDITION 12


// Returns 1: okay, match, proceed; *value is set for axis/plane (incl.
// exponential update).
// Returns 0: no match, stop immediately; *value is unchanged.
int
get_check_label (state_t * state, json_t * jinput, json_t * jselection,
		 int ltype, int xbin, int ybin, int planeidx, double *value)
{
  json_t *jvalue = NULL;
  double dvalue = NAN;

  {
    json_t *jlabel = json_object_get (jselection, "label");
    const char *labelname = json_string_value (jlabel);
    if (labelname == NULL)
      {
	if (ltype == LTYPE_AXIS || ltype == LTYPE_PLANE)
	  return 0;		// label required (already checked at startup)
	else
	  return 1;		// condition can be skipped if no label given
      }

    if (!strcmp (labelname, "autogenerated:time"))
      {
	struct timeval tv;
	gettimeofday (&tv, NULL);
	dvalue = (double) tv.tv_sec;
	dvalue += 0.000001 * (double) tv.tv_usec;
	state->lasttime_d = dvalue;

	jvalue = jlabel;	// fake for following checks
      }
    else
      {
	jvalue = json_object_get (jinput, labelname);
	if (json_is_number (jvalue))
	  dvalue = json_number_value (jvalue);
	// else dvalue stays NAN
      }
  }

  if (ltype == LTYPE_CONDITION)
    {
      int optional = 0;
      int forbidden = 0;

      const char *cp;
      cp = json_string_value (json_object_get (jselection, "presence"));
      if (cp != NULL && !strcmp (cp, "optional"))
	optional = 1;
      else if (cp != NULL && !strcmp (cp, "forbidden"))
	forbidden = 1;

      if (jvalue != NULL && forbidden)
	return 0;
      if (jvalue == NULL && forbidden)
	return 1;		// No further checks needed
      if (jvalue == NULL && optional)
	return 1;		// No further checks possible
    }

  // All further checks must at least have the json label present
  if (jvalue == NULL)
    return 0;

  // Axis/plane must always be numeric value
  if (isnan (dvalue) && (ltype == LTYPE_AXIS || ltype == LTYPE_PLANE))
    return 0;

  {
    json_t *jcond = json_object_get (jselection, "min");
    if (json_is_number (jcond))
      {
	double condval = json_number_value (jcond);
	if (isnan (dvalue) || dvalue < condval)
	  return 0;
      }
  }

  {
    json_t *jcond = json_object_get (jselection, "max");
    if (json_is_number (jcond))
      {
	double condval = json_number_value (jcond);
	if (isnan (dvalue) || dvalue > condval)
	  return 0;
      }
  }

  while (ltype == LTYPE_PLANE)	// Really "if" with easy "break" exit
    {
      double oldval;
      const char *repl =
	json_string_value (json_object_get (jselection, "replace"));

      if (repl == NULL)
	break;

      if (!strcmp (repl, "newest"))
	break;

      oldval = getdata (state, xbin, ybin, planeidx);
      if (isnan (oldval))
	break;			// Use new value immediately

      if (!strcmp (repl, "min"))
	{
	  if (dvalue > oldval)
	    return 0;
	  // When exactly equal, can still update other planes to newest value
	  else
	    break;
	}

      if (!strcmp (repl, "max"))
	{
	  if (dvalue < oldval)
	    return 0;
	  // When exactly equal, can still update other planes to newest value
	  else
	    break;
	}

      if (!strcmp (repl, "exponential"))
	{
	  double newfactor;
	  json_t *jnewfactor = json_object_get (jselection, "newfactor");

	  if (!json_is_number (jnewfactor))
	    break;		// just replace
	  newfactor = json_number_value (jnewfactor);
	  dvalue = newfactor * dvalue + (1. - newfactor) * oldval;
	  break;
	}

      break;
    }

  // Everything okay
  if (value != NULL)
    *value = dvalue;
  return 1;
}

void
receive_data (state_t * state, json_t * jinput)
{
  double x, y;
  int xbin, ybin;
  double planevals[state->jplanes_num];
  int i;
  int first_datapoint;

  if (!get_check_label (state, jinput, state->jaxes[0],
			LTYPE_AXIS, 0, 0, 0, &x))
    return;
  if (!get_check_label (state, jinput, state->jaxes[1],
			LTYPE_AXIS, 0, 0, 0, &y))
    return;

  xbin = lround (x / state->xbinsize);
  ybin = lround (y / state->ybinsize);

  for (i = 0; i < state->jplanes_num; i++)
    {
      if (!get_check_label (state, jinput, state->jplanes[i],
			    LTYPE_PLANE, xbin, ybin, i, &planevals[i]))
	return;
    }

  if (state->jconditions != NULL)
    {
      for (i = 0; i < json_array_size (state->jconditions); i++)
	{
	  json_t *jcond = json_array_get (state->jconditions, i);
	  if (!json_is_object (jcond))
	    continue;
	  if (!get_check_label (state, jinput, jcond,
				LTYPE_CONDITION, 0, 0, 0, NULL))
	    return;
	}
    }

  first_datapoint = (state->data == NULL);

  storedata_n (state, xbin, ybin, 0, state->jplanes_num, planevals);

  state->lastrecv_x = x;
  state->lastrecv_y = y;

  if (first_datapoint)
    mqtt_sendinfo (state);
}
