// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <stdio.h>
#include <string.h>
#include <jpeglib.h>
#include <jansson.h>

#include "process_image.h"
#include "../shared/mqttpub.h"


// 0 = error, 1 = okay
int
jpeg_getsize (state_t * state, char *filename, int *width, int *height)
{
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  FILE *infile;

  infile = fopen (filename, "rb");
  if (infile == NULL)
    {
      MDBG ("can't open %s\n", filename);
      return 0;
    }

  // NOTE: Any jpeg error will exit() immediately unless we handle specially
  // using longjmps etc.
  // See https://github.com/libjpeg-turbo/libjpeg-turbo/blob/master/example.txt

  cinfo.err = jpeg_std_error (&jerr);
  jpeg_create_decompress (&cinfo);

  jpeg_stdio_src (&cinfo, infile);

  jpeg_read_header (&cinfo, TRUE);

  *width = cinfo.image_width;
  *height = cinfo.image_height;

  jpeg_destroy_decompress (&cinfo);

  fclose (infile);

  return 1;
}


void
process_image (state_t * state, const struct mosquitto_message *message)
{
  char *cp;
  char *origfmtname;
  char pubtopic_base[TOPICLEN + 10];
  char filename_base[FNAMELEN + 10];
  char origfilename[FNAMELEN + 10];
  char outfilename[FNAMELEN + 10];
  char outfiletmp[FNAMELEN + 10];
  char outfilepub[FNAMELEN + 10];
  int res;
  FILE *f;
  int width, height;

  if (message->payloadlen > 10000000)
    {
      MDBG ("Payload too large: %d", message->payloadlen);
      return;
    }

  // Incoming topic like:
  //   scwips/cam/c0/img/orig/yuv4mpeg
  //   scwips/cam/c0/img/colfilt/pnm
  // Outgoing topic like:
  //   scwips/cam/c0/img/orig/jpeg/proc0/loc
  //   scwips/cam/c0/img/colfilt/jpeg/proc0/loc
  // Filename like:
  //   scwips_cam_c0_img_orig_yuv4mpeg-proc0.jpg
  //   scwips_cam_c0_img_colfilt_pnm-proc0.jpg

  if (strlen (message->topic) > TOPICLEN
      || strlen (state->imgpubdir) + strlen (message->topic)
      + strlen (state->name) > FNAMELEN - 100
      || strlen (state->imgpubpfx) + strlen (message->topic)
      + strlen (state->name) > FNAMELEN - 100)
    {
      MDBG ("topic too long, increase TOPICLEN and/or FNAMELEN");
      return;
    }

  strncpy (pubtopic_base, message->topic, TOPICLEN);
  pubtopic_base[TOPICLEN - 1] = '\0';
  cp = strrchr (pubtopic_base, '/');	// Find last topic level
  origfmtname = cp + 1;		// Save as format indicator
  if (cp != NULL)		// Strip from pubtopic
    *cp = '\0';

  strncpy (filename_base, message->topic, FNAMELEN);
  filename_base[FNAMELEN - 1] = '\0';
  for (cp = filename_base; *cp != '\0'; cp++)
    {
      if (*cp >= '0' && *cp <= '9')
	continue;
      if (*cp >= 'A' && *cp <= 'Z')
	continue;
      if (*cp >= 'a' && *cp <= 'z')
	continue;
      // else: replace char
      *cp = '_';
    }

  res =
    snprintf (origfilename, FNAMELEN, "%s/%s-%s.orig", state->imgpubdir,
	      filename_base, state->name_sane);
  if (res < 0 || res >= FNAMELEN)
    {
      MDBG ("Filename too long (1)");
      return;
    }

  res =
    snprintf (outfilename, FNAMELEN, "%s/%s-%s.jpg", state->imgpubdir,
	      filename_base, state->name_sane);
  if (res < 0 || res >= FNAMELEN)
    {
      MDBG ("Filename too long (3)");
      return;
    }

  res = snprintf (outfiletmp, FNAMELEN, "%s.tmp.jpg", outfilename);
  if (res < 0 || res >= FNAMELEN)
    {
      MDBG ("Filename too long (4)");
      return;
    }

  res =
    snprintf (outfilepub, FNAMELEN, "%s%s-%s.jpg", state->imgpubpfx,
	      filename_base, state->name_sane);
  if (res < 0 || res >= FNAMELEN)
    {
      MDBG ("Filename too long (5)");
      return;
    }

  // Dump image into origfile
  f = fopen (origfilename, "w");
  if (f == NULL)
    {
      MDBG ("Couldn't write to %s", origfilename);
      return;
    }
  fwrite (message->payload, message->payloadlen, 1, f);
  fclose (f);

  // Do processing depending on format to make jpeg output.
  // Could do scaling etc.; width/height are read back from output jpeg.

  setenv ("A", origfilename, 1);
  setenv ("B", outfiletmp, 1);

  if (!strcmp (origfmtname, "pnm"))
    {
      if (system ("cjpeg -quality 90 < \"$A\" > \"$B\"") != 0)
	{
	  MDBG ("Command cjpeg failed");
	  return;
	}
    }
  else if (!strcmp (origfmtname, "yuv4mpeg"))
    {
      if (system
	  ("ffmpeg -f yuv4mpegpipe -i \"$A\" -f image2 -vframes 1 -y \"$B\"")
	  != 0)
	{
	  MDBG ("Command ffmpeg failed");
	  return;
	}
    }
  else if (!strcmp (origfmtname, "yuv4mpegPOORMANS"))
    {
      char cmd[FNAMELEN + 10];

      //XXXTODO parse width/height --directly from payload memory
      width = 100;
      height = 100;

      // poor man's ffmpeg-less version...
      res =
	snprintf (cmd, FNAMELEN,
		  "( echo \"P5 %d %d 255\" ; tail --lines=+3 \"$A\" ) | cjpeg -quality 95 > \"$B\"",
		  width, 2 * height);
      if (res < 0 || res >= FNAMELEN)
	{
	  MDBG ("Command snprintf failed %d", res);
	  return;
	}
      if (system (cmd) != 0)
	{
	  MDBG ("Command echo|cjpeg failed");
	  return;
	}
    }
  else
    {
      MDBG ("Unknown format: %s", origfmtname);
      return;
    }

  rename (outfiletmp, outfilename);

  // Get width/height so that consumer can properly pre-set window/box size.
  width = 0;
  height = 0;
  if (jpeg_getsize (state, outfilename, &width, &height) != 1)
    {
      MDBG ("jpeg_getsize error");
      return;
    }

  mqttpubjson (state->mosq, pubtopic_base, "jpeg", state->name, "loc",
	       NULL,
	       json_pack ("{s:s,s:s,s:i,s:i}",
			  "file", outfilename, "url", outfilepub,
			  "pixw", width, "pixh", height));
}
