// Simple MQTT receiver that logs to stdout (i.e. >file)
// Also see --help.

// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#define PROGNAME  "mqttlogger"

#define _GNU_SOURCE		// for memmem()
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <mosquitto.h>

#include "../shared/version.h"
#include "../shared/error.h"
#include "../shared/mqttpub.h"

const char *progname = PROGNAME;

typedef struct
{
  struct mosquitto *mosq;

  char *svr_hostname;
  int svr_port;

  int maxtopiclen;
  int maxpayloadlen;
  char *shutdowntopic;

  char **topics;
  int topics_num;

  char *printbuf;		// malloc'd
  long printbuf_allocd;

  int connstate;
  time_t connect_nexttime;	/* Planned time of next connect attempt;
				   also stdin-EOF force-disconnect time */
} state_t;


void
print_help (FILE * f)
{
  fprintf (f, "\n"
	   "USAGE: %s [options] [--] topic [topics ...]  > logfile\n"
	   "\n"
	   "Simple MQTT logger to stdout.\n"
	   "\n"
	   "Options are:\n"
	   "    --host <name>        Hostname of mosquitto broker (default: localhost)\n"
	   "    --port <num>         Portname of mosquitto broker (default: 1883)\n"
	   "    --maxtopiclen <i>    Do not print topics longer than this (default: 10k)\n"
	   "    --maxpayloadlen <i>  Do not print payloads longer than this (default: 100k)\n"
	   "    --shutdowntopic <s>  Shutdown when receiving any message containing\n"
	   "                         \"shutdown\" (including quotes!) on this topic.\n"
	   "\n"
	   "    --help               Print this help\n"
	   "\n"
	   "Logs to stdout in this format:\n"
	   "\n"
	   "<timestamp>|<topic>|<payload>\\n\n"
	   "\n"
	   "Any non-printing byte, and also | and \\ and %%, will be replaced by \\xHH;\n"
	   "this includes sequences that could potentially be valid printable UTF-8.\n"
	   "This is usable as bash printf format string thanks to %% being escaped.\n"
	   "In real programming languages, printf will not expand backslashes (the compiler\n"
	   "will do, but in string constants only).\n"
	   "\n"
	   "Topics and payloads longer than given limits will be printed as \"(TOO-LONG)\".\n"
	   "\n"
	   "Reconnections to mqtt broker are handled gracefully.\n"
	   "DNS and connect are protected by alarm(), use external loop to restart.\n"
	   "\nPart of SCWIPS version %s\n\n", PROGNAME, SCWIPSVERSION);
}


/* Parse commandline parameters.
   Returns:
   0 = all okay, continue
   1 = immediately exit successfully (after --help)
   2 = immediately exit unsuccessfully
*/
int
parse_cmdline (int argc, char **argv, state_t * state)
{
  int i;

  // Rather simplistic & easy parameter parsing, adapted from earlier stuff
  for (i = 1; i < argc; i++)
    {
      if (!strcmp (argv[i], "--help"))
	{
	  print_help (stdout);
	  return 1;		// exit successfully
	}

      else if (!strcmp (argv[i], "--host"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --host requires argument");
	  state->svr_hostname = argv[i];
	}
      else if (!strcmp (argv[i], "--port"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --port requires argument");
	  state->svr_port = atoi (argv[i]);
	  if (state->svr_port < 1 || state->svr_port > 65535)
	    CMDLINEERR_RETURN2 ("parameter --port invalid value: %s",
				argv[i]);
	}
      else if (!strcmp (argv[i], "--maxtopiclen"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --maxtopiclen requires argument");
	  state->maxtopiclen = atoi (argv[i]);
	  if (state->maxtopiclen < 0)
	    CMDLINEERR_RETURN2 ("parameter --maxtopiclen invalid value: %s",
				argv[i]);
	}
      else if (!strcmp (argv[i], "--maxpayloadlen"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2
	      ("parameter --maxpayloadlen requires argument");
	  state->maxpayloadlen = atoi (argv[i]);
	  if (state->maxpayloadlen < 0)
	    CMDLINEERR_RETURN2 ("parameter --maxpayloadlen invalid value: %s",
				argv[i]);
	}
      else if (!strcmp (argv[i], "--shutdowntopic"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2
	      ("parameter --shutdowntopic requires argument");
	  state->shutdowntopic = argv[i];
	}

      else if (!strcmp (argv[i], "--"))
	{
	  i++;
	  break;
	}
      else if (argv[i][0] != '-')
	{
	  break;
	}

      else
	CMDLINEERR_RETURN2 ("unknown parameter: %s", argv[i]);
    }

  // The rest of arguments are supposed to be subscription topics.
  state->topics = argv + i;
  state->topics_num = argc - i;

  if (state->topics_num <= 0)
    CMDLINEERR_RETURN2 ("please specify subscription topics");

  // All should be okay now
  return 0;
}


// Returns pointer to char following date string
char *
add_time (char *bufp)
{
  struct timespec tspec = { 0, 0 };
  struct tm *tmp;
  int res;

  res = clock_gettime (CLOCK_REALTIME, &tspec);
  if (res == -1)
    {
      STDERR ("clock_gettime failed: %s", strerror (errno));
      exit (1);
    }

  tmp = localtime (&tspec.tv_sec);
  if (tmp == NULL)
    {
      STDERR ("localtime failed: %s", strerror (errno));
      exit (1);
    }

  res = strftime (bufp, 50, "%Y-%m-%dT%H:%M:%S.", tmp);
  if (res <= 0 || res >= 50)
    {
      STDERR ("strftime returned %d", res);
      exit (1);
    }
  bufp += res;

  res = snprintf (bufp, 50, "%09ld", tspec.tv_nsec);
  if (res <= 0 || res >= 50)
    {
      STDERR ("snprintf nanoseconds returned %d", res);
      exit (1);
    }
  bufp += res;

  return bufp;
}

// Returns pointer to char following escaped string
char *
add_escapedstr (char *bufp, unsigned char *inp, int inlen)
{
  int i;
  unsigned char *bufp_u;
  char *hexchar = "0123456789ABCDEF";

  bufp_u = (unsigned char *) bufp;

  for (i = 0; i < inlen; i++, inp++)
    {
      if ((*inp >= ' ' && *inp <= '$') ||
	  (*inp >= '&' && *inp <= '[') ||
	  (*inp >= ']' && *inp <= '{') || (*inp >= '}' && *inp <= '~'))
	{
	  *bufp_u = *inp;
	  bufp_u++;
	}
      else
	{
	  *bufp_u = '\\';
	  bufp_u++;
	  *bufp_u = 'x';
	  bufp_u++;
	  *bufp_u = hexchar[(*inp >> 4) & 0x0F];
	  bufp_u++;
	  *bufp_u = hexchar[*inp & 0x0F];
	  bufp_u++;
	}
    }

  return (char *) bufp_u;
}


void
msg_to_stdout (state_t * state, const struct mosquitto_message *message)
{
  int topiclen;
  int requiredlen;
  char *bufp;
  int res;

  requiredlen = 100;		// timestamp

  topiclen = strlen (message->topic);
  if (topiclen <= state->maxtopiclen)
    requiredlen += topiclen * 4;	// \xHH worstcase
  else
    requiredlen += 100;

  if (message->payloadlen <= state->maxpayloadlen)
    requiredlen += message->payloadlen * 4;	// \xHH worstcase
  else
    requiredlen += 100;

  if (requiredlen > state->printbuf_allocd)
    {
      state->printbuf_allocd *= 2;	// efficient case
      if (requiredlen > state->printbuf_allocd)
	state->printbuf_allocd = requiredlen;	// still not enough

      // No realloc() since we don't need copying of contents.
      if (state->printbuf != NULL)
	free (state->printbuf);
      state->printbuf = malloc (state->printbuf_allocd);
      if (state->printbuf == NULL)
	{
	  STDERR ("malloc failed");
	  exit (1);
	}
    }

  // Now we know that printbuf is large enough to hold everything,
  // no further checks needed.

  bufp = state->printbuf;

  bufp = add_time (bufp);

  *bufp = '|';
  bufp++;

  if (topiclen <= state->maxtopiclen)
    bufp = add_escapedstr (bufp, (unsigned char *) message->topic, topiclen);
  else
    {
      res = snprintf (bufp, 50, "(TOO-LONG)");
      if (res <= 0 || res >= 50)
	{
	  STDERR ("snprintf returned %d", res);
	  exit (1);
	}
      bufp += res;
    }

  *bufp = '|';
  bufp++;

  if (message->payloadlen <= state->maxpayloadlen)
    bufp =
      add_escapedstr (bufp, (unsigned char *) message->payload,
		      message->payloadlen);
  else
    {
      res = snprintf (bufp, 50, "(TOO-LONG)");
      if (res <= 0 || res >= 50)
	{
	  STDERR ("snprintf returned %d", res);
	  exit (1);
	}
      bufp += res;
    }

  *bufp = '\n';
  bufp++;

  // Use unbuffered write, outputfile will still go via writecache.
  write (1, state->printbuf, bufp - state->printbuf);
}


/* This connect callback is called when CONNACK is received, i.e. when
   the broker has actively allowed us access.
*/
void
mqttconnect_callback (struct mosquitto *mosq, void *usrobj, int result)
{
  state_t *state = usrobj;
  int i;

  if (result != 0)
    {
      // Access actively refused.
      // Socket will be closed by library via
      // mosquitto_loop_read() -> _mosquitto_loop_rc_handle()
      return;
    }

  state->connstate = CS_CONNECTED;

  // Send subscriptions
  if (state->shutdowntopic != NULL)
    mosquitto_subscribe (state->mosq, NULL, state->shutdowntopic, 0);

  for (i = 0; i < state->topics_num; i++)
    mosquitto_subscribe (state->mosq, NULL, state->topics[i], 0);
}

/* This disconnect callback is called whenever socket is (already) closed
   for whatever reason.
*/
void
mqttdisconnect_callback (struct mosquitto *mosq, void *usrobj, int reason)
{
  state_t *state = usrobj;

  // We are disconnected already, can't send any message.
  // But if we wanted disconnection, we already sent "shutdown".
  // And if we got disconnected without our request, the broker will
  // already have sent our Last Will = position-lost message.

  if (state->connstate == CS_EOFDISCONNECTING
      || state->connstate == CS_EOFDISCONNECTED)
    {
      state->connstate = CS_EOFDISCONNECTED;
    }
  else
    {
      state->connstate = CS_UNCONNECTED;
      state->connect_nexttime = time (NULL) + 5;
    }
}

/* This message callback is called whenever a message is received; may be 
   called _multiple_ times from 1 mosquitto_loop_read() call.
*/
void
mqttmessage_callback (struct mosquitto *mosq, void *usrobj,
		      const struct mosquitto_message *message)
{
  state_t *state = usrobj;

  // We are non-threaded, and all messages can always be handled immediately
  // when they arrive, no queueing needed. So we just handle the message
  // right here inside the callback.

  msg_to_stdout (state, message);

  // Topic is zero-terminated, but nessages are NOT.
  if (state->shutdowntopic != NULL &&
      !strcmp (message->topic, state->shutdowntopic) &&
      message->payloadlen >= 10 &&
      memmem (message->payload, message->payloadlen, "\"shutdown\"",
	      10) != NULL)
    {
      // Initiate mqtt disconnect, will lead to shutdown after disconnected
      // by broker, or after timeout; handled in mainloop().

      if (state->connstate == CS_CONNECTED)
	{
	  state->connstate = CS_EOFDISCONNECTING;
	  state->connect_nexttime = time (NULL) + 5;
	  mosquitto_disconnect (state->mosq);	// just queues DISCONNECT
	}
      else if (state->connstate != CS_EOFDISCONNECTING
	       && state->connstate != CS_EOFDISCONNECTED)
	{
	  // Not connected at all, so sending DISCONNECT would be
	  // useless.
	  // (This should not happen in practice, since we are handling
	  // the shutdown command that arrived via mqtt.)
	  state->connstate = CS_EOFDISCONNECTED;
	  // This will cause loop break immediately in next round.
	}
    }

  // *message is free()'d by the library after this callback returns.
}


void
mainloop (state_t * state)
{
  time_t now;
  fd_set readfds, writefds;
  int mosqsock, maxfd;
  struct timeval tv;
  int selectres;

  while (1)
    {
      alarm (5);

      now = time (NULL);

      // Server has disconnected after we asked
      if (state->connstate == CS_EOFDISCONNECTED)
	{
	  break;
	}

      // Server did not answer after we asked DISCONNECT
      // (Test: suspend broker with Ctrl-Z just before DISCONNECT.)
      if (state->connstate == CS_EOFDISCONNECTING
	  && now >= state->connect_nexttime)
	{
	  break;
	}

      // If disconnected without our asking, attempt to reconnect periodically
      // (=also immediately on startup).
      if (state->connstate == CS_UNCONNECTED
	  && now >= state->connect_nexttime)
	{
	  state->connstate = CS_CONNECTING;

	  if (mosquitto_connect
	      (state->mosq, state->svr_hostname, state->svr_port,
	       10) != MOSQ_ERR_SUCCESS)
	    {
	      // Failed, try again after some time
	      now = time (NULL);	// DNS/connect may have taken time
	      state->connstate = CS_UNCONNECTED;
	      state->connect_nexttime = now + 5;
	      continue;
	    }

	  // mosquitto_connect() returns after TCP connect() finished, with
	  // CONNECT packet queued to be sent by next mosquitto_loop_write().
	  // DNS resolving and connect() is protected by alarm() above.

	  // We keep ->connstate in CS_CONNECTING until CONNACK received
	  // via mqttconnect_callback().

	  // CONNACK is waited for up to keepalive time, checked in
	  // _mosquitto_check_keepalive() via mosquitto_loop_misc().
	  // (state->mosq->state is not cs_connected yet; _no_ pings are sent.)
	  // Waiting-for-CONNACK is easily testable: Ctrl-S on broker.

	  // DNS/connect may have taken long, need new "now" value
	  continue;
	}

      FD_ZERO (&readfds);
      FD_ZERO (&writefds);
      maxfd = 0;

      mosqsock = mosquitto_socket (state->mosq);
      if (mosqsock != -1)
	{
	  FD_SET (mosqsock, &readfds);
	  if (mosquitto_want_write (state->mosq) == true)
	    FD_SET (mosqsock, &writefds);
	  if (mosqsock > maxfd)
	    maxfd = mosqsock;
	}

      tv.tv_sec = 1;
      tv.tv_usec = 0;

      selectres = select (maxfd + 1, &readfds, &writefds, NULL, &tv);
      // We could handle errno==EINTR, but shouldn't occur in practice.
      if (selectres == -1)
	break;			// Stop on any error

      // Handle mqtt network input and output.

      if (mosqsock != -1 && FD_ISSET (mosqsock, &readfds))
	{
	  mosquitto_loop_read (state->mosq, 1);
	  // Most callbacks are called from here, including
	  // mqttmessage_callback() which handles any incoming messages
	  // (possibly multiple, if piled up in tcp buffer),
	  // and starts new foreground communication if possible.

	  // Don't mind exitcode; socket already disconnected in case of
	  // trouble, then also mqttdisconnect_callback() was called.
	}

      if (mosqsock != -1 && FD_ISSET (mosqsock, &writefds))
	{
	  mosquitto_loop_write (state->mosq, 1);
	}

      mosquitto_loop_misc (state->mosq);

    }

  // This is reached only after our explicit shutdown.
  // If we were connected, DISCONNECT has been sent from inside the loop,
  // and broker hopefully has neatly disconnected us within our timeout.
  // (Else: don't care, we'll close the socket anyway.)

  //mosquitto_destroy() -> in main

  return;
}


int
main (int argc, char **argv)
{
  state_t state;
  int res;
  char clientid[MOSQ_MQTT_ID_MAX_LENGTH + 5];
  char hostname[TOPICLEN + 5];

  alarm (5);

  memset (&state, 0, sizeof (state));

  // Set default configuration
  state.svr_hostname = "127.0.0.1";
  state.svr_port = 1883;
  state.maxtopiclen = 10240;
  state.maxpayloadlen = 102400;

  res = parse_cmdline (argc, argv, &state);
  if (res >= 1)
    exit (res - 1);

  // Note that broker kicks any connected client with equal clientid.
  // Clientid is max 23 chars.
  if (gethostname (hostname, TOPICLEN) != 0)
    hostname[0] = '\0';
  res =
    snprintf (clientid, MOSQ_MQTT_ID_MAX_LENGTH, "%d/%s", getpid (),
	      hostname);
  // Fake resultvalue handling to make gcc stop complaining about the
  // intended truncation, which will actually always have res LARGER than 1.
  if (res < 1)
    exit (1);

  mosquitto_lib_init ();	// cannot return errors

  state.mosq = mosquitto_new (clientid, true, &state);
  if (state.mosq == NULL)
    {
      STDERR ("mosquitto_new failed: %s", strerror (errno));
      return 1;
    }

  mosquitto_connect_callback_set (state.mosq, mqttconnect_callback);
  mosquitto_disconnect_callback_set (state.mosq, mqttdisconnect_callback);
  mosquitto_message_callback_set (state.mosq, mqttmessage_callback);

  mainloop (&state);
  // Returns after "shutdown"

  mosquitto_destroy (state.mosq);
  // This is the _only_ way an mqtt client can forcibly close the socket.

  return 0;
}
