// Data position annotator JSON/MQTT -> JSON/MQTT
// Also see --help.

// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#define PROGNAME  "mqttpositioner"

#include <string.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/timerfd.h>
#include <mosquitto.h>
#include <jansson.h>

#include "../shared/version.h"
#include "../shared/error.h"
#include "../shared/mqttpub.h"

const char *progname = PROGNAME;

typedef struct
{
  struct mosquitto *mosq;

  char *svr_hostname;
  int svr_port;
  int connstate;
  time_t connect_nexttime;	/* Planned time of next connect attempt;
				   also stdin-EOF force-disconnect time */
  char *positiontopic;
  char *datatopic;
  char *pubtopic;
  char *shutdowntopic;
  char *must_have_key;
  char *must_not_have_key;

  json_t *jdata;

  struct itimerspec zero_itimval;
  struct itimerspec datavalid_itimval;
  int datavalid_timerfd;

} state_t;


void
print_help (FILE * f)
{
  fprintf (f, "\n"
	   "USAGE: %s [options]\n"
	   "\n"
	   "Combine MQTT JSON position stream and JSON data stream into one.\n"
	   "\n"
	   "Options are:\n"
	   "    --host <name>        Hostname of mosquitto broker (default: localhost)\n"
	   "    --port <num>         Portname of mosquitto broker (default: 1883)\n"
	   "    --positiontopic <s>  Receive position from this topic.\n"
	   "    --datatopic <s>      Receive data from this topic.\n"
	   "    --pubtopic <s>       Publish position-annotated data to this topic.\n"
	   "    --shutdowntopic <s>  Shutdown when received \"cmd\":\"shutdown\" on this topic.\n"
	   "    --must-have-key <s>  Data is only accepted if it has this json key.\n"
	   "    --must-not-have-key <s>  Data is not accepted if it has this json key.\n"
	   "\n"
	   "    --help               Print this help\n"
	   "\n"
	   "Republishes any data with a position annotation.\n"
	   "Data is buffered at most one second waiting for next position fix.\n"
	   "Any received position fix is either used to publish the last buffered data,\n"
	   "or discarded immediately when no data was buffered.\n"
	   "Data and position must be valid JSON objects. Only \"x\", \"y\", \"z\" numeric\n"
	   "values will be inserted into data object, or overwritten if already present.\n"
	   "Any one of \"x\", \"y\", \"z\" will suffice to post, otherwise data is discarded.\n"
	   "\n"
	   "Reconnections to mqtt broker are handled gracefully.\n"
	   "DNS and connect are protected by alarm(), use external loop to restart.\n"
	   "\nPart of SCWIPS version %s\n\n", PROGNAME, SCWIPSVERSION);
}


/* Parse commandline parameters.
   Returns:
   0 = all okay, continue
   1 = immediately exit successfully (after --help)
   2 = immediately exit unsuccessfully
*/
int
parse_cmdline (int argc, char **argv, state_t * state)
{
  int i;

  // Rather simplistic & easy parameter parsing, adapted from earlier stuff
  for (i = 1; i < argc; i++)
    {
      if (!strcmp (argv[i], "--help"))
	{
	  print_help (stdout);
	  return 1;		// exit successfully
	}

      else if (!strcmp (argv[i], "--host"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --host requires argument");
	  state->svr_hostname = argv[i];
	}
      else if (!strcmp (argv[i], "--port"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --port requires argument");
	  state->svr_port = atoi (argv[i]);
	  if (state->svr_port < 1 || state->svr_port > 65535)
	    CMDLINEERR_RETURN2 ("parameter --port invalid value: %s",
				argv[i]);
	}
      else if (!strcmp (argv[i], "--positiontopic"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2
	      ("parameter --positiontopic requires argument");
	  state->positiontopic = argv[i];
	}
      else if (!strcmp (argv[i], "--datatopic"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --datatopic requires argument");
	  state->datatopic = argv[i];
	}
      else if (!strcmp (argv[i], "--pubtopic"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --pubtopic requires argument");
	  state->pubtopic = argv[i];
	}
      else if (!strcmp (argv[i], "--shutdowntopic"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2
	      ("parameter --shutdowntopic requires argument");
	  state->shutdowntopic = argv[i];
	}
      else if (!strcmp (argv[i], "--must-have-key"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2
	      ("parameter --must-have-key requires argument");
	  state->must_have_key = argv[i];
	}
      else if (!strcmp (argv[i], "--must-not-have-key"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2
	      ("parameter --must-not-have-key requires argument");
	  state->must_not_have_key = argv[i];
	}

      else
	CMDLINEERR_RETURN2 ("unknown parameter: %s", argv[i]);
    }

  if (state->positiontopic == NULL)
    CMDLINEERR_RETURN2 ("please specify --positiontopic");
  if (state->datatopic == NULL)
    CMDLINEERR_RETURN2 ("please specify --datatopic");
  if (state->pubtopic == NULL)
    CMDLINEERR_RETURN2 ("please specify --pubtopic");

  // All should be okay now
  return 0;
}


void
recv_data (state_t * state, json_t * jroot)
{
  if (state->jdata != NULL)
    json_decref (state->jdata);
  state->jdata = NULL;

  // Stop invalidation timeout.
  // This is safe: will drop any pending select() triggers.
  timerfd_settime (state->datavalid_timerfd, 0, &state->zero_itimval, NULL);

  // Only *objects* are accepted (else can't add position)
  if (!json_is_object (jroot))
    return;

  if (state->must_have_key != NULL)
    {
      if (json_object_get (jroot, state->must_have_key) == NULL)
	return;
    }

  if (state->must_not_have_key != NULL)
    {
      if (json_object_get (jroot, state->must_not_have_key) != NULL)
	return;
    }

  // All okay; remember this data for publishing once position received
  state->jdata = json_incref (jroot);

  // Start invalidation timeout.
  timerfd_settime (state->datavalid_timerfd, 0,
		   &state->datavalid_itimval, NULL);
}

void
pub_positiondata (state_t * state, json_t * jposition)
{
  int okay = 0;
  json_t *jval;

  if (state->jdata == NULL)
    // No pending data to post
    return;

  jval = json_object_get (jposition, "x");
  if (json_is_number (jval))
    {
      json_object_set (state->jdata, "x", jval);
      okay = 1;
    }

  jval = json_object_get (jposition, "y");
  if (json_is_number (jval))
    {
      json_object_set (state->jdata, "y", jval);
      okay = 1;
    }

  jval = json_object_get (jposition, "z");
  if (json_is_number (jval))
    {
      json_object_set (state->jdata, "z", jval);
      okay = 1;
    }

  if (okay)
    {
      char *jstring;
      jstring = json_dumps (state->jdata, JSON_INDENT (2));
      if (jstring != NULL)
	{
	  mosquitto_publish (state->mosq, NULL, state->pubtopic,
			     strlen (jstring), jstring, 0, 0);
	  free (jstring);
	}
    }

  json_decref (state->jdata);
  state->jdata = NULL;
}


/* This connect callback is called when CONNACK is received, i.e. when
   the broker has actively allowed us access.
*/
void
mqttconnect_callback (struct mosquitto *mosq, void *usrobj, int result)
{
  state_t *state = usrobj;

  if (result != 0)
    {
      // Access actively refused.
      // Socket will be closed by library via
      // mosquitto_loop_read() -> _mosquitto_loop_rc_handle()
      return;
    }

  state->connstate = CS_CONNECTED;

  // Send subscriptions
  mosquitto_subscribe (state->mosq, NULL, state->positiontopic, 0);
  mosquitto_subscribe (state->mosq, NULL, state->datatopic, 0);
  if (state->shutdowntopic != NULL)
    mosquitto_subscribe (state->mosq, NULL, state->shutdowntopic, 0);
}

/* This disconnect callback is called whenever socket is (already) closed
   for whatever reason.
*/
void
mqttdisconnect_callback (struct mosquitto *mosq, void *usrobj, int reason)
{
  state_t *state = usrobj;

  // We are disconnected already, can't send any message.
  // But if we wanted disconnection, we already sent "shutdown".
  // And if we got disconnected without our request, the broker will
  // already have sent our Last Will = position-lost message.

  if (state->connstate == CS_EOFDISCONNECTING
      || state->connstate == CS_EOFDISCONNECTED)
    {
      state->connstate = CS_EOFDISCONNECTED;
    }
  else
    {
      state->connstate = CS_UNCONNECTED;
      state->connect_nexttime = time (NULL) + 5;
    }
}

/* This message callback is called whenever a message is received; may be 
   called _multiple_ times from 1 mosquitto_loop_read() call.
*/
void
mqttmessage_callback (struct mosquitto *mosq, void *usrobj,
		      const struct mosquitto_message *message)
{
  state_t *state = usrobj;
  json_t *jroot;
  json_error_t jerror;

  // We are non-threaded, and all messages can always be handled immediately
  // when they arrive, no queueing needed. So we just handle the message
  // right here inside the callback.

  // All messages are json, non-json silently ignored
  jroot = json_loadb (message->payload, message->payloadlen, 0, &jerror);
  if (jroot == NULL)
    return;

  do
    {
      // Easy "break" for decref & return

      if (!strcmp (message->topic, state->datatopic))
	{
	  recv_data (state, jroot);
	  break;
	}

      if (!strcmp (message->topic, state->positiontopic))
	{
	  pub_positiondata (state, jroot);
	  break;
	}

      if (state->shutdowntopic != NULL &&
	  !strcmp (message->topic, state->shutdowntopic))
	{
	  const char *cmd;
	  cmd = json_string_value (json_object_get (jroot, "cmd"));
	  if (cmd == NULL)
	    break;
	  if (!strcmp (cmd, "shutdown"))
	    {
	      // Initiate mqtt disconnect, will lead to shutdown after disconnected
	      // by broker, or after timeout; handled in mainloop().

	      if (state->connstate == CS_CONNECTED)
		{
		  state->connstate = CS_EOFDISCONNECTING;
		  state->connect_nexttime = time (NULL) + 5;
		  mosquitto_disconnect (state->mosq);	// just queues DISCONNECT
		}
	      else if (state->connstate != CS_EOFDISCONNECTING
		       && state->connstate != CS_EOFDISCONNECTED)
		{
		  // Not connected at all, so sending DISCONNECT would be
		  // useless.
		  // (This should not happen in practice, since we are handling
		  // the shutdown command that arrived via mqtt.)
		  state->connstate = CS_EOFDISCONNECTED;
		  // This will cause loop break immediately in next round.
		}
	    }
	  break;
	}

    }
  while (0);

  json_decref (jroot);
  // *message is free()'d by the library after this callback returns.
}


void
mainloop (state_t * state)
{
  time_t now;
  fd_set readfds, writefds;
  int mosqsock, maxfd;
  struct timeval tv;
  int selectres;

  while (1)
    {
      alarm (5);

      now = time (NULL);

      // Server has disconnected after we asked
      if (state->connstate == CS_EOFDISCONNECTED)
	{
	  break;
	}

      // Server did not answer after we asked DISCONNECT
      // (Test: suspend broker with Ctrl-Z just before DISCONNECT.)
      if (state->connstate == CS_EOFDISCONNECTING
	  && now >= state->connect_nexttime)
	{
	  break;
	}

      // If disconnected without our asking, attempt to reconnect periodically
      // (=also immediately on startup).
      if (state->connstate == CS_UNCONNECTED
	  && now >= state->connect_nexttime)
	{
	  state->connstate = CS_CONNECTING;

	  if (mosquitto_connect
	      (state->mosq, state->svr_hostname, state->svr_port,
	       10) != MOSQ_ERR_SUCCESS)
	    {
	      // Failed, try again after some time
	      now = time (NULL);	// DNS/connect may have taken time
	      state->connstate = CS_UNCONNECTED;
	      state->connect_nexttime = now + 5;
	      continue;
	    }

	  // mosquitto_connect() returns after TCP connect() finished, with
	  // CONNECT packet queued to be sent by next mosquitto_loop_write().
	  // DNS resolving and connect() is protected by alarm() above.

	  // We keep ->connstate in CS_CONNECTING until CONNACK received
	  // via mqttconnect_callback().

	  // CONNACK is waited for up to keepalive time, checked in
	  // _mosquitto_check_keepalive() via mosquitto_loop_misc().
	  // (state->mosq->state is not cs_connected yet; _no_ pings are sent.)
	  // Waiting-for-CONNACK is easily testable: Ctrl-S on broker.

	  // DNS/connect may have taken long, need new "now" value
	  continue;
	}

      FD_ZERO (&readfds);
      FD_ZERO (&writefds);
      maxfd = 0;

      mosqsock = mosquitto_socket (state->mosq);
      if (mosqsock != -1)
	{
	  FD_SET (mosqsock, &readfds);
	  if (mosquitto_want_write (state->mosq) == true)
	    FD_SET (mosqsock, &writefds);
	  if (mosqsock > maxfd)
	    maxfd = mosqsock;
	}

      FD_SET (state->datavalid_timerfd, &readfds);
      if (state->datavalid_timerfd > maxfd)
	maxfd = state->datavalid_timerfd;

      tv.tv_sec = 1;
      tv.tv_usec = 0;

      selectres = select (maxfd + 1, &readfds, &writefds, NULL, &tv);
      // We could handle errno==EINTR, but shouldn't occur in practice.
      if (selectres == -1)
	break;			// Stop on any error

      // Handle mqtt network input and output.

      if (mosqsock != -1 && FD_ISSET (mosqsock, &readfds))
	{
	  mosquitto_loop_read (state->mosq, 1);
	  // Most callbacks are called from here, including
	  // mqttmessage_callback() which handles any incoming messages
	  // (possibly multiple, if piled up in tcp buffer),
	  // and starts new foreground communication if possible.

	  // Don't mind exitcode; socket already disconnected in case of
	  // trouble, then also mqttdisconnect_callback() was called.
	}

      if (mosqsock != -1 && FD_ISSET (mosqsock, &writefds))
	{
	  mosquitto_loop_write (state->mosq, 1);
	}

      mosquitto_loop_misc (state->mosq);

      if (FD_ISSET (state->datavalid_timerfd, &readfds))
	{
	  // Invalidate any data
	  if (state->jdata != NULL)
	    {
	      json_decref (state->jdata);
	      state->jdata = NULL;
	    }

	  // _settime will drop pending triggers, no need to read()
	  timerfd_settime (state->datavalid_timerfd, 0,
			   &state->zero_itimval, NULL);
	}

    }

  // This is reached only after our explicit shutdown.
  // If we were connected, DISCONNECT has been sent from inside the loop,
  // and broker hopefully has neatly disconnected us within our timeout.
  // (Else: don't care, we'll close the socket anyway.)

  //mosquitto_destroy() -> in main

  return;
}


int
main (int argc, char **argv)
{
  state_t state;
  int res;
  char clientid[MOSQ_MQTT_ID_MAX_LENGTH + 5];
  char hostname[TOPICLEN + 5];

  alarm (5);

  memset (&state, 0, sizeof (state));

  // Set default configuration
  state.svr_hostname = "127.0.0.1";
  state.svr_port = 1883;

  res = parse_cmdline (argc, argv, &state);
  if (res >= 1)
    exit (res - 1);

  // Note that broker kicks any connected client with equal clientid.
  // Clientid is max 23 chars.
  if (gethostname (hostname, TOPICLEN) != 0)
    hostname[0] = '\0';
  res =
    snprintf (clientid, MOSQ_MQTT_ID_MAX_LENGTH, "%d/%s", getpid (),
	      hostname);
  // Fake resultvalue handling to make gcc stop complaining about the
  // intended truncation, which will actually always have res LARGER than 1.
  if (res < 1)
    exit (1);

  mosquitto_lib_init ();	// cannot return errors

  state.mosq = mosquitto_new (clientid, true, &state);
  if (state.mosq == NULL)
    {
      STDERR ("mosquitto_new failed: %s", strerror (errno));
      return 1;
    }

  mosquitto_connect_callback_set (state.mosq, mqttconnect_callback);
  mosquitto_disconnect_callback_set (state.mosq, mqttdisconnect_callback);
  mosquitto_message_callback_set (state.mosq, mqttmessage_callback);

  // Init timer
  state.datavalid_timerfd = timerfd_create (CLOCK_MONOTONIC, 0);
  if (state.datavalid_timerfd == -1)
    {
      STDERR ("failed to create timerfd: %s", strerror (errno));
      exit (1);
    }

  state.zero_itimval.it_value.tv_sec = 0;
  state.zero_itimval.it_value.tv_nsec = 0;
  state.zero_itimval.it_interval.tv_sec = 0;
  state.zero_itimval.it_interval.tv_nsec = 0;

  // Timeout to invalidate data if no position was reported
  state.datavalid_itimval.it_value.tv_sec = 1;	// 1 second
  state.datavalid_itimval.it_value.tv_nsec = 0;
  state.datavalid_itimval.it_interval.tv_sec = 0;
  state.datavalid_itimval.it_interval.tv_nsec = 0;

  mainloop (&state);
  // Returns after "shutdown"

  mosquitto_destroy (state.mosq);
  // This is the _only_ way an mqtt client can forcibly close the socket.

  return 0;
}
