// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/timerfd.h>

#include "state.h"
#include "calc.h"
#include "../shared/error.h"
#include "../shared/mqttpub.h"
#include "../camfuncs/camfuncs.h"
#include "../camfuncs/camhelpers.h"
#include "../cmpfit/mpfit.h"


// Note: *MASSIVE* output.
// MDBG is not useful: everything will get buffered, only sent (all at once)
// after calc finished.
//#define FITDBG(args...)  MDBG(args)
//#define FITDBG(args...)  STDERR(args)
#define FITDBG(args...)  do { } while(0)

// For easily commenting out individual MDBG's: replace by noMDBG
#define noMDBG(args...)  do { } while(0)


// Duplicate observations when one cam is really close to another cam
double dist_duplicate_min = 0.30;	// full duplicate within this distance
double dist_duplicate_max = 0.50;	// no duplicate outside this distance

// Duplicate observations when object is almost on cam-to-cam line
double oneline_duplicate_min = 0.060;	// -1..+1 scale, 1 pixel is ~0.003
double oneline_duplicate_max = 0.120;



typedef struct
{
  char *name;			// mallocd!
  int haveinfo;			// valid info received (0 = useless cam)
  int inscene;			// valid scene result (0 = useless cam)

  int camtype;			// index to camtype[]
  double *typeparam;		// malloc'd for specified type
  int camdistortion;		// index to camdistortion[]
  double *distortionparam;	// malloc'd for specified distortion
  double dist_lin_radius;	// distortion linearized beyond this limit

  int w;			// camera #pixels
  int h;
  double xscale;		// to convert pixel into -1..+1 coords
  double xoffset;
  double yscale;
  double yoffset;

  double posrot[6];		// position + rotation: x,y,z,a,b,g

  int curvalid;			// current observation is valid
  double curx;			// current observation, -1..+1 coords
  double cury;
  double residscalex;		// 1/(pixelsize * #blinkpix)
  double residscaley;
  double backcalcx;		// where current pixel "should have been"
  double backcalcy;
  double duplicates;
  double residscale;		// = 1/duplicates

  int pixinvalid_timerfd;
} cam_t;

// calc_campix requires cam_t to be defined
#include "../shared/calc_campix.h"

// parse_cam_json requires cam_t to be defined
#define CAMERR_RETURN(args...)  do { MDBG(args); return 1; } while(0)
#include "../shared/parse_cam_json.h"
#undef CAMERR_RETURN


typedef struct
{
  double x;			// campix2cam[N][M] is where cam[N] sees [M]
  double y;
  double dist;
  double dist_duplicate;
} cam2cam_t;


cam_t *cam = NULL;		// mallocd
int cam_allocd = 0;
int cam_used = 0;
int cams = 0;

cam2cam_t *cam2cam = NULL;	// mallocd
int cam2cam_n = 0;

double curpos[3];		// current object position x/y/z


// Used on scwips/cam/c0/info announce, and for scenecalc objects
int
find_make_cam (char *name)
{
  int i;

  for (i = 0; i < cam_used; i++)
    {
      if (cam[i].name != NULL)
	if (!strcmp (cam[i].name, name))
	  return i;
    }

  // If still here, then not found: add it.

  if (cam_used >= cam_allocd)
    {
      cam_allocd = cam_used * 2;
      if (cam_allocd < 100)
	cam_allocd = 100;
      cam = realloc (cam, sizeof (*cam) * cam_allocd);
      if (cam == NULL)
	{
	  // Now we have a big problem. Can't assume mqtt will still work.
	  STDERR ("realloc failed");
	  exit (1);
	}
      // memset(0) is done below; is required for re-use as well.
    }

  memset (&cam[cam_used], 0, sizeof (cam[cam_used]));
  cam[cam_used].name = strdup (name);	// =malloc
  cam[cam_used].pixinvalid_timerfd = -1;

  cams = cam_used + 1;

  cam_used++;
  return cam_used - 1;
}

// Used for scwips/cam/c0/pix : camera must already be known
int
find_cam (char *name)
{
  int i;

  for (i = 0; i < cams; i++)
    {
      if (cam[i].name != NULL)
	if (!strcmp (cam[i].name, name))
	  return i;
    }
  MDBG ("name %s is not known, skipping", name);
  return -1;
}


// From scene result, find out where each camera sees each other,
// and how far apart they are.
void
update_cam2cam ()
{
  int i, j;
  double dist, distdup;

  if (!state->sceneokay)
    return;

  if (cam2cam != NULL)
    free (cam2cam);

  cam2cam_n = cams;

  cam2cam = calloc (cam2cam_n * cam2cam_n, sizeof (cam2cam_t));
  if (cam2cam == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  for (i = 0; i < cam2cam_n; i++)
    {
      if (!cam[i].inscene || !cam[i].haveinfo)
	continue;

      for (j = 0; j < cam2cam_n; j++)
	{
	  if (i == j)
	    continue;

	  if (!cam[j].inscene || !cam[j].haveinfo)
	    continue;

	  calc_campix (&cam[i], cam[i].posrot, cam[j].posrot,
		       &cam2cam[i * cam2cam_n + j].x,
		       &cam2cam[i * cam2cam_n + j].y);

	  // Pixel can get very large, esp. if *behind* other cam;
	  // maximize into useful values for -1..+1 scale...
	  if (cam2cam[i * cam2cam_n + j].x < -100.)
	    cam2cam[i * cam2cam_n + j].x = -100.;
	  if (cam2cam[i * cam2cam_n + j].x > 100.)
	    cam2cam[i * cam2cam_n + j].x = 100.;
	  if (cam2cam[i * cam2cam_n + j].y < -100.)
	    cam2cam[i * cam2cam_n + j].y = -100.;
	  if (cam2cam[i * cam2cam_n + j].y > 100.)
	    cam2cam[i * cam2cam_n + j].y = 100.;

	  dist = distance_3d (cam[i].posrot[0], cam[i].posrot[1],
			      cam[i].posrot[2],
			      cam[j].posrot[0], cam[j].posrot[1],
			      cam[j].posrot[2]);

	  cam2cam[i * cam2cam_n + j].dist = dist;

	  // distdup is 1 for dist <= dist_duplicate_min
	  // dustdup is 0 for dist >= dist_duplicate_max
	  // and linear in-between
	  distdup = 1. - (dist - dist_duplicate_min) /
	    (dist_duplicate_max - dist_duplicate_min);
	  if (distdup <= 0.)
	    distdup = 0.;
	  else if (distdup >= 1.)
	    distdup = 1.;

	  cam2cam[i * cam2cam_n + j].dist_duplicate = distdup;

	  MDBG ("cam2cam %d to %d: dist %f dup %f x %f y %f = pix %d %d", i,
		j, cam2cam[i * cam2cam_n + j].dist,
		cam2cam[i * cam2cam_n + j].dist_duplicate,
		cam2cam[i * cam2cam_n + j].x, cam2cam[i * cam2cam_n + j].y,
		lround ((cam2cam[i * cam2cam_n + j].x / cam[i].xscale) +
			cam[i].xoffset),
		lround ((cam2cam[i * cam2cam_n + j].y / cam[i].yscale) +
			cam[i].yoffset));
	}
    }
}


/*
  Purpose: we want to semi-ignore duplicate observations in object calc.

  A camera observation puts the object on a 3D line, without distance.
  When a second camera produces (nearly) the same 3D line, it does not
  add any useful information: the distance still can not be calculated.

  - Two cameras very close together (multi-camera panorama), i.e.
  they look from the same point to the same object.

  - Two cameras that see each other, and the object is on the line
  between the cameras.

  Method: we keep a "duplicate observations" counter per camera.
  Starts at 1; goes to 2 if there is a duplicate observation, goes to 3
  if there is another duplicate; etc.

  Then during fitting, we multiply the pixel errors by either 1/1 (normal),
  or 1/2 or 1/3 etc. This way, the total summed residual of all 2 or 3
  duplicates is weighed the same as a normal single camera.

  Also, this provides the entropy measure: we need two really different
  3D lines to find a fix.

  Close-together camera's are symmetrical; camera-camera line is not:
  object may be very close to cam2: then cam1 will see as on-line but cam2
  will see very much off-line -- then only cam1 will reduce its importance,
  which is mostly as desired since it is at a much greater distance.

  Returns: 1 = okay; 0 = not enough entropy, do not calculate.
*/
int
calc_duplicates ()
{
  int i, j;
  double entropy;

  // cam2cam_n is set on scene result
  // so any cam[i].inscene camera must be in it.

  for (i = 0; i < cam2cam_n; i++)
    {
      cam[i].duplicates = 1.;
      cam[i].residscale = 1.;
    }

  for (i = 0; i < cam2cam_n; i++)
    {
      if (!cam[i].inscene || !cam[i].haveinfo || !cam[i].curvalid)
	continue;

      for (j = 0; j < cam2cam_n; j++)
	{
	  double pixdist, pixdup;

	  if (i == j)
	    continue;

	  if (!cam[j].inscene || !cam[j].haveinfo || !cam[j].curvalid)
	    continue;

	  // Now we have valid observations for both cam i and j,
	  // see if they are duplicates.

	  // Object on camera-to-camera line is detected as being
	  // "too close" to other-camera pixel.
	  pixdist = hypot (cam[i].curx - cam2cam[i * cam2cam_n + j].x,
			   cam[i].cury - cam2cam[i * cam2cam_n + j].y);
	  // Formula: see update_cam2cam()
	  pixdup = 1. - (pixdist - oneline_duplicate_min) /
	    (oneline_duplicate_max - oneline_duplicate_min);
	  if (pixdup <= 0.)
	    pixdup = 0.;
	  else if (pixdup >= 1.)
	    pixdup = 1.;

	  // This camera being "too close" to the other camera can be
	  // looked up from precalculated matrix.
	  pixdup += cam2cam[i * cam2cam_n + j].dist_duplicate;

	  // But maximally add 1 per "other" camera to our duplicate count.
	  // (E.g. weird setup of really-close & face-to-face cameras,
	  // and object between them.)
	  if (pixdup >= 1.)
	    pixdup = 1.;

	  cam[i].duplicates += pixdup;
	}
    }

  entropy = 0.;
  for (i = 0; i < cam2cam_n; i++)
    {
      if (!cam[i].inscene || !cam[i].haveinfo || !cam[i].curvalid)
	continue;

      cam[i].residscale = 1 / cam[i].duplicates;

      // Cams produce two independent results: pixx and pixy
      entropy += 2. * cam[i].residscale;

      noMDBG ("cam %s dupl %f scale %f", cam[i].name,
	      cam[i].duplicates, cam[i].residscale);
    }

  // Tracking coordinate requires three independent results: x, y, z.
  if (entropy > 3.5)
    return 1;
  else
    {
      MDBG ("entropy %f too low to calculate", entropy);
      return 0;
    }
}


void
fixlost (char *cause)
{
  int i;
  json_t *jmsg;

  if (!state->fixreported)
    return;
  state->fixreported = 0;

  jmsg = json_pack ("{s:s, s:s}", "lost", "fix", "cause", cause);

  mqttpubjson (state->mosq, state->topicprefix, "objtrack/result",
	       NULL, json_incref (jmsg));

  for (i = 0; i < cams; i++)
    {
      mqttpubjson (state->mosq, state->topicprefix, "cam", cam[i].name,
		   "calcpix", NULL, json_incref (jmsg));
    }

  json_decref (jmsg);
}


// Adapted from testmpfit.c
void
mp_printresult (double *x, mp_result * result)
{
  int i;

  if ((x == 0) || (result == 0))
    return;
  printf ("  CHI-SQUARE = %f    (%d DOF)\n",
	  result->bestnorm, result->nfunc - result->nfree);
  printf ("   eqv resid = %f\n", sqrt (result->bestnorm / result->nfunc));
  // This is not the same as average residual, since residuals are weighted
  // by themselves, i.e. larger residuals weigh more in the sum-of-squares.
  // When residual values are really spread out, there will be relatively more
  // large residuals with large weights, so the difference will get more
  // pronounced.

  printf ("  params     = %d\n", result->npar);
  printf ("   - free    = %d\n", result->nfree);
  printf ("   - pegged  = %d\n", result->npegged);
  printf ("  iterations = %d\n", result->niter);
  printf ("  func evals = %d\n", result->nfev);
  for (i = 0; i < result->npar; i++)
    {
      printf ("  P[%d] = %f +/- %f\n", i, x[i], result->xerror[i]);
    }
  if (result->resid != NULL)
    {
      double sum = 0.;
      double abssum = 0.;
      double sqsum = 0.;
      double absmax = 0.;
      for (i = 0; i < result->nfunc; i++)
	{
	  sum += result->resid[i];
	  abssum += fabs (result->resid[i]);
	  sqsum += result->resid[i] * result->resid[i];
	  if (fabs (result->resid[i]) > absmax)
	    absmax = fabs (result->resid[i]);
	  printf ("  resid[%d] = %f\n", i, result->resid[i]);
	}
      printf ("  resid avg = %f   avg-of-abs = %f   sum-of-squares = %f\n",
	      sum / result->nfunc, abssum / result->nfunc, sqsum);
      printf ("  resid max-of-abs = %f\n", absmax);
    }
}

void
mqtt_mp_printresult (mp_result * result, char *addstr)
{
  int i;
  double sum = 0.;
  double abssum = 0.;
  double absmax = 0.;

  if (result == NULL)
    {
      MDBG ("called with zeropointer");
      return;
    }

  if (result->resid != NULL)
    for (i = 0; i < result->nfunc; i++)
      {
	sum += result->resid[i];
	abssum += fabs (result->resid[i]);
	if (fabs (result->resid[i]) > absmax)
	  absmax = fabs (result->resid[i]);
      }

  mqttprintf (state->mosq, state->topicprefix, "objtrack/log", NULL,
	      "status %d resid num %d sumsq %f avgabs %f maxabs %f "
	      "iter %d func %d "
	      "param num %d free %d pegged %d dof %d%s",
	      result->status,
	      result->nfunc, result->bestnorm,
	      abssum / result->nfunc, absmax,
	      result->niter, result->nfev,
	      result->npar, result->nfree, result->npegged,
	      result->nfunc - result->nfree, addstr);
  // See comments in mp_printresult
}



// Fit functions work like this:
//
// Caller has guessed param[nparam], and this fit function must calculate
// how correct those guesses are.
// It does so by assuming the guesses are fully correct, and calculates what
// the results would have been of the nmeas measurements that were done.
// Of course, those calculated results will be different from the measured
// results, and these differences must be stored in residual[nmeas].
// Caller will then use these residuals to adjust the param guesses.
//
// *private can be used to pass any data that is needed, such as the measured
// results to compare against.
//
// (derivvec[])[], if non-NULL, can be used to store derivatives if calculable.
//
// Returns 0 = success.


int
objpos_fitfunc (int nmeas, int nparam, double *param, double *residual,
		double **derivvec, void *private)
{
  // We do all valid camera pixels, no private data needed.
  // Always double handling: pixx and pixy.
  int i;
  int meascnt;

  meascnt = 0;

  for (i = 0; i < cams; i++)
    if (cam[i].haveinfo && cam[i].inscene && cam[i].curvalid)
      {
	double campixx, campixy;

	FITDBG ("  cam %d pixel", i);

	// param[0/1/2] is always object position guess

	calc_campix (&cam[i], cam[i].posrot, param, &campixx, &campixy);

	residual[meascnt] =
	  (cam[i].curx - campixx) * cam[i].residscalex * cam[i].residscale;

	residual[meascnt + 1] =
	  (cam[i].cury - campixy) * cam[i].residscaley * cam[i].residscale;

	FITDBG ("  wanted x %f y %f  resid x %f y %f",
		cam[i].curx, cam[i].cury,
		residual[meascnt], residual[meascnt + 1]);

	meascnt += 2;
      }

  // Doublecheck that nmeas is always exactly done now,
  // i.e. that we never get called with a meas subset;
  // that's required for pixx/pixy combi to keep working.
  if (meascnt != nmeas)
    {
      // Fatal error, need to exit, i.e. mqtt will never get posted
      STDERR ("FATAL: called with nmeas = %d instead of expected %d, "
	      "which is not currently handled", nmeas, meascnt);
      exit (1);
    }

  return 0;
}


void
calcpos ()
{
  int i;
  int nmeas;
  int nparam;
  double *resid;
  double *param;
  double *param_error;
  mp_par *param_settings;
  mp_result result;
  mp_config mpconfig;
  nmeas = 0;
  nparam = 0;
  int fitres;
  int totaliter;
  char invalidatedstr[1000];

  // Each camera, if valid, supplies two meas (x/y)
  for (i = 0; i < cams; i++)
    if (cam[i].inscene && cam[i].haveinfo && cam[i].curvalid)
      nmeas += 2;

  // Always one object to guess
  nparam = 3;

  // Now we can allocate the arrays
  resid = calloc (nmeas, sizeof (double));
  param = calloc (nparam, sizeof (double));
  param_error = calloc (nparam, sizeof (double));
  param_settings = calloc (nparam, sizeof (mp_par));

  if (resid == NULL || param == NULL || param_error == NULL
      || param_settings == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  // Initialize starting guesses
  param[0] = curpos[0];
  param[1] = curpos[1];
  param[2] = curpos[2];

  param_settings[0].step = 0.001;
  param_settings[1].step = 0.001;
  param_settings[2].step = 0.001;

  param_settings[0].advlimited = 1;
  param_settings[0].advlimit = 0.5;
  param_settings[1].advlimited = 1;
  param_settings[1].advlimit = 0.5;
  param_settings[2].advlimited = 1;
  param_settings[2].advlimit = 0.5;

#if 1
  for (i = 0; i < nparam; i++)
    FITDBG ("  startguess param[%d]=%f  fixed=%d",
	    i, param[i], param_settings[i].fixed);
#endif

  noMDBG ("Calculating object position");

  memset (&mpconfig, 0, sizeof (mpconfig));
  mpconfig.maxiter = 200;

  memset (&result, 0, sizeof (result));
  result.xerror = param_error;
  result.resid = resid;

  fitres = mpfit (objpos_fitfunc, nmeas, nparam, param, param_settings,
		  &mpconfig, NULL, &result);
  FITDBG ("  mpfit status = %d", i);
  //mp_printresult (param, &result);

  totaliter = result.niter;

  invalidatedstr[0] = '\0';

  // Now see if any one cam has much more error than all others,
  // then consider outlier: invalidate and try again.
  // If some camera was moved after the scene calibration,
  // this wil happen all the time...

  if (fitres >= 1 && fitres <= 4)
    {
      // mpfit has reliable result

      int meascnt;
      double highestresid;
      double secondhighestresid;
      int highestresidcam;

      // First find the highest residual.

      highestresid = -1.;
      highestresidcam = -1;
      meascnt = 0;
      for (i = 0; i < cams; i++)
	{
	  if (!cam[i].inscene || !cam[i].haveinfo || !cam[i].curvalid)
	    continue;

	  if (fabs (resid[meascnt]) > highestresid)
	    {
	      highestresid = fabs (resid[meascnt]);
	      highestresidcam = i;
	    }
	  if (fabs (resid[meascnt + 1]) > highestresid)
	    {
	      highestresid = fabs (resid[meascnt + 1]);
	      highestresidcam = i;
	    }

	  meascnt += 2;
	}

      // Then first second-highest residual.
      // (Could also do in same loop as above, but that gets quite
      //  unreadable.)

      secondhighestresid = -1.;
      meascnt = 0;
      for (i = 0; i < cams; i++)
	{
	  if (!cam[i].inscene || !cam[i].haveinfo || !cam[i].curvalid)
	    continue;

	  if (i != highestresidcam)
	    {
	      if (fabs (resid[meascnt]) > secondhighestresid)
		{
		  secondhighestresid = fabs (resid[meascnt]);
		}
	      if (fabs (resid[meascnt + 1]) > secondhighestresid)
		{
		  secondhighestresid = fabs (resid[meascnt + 1]);
		}
	    }

	  meascnt += 2;
	}

      // Note: Outlier removal can be done much more robustly,
      // especially in the case that very many measurements
      // are available (i.e. many cameras). For example:
      // http://www.biomedcentral.com/content/pdf/1471-2105-7-123.pdf
      // Motulsky, Brown, "Detecting outliers when fitting data with
      // nonlinear regression ..."

      if (highestresid > secondhighestresid * 4.)
	{
	  MDBG ("Invalidating outlier camera %s with resid %f"
		" much higher than next %f",
		cam[highestresidcam].name, highestresid, secondhighestresid);

	  cam[highestresidcam].curvalid = 0;

	  // Then do a refit if still possible
	  if (calc_duplicates ())
	    {
	      // One cam was removed.
	      // Existing resid malloc is now too big, no problem.
	      nmeas -= 2;

	      // Since new result may be way off,
	      // restart with previous values.
	      param[0] = curpos[0];
	      param[1] = curpos[1];
	      param[2] = curpos[2];

	      memset (&result, 0, sizeof (result));
	      result.xerror = param_error;
	      result.resid = resid;

	      fitres =
		mpfit (objpos_fitfunc, nmeas, nparam, param, param_settings,
		       &mpconfig, NULL, &result);
	      FITDBG ("  mpfit status = %d", i);
	      //mp_printresult (param, &result);

	      totaliter = result.niter;

	      snprintf (invalidatedstr, sizeof (invalidatedstr),
			" invalidated %s", cam[highestresidcam].name);
	    }
	  else
	    {
	      snprintf (invalidatedstr, sizeof (invalidatedstr),
			" no-entropy-after-invalidating %s",
			cam[highestresidcam].name);
	    }

	  // Leave highestresidcam invalid; will probably send
	  // new pixel soon enough.
	  // Below will still back-calculate correct pixel for all
	  // invalid cams.
	}
    }

  mqtt_mp_printresult (&result, invalidatedstr);

  if (fitres >= 1 && fitres <= 4)
    {
      // mpfit has reliable result
      // (hmm, 4 sometimes okay, sometimes not...)

      // Remember position and report it.
      curpos[0] = param[0];
      curpos[1] = param[1];
      curpos[2] = param[2];

      mqttpubjson (state->mosq, state->topicprefix, "objtrack/result", NULL,
		   json_pack ("{s:f, s:f, s:f, s:f, s:i}",
			      "x", curpos[0], "y", curpos[1], "z", curpos[2],
			      "sumofsquares", result.bestnorm,
			      "rounds", totaliter));
      state->fixreported = 1;

      // Start (or restart) the timeout for sending fix-lost if we
      // do not come here again quickly enough.
      timerfd_settime (state->fixlost_timerfd, 0,
		       &state->fixlost_itimval, NULL);

      // Calculate and report the pixels that each cam _should_ be seeing.
      for (i = 0; i < cams; i++)
	{
	  if (!cam[i].inscene || !cam[i].haveinfo)
	    continue;

	  calc_campix (&cam[i], cam[i].posrot, curpos,
		       &cam[i].backcalcx, &cam[i].backcalcy);

	  // Pixel can get very large, esp. if *behind* cam;
	  // maximize into useful values for -1..+1 scale...
	  if (cam[i].backcalcx < -20.)
	    cam[i].backcalcx = -20.;
	  if (cam[i].backcalcx > 20.)
	    cam[i].backcalcx = 20.;
	  if (cam[i].backcalcy < -20.)
	    cam[i].backcalcy = -20.;
	  if (cam[i].backcalcy > 20.)
	    cam[i].backcalcy = 20.;

	  mqttprintf (state->mosq, state->topicprefix, "cam", cam[i].name,
		      "calcpix", NULL, "{\"pixx\": %ld, \"pixy\": %ld}",
		      lround ((cam[i].backcalcx / cam[i].xscale) +
			      cam[i].xoffset),
		      lround ((cam[i].backcalcy / cam[i].yscale) +
			      cam[i].yoffset));

	  noMDBG
	    ("Cam %s backcalc x %f y %f = pix %ld %ld (had %ld %ld)",
	     cam[i].name, cam[i].backcalcx, cam[i].backcalcy,
	     lround ((cam[i].backcalcx / cam[i].xscale)
		     + cam[i].xoffset),
	     lround ((cam[i].backcalcy / cam[i].yscale)
		     + cam[i].yoffset),
	     cam[i].curvalid ?
	     lround ((cam[i].curx / cam[i].xscale)
		     + cam[i].xoffset) : -1,
	     cam[i].curvalid ?
	     lround ((cam[i].cury / cam[i].yscale) + cam[i].yoffset) : -1);
	}

    }
  else
    {
      // mpfit has no reliable result
      fixlost ("failed fit");

      timerfd_settime (state->fixlost_timerfd, 0, &state->zero_itimval, NULL);
    }

  free (resid);
  free (param);
  free (param_error);
  free (param_settings);
}


void
receive_caminfo (char *camname, json_t * jroot)
{
  int oi;			// object index

  oi = find_make_cam (camname);

  // Note: this can be new cam, but also new params for existing cam.
  // Scene result still assumed to be valid.

  // forget old params
  cam[oi].haveinfo = 0;

  free (cam[oi].typeparam);
  cam[oi].typeparam = NULL;

  free (cam[oi].distortionparam);
  cam[oi].distortionparam = NULL;

  if (cam[oi].pixinvalid_timerfd == -1)
    {
      int res;
      cam[oi].pixinvalid_timerfd = timerfd_create (CLOCK_MONOTONIC, 0);
      if (cam[oi].pixinvalid_timerfd == -1)
	{
	  STDERR ("failed to create timerfd: %s", strerror (errno));
	  exit (1);
	}
      res = fcntl (cam[oi].pixinvalid_timerfd, F_SETFL, O_NONBLOCK);
      if (res == -1)
	{
	  STDERR ("failed to fcntl timerfd: %s", strerror (errno));
	  exit (1);
	}
      timerfd_settime (cam[oi].pixinvalid_timerfd, 0,
		       &state->zero_itimval, NULL);
    }

  if (parse_cam_json (jroot, &cam[oi], 1) > 0)
    {
      MDBG ("Cam %s info incomplete, disabling", camname);

      cam[oi].haveinfo = 0;

      if (cam[oi].typeparam != NULL)
	free (cam[oi].typeparam);
      cam[oi].typeparam = NULL;

      if (cam[oi].distortionparam != NULL)
	free (cam[oi].distortionparam);
      cam[oi].distortionparam = NULL;

      return;
    }

  dist_lin_precalc (&cam[oi]);

  cam[oi].haveinfo = 1;

  update_cam2cam ();
}

void
receive_campix (char *camname, json_t * jroot)
{
  int oi;
  double px = -1., py = -1., pw = 1., ph = 1.;
  int res;

  oi = find_cam (camname);
  if (oi < 0)
    {
      // MDBG already done by find_cam()
      return;
    }

  res = json_unpack (jroot, "{s:F, s:F, s?:F, s?:F}",
		     "pixx", &px, "pixy", &py, "pixw", &pw, "pixh", &ph);
  if (res != 0)
    {
      // required pixx and/or pixy missing, probably {"lost":"fix"}

      cam[oi].curvalid = 0;

      timerfd_settime (cam[oi].pixinvalid_timerfd, 0,
		       &state->zero_itimval, NULL);
      return;
    }

  cam[oi].curx = (px - cam[oi].xoffset) * cam[oi].xscale;
  cam[oi].cury = (py - cam[oi].yoffset) * cam[oi].yscale;

  if (pw < .5)
    pw = 1.;
  if (ph < .5)
    ph = 1.;
  if (pw > cam[oi].w)
    pw = cam[oi].w;
  if (ph > cam[oi].h)
    ph = cam[oi].h;

  // The bigger the pixel, the larger the uncertainty where the object
  // really is: on rescaled -1..+1 coords, a bigger residual is acceptable
  // since that still wouldn't move detection to the next pixel.
  //
  // The bigger the detected blink: idem, multiplied. We cannot be quite
  // sure that our weighted-average pixel is actually pointing to the exact
  // geometric center of the object. Only for a 1-pixel blink, it certainly is.
  //
  // So calculation residuals must be scaled by 1/(pixelsize * #blinkpix),
  // which will then fairly compare 640x480 to full-HD.
  // 
  // Pixelsize = x/yscale, these are about 0.002 (1080p) to 0.004 (480p).
  // Multiply by 0.005 to keep scale factor somewhat close to 1.

  cam[oi].residscalex = 0.005 / (pw * cam[oi].xscale);
  cam[oi].residscaley = 0.005 / (ph * cam[oi].yscale);

  cam[oi].curvalid = 1;


  // Start or restart the validity timeout for this camera.
  // This is safe: drops all pending select() triggers (as in iob2mqtt.c)
  timerfd_settime (cam[oi].pixinvalid_timerfd, 0,
		   &state->pixinval_itimval, NULL);

  // If this is the first pixel report since last calc, start recalc in
  // max 0.1 sec (...if calib was okay and this cam was actually calib'd).
  // This leaves some time for other cams to report their pixels too.
  if (!state->morepixwait_running &&
      state->sceneokay && cam[oi].inscene && cam[oi].haveinfo)
    {
      timerfd_settime (state->morepixwait_timerfd, 0,
		       &state->morepixwait_itimval, NULL);
      state->morepixwait_running = 1;
    }
}

void
receive_sceneresult (json_t * jroot)
{
  int i;
  int arrayidx;

  state->sceneokay = 0;
  for (i = 0; i < cams; i++)
    {
      cam[i].inscene = 0;
    }

  for (arrayidx = 0; arrayidx < json_array_size (jroot); arrayidx++)
    {
      json_t *jitem;
      int res;
      char *camname;
      double x = 0., y = 0., z = 0., a = 0., b = 0., g = 0.;
      int oi;			// object index

      jitem = json_array_get (jroot, arrayidx);
      if (!json_is_object (jitem))
	continue;

      res = json_unpack (jitem, "{s:s, s:F, s:F, s:F, s:F, s:F, s:F}",
			 "cam", &camname,
			 "x", &x, "y", &y, "z", &z,
			 "a", &a, "b", &b, "g", &g);
      if (res != 0)
	continue;		// e.g. "obj" or "global"

      // On delayed start, sceneresult repost may arrive before
      // caminfo repost. Create new cam if not exists.

      oi = find_make_cam (camname);

      cam[oi].posrot[0] = x;
      cam[oi].posrot[1] = y;
      cam[oi].posrot[2] = z;
      cam[oi].posrot[3] = a;
      cam[oi].posrot[4] = b;
      cam[oi].posrot[5] = g;

      cam[oi].inscene = 1;
    }

  state->sceneokay = 1;

  update_cam2cam ();
}

void
calc_init ()
{
  int i;

  curpos[0] = 0.1;
  curpos[1] = 0.2;
  curpos[2] = 0.3;

  state->zero_itimval.it_value.tv_sec = 0;
  state->zero_itimval.it_value.tv_nsec = 0;
  state->zero_itimval.it_interval.tv_sec = 0;
  state->zero_itimval.it_interval.tv_nsec = 0;

  // Timeout to forget any cam's pixel if it didn't report a new one.
  state->pixinval_itimval.it_value.tv_sec = 0;
  state->pixinval_itimval.it_value.tv_nsec = 900000000;	// 0.9sec
  state->pixinval_itimval.it_interval.tv_sec = 0;
  state->pixinval_itimval.it_interval.tv_nsec = 0;

  // Timeout to start new position calc after receiving the first
  // cam pixel following the previous calc.
  // (I.e. in this time, all other cams can send their new pixels too.)
  // This limits how often we can do a recalc per second.
  state->morepixwait_itimval.it_value.tv_sec = 0;
  state->morepixwait_itimval.it_value.tv_nsec = 100000000;	// 0.1sec
  state->morepixwait_itimval.it_interval.tv_sec = 0;
  state->morepixwait_itimval.it_interval.tv_nsec = 0;

  // Timeout to annouce fix-lost after last successful recalc.
  state->fixlost_itimval.it_value.tv_sec = 2;
  state->fixlost_itimval.it_value.tv_nsec = 0;
  state->fixlost_itimval.it_interval.tv_sec = 0;
  state->fixlost_itimval.it_interval.tv_nsec = 0;


  // Init morepixwait timer
  state->morepixwait_timerfd = timerfd_create (CLOCK_MONOTONIC, 0);
  if (state->morepixwait_timerfd == -1)
    {
      STDERR ("failed to create timerfd: %s", strerror (errno));
      exit (1);
    }
  i = fcntl (state->morepixwait_timerfd, F_SETFL, O_NONBLOCK);
  if (i == -1)
    {
      STDERR ("failed to fcntl timerfd: %s", strerror (errno));
      exit (1);
    }
  timerfd_settime (state->morepixwait_timerfd, 0, &state->zero_itimval, NULL);
  state->morepixwait_running = 0;


  // Init fixlost timer
  state->fixlost_timerfd = timerfd_create (CLOCK_MONOTONIC, 0);
  if (state->fixlost_timerfd == -1)
    {
      STDERR ("failed to create timerfd: %s", strerror (errno));
      exit (1);
    }
  i = fcntl (state->fixlost_timerfd, F_SETFL, O_NONBLOCK);
  if (i == -1)
    {
      STDERR ("failed to fcntl timerfd: %s", strerror (errno));
      exit (1);
    }
  timerfd_settime (state->fixlost_timerfd, 0, &state->zero_itimval, NULL);
}

// This is called to add fds to the select sets that are interesting
// for the scenecalc handling, in particular timerfds.
void
select_addcalc (int *maxfd, fd_set * readfds, fd_set * writefds)
{
  int i;

  for (i = 0; i < cams; i++)
    if (cam[i].pixinvalid_timerfd != -1)
      {
	FD_SET (cam[i].pixinvalid_timerfd, readfds);
	if (cam[i].pixinvalid_timerfd > *maxfd)
	  *maxfd = cam[i].pixinvalid_timerfd;
      }

  if (state->morepixwait_running)
    {
      FD_SET (state->morepixwait_timerfd, readfds);
      if (state->morepixwait_timerfd > *maxfd)
	*maxfd = state->morepixwait_timerfd;
    }

  FD_SET (state->fixlost_timerfd, readfds);
  if (state->fixlost_timerfd > *maxfd)
    *maxfd = state->fixlost_timerfd;
}

// This is called after the mainloop select to handle the fds that are
// interesting for the scenecalc handling, in particular timerfds.
void
select_handlecalc (fd_set * readfds, fd_set * writefds)
{
  int i;

  // Check if any can has stopped sending pixels
  for (i = 0; i < cams; i++)
    if (cam[i].pixinvalid_timerfd != -1 &&
	FD_ISSET (cam[i].pixinvalid_timerfd, readfds))
      {
	// Camera timeout
	MDBG ("cam %s timeout, invalidating pixel", cam[i].name);

	cam[i].curvalid = 0;

	// _settime will drop pending triggers, no need to read()
	timerfd_settime (cam[i].pixinvalid_timerfd, 0,
			 &state->zero_itimval, NULL);
      }

  // Check if waited long enough for doing objectpos calc
  if (state->morepixwait_timerfd != -1 &&
      FD_ISSET (state->morepixwait_timerfd, readfds))
    {
      if (state->sceneokay)
	{
	  if (calc_duplicates ())
	    {
	      calcpos ();
	    }
	  else
	    {
	      fixlost ("insufficient entropy");

	      // _settime will drop pending triggers, no need to read()
	      timerfd_settime (state->fixlost_timerfd, 0,
			       &state->zero_itimval, NULL);
	    }
	}

      // _settime will drop pending triggers, no need to read()
      timerfd_settime (state->morepixwait_timerfd, 0,
		       &state->zero_itimval, NULL);
      state->morepixwait_running = 0;
    }

  // Check if last position was too long ago
  if (state->fixlost_timerfd != -1 &&
      FD_ISSET (state->fixlost_timerfd, readfds))
    {
      fixlost ("no cams reporting");

      // _settime will drop pending triggers, no need to read()
      timerfd_settime (state->fixlost_timerfd, 0, &state->zero_itimval, NULL);
    }
}
