// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <jansson.h>

#include "recv_pub.h"
#include "idletimer.h"
#include "mathgl_prepare.h"
#include "render_csv.h"
#include "render_tile.h"
#include "../shared/error.h"


// Caller must supply destfname and tmpfname buffers of at least FNAMELEN size.
// Returns 1 = okay, 0 = fail.
int
makefilenames (state_t * state, const char *rendername_sane,
	       const char *exten, char *destfname, char *tmpfname)
{
  int res;

  res =
    snprintf (destfname, FNAMELEN, "%s/%s-%s.%s", state->pubdir,
	      rendername_sane, state->name_sane, exten);
  if (res <= 0 || res >= FNAMELEN)
    {
      MDBG ("filename too long, increase FNAMELEN");
      return 0;
    }

  // Make sure that extension is same: mathgl looks at it...
  res = snprintf (tmpfname, FNAMELEN, "%s.tmp.%s", destfname, exten);
  if (res <= 0 || res >= FNAMELEN)
    {
      MDBG ("filename too long, increase FNAMELEN");
      return 0;
    }

  return 1;
}

void
announcefile (state_t * state, const char *collecttopic,
	      const char *rendername, const char *rendername_sane,
	      const char *exten, const char *filetype, int width, int height)
{
  char outfilename[FNAMELEN + 10];
  char outfilepub[FNAMELEN + 10];
  int res;

  res =
    snprintf (outfilename, FNAMELEN, "%s/%s-%s.%s", state->pubdir,
	      rendername_sane, state->name_sane, exten);
  if (res <= 0 || res >= FNAMELEN)
    {
      MDBG ("filename too long, increase FNAMELEN");
      return;
    }

  res =
    snprintf (outfilepub, FNAMELEN, "%s%s-%s.%s", state->pubpfx,
	      rendername_sane, state->name_sane, exten);
  if (res <= 0 || res >= FNAMELEN)
    {
      MDBG ("pubname too long, increase FNAMELEN");
      return;
    }

  mqttpubjson (state->mosq, collecttopic, "img", rendername,
	       filetype, state->name, "loc", NULL,
	       json_pack ("{s:s,s:s,s:i,s:i}",
			  "file", outfilename, "url", outfilepub,
			  "pixw", width, "pixh", height));
}

// Returns -1: not found
int
lookup_plane (data_t * data, const char *searchlabel)
{
  int i;
  for (i = 0; i < data->planes_num; i++)
    if (!strcmp (data->planes[i].label, searchlabel))
      return i;
  return -1;
}


void
recv_bindata (state_t * state, char *bindatatopic, unsigned char *payload,
	      size_t payloadlen)
{
  uint32_t magicheader = 0x0000DA7A;
  double magic_double_7654321 = 765.3421;
  double magic_double_NAN = NAN;

  char *cp;
  uint32_t json_header_len;
  json_t *jroot;
  json_error_t jerror;
  data_t data;

  memset (&data, 0, sizeof (data));


  // Topic handling

  if (strlen (bindatatopic) >= TOPICLEN)
    return;			// Can't rewrite

  strcpy (data.collecttopic, bindatatopic);
  cp = strrchr (data.collecttopic, '/');	// Strip "/bindata" (if present)
  if (cp != NULL && !strcmp (cp, "/bindata"))
    *cp = '\0';


  // JSON header decoding and basic checks

  if (payloadlen < 10)
    return;			// Can't be interesting

  // Bindata header check
  {
    uint32_t magiccheck;
    memcpy (&magiccheck, payload, 4);
    if (magiccheck != magicheader)
      return;			// Not bindata
  }

  memcpy (&json_header_len, payload + 4, 4);
  if (json_header_len < 2 || json_header_len > 100000)
    return;			// Really insane

  if (payloadlen < json_header_len + 8 + 3 * sizeof (double))
    return;			// Truncated

  jroot = json_loadb ((void *) (payload + 8), json_header_len, 0, &jerror);
  if (jroot == NULL)
    return;

  {
    const char *format =
      json_string_value (json_object_get (jroot, "format"));
    if (format == NULL || strcmp (format, "xy,data:interleaved"))
      return;			// Unsupported
  }

  // Axes parse
  {
    int i;
    json_t *jaxes = json_object_get (jroot, "axes");
    if (json_array_size (jaxes) != 2)
      {
	MDBG ("No two axes");
	return;
      }
    for (i = 0; i < 2; i++)
      {
	json_t *jaxis = json_array_get (jaxes, i);
	data.axes[i].label =
	  json_string_value (json_object_get (jaxis, "label"));
	data.axes[i].items =
	  json_integer_value (json_object_get (jaxis, "items"));
	if (data.axes[i].items < 1)
	  {
	    MDBG ("Axis %d too few items", i);
	    return;
	  }
	data.axes[i].interval =
	  json_number_value (json_object_get (jaxis, "interval"));
	if (data.axes[i].interval <= 0.0000001)
	  {
	    MDBG ("Axis %d too small interval", i);
	    return;
	  }
	data.axes[i].firstitem =
	  json_number_value (json_object_get (jaxis, "firstitem"));
      }
  }

  // Planes parse
  {
    int i;
    json_t *jplanes = json_object_get (jroot, "planes");
    data.planes_num = json_array_size (jplanes);
    if (data.planes_num < 1)
      {
	MDBG ("Too few planes");
	return;
      }
    data.planes = calloc (data.planes_num, sizeof (*data.planes));
    if (data.planes == NULL)
      {
	STDERR ("malloc failed");
	exit (1);
      }
    for (i = 0; i < data.planes_num; i++)
      {
	json_t *jplane = json_array_get (jplanes, i);
	data.planes[i].label =
	  json_string_value (json_object_get (jplane, "label"));
	if (data.planes[i].label == NULL)
	  {
	    MDBG ("plane %d has no label", i);
	    free (data.planes);
	    return;		// Unsupported
	  }

	if (!strcmp (data.planes[i].label, "autogenerated:time"))
	  {
	    data.planes[i].is_autogen_time = 1;
	    data.planes[i].time_last =
	      json_number_value (json_object_get
				 (jplane, "autogenerated:time-last"));
	    data.planes[i].time_now =
	      json_number_value (json_object_get
				 (jplane, "autogenerated:time-now"));
	  }

	const char *itemformat =
	  json_string_value (json_object_get (jplane, "itemformat"));
	if (itemformat == NULL || strcmp (itemformat, "double-x86"))
	  {
	    free (data.planes);
	    return;		// Unsupported
	  }

	data.planes[i].rounddata =
	  json_number_value (json_object_get (jplane, "rounddata"));
      }
  }

  data.lastreceived.valid = 0;
  while (1)			// Easy "break" exit
    {
      json_t *jlastrecv = json_object_get (jroot, "last-received");
      json_t *jval;
      if (jlastrecv == NULL)
	break;

      jval = json_object_get (jlastrecv, "x");
      if (!json_is_number (jval))
	break;
      data.lastreceived.x = json_number_value (jval);

      jval = json_object_get (jlastrecv, "y");
      if (!json_is_number (jval))
	break;
      data.lastreceived.y = json_number_value (jval);

      data.lastreceived.valid = 1;
      break;
    }

  if (payloadlen !=
      8 + json_header_len + (2 +
			     data.planes_num * data.axes[0].items *
			     data.axes[1].items) * sizeof (double))
    {
      MDBG ("payload size mismatch");
      free (data.planes);
      return;
    }

  // Double-data header check
  {
    double d;

    memcpy (&d, payload + 8 + json_header_len, sizeof (double));
    if (memcmp (&d, &magic_double_7654321, sizeof (double)))
      {
	MDBG ("data header magic1 mismatch: %f instead of %f", d,
	      magic_double_7654321);
	free (data.planes);
	return;
      }

    memcpy (&d, payload + 8 + json_header_len + sizeof (double),
	    sizeof (double));
    if (memcmp (&d, &magic_double_NAN, sizeof (double)))
      {
	MDBG ("data header magic2 mismatch");
	free (data.planes);
	return;
      }
  }

  data.data = (void *) (payload + 8 + json_header_len + 2 * sizeof (double));


  if (state->idletimer_running)
    {
      // We are probably being overly active
      state->overactive_counter++;
      if (state->overactive_counter > 5)
	{
	  // Just skip this one,
	  // and restart timer to skip anything else that comes too quickly again.
	  MDBG ("skipping due to prolonged overload");
	  idletimer_start (state);
	  free (data.planes);
	  return;
	}
    }
  else
    state->overactive_counter = 0;

  idletimer_stop (state);
  // This is inside message callback, can not usefully send "processing":
  // will only get queued until everything is done.

  // Renders parse
  {
    int i;
    json_t *jrenders = json_object_get (jroot, "renders");
    for (i = 0; i < json_array_size (jrenders); i++)
      {
	json_t *jrender = json_array_get (jrenders, i);
	const char *rendername =
	  json_string_value (json_object_get (jrender, "name"));
	const char *rendertype =
	  json_string_value (json_object_get (jrender, "type"));
	char rendername_sane[TOPICLEN + 10];

	if (rendername == NULL || rendertype == NULL)
	  continue;		// Probably comment etc.

	if (strchr (rendername, '/') != NULL)
	  {
	    MDBG ("render name may not contain slash");
	    continue;
	  }
	if (strlen (rendername) >= TOPICLEN)
	  {
	    MDBG ("render name too long, increase TOPICLEN");
	    continue;
	  }

	strncpy (rendername_sane, rendername, TOPICLEN);
	rendername_sane[TOPICLEN - 1] = '\0';
	for (cp = rendername_sane; *cp != '\0'; cp++)
	  {
	    if (*cp >= '0' && *cp <= '9')
	      continue;
	    if (*cp >= 'A' && *cp <= 'Z')
	      continue;
	    if (*cp >= 'a' && *cp <= 'z')
	      continue;
	    // else: replace char
	    *cp = '_';
	  }

	if (!strcmp (rendertype, "csv"))
	  render_csv (state, &data, jrender, rendername, rendername_sane);
	else if (!strcmp (rendertype, "mathgl-tile"))
	  render_tile (state, &data, jrender, rendername, rendername_sane);
      }
  }

  mathgl_delete_data (state, &data);
  free (data.planes);

  idletimer_start (state);
}
