// Render planar visualizations
// Also see --help.

// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#define PROGNAME  "render2d"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <math.h>
#include <sys/timerfd.h>
#include <jansson.h>

#include "common.h"
#include "recv_pub.h"
#include "idletimer.h"
#include "../shared/version.h"
#include "../shared/error.h"

const char *progname = PROGNAME;


void
print_help (FILE * f)
{
  fprintf (f, "\n"
	   "USAGE: %s [options] [--] <topic [...]>\n"
	   "\n"
	   "SCWIPS Render planar visualizations.\n"
	   "\n"
	   "Options are:\n"
	   "    --host <name>        Hostname of mosquitto broker (default: localhost)\n"
	   "    --port <int>         Portname of mosquitto broker (default: 1883)\n"
	   "    --topicprefix <s>    Scwips instance mqtt topic prefix.\n"
	   "    --name <s>           Name of this instance for publish subtopic.\n"
	   "    --pubdir <s>         Directory to publish images in.\n"
	   "    --pubpfx <s>         Prefix/directory to mention in ready messages.\n"
	   "    <topic [...]>        Topics to receive data from; scwips/collect/+/bindata\n"
	   "\n"
	   "    --help               Print this help\n"
	   "\n"
	   "Receives bindata on given topics and produces visualization renderings.\n"
	   "Renderings are produced as (image) files and announced to corresponding\n"
	   "subtopic.\n"
	   "\n"
	   "Reconnections to mqtt broker are handled gracefully.\n"
	   "DNS and connect are protected by alarm(), use external loop to restart.\n"
	   "\nPart of SCWIPS version %s\n\n", PROGNAME, SCWIPSVERSION);
}


/* Parse commandline parameters.
   Returns:
   0 = all okay, continue
   1 = immediately exit successfully (after --help)
   2 = immediately exit unsuccessfully
*/
int
parse_cmdline (int argc, char **argv, state_t * state)
{
  int i;
  char *cp;

  // Rather simplistic & easy parameter parsing, adapted from earlier stuff
  for (i = 1; i < argc; i++)
    {
      if (!strcmp (argv[i], "--help"))
	{
	  print_help (stdout);
	  return 1;		// exit successfully
	}

      else if (!strcmp (argv[i], "--host"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --host requires argument");
	  state->svr_hostname = argv[i];
	}
      else if (!strcmp (argv[i], "--port"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --port requires argument");
	  state->svr_port = atoi (argv[i]);
	  if (state->svr_port < 1 || state->svr_port > 65535)
	    CMDLINEERR_RETURN2 ("parameter --port invalid value: %s",
				argv[i]);
	}
      else if (!strcmp (argv[i], "--topicprefix"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --topicprefix requires argument");
	  state->topicprefix = argv[i];
	}
      else if (!strcmp (argv[i], "--name"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --name requires argument");
	  state->name = argv[i];
	}
      else if (!strcmp (argv[i], "--pubdir"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --pubdir requires argument");
	  state->pubdir = argv[i];
	}
      else if (!strcmp (argv[i], "--pubpfx"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --pubpfx requires argument");
	  state->pubpfx = argv[i];
	}

      else if (!strcmp (argv[i], "--"))
	{
	  i++;
	  break;
	}
      else if (argv[i][0] != '-')
	{
	  break;
	}

      else
	CMDLINEERR_RETURN2 ("unknown parameter: %s", argv[i]);
    }

  // Rest of arguments are bindata receive topics
  state->recv_topics = argv + i;
  state->recv_topics_num = argc - i;

  if (state->topicprefix == NULL)
    CMDLINEERR_RETURN2 ("please specify --topicprefix");
  if (state->name == NULL)
    CMDLINEERR_RETURN2 ("please specify --name");
  if (state->pubdir == NULL)
    CMDLINEERR_RETURN2 ("please specify --pubdir");
  if (state->pubpfx == NULL)
    CMDLINEERR_RETURN2 ("please specify --pubpfx");
  if (state->recv_topics_num < 1)
    CMDLINEERR_RETURN2 ("please specify receive topics");

  if (strlen (state->topicprefix) + strlen (state->name) > TOPICLEN - 200)
    CMDLINEERR_RETURN2
      ("parameter --topicprefix + --name too long, increase TOPICLEN");

  // Name sanity, may only be one topic level
  if (strchr (state->name, '/') != NULL)
    CMDLINEERR_RETURN2 ("--name is not allowed to contain slash");

  strncpy (state->name_sane, state->name, TOPICLEN);
  state->name_sane[TOPICLEN - 1] = '\0';
  for (cp = state->name_sane; *cp != '\0'; cp++)
    {
      if (*cp >= '0' && *cp <= '9')
	continue;
      if (*cp >= 'A' && *cp <= 'Z')
	continue;
      if (*cp >= 'a' && *cp <= 'z')
	continue;
      // else: replace char
      *cp = '_';
    }

  snprintf (state->broadcmd_topic, TOPICLEN, "%s/cmd", state->topicprefix);
  snprintf (state->rendercmd_topic, TOPICLEN, "%s/render/cmd",
	    state->topicprefix);
  snprintf (state->ourcmd_topic, TOPICLEN, "%s/render/%s/cmd",
	    state->topicprefix, state->name);

  // All should be okay now
  return 0;
}


/* This connect callback is called when CONNACK is received, i.e. when
   the broker has actively allowed us access.
*/
void
mqttconnect_callback (struct mosquitto *mosq, void *usrobj, int result)
{
  state_t *state = usrobj;
  int i;

  if (result != 0)
    {
      // Access actively refused.
      // Socket will be closed by library via
      // mosquitto_loop_read() -> _mosquitto_loop_rc_handle()
      return;
    }

  state->connstate = CS_CONNECTED;

  // Send subscriptions
  mosquitto_subscribe (state->mosq, NULL, state->ourcmd_topic, 0);
  mosquitto_subscribe (state->mosq, NULL, state->rendercmd_topic, 0);
  mosquitto_subscribe (state->mosq, NULL, state->broadcmd_topic, 0);

  for (i = 0; i < state->recv_topics_num; i++)
    mosquitto_subscribe (state->mosq, NULL, state->recv_topics[i], 0);

  MDBG ("connected and subscribed");

  // All processing is done from message callback, so we can never be here
  // when actually processing.
  //idletimer_start (state);
  // Actually, do *not* announce, will cause unsyncedness in many cases.
  // Any observer must apply timeout anyway.
}

/* This disconnect callback is called whenever socket is (already) closed
   for whatever reason.
*/
void
mqttdisconnect_callback (struct mosquitto *mosq, void *usrobj, int reason)
{
  state_t *state = usrobj;

  // We are disconnected already, can't send any message.
  // But if we wanted disconnection, we already sent "shutdown".
  // And if we got disconnected without our request, the broker will
  // already have sent our Last Will = position-lost message.

  if (state->connstate == CS_EOFDISCONNECTING
      || state->connstate == CS_EOFDISCONNECTED)
    {
      state->connstate = CS_EOFDISCONNECTED;
    }
  else
    {
      state->connstate = CS_UNCONNECTED;
      state->connect_nexttime = time (NULL) + 5;
    }

  idletimer_stop (state);
}

/* This message callback is called whenever a message is received; may be 
   called _multiple_ times from 1 mosquitto_loop_read() call.
*/
void
mqttmessage_callback (struct mosquitto *mosq, void *usrobj,
		      const struct mosquitto_message *message)
{
  state_t *state = usrobj;
  json_t *jroot;
  json_error_t jerror;

  // We are non-threaded, and all messages can always be handled immediately
  // when they arrive, no queueing needed. So we just handle the message
  // right here inside the callback.

  // We don't want topicless or payloadless messages.
  if (message->topic[0] == '\0' || message->payloadlen == 0)
    return;

  if (!strcmp (message->topic, state->ourcmd_topic) ||
      !strcmp (message->topic, state->rendercmd_topic) ||
      !strcmp (message->topic, state->broadcmd_topic))
    {
      jroot = json_loadb (message->payload, message->payloadlen, 0, &jerror);
      if (jroot == NULL)
	return;

      do
	{
	  // Easy "break" for decref & return

	  const char *cmd;

	  cmd = json_string_value (json_object_get (jroot, "cmd"));
	  if (cmd == NULL)
	    break;

	  if (!strcmp (cmd, "shutdown"))
	    {
	      // Initiate mqtt disconnect, will lead to shutdown after disconnected
	      // by broker, or after timeout; handled in mainloop().

	      if (state->connstate == CS_CONNECTED)
		{
		  state->connstate = CS_EOFDISCONNECTING;
		  state->connect_nexttime = time (NULL) + 5;
		  mosquitto_disconnect (state->mosq);	// just queues DISCONNECT
		}
	      else if (state->connstate != CS_EOFDISCONNECTING
		       && state->connstate != CS_EOFDISCONNECTED)
		{
		  // Not connected at all, so sending DISCONNECT would be
		  // useless.
		  // (This should not happen in practice, since we are handling
		  // the shutdown command that arrived via mqtt.)
		  state->connstate = CS_EOFDISCONNECTED;
		  // This will cause loop break immediately in next round.
		}

	      break;
	    }
	}
      while (0);
      json_decref (jroot);
    }
  else
    {
      // Any other topic must be bindata

      recv_bindata (state, message->topic, message->payload,
		    message->payloadlen);
    }

  // *message is free()'d by the library after this callback returns.
}


void
mainloop (state_t * state)
{
  time_t now;
  fd_set readfds, writefds;
  int mosqsock, maxfd;
  struct timeval tv;
  int selectres;

  while (1)
    {
      alarm (10);		// Includes visualization processing time

      now = time (NULL);

      // Server has disconnected after we asked
      if (state->connstate == CS_EOFDISCONNECTED)
	{
	  break;
	}

      // Server did not answer after we asked DISCONNECT
      // (Test: suspend broker with Ctrl-Z just before DISCONNECT.)
      if (state->connstate == CS_EOFDISCONNECTING
	  && now >= state->connect_nexttime)
	{
	  break;
	}

      // If disconnected without our asking, attempt to reconnect periodically
      // (=also immediately on startup).
      if (state->connstate == CS_UNCONNECTED
	  && now >= state->connect_nexttime)
	{
	  state->connstate = CS_CONNECTING;

	  if (mosquitto_connect
	      (state->mosq, state->svr_hostname, state->svr_port,
	       10) != MOSQ_ERR_SUCCESS)
	    {
	      // Failed, try again after some time
	      now = time (NULL);	// DNS/connect may have taken time
	      state->connstate = CS_UNCONNECTED;
	      state->connect_nexttime = now + 5;
	      continue;
	    }

	  // mosquitto_connect() returns after TCP connect() finished, with
	  // CONNECT packet queued to be sent by next mosquitto_loop_write().
	  // DNS resolving and connect() is protected by alarm() above.

	  // We keep ->connstate in CS_CONNECTING until CONNACK received
	  // via mqttconnect_callback().

	  // CONNACK is waited for up to keepalive time, checked in
	  // _mosquitto_check_keepalive() via mosquitto_loop_misc().
	  // (state->mosq->state is not cs_connected yet; _no_ pings are sent.)
	  // Waiting-for-CONNACK is easily testable: Ctrl-S on broker.

	  // DNS/connect may have taken long, need new "now" value
	  continue;
	}

      FD_ZERO (&readfds);
      FD_ZERO (&writefds);
      maxfd = 0;

      mosqsock = mosquitto_socket (state->mosq);
      if (mosqsock != -1)
	{
	  FD_SET (mosqsock, &readfds);
	  if (mosquitto_want_write (state->mosq) == true)
	    FD_SET (mosqsock, &writefds);
	  if (mosqsock > maxfd)
	    maxfd = mosqsock;
	}

      if (state->idletimer_running)
	{
	  FD_SET (state->idletimer_fd, &readfds);
	  if (state->idletimer_fd > maxfd)
	    maxfd = state->idletimer_fd;
	}

      tv.tv_sec = 1;
      tv.tv_usec = 0;

      selectres = select (maxfd + 1, &readfds, &writefds, NULL, &tv);
      // We could handle errno==EINTR, but shouldn't occur in practice.
      if (selectres == -1)
	break;			// Stop on any error

      // Handle mqtt network input and output.

      if (mosqsock != -1 && FD_ISSET (mosqsock, &readfds))
	{
	  mosquitto_loop_read (state->mosq, 1);
	  // Most callbacks are called from here, including
	  // mqttmessage_callback() which handles any incoming messages
	  // (possibly multiple, if piled up in tcp buffer),
	  // and starts new foreground communication if possible.

	  // Don't mind exitcode; socket already disconnected in case of
	  // trouble, then also mqttdisconnect_callback() was called.
	}

      if (mosqsock != -1 && FD_ISSET (mosqsock, &writefds))
	{
	  mosquitto_loop_write (state->mosq, 1);
	}

      mosquitto_loop_misc (state->mosq);

      if (state->idletimer_running &&
	  FD_ISSET (state->idletimer_fd, &readfds))
	{
	  idletimer_expired (state);
	}
    }

  // This is reached only after our explicit shutdown.
  // If we were connected, DISCONNECT has been sent from inside the loop,
  // and broker hopefully has neatly disconnected us within our timeout.
  // (Else: don't care, we'll close the socket anyway.)

  //mosquitto_destroy() -> in main

  return;
}


int
main (int argc, char **argv)
{
  state_t state;
  int res;
  char clientid[MOSQ_MQTT_ID_MAX_LENGTH + 5];
  char hostname[TOPICLEN + 5];

  alarm (5);

  // Initialize
  memset (&state, 0, sizeof (state));

  // Set default configuration
  state.svr_hostname = "127.0.0.1";
  state.svr_port = 1883;

  res = parse_cmdline (argc, argv, &state);
  if (res >= 1)
    exit (res - 1);

  // Note that broker kicks any connected client with equal clientid.
  // Clientid is max 23 chars.
  if (gethostname (hostname, TOPICLEN) != 0)
    hostname[0] = '\0';
  res =
    snprintf (clientid, MOSQ_MQTT_ID_MAX_LENGTH, "%d/%s", getpid (),
	      hostname);
  // Fake resultvalue handling to make gcc stop complaining about the
  // intended truncation, which will actually always have res LARGER than 1.
  if (res < 1)
    exit (1);

  mosquitto_lib_init ();	// cannot return errors

  state.mosq = mosquitto_new (clientid, true, &state);
  if (state.mosq == NULL)
    {
      STDERR ("mosquitto_new failed: %s", strerror (errno));
      return 1;
    }

  mosquitto_connect_callback_set (state.mosq, mqttconnect_callback);
  mosquitto_disconnect_callback_set (state.mosq, mqttdisconnect_callback);
  mosquitto_message_callback_set (state.mosq, mqttmessage_callback);

  // Init timer
  state.idletimer_fd = timerfd_create (CLOCK_MONOTONIC, 0);
  if (state.idletimer_fd == -1)
    {
      STDERR ("failed to create timerfd: %s", strerror (errno));
      exit (1);
    }

  mainloop (&state);
  // Returns after "shutdown"

  mosquitto_destroy (state.mosq);
  // This is the _only_ way an mqtt client can forcibly close the socket.

  return 0;
}
