// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <unistd.h>

#include "render_tile.h"
#include "recv_pub.h"
#include "mathgl_prepare.h"

void
render_tile (state_t * state, data_t * data, json_t * jrender,
	     const char *rendername, const char *rendername_sane)
{
  const char *planelabel;
  int planeidx;
  int imgw, imgh;
  char destfname[FNAMELEN + 10];
  char tmpfname[FNAMELEN + 10];
  plane_t *plane;
  HMGL gr;
  double zmarks, zminround, zmaxround;

  planelabel = json_string_value (json_object_get (jrender, "plane-color"));
  if (planelabel == NULL)
    {
      MDBG ("render object has no plane label");
      return;
    }

  planeidx = lookup_plane (data, planelabel);
  if (planeidx < 0)
    {
      MDBG ("plane %s not found", planelabel);
      return;
    }
  plane = &data->planes[planeidx];

  imgw = json_integer_value (json_object_get (jrender, "pixw"));
  imgh = json_integer_value (json_object_get (jrender, "pixh"));
  if (imgw <= 0 || imgh <= 0)
    {
      MDBG ("can't very well make %d x %d picture", imgw, imgh);
      return;
    }

  if (!makefilenames (state, rendername_sane, "png", destfname, tmpfname))
    return;

  mathgl_init_plane_1extra (state, data, planeidx);

  gr = mgl_create_graph (imgw, imgh);

  //mgl_subplot (gr, 1, 1, 0, "");      // less white borders
  // almost no borders, colorbar takes space at right
  mgl_inplot (gr, -.22, 1.13, -.22, 1.22);

  mgl_set_range_val (gr, 'x',
		     ((double) (-1 - data->axes[0].mgl_extra_items_at_front))
		     * data->axes[0].interval + data->axes[0].firstitem,
		     ((double) (data->axes[0].mgl_items
				- data->axes[0].mgl_extra_items_at_front))
		     * data->axes[0].interval + data->axes[0].firstitem);

  mgl_set_range_val (gr, 'y',
		     ((double) (-1 - data->axes[1].mgl_extra_items_at_front))
		     * data->axes[1].interval + data->axes[1].firstitem,
		     ((double) (data->axes[1].mgl_items
				- data->axes[1].mgl_extra_items_at_front))
		     * data->axes[1].interval + data->axes[1].firstitem);

  zmarks = plane->mathgl_plane_1extra.mmax + 0.02;
  zminround = floor (plane->mathgl_plane_1extra.mmin - 0.02);
  zmaxround = ceil (zmarks + 0.01);

  mgl_set_range_val (gr, 'z', zminround, zmaxround);
  mgl_set_range_val (gr, 'c', zminround, zmaxround);

  //mgl_aspect (gr, NAN, 1, 0.01);      // force square grid
  // Hmm. Buggy. xintv/yintv 0.5, data "1 2 3", "20 1 3" will give totally
  // wrong aspect; only "20 6 3" will fix again...
  // Old way via
  // https://sourceforge.net/p/mathgl/discussion/508393/thread/84f74c00/
  //mgl_aspect (gr, imgh * (data->axes[0].mgl_items + 1),
  //              imgw * (data->axes[1].mgl_items + 1), 1.0);
  // Works much better ...but now grid is always *over* data
  mgl_aspect (gr, imgh * (data->axes[0].mgl_items + 1),
	      imgw * (data->axes[1].mgl_items + 1), imgw);
  // Good, now only sometimes lowest tile will get crossed by grid,
  // seems unavoidable.

  // axes/grid position
  if (plane->mathgl_plane_1extra.filled_factor < 0.20)	// ~~non-interpolated
    {
      mgl_set_origin (gr, 0, 0, NAN);
      // z=NAN will always set at bottom of graph
      //mgl_set_origin (gr, 0, 0, plane->mathgl_plane_1extra.mmin - 0.01);
      // Hmm, doesn't really help...
    }
  else
    {
      mgl_set_origin (gr, 0, 0, plane->mathgl_plane_1extra.mmax + 0.01);
    }

  // force ticks at integer intervals
  mgl_set_ticks (gr, 'x', 1., 0, 0);
  mgl_set_ticks (gr, 'y', 1., 0, 0);
  mgl_set_ticks (gr, 'z', 1., 0, 0);
  mgl_set_ticks (gr, 'c', 1., 0, 0);

  mgl_load_font (gr, "heros", "/usr/share/mathgl/fonts");

// Colorscheme must be #define: need to attach strings to it
//#define COLORSCHEME "BbcyrR"
#define COLORSCHEME "UbcyqR"

  mgl_tile_xy (gr,
	       data->axes[0].mgl_list_minushalf_1extra,
	       data->axes[1].mgl_list_minushalf_1extra,
	       plane->mathgl_plane_1extra.mdata, COLORSCHEME, "");

  mgl_set_font_size (gr, 1.5);
  //mgl_colorbar_ext (gr, COLORSCHEME, 0.80, 0.20, 0.5, 0.5);
  mgl_colorbar (gr, COLORSCHEME ">I");

  mgl_set_font_size (gr, 1.5);
  mgl_axis (gr, "xyU", "k1-", "");
  mgl_axis_grid (gr, "xy", "W0-", "");

  if (data->lastreceived.valid)
    {
      // Current position marker, thick-white + thin-magenta
      mgl_ellipse (gr,
		   data->lastreceived.x, data->lastreceived.y, zmarks,
		   data->lastreceived.x, data->lastreceived.y, zmarks,
		   (data->axes[0].interval + data->axes[1].interval) * 0.5,
		   "w5-#");
      mgl_ellipse (gr,
		   data->lastreceived.x, data->lastreceived.y, zmarks,
		   data->lastreceived.x, data->lastreceived.y, zmarks,
		   (data->axes[0].interval + data->axes[1].interval) * 0.5,
		   "m2-#");
    }

  unlink (tmpfname);
  mgl_write_frame (gr, tmpfname, "");

  mgl_delete_graph (gr);

  if (rename (tmpfname, destfname) != 0)
    return;

  announcefile (state, data->collecttopic, rendername, rendername_sane,
		"png", "png", imgw, imgh);

#undef COLORSCHEME
}
