// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/timerfd.h>

#include "state.h"
#include "calc.h"
#include "../shared/error.h"
#include "../shared/mqttpub.h"
#include "../camfuncs/camfuncs.h"
#include "../camfuncs/camhelpers.h"
#include "../cmpfit/mpfit.h"


// At reasonably large distances, a camera pixel that would itself be
// 0.003-0.005 in the rescaled -1..+1 coord, will equal some 1-5 cm
// (0.01-0.05m) depending on camera type (zoom).
// These factors multiply the height/distance measurement errors to be
// more comparable to pixel errors.
#define RESIDFACT_HEIGHT   0.1
#define RESIDFACT_DIST     0.1

// For most *correct* scene fit results, sum-of-squares will be less
// than 0.0005.
#define DEFAULT_MAXSUMSQ  0.050


// Note: *MASSIVE* output.
// MDBG is not useful: everything will get buffered, only sent (all at once)
// after calc finished.
//#define FITDBG(args...)  MDBG(args)
//#define FITDBG(args...)  STDERR(args)
#define FITDBG(args...)  do { } while(0)

// MDBG is sent only in next select() round, i.e. al at once AFTER calc is done.
// For easier debug, switch to stderr
//#undef MDBG
//#define MDBG(args...)  STDERR(args)
//#define MDBG(args...)  do { } while(0)

// For easily commenting out individual MDBG's: replace by noMDBG
#define noMDBG(args...)  do { } while(0)


#define OBJ_NONE    0
#define OBJ_OBJ     10
#define OBJ_CAM     11

// This struct is used for both cameras and object positions.
// Objects have only name and x/y/z/fitparamidx.
// Cameras are added as they announce themselves on scwips/cam/<name>/info;
// are never removed.
// Objects are only added temporarily during scene calculation, then removed
// immediately.
// So the first few items will always be all cameras.
typedef struct
{
  int objtype;			// OBJ_NONE, OBJ_OBJ, OBJ_CAM
  char *name;			// mallocd!
  int sceneref;			// #refs in scene (0=not in scene, useless)

  int camtype;			// index to camtype[]
  double *typeparam;		// malloc'd for specified type
  int camdistortion;		// index to camdistortion[]
  double *distortionparam;	// malloc'd for specified distortion
  double dist_lin_radius;	// distortion linearized beyond this limit

  int w;			// camera #pixels
  int h;
  double xscale;		// to convert pixel into -1..+1 coords
  double xoffset;
  double yscale;
  double yoffset;

  double in_posrot[6];		// position + (cam-only) rotation: x,y,z,a,b,g
  int heightmeasured;

  double work_posrot[6];
  int pos_fitparamidx;		// index where x,y,z(a,b,g) are in param array
  int posrot_fixonce[6];
  int posrot_fixalways[6];
  int has_fix;

  double residscalex;		// 1/(pixelsize * #blinkpix)
  double residscaley;

  int selected;			// used in fitting

  int cnt_camsseen;		// selection criteria
  int cnt_pixobserv;
  int cnt_distobserv;
  int cnt_heightobserv;
  int cnt_obj1;
  int cnt_obj2;
  int selectable;
} cam_obj_t;


#define cam_t cam_obj_t

// calc_campix requires cam_t to be defined
#include "../shared/calc_campix.h"

// parse_cam_json requires cam_t to be defined
#define CAMERR_RETURN(args...)  do { MDBG(args); return 1; } while(0)
#include "../shared/parse_cam_json.h"
#undef CAMERR_RETURN

#undef cam_t


#define OBS_PIX     2
#define OBS_DIST    3
#define OBS_HEIGHT  4

// This holds all scene observations;
// which fields are used depends on type.
typedef struct
{
  int obs_type;			// PIX / DIST / HEIGHT
  int obj1_num;
  int obj2_num;			// if PIX then obj2 must be camera
  double pixx;			// -1..+1 coords (if PIX)
  double pixy;
  double pixw_inv;		// 1/width for residual scaling
  double pixh_inv;		// 1/height
  double meas;			// distance or height

  int selected;			// used in fitting
} observ_t;


cam_obj_t *cam_obj = NULL;	// mallocd
int cam_obj_allocd = 0;
int cam_obj_used = 0;
int cam_obj_cams = 0;		// 0..N-1 are cams, N..used are objects

observ_t *observ = NULL;	// mallocd
int observ_allocd = 0;
int observ_used = 0;


// Used on scwips/cam/c0/info announce, and for scenecalc objects
int
find_make_cam_obj (const char *name)
{
  int i;

  for (i = 0; i < cam_obj_used; i++)
    {
      if (cam_obj[i].name != NULL)
	if (!strcmp (cam_obj[i].name, name))
	  return i;
    }

  // If still here, then not found: add it.

  if (cam_obj_used >= cam_obj_allocd)
    {
      cam_obj_allocd = cam_obj_used * 2;
      if (cam_obj_allocd < 100)
	cam_obj_allocd = 100;
      cam_obj = realloc (cam_obj, sizeof (*cam_obj) * cam_obj_allocd);
      if (cam_obj == NULL)
	{
	  // Now we have a big problem. Can't assume mqtt will still work.
	  STDERR ("realloc failed");
	  exit (1);
	}
      // memset(0) is done below; is required for re-use as well.
    }

  memset (&cam_obj[cam_obj_used], 0, sizeof (cam_obj[cam_obj_used]));
  cam_obj[cam_obj_used].name = strdup (name);	// =malloc
  cam_obj[cam_obj_used].objtype = OBJ_NONE;

  //cam_obj[cam_obj_used].objtype = OBJ_OBJ;  --leave to caller
  //cam_obj_cams = cam_obj_used + 1;  --leave to caller

  cam_obj_used++;
  return cam_obj_used - 1;
}

// Cam must already be known
int
find_cam (char *name)
{
  int i;

  for (i = 0; i < cam_obj_cams; i++)
    {
      if (cam_obj[i].name != NULL)
	if (!strcmp (cam_obj[i].name, name))
	  {
	    if (cam_obj[i].objtype == OBJ_CAM)
	      return i;
	    else
	      {
		MDBG ("name %s is known but not a camera %d, skipping",
		      name, cam_obj[i].objtype);
		return -1;
	      }
	  }
    }
  MDBG ("name %s is not known, skipping", name);
  return -1;
}

inline int
new_observ ()
{
  if (observ_used >= observ_allocd)
    {
      observ_allocd = observ_used * 2;
      if (observ_allocd < 100)
	observ_allocd = 100;
      observ = realloc (observ, sizeof (*observ) * observ_allocd);
      if (observ == NULL)
	{
	  // Now we have a big problem. Can't assume mqtt will still work.
	  STDERR ("realloc failed");
	  exit (1);
	}
      // memset(0) is done below; is required for re-use as well.
    }

  memset (&observ[observ_used], 0, sizeof (observ[observ_used]));

  observ_used++;
  return observ_used - 1;
}


// Adapted from testmpfit.c
void
mp_printresult (double *x, mp_result * result)
{
  int i;

  if ((x == 0) || (result == 0))
    return;
  printf ("  CHI-SQUARE = %f    (%d DOF)\n",
	  result->bestnorm, result->nfunc - result->nfree);
  printf ("   eqv resid = %f\n", sqrt (result->bestnorm / result->nfunc));
  // This is not the same as average residual, since residuals are weighted
  // by themselves, i.e. larger residuals weigh more in the sum-of-squares.
  // When residual values are really spread out, there will be relatively more
  // large residuals with large weights, so the difference will get more
  // pronounced.

  printf ("  params     = %d\n", result->npar);
  printf ("   - free    = %d\n", result->nfree);
  printf ("   - pegged  = %d\n", result->npegged);
  printf ("  iterations = %d\n", result->niter);
  printf ("  func evals = %d\n", result->nfev);
  for (i = 0; i < result->npar; i++)
    {
      printf ("  P[%d] = %f +/- %f\n", i, x[i], result->xerror[i]);
    }
  if (result->resid != NULL)
    {
      double sum = 0.;
      double abssum = 0.;
      double sqsum = 0.;
      double absmax = 0.;
      for (i = 0; i < result->nfunc; i++)
	{
	  sum += result->resid[i];
	  abssum += fabs (result->resid[i]);
	  sqsum += result->resid[i] * result->resid[i];
	  if (fabs (result->resid[i]) > absmax)
	    absmax = fabs (result->resid[i]);
	  printf ("  resid[%d] = %f\n", i, result->resid[i]);
	}
      printf ("  resid avg = %f   avg-of-abs = %f   sum-of-squares = %f\n",
	      sum / result->nfunc, abssum / result->nfunc, sqsum);
      printf ("  resid max-of-abs = %f\n", absmax);
    }
}

void
mqtt_mp_printresult (mp_result * result, char *addstr)
{
  int i;
  double sum = 0.;
  double abssum = 0.;
  double absmax = 0.;

  if (result == NULL)
    {
      MDBG ("called with zeropointer");
      return;
    }

  if (result->resid != NULL)
    for (i = 0; i < result->nfunc; i++)
      {
	sum += result->resid[i];
	abssum += fabs (result->resid[i]);
	if (fabs (result->resid[i]) > absmax)
	  absmax = fabs (result->resid[i]);
      }

  mqttprintf (state->mosq, state->topicprefix, "scene", state->name, "log",
	      NULL,
	      "status %d resid num %d sumsq %f avgabs %f maxabs %f "
	      "iter %d func %d " "param num %d free %d pegged %d dof %d%s",
	      result->status, result->nfunc, result->bestnorm,
	      abssum / result->nfunc, absmax, result->niter, result->nfev,
	      result->npar, result->nfree, result->npegged,
	      result->nfunc - result->nfree, addstr);
  // See comments in mp_printresult
}


void
mqtt_postresult ()
{
  json_t *jroot;
  int i;

  if (!state->resultokay)
    return;

  jroot = json_array ();
  json_array_append_new (jroot, json_pack ("{s:{s:s, s:i, s:f, s:f, s:i}}",
					   "global",
					   "instance", state->name,
					   "attempts", state->attempts,
					   "sumsquares", state->res_sumsq,
					   "sumsquares-limit",
					   state->res_maxsumsq,
					   "rounds", state->res_rounds));

  for (i = 0; i < cam_obj_used; i++)
    {
      if (!cam_obj[i].selected)
	continue;

      if (cam_obj[i].objtype == OBJ_CAM)
	{
	  json_array_append_new (jroot,
				 json_pack
				 ("{s:s, s:f, s:f, s:f, s:f, s:f, s:f}",
				  "cam", cam_obj[i].name, "x",
				  cam_obj[i].work_posrot[0], "y",
				  cam_obj[i].work_posrot[1], "z",
				  cam_obj[i].work_posrot[2], "a",
				  cam_obj[i].work_posrot[3], "b",
				  cam_obj[i].work_posrot[4], "g",
				  cam_obj[i].work_posrot[5]));
	}
      else
	{
	  json_array_append_new (jroot, json_pack ("{s:s, s:f, s:f, s:f}",
						   "obj", cam_obj[i].name,
						   "x",
						   cam_obj[i].work_posrot[0],
						   "y",
						   cam_obj[i].work_posrot[1],
						   "z",
						   cam_obj[i].work_posrot
						   [2]));
	}
    }

  mqttpubjson (state->mosq, state->topicprefix, "scene/result", NULL, jroot);
  // already decref's jroot
}


// Fit functions work like this:
//
// Caller has guessed param[nparam], and this fit function must calculate
// how correct those guesses are.
// It does so by assuming the guesses are fully correct, and calculates what
// the results would have been of the nmeas measurements that were done.
// Of course, those calculated results will be different from the measured
// results, and these differences must be stored in residual[nmeas].
// Caller will then use these residuals to adjust the param guesses.
//
// *private can be used to pass any data that is needed, such as the measured
// results to compare against.
//
// (derivvec[])[], if non-NULL, can be used to store derivatives if calculable.
//
// Returns 0 = success.


int
scenepos_fitfunc (int nmeas, int nparam, double *param, double *residual,
		  double **derivvec, void *private)
{
  // We do all the observ entries, as far as selected; no private data needed.
  // Sometimes two results per entry: pixx and pixy.
  int i;
  int meascnt;

  meascnt = 0;
  for (i = 0; i < observ_used; i++)
    {
      if (!observ[i].selected)
	continue;

      if (observ[i].obs_type == OBS_HEIGHT)
	{
	  int o1;
	  int p_pos1;
	  double objz;

	  o1 = observ[i].obj1_num;
	  p_pos1 = cam_obj[o1].pos_fitparamidx;
	  objz = param[p_pos1 + 2];

	  FITDBG ("  observ %d = height obj %d, guess %f", i, o1, objz);

	  residual[meascnt] = (observ[i].meas - objz) * RESIDFACT_HEIGHT;

	  FITDBG ("  wanted height %f  resid %f",
		  observ[i].meas, residual[meascnt]);

	  meascnt++;
	}

      else if (observ[i].obs_type == OBS_DIST)
	{
	  int o1, o2;
	  int p_pos1, p_pos2;
	  double dist;

	  o1 = observ[i].obj1_num;
	  o2 = observ[i].obj2_num;
	  p_pos1 = cam_obj[o1].pos_fitparamidx;
	  p_pos2 = cam_obj[o2].pos_fitparamidx;

	  FITDBG ("  observ %d = dist obj %d to %d", i, o1, o2);

	  FITDBG ("  position guess %d x %f y %f z %f", o1,
		  param[p_pos1], param[p_pos1 + 1], param[p_pos1 + 2]);

	  FITDBG ("  position guess %d x %f y %f z %f", o2,
		  param[p_pos2], param[p_pos2 + 1], param[p_pos2 + 2]);

	  dist =
	    distance_3d (param[p_pos1], param[p_pos1 + 1], param[p_pos1 + 2],
			 param[p_pos2], param[p_pos2 + 1], param[p_pos2 + 2]);

	  FITDBG ("  dist guess %f", dist);

	  residual[meascnt] = (observ[i].meas - dist) * RESIDFACT_DIST;

	  FITDBG ("  wanted dist %f  resid %f",
		  observ[i].meas, residual[meascnt]);

	  meascnt++;
	}

      else if (observ[i].obs_type == OBS_PIX)
	{
	  int o1, o2;
	  int p_pos1, p_pos2;
	  double campixx, campixy;

	  o1 = observ[i].obj1_num;
	  o2 = observ[i].obj2_num;
	  p_pos1 = cam_obj[o1].pos_fitparamidx;
	  p_pos2 = cam_obj[o2].pos_fitparamidx;

	  FITDBG ("  observ %d = obj %d to cam %d pixel", i, o1, o2);

	  calc_campix (&cam_obj[o2], &param[p_pos2],
		       &param[p_pos1], &campixx, &campixy);

	  residual[meascnt] =
	    (observ[i].pixx - campixx)
	    * cam_obj[o2].residscalex * observ[i].pixw_inv;

	  residual[meascnt + 1] =
	    (observ[i].pixy - campixy)
	    * cam_obj[o2].residscaley * observ[i].pixh_inv;

	  FITDBG ("  wanted x %f y %f  resid x %f y %f",
		  observ[i].pixx, observ[i].pixy,
		  residual[meascnt], residual[meascnt + 1]);

	  meascnt += 2;
	}
    }

  // Doublecheck that nmeas is always exactly done now,
  // i.e. that we never get called with a meas subset;
  // that's required for pixx/pixy combi to keep working.
  if (meascnt != nmeas)
    {
      // Fatal error, need to exit, i.e. mqtt will never get posted
      STDERR ("FATAL: called with nmeas = %d instead of expected %d, "
	      "which is not currently handled", nmeas, meascnt);
      exit (1);
    }

  return 0;
}


// Returns:
// 1: okay result
// 0: rejected result
// -1: parameter error
int
calc_scene ()
{
  int i;
  int nmeas;
  int nparam;
  double *resid;
  double *param;
  double *param_error;
  mp_par *param_settings;
  mp_result result;
  mp_config mpconfig;
  int fitres;
  int returnval = -1;

  nmeas = 0;
  nparam = 0;

  MDBG ("Calculating camera/object positions/rotations");

  // Prepare residual scaling, see receive_campix() for explanation.
  // the pixelwidth/height go via observ pixw_inv/pixh_inv.
  for (i = 0; i < cam_obj_cams; i++)
    {
      cam_obj[i].residscalex = 0.005 / cam_obj[i].xscale;
      cam_obj[i].residscaley = 0.005 / cam_obj[i].yscale;
    }

  // height/dist supply one meas, campix supplies two.
  for (i = 0; i < observ_used; i++)
    {
      if (!observ[i].selected)
	continue;

      if (observ[i].obs_type == OBS_HEIGHT || observ[i].obs_type == OBS_DIST)
	nmeas++;
      else if (observ[i].obs_type == OBS_PIX)
	nmeas += 2;
      else
	{
	  MDBG ("observ[%d] has illegal type %d, can't calculate.",
		i, observ[i].obs_type);
	  return -1;
	}
    }

  // cam needs x/y/z/a/b/g, obj needs x/y/z
  for (i = 0; i < cam_obj_used; i++)
    {
      if (!cam_obj[i].selected)
	continue;

      if (cam_obj[i].objtype == OBJ_CAM)
	{
	  cam_obj[i].pos_fitparamidx = nparam;
	  nparam += 6;
	}
      else if (cam_obj[i].objtype == OBJ_OBJ)
	{
	  cam_obj[i].pos_fitparamidx = nparam;
	  nparam += 3;
	}
      else if (cam_obj[i].objtype != OBJ_NONE)
	{
	  MDBG ("cam_obj[%d] has illegal type %d, can't calculate.",
		i, cam_obj[i].objtype);
	  return -1;
	}
    }

  // Now we can allocate the arrays
  resid = calloc (nmeas, sizeof (double));
  param = calloc (nparam, sizeof (double));
  param_error = calloc (nparam, sizeof (double));
  param_settings = calloc (nparam, sizeof (mp_par));

  if (resid == NULL || param == NULL || param_error == NULL
      || param_settings == NULL)
    {
      STDERR ("malloc failed");
      exit (1);
    }

  // Initialize from starting guesses

  for (i = 0; i < cam_obj_used; i++)
    {
      int p, j;

      if (!cam_obj[i].selected)
	continue;

      p = cam_obj[i].pos_fitparamidx;

      for (j = 0; j < (cam_obj[i].objtype == OBJ_CAM ? 6 : 3); j++)
	{
	  param[p + j] = cam_obj[i].work_posrot[j];

	  param_settings[p + j].fixed =
	    !!(cam_obj[i].posrot_fixonce[j] ||
	       cam_obj[i].posrot_fixalways[j]);
	}

      param_settings[p].step = 0.001;
      param_settings[p + 1].step = 0.001;
      param_settings[p + 2].step = 0.001;

      param_settings[p].advlimited = 1;
      param_settings[p].advlimit = 0.5;
      param_settings[p + 1].advlimited = 1;
      param_settings[p + 1].advlimit = 0.5;
      param_settings[p + 2].advlimited = 1;
      param_settings[p + 2].advlimit = 0.5;

      if (cam_obj[i].objtype == OBJ_CAM)
	{
	  // wraps into -period/2 .. period/2
	  param_settings[p + 3].periodic = 1;
	  param_settings[p + 3].period = 2. * M_PI;
	  param_settings[p + 3].advlimited = 1;
	  param_settings[p + 3].advlimit = 2.;
	  param_settings[p + 3].step = 0.001;

	  param_settings[p + 4].periodic = 1;
	  param_settings[p + 4].period = 2. * M_PI;
	  param_settings[p + 4].advlimited = 1;
	  param_settings[p + 4].advlimit = 2.;
	  param_settings[p + 4].step = 0.001;

	  param_settings[p + 5].periodic = 1;
	  param_settings[p + 5].period = 2. * M_PI;
	  param_settings[p + 5].advlimited = 1;
	  param_settings[p + 5].advlimit = 2.;
	  param_settings[p + 5].step = 0.001;
	}
    }

#if 0
  for (i = 0; i < nparam; i++)
    //or FITDBG
    MDBG ("  startguess param[%d]=%f  fixed=%d",
	  i, param[i], param_settings[i].fixed);
#endif

  memset (&mpconfig, 0, sizeof (mpconfig));
  mpconfig.maxiter = 2000;	// usually *good* result needs <200

  memset (&result, 0, sizeof (result));
  result.xerror = param_error;
  result.resid = resid;

  fitres = mpfit (scenepos_fitfunc, nmeas, nparam, param, param_settings,
		  &mpconfig, NULL, &result);
  FITDBG ("  mpfit status = %d", fitres);
  //mp_printresult (param, &result);
  mqtt_mp_printresult (&result, "");

  state->res_sumsq = result.bestnorm;
  state->res_rounds += result.niter;

  // We accept worse result here, since height and distance measurements
  // probably lead to some inaccuracies that refinement will fix better.
  if (fitres >= 1 && fitres <= 4 &&
      result.bestnorm < 10. * state->res_maxsumsq)
    {
      // mpfit has reliable result
      //
      // mqttlog-20181013-212413.scenes.ae gives 4 with useful result,
      // earlier for first round, now for second round.

      MDBG ("Refining camera/object positions/rotations without fixes");

      for (i = 0; i < cam_obj_used; i++)
	{
	  int p, j;

	  if (!cam_obj[i].selected)
	    continue;

	  p = cam_obj[i].pos_fitparamidx;

	  for (j = 0; j < (cam_obj[i].objtype == OBJ_CAM ? 6 : 3); j++)
	    {
	      param_settings[p + j].fixed =
		!!(cam_obj[i].posrot_fixalways[j]);

	      // Note: _fixonce is left for future attempts
	      // in case this refinement doesn't work out.
	    }
	}

#if 0
      for (i = 0; i < nparam; i++)
	//or FITDBG
	MDBG ("  startguess param[%d]=%f  fixed=%d",
	      i, param[i], param_settings[i].fixed);
#endif

      memset (&mpconfig, 0, sizeof (mpconfig));
      mpconfig.maxiter = 500;

      memset (&result, 0, sizeof (result));
      result.xerror = param_error;
      result.resid = resid;

      fitres = mpfit (scenepos_fitfunc, nmeas, nparam, param, param_settings,
		      &mpconfig, NULL, &result);
      FITDBG ("  mpfit status = %d", fitres);
      //mp_printresult (param, &result);
      mqtt_mp_printresult (&result, "");

      state->res_sumsq = result.bestnorm;
      state->res_rounds += result.niter;
    }

  if (fitres >= 1 && fitres <= 4 && result.bestnorm < state->res_maxsumsq)
    {
      // mpfit has reliable result

      for (i = 0; i < cam_obj_used; i++)
	{
	  int p;

	  if (!cam_obj[i].selected)
	    continue;

	  p = cam_obj[i].pos_fitparamidx;

	  cam_obj[i].work_posrot[0] = param[p];
	  cam_obj[i].work_posrot[1] = param[p + 1];
	  cam_obj[i].work_posrot[2] = param[p + 2];

	  if (cam_obj[i].objtype == OBJ_CAM)
	    {
	      cam_obj[i].work_posrot[3] = param[p + 3];
	      cam_obj[i].work_posrot[4] = param[p + 4];
	      cam_obj[i].work_posrot[5] = param[p + 5];
	    }
	}

#if 0
      // debug: print all residuals
      if (result.resid != NULL)
	{
	  int measidx = 0;

	  for (i = 0; i < observ_used; i++)
	    {
	      if (!observ[i].selected)
		continue;

	      if (observ[i].obs_type == OBS_HEIGHT)
		{
		  MDBG ("  resid[%d]=%f  height %s",
			measidx, result.resid[measidx],
			cam_obj[observ[i].obj1_num].name);
		  measidx++;
		}
	      else if (observ[i].obs_type == OBS_DIST)
		{
		  MDBG ("  resid[%d]=%f  dist %s %s",
			measidx, result.resid[measidx],
			cam_obj[observ[i].obj1_num].name,
			cam_obj[observ[i].obj2_num].name);
		  measidx++;
		}
	      else if (observ[i].obs_type == OBS_PIX)
		{
		  MDBG ("  resid[%d]=%f  pix %s %s",
			measidx, result.resid[measidx],
			cam_obj[observ[i].obj1_num].name,
			cam_obj[observ[i].obj2_num].name);
		  measidx++;
		  MDBG ("  resid[%d]=%f  pix %s %s",
			measidx, result.resid[measidx],
			cam_obj[observ[i].obj1_num].name,
			cam_obj[observ[i].obj2_num].name);
		  measidx++;
		}
	    }
	}
#endif

      returnval = 1;
    }
  else if (fitres <= 0)
    {
      MDBG ("Fitting parameter error %d, stopping", fitres);
      returnval = -1;
    }
  else
    {
      // mpfit has result, but not reliable
      MDBG ("Result rejected");
      returnval = 0;
    }

  free (resid);
  free (param);
  free (param_error);
  free (param_settings);

  return returnval;
}


void
select_corresponding_objs_observs ()
{
  int i;

  // Reset selections for reference objects only
  for (i = cam_obj_cams; i < cam_obj_used; i++)
    cam_obj[i].selected = 0;

  // Reset selections for observations
  for (i = 0; i < observ_used; i++)
    observ[i].selected = 0;

  for (i = 0; i < cam_obj_used; i++)
    cam_obj[i].cnt_camsseen = 0;

  // Count reference-objects as visible from selected cameras
  for (i = 0; i < observ_used; i++)
    {
      if (observ[i].obs_type == OBS_PIX)
	{
	  // obj2 is always a camera
	  if (cam_obj[observ[i].obj2_num].selected)
	    cam_obj[observ[i].obj1_num].cnt_camsseen++;
	}
    }

  // Select all useful reference objects
  for (i = cam_obj_cams; i < cam_obj_used; i++)
    {
      if (cam_obj[i].objtype == OBJ_OBJ &&
	  cam_obj[i].sceneref >= 2 && cam_obj[i].cnt_camsseen >= 2)
	cam_obj[i].selected = 1;
    }

  // Select all observs that correspond to selected cameras / reference objects
  for (i = 0; i < observ_used; i++)
    {
      if (observ[i].obs_type != OBS_HEIGHT &&
	  observ[i].obs_type != OBS_DIST && observ[i].obs_type != OBS_PIX)
	continue;

      if (!cam_obj[observ[i].obj1_num].selected)
	continue;

      if (observ[i].obs_type == OBS_DIST || observ[i].obs_type == OBS_PIX)
	{
	  if (!cam_obj[observ[i].obj2_num].selected)
	    continue;
	}

      observ[i].selected = 1;
    }
}

void
select_all_useful_entries ()
{
  int i;

  for (i = 0; i < cam_obj_used; i++)
    cam_obj[i].selected = 0;

  // Select all useful cameras
  for (i = 0; i < cam_obj_used; i++)
    {
      // Cam needs at least 3 pixel observs to be calculable,
      // or some other combinations with more observs.
      if (cam_obj[i].objtype == OBJ_CAM && cam_obj[i].sceneref >= 3)
	cam_obj[i].selected = 1;
    }

  select_corresponding_objs_observs ();
}

// Set all cams selectable that see given reference object
// (mostly for first and second object)
// Returns number of newly selectable items.
int
set_selectable_when_seeing_obj (int obj_searched)
{
  int i;
  int num = 0;

  for (i = 0; i < cam_obj_used; i++)
    cam_obj[i].selectable = 0;

  for (i = 0; i < observ_used; i++)
    {
      if (observ[i].obs_type == OBS_PIX)
	{
	  // Obj2 is always a camera, so obj1 must be what we're looking for.
	  // Also check that camera is useful at all.
	  if (observ[i].obj1_num == obj_searched &&
	      cam_obj[observ[i].obj2_num].objtype == OBJ_CAM &&
	      cam_obj[observ[i].obj2_num].sceneref >= 3 &&
	      !cam_obj[observ[i].obj2_num].selected)
	    {
	      cam_obj[observ[i].obj2_num].selectable = 1;
	      num++;
	    }
	}
    }

  return num;
}

int
set_selectable_when_seeing_two_obj (int obj_searched1, int obj_searched2)
{
  int i;
  int num = 0;

  for (i = 0; i < cam_obj_used; i++)
    {
      cam_obj[i].selectable = 0;
      cam_obj[i].cnt_obj1 = 0;
      cam_obj[i].cnt_obj2 = 0;
    }

  for (i = 0; i < observ_used; i++)
    {
      if (observ[i].obs_type == OBS_PIX)
	{
	  // Obj2 is always a camera, so obj1 must be what we're looking for.
	  if (observ[i].obj1_num == obj_searched1)
	    cam_obj[observ[i].obj2_num].cnt_obj1++;

	  if (observ[i].obj1_num == obj_searched2)
	    cam_obj[observ[i].obj2_num].cnt_obj2++;
	}
    }

  for (i = 0; i < cam_obj_cams; i++)
    {
      if (cam_obj[i].cnt_obj1 >= 1 &&
	  cam_obj[i].cnt_obj2 >= 1 &&
	  cam_obj[i].objtype == OBJ_CAM &&
	  cam_obj[i].sceneref >= 3 && !cam_obj[i].selected)
	{
	  cam_obj[i].selectable = 1;
	  num++;
	}
    }

  return num;
}

// Set all cams selectable that have 3+ reference points connecting to
// already-selected cams.
// Returns number of newly selectable items.
int
set_selectable_when_connected ()
{
  int i;
  int num = 0;

  for (i = 0; i < cam_obj_used; i++)
    {
      cam_obj[i].selectable = 0;
      cam_obj[i].cnt_camsseen = 0;
      cam_obj[i].cnt_pixobserv = 0;
    }

  // Count reference-objects seen by selected cameras
  for (i = 0; i < observ_used; i++)
    {
      if (observ[i].obs_type == OBS_PIX)
	{
	  // obj2 is always a camera
	  if (cam_obj[observ[i].obj2_num].selected)
	    cam_obj[observ[i].obj1_num].cnt_camsseen++;
	}
    }

  // Count cameras seeing marked reference-objects
  for (i = 0; i < observ_used; i++)
    {
      if (observ[i].obs_type == OBS_PIX)
	{
	  // obj2 is always a camera
	  if (cam_obj[observ[i].obj1_num].cnt_camsseen >= 1)
	    cam_obj[observ[i].obj2_num].cnt_pixobserv++;
	}
    }

  for (i = 0; i < cam_obj_cams; i++)
    {
      if (cam_obj[i].cnt_pixobserv >= 3 &&
	  cam_obj[i].objtype == OBJ_CAM &&
	  cam_obj[i].sceneref >= 3 && !cam_obj[i].selected)
	{
	  cam_obj[i].selectable = 1;
	  num++;
	}
    }

  return num;
}

// Set all cams selectable that are not yet selected.
// Mostly last-resort.
// Returns number of newly selectable items.
int
set_selectable_all_unselected ()
{
  int i;
  int num = 0;

  for (i = 0; i < cam_obj_used; i++)
    {
      cam_obj[i].selectable = 0;
    }

  for (i = 0; i < cam_obj_cams; i++)
    {
      if (cam_obj[i].objtype == OBJ_CAM &&
	  cam_obj[i].sceneref >= 3 && !cam_obj[i].selected)
	{
	  cam_obj[i].selectable = 1;
	  num++;
	}
    }

  return num;
}

int
num_selected_cams_seeing_obj (int obj_searched)
{
  int i;
  int num = 0;

  for (i = 0; i < observ_used; i++)
    {
      if (observ[i].obs_type == OBS_PIX)
	{
	  // Obj2 is always a camera, so obj1 must be what we're looking for.
	  // Also check that camera is useful at all.
	  if (observ[i].obj1_num == obj_searched &&
	      cam_obj[observ[i].obj2_num].objtype == OBJ_CAM &&
	      cam_obj[observ[i].obj2_num].selected)
	    {
	      num++;
	    }
	}
    }

  return num;
}

// Random number 0..max inclusive
long
random_with_max (long max)
{
  // max * random() / RAND_MAX      will overflow when RAND_MAX = 2G
  // random() * ( max / RAND_MAX )  will always be 0
  // random() / ( RAND_MAX / max )  will work when max not too high
  // random() / ( RAND_MAX / (max+1) )  gets rounding correct

  long randnum, div;

  // Direct attempts when max is high enough
  if (max > RAND_MAX / 4)
    {
      do
	{
	  randnum = random ();
	}
      while (randnum > max);	// Loops <=4 times average
      return randnum;
    }

  // Use divisor when max is low enough that division will work;
  // rounding makes sure some overshoot is created to remain uniform.
  div = RAND_MAX / (max + 1);
  do
    {
      randnum = random () / div;
    }
  while (randnum > max);	// Loops <=2 times average
  return randnum;
}


// From all selectable camobj's, choose one randomly
// Returns: index
// or -1 if no selectable items at all.
int
select_random_from_selectable ()
{
  int i, num;
  int randnum;

  num = 0;
  for (i = 0; i < cam_obj_used; i++)
    if (cam_obj[i].selectable)
      num++;

  if (num <= 0)
    return -1;

  randnum = random_with_max (num - 1);	// 0..(num-1)

  num = 0;
  for (i = 0; i < cam_obj_used; i++)
    if (cam_obj[i].selectable)
      {
	if (num == randnum)
	  {
	    MDBG ("Selected %s", cam_obj[i].name);
	    cam_obj[i].selected = 1;
	    return i;
	  }
	num++;
      }

  // should never get here
  return -1;
}

// Returns: 1: success, 0: no cams any more
int
select_additional_cam ()
{
  int res;

  res = set_selectable_when_connected ();
  if (res <= 0)
    res = set_selectable_all_unselected ();
  if (res <= 0)
    return 0;

  res = select_random_from_selectable ();
  if (res >= 0)
    return 1;
  else
    return 0;
}

void
copy_selected_to_hasfix ()
{
  int i;

  for (i = 0; i < cam_obj_used; i++)
    cam_obj[i].has_fix = cam_obj[i].selected;
}

void
copy_hasfix_to_selected ()
{
  int i;

  for (i = 0; i < cam_obj_used; i++)
    cam_obj[i].selected = cam_obj[i].has_fix;
}

void
set_fixonce_if_hasfix ()
{
  int i;

  for (i = 0; i < cam_obj_used; i++)
    if (cam_obj[i].has_fix)
      {
	cam_obj[i].posrot_fixonce[0] = 1;
	cam_obj[i].posrot_fixonce[1] = 1;
	cam_obj[i].posrot_fixonce[2] = 1;

	if (cam_obj[i].objtype == OBJ_CAM)
	  {
	    cam_obj[i].posrot_fixonce[3] = 1;
	    cam_obj[i].posrot_fixonce[4] = 1;
	    cam_obj[i].posrot_fixonce[5] = 1;
	  }
      }
}


// Find out if enough data selected for possibly getting some solution
int
is_calculable ()
{
  int i;
  int cnt_unknowns = 0;
  int cnt_meas = 0;
  int cnt_heights = 0;
  int cnt_dists = 0;

  // Reset and count cameras / reference objects
  for (i = 0; i < cam_obj_used; i++)
    {
      cam_obj[i].cnt_pixobserv = 0;
      cam_obj[i].cnt_distobserv = 0;
      cam_obj[i].cnt_heightobserv = 0;

      if (!cam_obj[i].selected)
	continue;

      if (cam_obj[i].objtype == OBJ_CAM)
	cnt_unknowns += 6;
      else
	cnt_unknowns += 3;
    }

  // Count observations
  for (i = 0; i < observ_used; i++)
    {
      if (!observ[i].selected)
	continue;

      if (observ[i].obs_type == OBS_PIX)
	{
	  cam_obj[observ[i].obj1_num].cnt_pixobserv++;
	  cam_obj[observ[i].obj2_num].cnt_pixobserv++;
	  cnt_meas += 2;
	}
      else if (observ[i].obs_type == OBS_DIST)
	{
	  cam_obj[observ[i].obj1_num].cnt_distobserv++;
	  cam_obj[observ[i].obj2_num].cnt_distobserv++;
	  cnt_meas += 1;
	  cnt_dists++;
	}
      else if (observ[i].obs_type == OBS_HEIGHT)
	{
	  cam_obj[observ[i].obj1_num].cnt_heightobserv++;
	  cnt_meas += 1;
	  cnt_heights++;
	}
    }

  // Evaluate cameras / reference objects
  for (i = 0; i < cam_obj_used; i++)
    {
      if (!cam_obj[i].selected)
	continue;

      if (cam_obj[i].cnt_heightobserv > 1)
	cam_obj[i].cnt_heightobserv = 1;	// Only one height can be used

      if (cam_obj[i].objtype == OBJ_CAM)
	{
	  // Each camera has 6 DoF total.
	  // Rotation 3 DoF can only be provided by pixel observs,
	  // so requires at least 2 pixels.
	  // Plus either another pixel or two measurements (one of which
	  // may be height).

	  if (cam_obj[i].cnt_pixobserv >= 3)
	    continue;		// okay

	  if (cam_obj[i].cnt_pixobserv >= 2 &&
	      cam_obj[i].cnt_distobserv + cam_obj[i].cnt_heightobserv >= 2)
	    continue;		// okay

	  FITDBG ("camera %s is not calculable: pix %d dist %d height %d",
		  cam_obj[i].name, cam_obj[i].cnt_pixobserv,
		  cam_obj[i].cnt_distobserv, cam_obj[i].cnt_heightobserv);
	  return 0;
	}
      else
	{
	  // Each reference object has 3 DoF,
	  // where each pixels observ supplies 2,
	  // and each distance/height supplies 1.

	  if (cam_obj[i].cnt_pixobserv * 2 +
	      cam_obj[i].cnt_distobserv + cam_obj[i].cnt_heightobserv >= 3)
	    continue;		// okay

	  FITDBG ("object %s is not calculable: pix %d dist %d height %d",
		  cam_obj[i].name, cam_obj[i].cnt_pixobserv,
		  cam_obj[i].cnt_distobserv, cam_obj[i].cnt_heightobserv);
	  return 0;
	}
    }

  // Pixel and distance can only supply 2 resp. 1 DoF *in total*.
  // Assuming first+second reference points are always included (checked
  // below), x=y=0 and y=0 are not actually unknown, so subtract 3.
  if (cnt_meas < cnt_unknowns - 3)
    {
      FITDBG ("overall not calculable: meas %d < unknowns %d - 3", cnt_meas,
	      cnt_unknowns);
      return 0;
    }

  // Each reference point needs to be seen by at least two cams to be useful
  // --> already selected that way by select_corresponding_objs_observs()

  // First and second reference objects need to be selected,
  // and seen by at least two cameras.
  for (i = cam_obj_cams; i <= cam_obj_cams + 1; i++)
    {
      if (!cam_obj[i].selected || cam_obj[i].cnt_pixobserv < 2)
	{
	  FITDBG ("object %s pix %d < 2", cam_obj[i].name,
		  cam_obj[i].cnt_pixobserv);
	  return 0;
	}
    }

  // At least 3 heights required for balancing the entire scene in 3D space.
  if (cnt_heights < 3)
    {
      FITDBG ("overall not calculable: heights %d < 3", cnt_heights);
      return 0;
    }

  // At least 1 distance required for scale.
  if (cnt_dists < 1)
    {
      FITDBG ("overall not calculable: dists %d < 1", cnt_dists);
      return 0;
    }

  // If still here, then no problem detected.
  return 1;
}

void
random_init ()
{
  if (state->random_inited == 0)
    {
      unsigned int seed = 0;
      int fd;

      fd = open ("/dev/urandom", O_RDONLY);
      if (fd != -1)
	{
	  if (read (fd, &seed, sizeof (seed)) != sizeof (seed))
	    seed = 0;
	}
      close (fd);
      if (seed == 0)
	{
	  MDBG
	    ("Random init failed, parallel processing useless. Please go fix your stuff");
	  exit (1);
	}

      srandom (seed);
      MDBG ("Using random param init with seed %u", seed);
      state->random_inited = 1;
    }
}

double
find_firstseconddist ()
{
  // See if there is a distance between first and second *object*,
  // since these determine the x-axis, and we can fix second object's
  // position to positive coordinate.

  int i;
  int firstobject = cam_obj_cams;
  int secondobject = cam_obj_cams + 1;

  for (i = 0; i < observ_used; i++)
    {
      if (observ[i].obs_type == OBS_DIST &&
	  observ[i].obj1_num == firstobject &&
	  observ[i].obj2_num == secondobject)
	return observ[i].meas;

      if (observ[i].obs_type == OBS_DIST &&
	  observ[i].obj1_num == secondobject &&
	  observ[i].obj2_num == firstobject)
	return observ[i].meas;
    }

  return -1.;
}

void
nonfix_set_startingguesses ()
{
  int i;
  int firstobject = cam_obj_cams;
  int secondobject = cam_obj_cams + 1;
  double firstseconddist = find_firstseconddist ();

  for (i = 0; i < cam_obj_used; i++)
    {
      if (cam_obj[i].has_fix)
	continue;

      memset (cam_obj[i].posrot_fixonce, 0,
	      sizeof (cam_obj[i].posrot_fixonce));
      memset (cam_obj[i].posrot_fixalways, 0,
	      sizeof (cam_obj[i].posrot_fixalways));

      if (cam_obj[i].objtype == OBJ_CAM)
	{
	  // Init cameras in circle 5 .. 50 around origin, z in +1 .. +3,
	  // look at origin with +/-2.0 rad pan, +/-1.0 rad tilt/oblique.
	  // For orientation see campixtest/k180test1.json.

	  double radius = 5. + (45. * (double) (random ())) /
	    ((double) RAND_MAX);
	  double angl = (M_PI * 2. * (double) (random ())) /
	    ((double) RAND_MAX);

	  cam_obj[i].work_posrot[0] = -radius * sin (angl);
	  cam_obj[i].work_posrot[1] = -radius * cos (angl);
	  cam_obj[i].work_posrot[2] =
	    1. + (2. * (double) (random ())) / ((double) RAND_MAX);

	  cam_obj[i].work_posrot[3] =
	    angl - 2.0 + (4.0 * (double) (random ())) / ((double) RAND_MAX);
	  cam_obj[i].work_posrot[4] =
	    -1.0 + (2.0 * (double) (random ())) / ((double) RAND_MAX);
	  cam_obj[i].work_posrot[5] =
	    -1.0 + (2.0 * (double) (random ())) / ((double) RAND_MAX);
	}
      else
	{
	  // Init objects x,y in -25 .. +25; z in +0.1 .. +3.1
	  cam_obj[i].work_posrot[0] =
	    -25. + (50. * (double) (random ())) / ((double) RAND_MAX);
	  cam_obj[i].work_posrot[1] =
	    -25. + (50. * (double) (random ())) / ((double) RAND_MAX);
	  cam_obj[i].work_posrot[2] =
	    0.1 + (3. * (double) (random ())) / ((double) RAND_MAX);
	}

      // Override starting guesses from any non-zero input values
      {
	int j;
	for (j = 0; j < (cam_obj[i].objtype == OBJ_CAM ? 6 : 3); j++)
	  {
	    if (fabs (cam_obj[i].in_posrot[j]) > 0.00001)
	      cam_obj[i].work_posrot[j] = cam_obj[i].in_posrot[j];
	  }
      }

      // Force height
      if (cam_obj[i].heightmeasured)
	{
	  // The guessed height is always compared to measured height
	  // and error added to residual squared sum.
	  //
	  // For speed and less guesswork, height is fixed in first round,
	  // i.e. residuals will always be 0.
	  // Will then be unfixed for further refinement: there will
	  // always be slight inconsistencies if >3 heights are fixed.

	  // Already copied above, except in case height is exactly zero
	  cam_obj[i].work_posrot[2] = cam_obj[i].in_posrot[2];

	  cam_obj[i].posrot_fixonce[2] = 1;
	}

      // Fix first reference object to x=0, y=0 (origin)
      if (i == firstobject)
	{
	  cam_obj[i].work_posrot[0] = 0.;
	  cam_obj[i].work_posrot[1] = 0.;
	  cam_obj[i].posrot_fixalways[0] = 1;
	  cam_obj[i].posrot_fixalways[1] = 1;
	}

      // Fix second reference object to y=0 (x-axis),
      // and also x=fixed if distance is available
      if (i == secondobject)
	{
	  cam_obj[i].work_posrot[1] = 0.;
	  cam_obj[i].posrot_fixalways[1] = 1;

	  if (firstseconddist > 0.)
	    {
	      cam_obj[i].work_posrot[0] = firstseconddist;
	      cam_obj[i].posrot_fixonce[0] = 1;
	    }
	}
    }
}


void
calculate_once ()
{
  int i;
  int calc_res;
  int firstobject = cam_obj_cams;
  int secondobject = cam_obj_cams + 1;

  if (state->calc_full_restart)
    {
      state->calc_full_restart = 0;

      state->attempts++;
      state->res_rounds = 0;
      state->calc_had_fix_previously = 0;

      for (i = 0; i < cam_obj_used; i++)
	{
	  cam_obj[i].has_fix = 0;
	  cam_obj[i].selected = 0;
	}

      nonfix_set_startingguesses ();

      // Preferably, add two cameras that see both origin and y-axis point.
      // Failing that, add more cameras that see either origin or y-axis point,
      // such that both are seen twice.
      for (i = 0; i < 2; i++)
	{
	  if (set_selectable_when_seeing_two_obj (firstobject, secondobject)
	      <= 0)
	    break;

	  select_random_from_selectable ();
	}
      for (i = 0; i < 2; i++)
	{
	  if (num_selected_cams_seeing_obj (firstobject) >= 2)
	    break;

	  if (set_selectable_when_seeing_obj (firstobject) <= 0)
	    break;

	  select_random_from_selectable ();
	}
      for (i = 0; i < 2; i++)
	{
	  if (num_selected_cams_seeing_obj (secondobject) >= 2)
	    break;

	  if (set_selectable_when_seeing_obj (secondobject) <= 0)
	    break;

	  select_random_from_selectable ();
	}

      if (num_selected_cams_seeing_obj (firstobject) < 2 ||
	  num_selected_cams_seeing_obj (secondobject) < 2)
	{
	  // Earlier is_calculable() should prevent this from ever happening
	  MDBG ("First selection is impossible, stopping.");
	  state->resultokay = 0;
	  stop_calculating ();
	  return;
	}
    }

/***
  if calculable:
    calculate for current selection

  if never fix yet:
    - 50% chance: stop  (to next round: start all over again)
    - 50% chance: add cam, repeat
	but when no cam addable any more: stop  (next round)

  if fix now:
    - add cam, repeat
	but when no cam addable any more: final round (all_useful)

  if fix previously:
    - 10% chance: stop  (next round, to get possibly different initial fix)
    - 50% chance: forget all nonfix-added cams, also new starting guesses
	then:
    - add cam, repeat
	but when no cam addable any more:
		  - forget all nonfix-added cams, also new starting guesses
		  - add cam, repeat
		    (always possible: fix _previously_, i.e. some cam was
		     added after that fix was obtained.)
***/

  // Note that loop is externalized for better mqtt checks.
  // We get called repeatedly, and all state is still as we left it last time.
  // While() is just for easy "break" on success.
  while (1)
    {
      int has_fix_now = 0;

      select_corresponding_objs_observs ();
      set_fixonce_if_hasfix ();

      if (is_calculable ())
	{
	  calc_res = calc_scene ();

	  if (calc_res > 0)
	    {
	      has_fix_now = 1;
	      state->calc_had_fix_previously = 1;
	      copy_selected_to_hasfix ();
	    }
	  // (calc_res < 0): ignored, may be due to subset selection.
	}
      else
	{
	  MDBG ("This selection is not calculable");
	}

      if (!state->calc_had_fix_previously)
	{
	  int randnum = random_with_max (1);
	  if (randnum == 0)
	    {
	      MDBG ("Never fix yet, better try all over again");
	      state->calc_full_restart = 1;
	      return;
	    }

	  MDBG ("Never fix yet, but try adding another cam");
	  if (!select_additional_cam ())
	    {
	      MDBG ("No cams left to be added, try all over again");
	      state->calc_full_restart = 1;
	      return;
	    }

	  return;
	}

      if (has_fix_now)
	{
	  MDBG ("Fix now, try adding another cam");
	  if (!select_additional_cam ())
	    {
	      // MDBG() below loop
	      break;
	    }

	  return;
	}

      if (state->calc_had_fix_previously)	// always, at this point
	{
	  int randnum = random_with_max (9);
	  if (randnum < 1)
	    {
	      MDBG ("Had fix earlier, seems not working out, "
		    "better try all over again");
	      state->calc_full_restart = 1;
	      return;
	    }
	  else if (randnum < 6)
	    {
	      MDBG ("New cams seem not working out, trying from last fix");
	      copy_hasfix_to_selected ();
	      nonfix_set_startingguesses ();
	    }
	  else
	    {
	      MDBG ("Had fix earlier, try adding another cam");
	    }

	  if (!select_additional_cam ())
	    {
	      MDBG ("No cams left to be added, trying from last fix");
	      copy_hasfix_to_selected ();
	      nonfix_set_startingguesses ();
	      if (!select_additional_cam ())
		{
		  MDBG ("No cams can be added, "
			"but we must have done so previously??");
		  state->calc_full_restart = 1;
		  return;
		}
	    }

	  return;
	}

      // should never get here
      break;
    }

  MDBG ("No cameras left to be added, do final round for all useful entries");

  // Final attempt. Normally should be here with all cams already selected,
  // but criteria may have been slightly different.
  select_all_useful_entries ();

  // Clear left-over fixonce
  for (i = 0; i < cam_obj_used; i++)
    memset (cam_obj[i].posrot_fixonce, 0, sizeof (cam_obj[i].posrot_fixonce));

  calc_res = calc_scene ();

  if (calc_res < 0)
    {
      // Some error in input, can't calculate at all
      state->resultokay = 0;
      stop_calculating ();
      return;
    }
  else if (calc_res == 0)
    {
      // Eh, no fix anyway? Shouldn't be possible.
      // But do try again in next round.
      state->calc_full_restart = 1;
      return;
    }
  else if (calc_res > 0)	// always, if here
    {
      state->resultokay = 1;
      state->calculating = 0;

      // Stop all other calcs
      // Note that this will get buffered and only posted in next select() round
      mqttprintf (state->mosq, state->topicprefix, "scene/cmd", NULL,
		  "{\"cmd\":\"stopcalc\"}");

      // Publish result with objects still included
      mqtt_postresult ();

      stop_calculating ();
    }
}


void
stop_calculating ()
{
  int i;

  state->calculating = 0;

  MDBG ("Calculation stopped");

  // Is safe to call even when not actually calculating.

  // Clear out all objects; leave only cameras
  for (i = cam_obj_cams; i < cam_obj_used; i++)
    {
      // Should really only be objs here
      if (cam_obj[i].objtype != OBJ_OBJ && cam_obj[i].objtype != OBJ_NONE)
	MDBG ("Cleaning unexpected obj %s type %d, may leak memory",
	      cam_obj[i].name, cam_obj[i].objtype);

      free (cam_obj[i].name);
    }
  cam_obj_used = cam_obj_cams;
}


void
receive_sceneobserv (json_t * jroot)
{
  int i;

  if (!json_is_array (jroot))
    return;

  // Forget old observations, if any
  stop_calculating ();

  // We could seed at startup, but we want to log the seed,
  // and that's only reliable when we know that we are connected.
  random_init ();

  // Reset old stuff
  for (i = 0; i < cam_obj_used; i++)
    {
      cam_obj[i].sceneref = 0;
      cam_obj[i].heightmeasured = 0;
      memset (cam_obj[i].in_posrot, 0, sizeof (cam_obj[i].in_posrot));
      memset (cam_obj[i].work_posrot, 0, sizeof (cam_obj[i].work_posrot));
    }

  state->resultokay = 0;
  state->res_maxsumsq = DEFAULT_MAXSUMSQ;

  // Forget all previous observs; no free()s needed.
  observ_used = 0;

  for (i = 0; i < json_array_size (jroot); i++)
    {
      json_t *jitem;
      char *objname = NULL;
      char *obj2name = NULL;
      char *camname = NULL;
      double z = NAN;
      double d = NAN;
      double x = NAN;
      double y = NAN;
      double a = NAN;
      double b = NAN;
      double g = NAN;
      double pixx = NAN;
      double pixy = NAN;
      double pixw = 1.;
      double pixh = 1.;
      double maxsumsq = NAN;

      jitem = json_array_get (jroot, i);
      if (!json_is_object (jitem))
	continue;

      json_unpack (jitem,
		   "{s?:s, s?:s, s?:s, s?:F, s?:F,"
		   " s?:F, s?:F, s?:F, s?:F, s?:F,"
		   " s?:F, s?:F, s?:F, s?:F, s?:F}",
		   "obj", &objname,
		   "obj2", &obj2name,
		   "cam", &camname,
		   "z", &z,
		   "d", &d,
		   "x", &x, "y", &y, "a", &a, "b", &b, "g", &g,
		   "pixx", &pixx, "pixy", &pixy,
		   "pixw", &pixw, "pixh", &pixh,
		   "sumsquares-limit", &maxsumsq);

      if (!isnan (maxsumsq))
	{
	  state->res_maxsumsq = maxsumsq;
	  continue;
	}

      if (objname == NULL || objname[0] == '\0')
	{
	  MDBG ("No object in item %d, skipping", i);
	  continue;
	}

      // Height
      if (!isnan (z))
	{
	  int o1;
	  int obs;

	  o1 = find_make_cam_obj (objname);
	  if (cam_obj[o1].objtype == OBJ_NONE)
	    {
	      // We can only make new objs, never cams
	      cam_obj[o1].objtype = OBJ_OBJ;
	    }

	  obs = new_observ ();
	  observ[obs].obs_type = OBS_HEIGHT;
	  observ[obs].obj1_num = o1;
	  observ[obs].meas = z;

	  cam_obj[o1].sceneref++;

	  // provide initial guess
	  cam_obj[o1].in_posrot[2] = z;
	  cam_obj[o1].heightmeasured = 1;
	}

      // x/y/a/b/g are only starting guesses (to force non-random)
      if (!isnan (x) || !isnan (y) || !isnan (a) || !isnan (b) || !isnan (g))
	{
	  int o1;

	  o1 = find_make_cam_obj (objname);
	  if (cam_obj[o1].objtype == OBJ_NONE)
	    {
	      // We can only make new objs, never cams
	      cam_obj[o1].objtype = OBJ_OBJ;
	    }

	  // provide initial guess
	  if (!isnan (x))
	    cam_obj[o1].in_posrot[0] = x;
	  if (!isnan (y))
	    cam_obj[o1].in_posrot[1] = y;
	  if (!isnan (a))
	    cam_obj[o1].in_posrot[3] = a;
	  if (!isnan (b))
	    cam_obj[o1].in_posrot[4] = b;
	  if (!isnan (g))
	    cam_obj[o1].in_posrot[5] = g;
	}

      // Distance
      if (!isnan (d))
	{
	  int o1, o2;
	  int obs;

	  if (obj2name == NULL || obj2name[0] == '\0')
	    {
	      MDBG ("Distance but no second object in item %d, skipping", i);
	      continue;
	    }

	  o1 = find_make_cam_obj (objname);
	  if (cam_obj[o1].objtype == OBJ_NONE)
	    {
	      // We can only make new objs, never cams
	      cam_obj[o1].objtype = OBJ_OBJ;
	    }

	  o2 = find_make_cam_obj (obj2name);
	  if (cam_obj[o2].objtype == OBJ_NONE)
	    {
	      // We can only make new objs, never cams
	      cam_obj[o2].objtype = OBJ_OBJ;
	    }

	  obs = new_observ ();
	  observ[obs].obs_type = OBS_DIST;
	  observ[obs].obj1_num = o1;
	  observ[obs].obj2_num = o2;
	  observ[obs].meas = d;

	  cam_obj[o1].sceneref++;
	  cam_obj[o2].sceneref++;
	}

      // Camera pixel
      if (!isnan (pixx) && !isnan (pixy))
	{
	  int o1, o2;
	  int obs;

	  if (camname == NULL || camname[0] == '\0')
	    {
	      MDBG ("Pixel but no camera in item %d, skipping", i);
	      continue;
	    }

	  o1 = find_make_cam_obj (objname);
	  if (cam_obj[o1].objtype == OBJ_NONE)
	    {
	      // We can only make new objs, never cams
	      cam_obj[o1].objtype = OBJ_OBJ;
	    }

	  o2 = find_cam (camname);
	  if (o2 < 0 || cam_obj[o2].objtype != OBJ_CAM)
	    {
	      MDBG ("Unknown camera in item %d, skipping", i);
	      continue;
	    }

	  obs = new_observ ();
	  observ[obs].obs_type = OBS_PIX;
	  observ[obs].obj1_num = o1;
	  observ[obs].obj2_num = o2;
	  observ[obs].pixx =
	    (pixx - cam_obj[o2].xoffset) * cam_obj[o2].xscale;
	  observ[obs].pixy =
	    (pixy - cam_obj[o2].yoffset) * cam_obj[o2].yscale;
	  observ[obs].pixw_inv = 1. / pixw;
	  observ[obs].pixh_inv = 1. / pixh;

	  cam_obj[o1].sceneref++;
	  cam_obj[o2].sceneref++;
	}
    }


#if 0
  for (i = 0; i < cam_obj_used; i++)
    printf ("  cam_obj[%d]=type %d name %s\n", i,
	    cam_obj[i].objtype, cam_obj[i].name);

  for (i = 0; i < observ_used; i++)
    printf
      ("  observ[%d]=type %d 1num %d 2num %d pixx %f pixy %f winv %f hinv %f meas %f\n",
       i, observ[i].obs_type, observ[i].obj1_num, observ[i].obj2_num,
       observ[i].pixx, observ[i].pixy, observ[i].pixw_inv, observ[i].pixh_inv,
       observ[i].meas);
#endif

  state->calculating = 1;
  state->calc_full_restart = 1;
  state->attempts = 0;

  select_all_useful_entries ();
  if (!is_calculable ())
    {
      MDBG ("Sorry, not enough information to calculate scene");
      stop_calculating ();
    }
}


void
receive_caminfo (char *camname, json_t * jroot)
{
  int oi;			// object index

  if (state->calculating)
    {
      if (find_cam (camname) >= 0)
	{
	  MDBG ("Cam %s is already known, assuming unchanged and"
		" continuing calculation", camname);
	  return;
	}
      else
	{
	  // Our code assumes that all cameras are at front of object array,
	  // so must get rid of other objects first.
	  // Hopefully user will notice and re-submit scene observations...
	  MDBG ("New cam %s, stopping calculation to be able to store it",
		camname);
	  stop_calculating ();
	}
    }

  oi = find_make_cam_obj (camname);

  // Note: this can be new cam, but also new params for existing cam.

  if (cam_obj[oi].objtype == OBJ_CAM)
    {
      // forget old params
      free (cam_obj[oi].typeparam);
      cam_obj[oi].typeparam = NULL;

      free (cam_obj[oi].distortionparam);
      cam_obj[oi].distortionparam = NULL;
    }
  else if (cam_obj[oi].objtype != 0)
    {
      MDBG
	("Previous type %d now reused for cam %s, will probably get memleak",
	 cam_obj[oi].objtype, camname);
    }

  cam_obj[oi].objtype = OBJ_CAM;
  if (cam_obj_cams <= oi)
    cam_obj_cams = oi + 1;	// cam<>obj border

  if (parse_cam_json (jroot, &cam_obj[oi], 1) > 0)
    {
      MDBG ("Cam %s info incomplete, disabling", camname);

      cam_obj[oi].objtype = 0;
      // Name stays around (cannot delete), < cam_obj_cams, may work next time

      if (cam_obj[oi].typeparam != NULL)
	free (cam_obj[oi].typeparam);
      cam_obj[oi].typeparam = NULL;

      if (cam_obj[oi].distortionparam != NULL)
	free (cam_obj[oi].distortionparam);
      cam_obj[oi].distortionparam = NULL;

      return;
    }

  dist_lin_precalc (&cam_obj[oi]);
}
