// Calculate scene setup via MQTT.
// Also see --help.

// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#define PROGNAME  "scenecalc"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>
#include <time.h>
#include <stdarg.h>
#include <math.h>
#include <mosquitto.h>

#include "state.h"
#include "calc.h"
#include "../shared/version.h"
#include "../shared/error.h"
#include "../shared/mqttpub.h"

// Normally scene calc should be finished quickly
#define MQTT_KEEPALIVE 30
#define MAIN_ALARM     30
// But when doing debug, scene calc printf's may take very long
//#define MQTT_KEEPALIVE 120
//#define MAIN_ALARM     120

const char *progname = PROGNAME;

// Okay, global is not nice. But in calc.c it becomes a real pain
// to carry the state into *every single function* incl. mpfit callbacks...
state_t *state;


void
print_help (FILE * f)
{
  fprintf (f, "\n"
	   "USAGE: %s [options]\n"
	   "\n"
	   "SCWIPS Calculate scene and positions."
	   "\n"
	   "Options are:\n"
	   "    --host <name>      Hostname of mosquitto broker (default: localhost)\n"
	   "    --port <num>       Portname of mosquitto broker (default: 1883)\n"
	   "    --topicprefix <s>  Scwips instance mqtt topic prefix.\n"
	   "    --name <s>         Unique name for this scenecalc instance.\n"
	   "\n"
	   "    --help             Print this help\n"
	   "\n"
	   "Reconnections to mqtt broker are handled gracefully.\n"
	   "DNS and connect are protected by alarm(), use external loop to restart.\n"
	   "\nPart of SCWIPS version %s\n\n", PROGNAME, SCWIPSVERSION);
}


/* Parse commandline parameters.
   Returns:
   0 = all okay, continue
   1 = immediately exit successfully (after --help)
   2 = immediately exit unsuccessfully
*/
int
parse_cmdline (int argc, char **argv, state_t * state)
{
  int i;

  // Rather simplistic & easy parameter parsing, adapted from earlier stuff
  for (i = 1; i < argc; i++)
    {
      if (!strcmp (argv[i], "--help"))
	{
	  print_help (stdout);
	  return 1;		// exit successfully
	}

      else if (!strcmp (argv[i], "--host"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --host requires argument");
	  state->svr_hostname = argv[i];
	}
      else if (!strcmp (argv[i], "--port"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --port requires argument");
	  state->svr_port = atoi (argv[i]);
	  if (state->svr_port < 1 || state->svr_port > 65535)
	    CMDLINEERR_RETURN2 ("parameter --port invalid value: %s",
				argv[i]);
	}
      else if (!strcmp (argv[i], "--topicprefix"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --topicprefix requires argument");
	  state->topicprefix = argv[i];
	}
      else if (!strcmp (argv[i], "--name"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --name requires argument");
	  state->name = argv[i];
	}

      else
	CMDLINEERR_RETURN2 ("unknown parameter: %s", argv[i]);
    }

  // Prepare the subscription topics

  if (state->topicprefix[0] == '\0')
    CMDLINEERR_RETURN2 ("please specify --topicprefix");
  if (state->name[0] == '\0')
    CMDLINEERR_RETURN2 ("please specify --name");

  if (strlen (state->topicprefix) > TOPICLEN - 200)
    CMDLINEERR_RETURN2
      ("parameter --topicprefix too long, increase TOPICLEN");

  snprintf (state->broadcmd_topic, TOPICLEN, "%s/cmd", state->topicprefix);
  snprintf (state->scenecmd_topic, TOPICLEN, "%s/scene/cmd",
	    state->topicprefix);
  snprintf (state->ourcmd_topic, TOPICLEN, "%s/scene/%s/cmd", state->name,
	    state->topicprefix);

  snprintf (state->sceneobs_topic, TOPICLEN, "%s/scene/observ",
	    state->topicprefix);

  snprintf (state->caminfo_topic, TOPICLEN, "%s/cam/+/info",
	    state->topicprefix);
  snprintf (state->cam_topicmatch, TOPICLEN, "%s/cam/", state->topicprefix);
  state->cam_topicmatchlen = strlen (state->cam_topicmatch);
  // topicprefix was already checked for sensible length

  // All should be okay now
  return 0;
}


/* This connect callback is called when CONNACK is received, i.e. when
   the broker has actively allowed us access.
*/
void
mqttconnect_callback (struct mosquitto *mosq, void *usrobj, int result)
{
  state_t *state = usrobj;

  if (result != 0)
    {
      // Access actively refused.
      // Socket will be closed by library via
      // mosquitto_loop_read() -> _mosquitto_loop_rc_handle()
      return;
    }

  state->connstate = CS_CONNECTED;

  // Send subscriptions
  mosquitto_subscribe (state->mosq, NULL, state->broadcmd_topic, 0);
  mosquitto_subscribe (state->mosq, NULL, state->scenecmd_topic, 0);
  mosquitto_subscribe (state->mosq, NULL, state->ourcmd_topic, 0);

  mosquitto_subscribe (state->mosq, NULL, state->sceneobs_topic, 0);

  mosquitto_subscribe (state->mosq, NULL, state->caminfo_topic, 0);

  // Request info form all connected cams
  mqttprintf (state->mosq, state->topicprefix, "cam/cmd", NULL,
	      "{\"cmd\":\"postinfo\"}");

  MDBG ("connected and subscribed");
}

/* This disconnect callback is called whenever socket is (already) closed
   for whatever reason.
*/
void
mqttdisconnect_callback (struct mosquitto *mosq, void *usrobj, int reason)
{
  state_t *state = usrobj;

  // We are disconnected already, can't send any message.
  // But if we wanted disconnection, we already sent "shutdown".
  // And if we got disconnected without our request, the broker will
  // already have sent our Last Will = position-lost message.

  if (state->connstate == CS_EOFDISCONNECTING
      || state->connstate == CS_EOFDISCONNECTED)
    {
      state->connstate = CS_EOFDISCONNECTED;
    }
  else
    {
      state->connstate = CS_UNCONNECTED;
      state->connect_nexttime = time (NULL) + 5;
    }
}

/* This message callback is called whenever a message is received; may be 
   called _multiple_ times from 1 mosquitto_loop_read() call.
*/
void
mqttmessage_callback (struct mosquitto *mosq, void *usrobj,
		      const struct mosquitto_message *message)
{
  state_t *state = usrobj;
  json_t *jroot;
  json_error_t jerror;

  // We are non-threaded, and all messages can always be handled immediately
  // when they arrive, no queueing needed. So we just handle the message
  // right here inside the callback.

  // We don't want topicless or payloadless messages.
  if (message->topic[0] == '\0' || message->payloadlen == 0)
    return;

  // All messages are json, non-json silently ignored
  jroot = json_loadb (message->payload, message->payloadlen, 0, &jerror);
  if (jroot == NULL)
    return;

  do
    {
      // Easy "break" for decref & return

      if (!strcmp (message->topic, state->sceneobs_topic))
	{
	  receive_sceneobserv (jroot);
	  break;
	}

      if (!strncmp (message->topic, state->cam_topicmatch,
		    state->cam_topicmatchlen))
	{
	  // This is scwips/cam/+/info

	  char **subtopics;
	  int subtopics_count;

	  mosquitto_sub_topic_tokenise (message->topic +
					state->cam_topicmatchlen, &subtopics,
					&subtopics_count);
	  // Note that we do not include topicprefix (topicmatch) into
	  // tokenize: may be multi-level.

	  // Subtopics now c0/info

	  if (subtopics_count == 2 && !strcmp (subtopics[1], "info"))
	    receive_caminfo (subtopics[0], jroot);

	  mosquitto_sub_topic_tokens_free (&subtopics, subtopics_count);
	  break;
	}

      if (!strcmp (message->topic, state->broadcmd_topic) ||
	  !strcmp (message->topic, state->scenecmd_topic) ||
	  !strcmp (message->topic, state->ourcmd_topic))
	{
	  const char *cmd;
	  cmd = json_string_value (json_object_get (jroot, "cmd"));
	  if (cmd == NULL)
	    break;

	  if (!strcmp (cmd, "stopcalc"))
	    {
	      if (state->calculating)
		stop_calculating ();
	    }

	  else if (!strcmp (cmd, "postresult"))
	    {
	      if (state->resultokay && !state->calculating)
		mqtt_postresult ();
	    }

	  else if (!strcmp (cmd, "shutdown"))
	    {
	      // Initiate mqtt disconnect, will lead to shutdown after disconnected
	      // by broker, or after timeout; handled in mainloop().

	      if (state->connstate == CS_CONNECTED)
		{
		  state->connstate = CS_EOFDISCONNECTING;
		  state->connect_nexttime = time (NULL) + 5;
		  mosquitto_disconnect (state->mosq);	// just queues DISCONNECT
		}
	      else if (state->connstate != CS_EOFDISCONNECTING
		       && state->connstate != CS_EOFDISCONNECTED)
		{
		  // Not connected at all, so sending DISCONNECT would be
		  // useless.
		  // (This should not happen in practice, since we are handling
		  // the shutdown command that arrived via mqtt.)
		  state->connstate = CS_EOFDISCONNECTED;
		  // This will cause loop break immediately in next round.
		}
	    }

	  break;
	}


    }
  while (0);

  json_decref (jroot);
  // *message is free()'d by the library after this callback returns.
}


void
mainloop (state_t * state)
{
  time_t now;
  fd_set readfds, writefds;
  int mosqsock, maxfd;
  struct timeval tv;
  int selectres;
  int any_fd_handled;

  while (1)
    {
      alarm (MAIN_ALARM);

      now = time (NULL);

      // Server has disconnected after we asked
      if (state->connstate == CS_EOFDISCONNECTED)
	{
	  break;
	}

      // Server did not answer after we asked DISCONNECT
      // (Test: suspend broker with Ctrl-Z just before DISCONNECT.)
      if (state->connstate == CS_EOFDISCONNECTING
	  && now >= state->connect_nexttime)
	{
	  break;
	}

      // If disconnected without our asking, attempt to reconnect periodically
      // (=also immediately on startup).
      if (state->connstate == CS_UNCONNECTED
	  && now >= state->connect_nexttime)
	{
	  state->connstate = CS_CONNECTING;

	  if (mosquitto_connect
	      (state->mosq, state->svr_hostname, state->svr_port,
	       MQTT_KEEPALIVE) != MOSQ_ERR_SUCCESS)
	    {
	      // Failed, try again after some time
	      now = time (NULL);	// DNS/connect may have taken time
	      state->connstate = CS_UNCONNECTED;
	      state->connect_nexttime = now + 5;
	      continue;
	    }

	  // mosquitto_connect() returns after TCP connect() finished, with
	  // CONNECT packet queued to be sent by next mosquitto_loop_write().
	  // DNS resolving and connect() is protected by alarm() above.

	  // We keep ->connstate in CS_CONNECTING until CONNACK received
	  // via mqttconnect_callback().

	  // CONNACK is waited for up to keepalive time, checked in
	  // _mosquitto_check_keepalive() via mosquitto_loop_misc().
	  // (state->mosq->state is not cs_connected yet; _no_ pings are sent.)
	  // Waiting-for-CONNACK is easily testable: Ctrl-S on broker.

	  // DNS/connect may have taken long, need new "now" value
	  continue;
	}

      FD_ZERO (&readfds);
      FD_ZERO (&writefds);
      maxfd = 0;

      mosqsock = mosquitto_socket (state->mosq);
      if (mosqsock != -1)
	{
	  FD_SET (mosqsock, &readfds);
	  if (mosquitto_want_write (state->mosq) == true)
	    FD_SET (mosqsock, &writefds);
	  if (mosqsock > maxfd)
	    maxfd = mosqsock;
	}

      if (state->calculating)
	{
	  tv.tv_sec = 0;
	  tv.tv_usec = 0;
	}
      else
	{
	  tv.tv_sec = 1;
	  tv.tv_usec = 0;
	}

      selectres = select (maxfd + 1, &readfds, &writefds, NULL, &tv);
      // We could handle errno==EINTR, but shouldn't occur in practice.
      if (selectres == -1)
	break;			// Stop on any error

      any_fd_handled = 0;

      // Handle mqtt network input and output.

      if (mosqsock != -1 && FD_ISSET (mosqsock, &readfds))
	{
	  mosquitto_loop_read (state->mosq, 1);
	  // Most callbacks are called from here, including
	  // mqttmessage_callback() which handles any incoming messages
	  // (possibly multiple, if piled up in tcp buffer),
	  // and starts new foreground communication if possible.

	  // Don't mind exitcode; socket already disconnected in case of
	  // trouble, then also mqttdisconnect_callback() was called.

	  any_fd_handled = 1;
	}

      if (mosqsock != -1 && FD_ISSET (mosqsock, &writefds))
	{
	  mosquitto_loop_write (state->mosq, 1);

	  any_fd_handled = 1;
	}

      mosquitto_loop_misc (state->mosq);

      // Handle timeouts etc.
      //now = time (NULL);      // select() may have waited 1 sec

      if (state->calculating && !any_fd_handled)
	calculate_once ();
    }

  // This is reached only after our explicit shutdown.
  // If we were connected, DISCONNECT has been sent from inside the loop,
  // and broker hopefully has neatly disconnected us within our timeout.
  // (Else: don't care, we'll close the socket anyway.)

  //mosquitto_destroy() -> in main

  return;
}


int
main (int argc, char **argv)
{
  state_t realstate;
  int res;
  char clientid[MOSQ_MQTT_ID_MAX_LENGTH + 5];
  char hostname[TOPICLEN + 5];

  alarm (5);

  state = &realstate;		// state is global
  memset (state, 0, sizeof (*state));

  // Set default configuration
  state->svr_hostname = "127.0.0.1";
  state->svr_port = 1883;
  state->topicprefix = "";
  state->name = "";

  res = parse_cmdline (argc, argv, state);
  if (res >= 1)
    exit (res - 1);

  // Note that broker kicks any connected client with equal clientid.
  // Clientid is max 23 chars.
  if (gethostname (hostname, TOPICLEN) != 0)
    hostname[0] = '\0';
  res =
    snprintf (clientid, MOSQ_MQTT_ID_MAX_LENGTH, "%d/%s", getpid (),
	      hostname);
  // Fake resultvalue handling to make gcc stop complaining about the
  // intended truncation, which will actually always have res LARGER than 1.
  if (res < 1)
    exit (1);

  mosquitto_lib_init ();	// cannot return errors

  state->mosq = mosquitto_new (clientid, true, state);
  if (state->mosq == NULL)
    {
      STDERR ("mosquitto_new failed: %s", strerror (errno));
      return 1;
    }

  mosquitto_connect_callback_set (state->mosq, mqttconnect_callback);
  mosquitto_disconnect_callback_set (state->mosq, mqttdisconnect_callback);
  mosquitto_message_callback_set (state->mosq, mqttmessage_callback);

  mainloop (state);
  // Returns after "shutdown"

  mosquitto_destroy (state->mosq);
  // This is the _only_ way an mqtt client can forcibly close the socket.

  return 0;
}
