// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

// These are shared functions where cam_t can be extended.

void
dist_lin_precalc (cam_t * cam_p)
{
  distortion_func_t *distfunc;
  double xin, yin;
  double xout, yout;
  double prev_xin, prev_yin;
  double prev_xout, prev_yout;
  double xhighend, xlowend, yhighend, ylowend;
  double min;
  int first;

  // We want to end up with a "limiting" radius from the image center.
  // Points outside the radius are then calculated as on-radius value +
  // linear outside-of-radius extension --> distortion_wrapper().

  // We currently detect the radius from the minimum of the pos/neg x/y axes,
  // but that could be improved with diagonals if really needed.
  // Procedure: walk the distortion function from the center,
  // see where it starts moving pixels *inwards* from the previous.
  // That's the point where distortion stops being useful,
  // and where fitting troubles begin.
  // Also, extremely high slopes are detected in the same way.

  // Most cams will be less than 1:2.35, so can keep walk limited.
  // Take coarse steps at ~100 pix per half-width.

  distfunc = camdistortion[cam_p->camdistortion].func;

  prev_xin = 0.;
  prev_xout = 0.;
  prev_xout = 0.;
  prev_yout = 0.;
  first = 1;
  yin = 0.;
  for (xin = 0.; xin < 3.; xin += 0.01)
    {
      distfunc (xin, yin, &xout, &yout, cam_p->distortionparam);
      if (first)
	first = 0;
      else
	{
	  if (xout <= prev_xout)
	    break;		// stepping back
	  else if (xout >= prev_xout + 0.1)
	    break;		// 10x speed-up slope
	}
      prev_xin = xin;
      prev_xout = xout;
    }
  // Now either at problem point, or at end of walk;
  // use linear extrapolation from this point onwards.
  xhighend = prev_xin;

  prev_xin = 0.;
  prev_xout = 0.;
  prev_xout = 0.;
  prev_yout = 0.;
  first = 1;
  yin = 0.;
  for (xin = 0.; xin > -3.; xin -= 0.01)
    {
      distfunc (xin, yin, &xout, &yout, cam_p->distortionparam);
      if (first)
	first = 0;
      else
	{
	  if (xout >= prev_xout)
	    break;		// stepping back
	  else if (xout <= prev_xout - 0.1)
	    break;		// 10x speed-up slope
	}
      prev_xin = xin;
      prev_xout = xout;
    }
  // Now either at problem point, or at end of walk;
  // use linear extrapolation from this point onwards.
  xlowend = prev_xin;

  prev_xin = 0.;
  prev_xout = 0.;
  prev_xout = 0.;
  prev_yout = 0.;
  first = 1;
  xin = 0.;
  for (yin = 0.; yin < 3.; yin += 0.01)
    {
      distfunc (xin, yin, &xout, &yout, cam_p->distortionparam);
      if (first)
	first = 0;
      else
	{
	  if (yout <= prev_yout)
	    break;		// stepping back
	  else if (yout >= prev_yout + 0.1)
	    break;		// 10x speed-up slope
	}
      prev_yin = yin;
      prev_yout = yout;
    }
  // Now either at problem point, or at end of walk;
  // use linear extrapolation from this point onwards.
  yhighend = prev_yin;

  prev_xin = 0.;
  prev_xout = 0.;
  prev_xout = 0.;
  prev_yout = 0.;
  first = 1;
  xin = 0.;
  for (yin = 0.; yin > -3.; yin -= 0.01)
    {
      distfunc (xin, yin, &xout, &yout, cam_p->distortionparam);
      if (first)
	first = 0;
      else
	{
	  if (yout >= prev_yout)
	    break;		// stepping back
	  else if (yout <= prev_yout - 0.1)
	    break;		// 10x speed-up slope
	}
      prev_yin = yin;
      prev_yout = yout;
    }
  // Now either at problem point, or at end of walk;
  // use linear extrapolation from this point onwards.
  ylowend = prev_yin;

  min = xhighend;
  if (-xlowend < min)
    min = -xlowend;
  if (yhighend < min)
    min = yhighend;
  if (-ylowend < min)
    min = -ylowend;

  cam_p->dist_lin_radius = min;

  MDBG ("Cam %s distortion useful limits x %f %f y %f %f  radius %f",
	cam_p->name, xhighend, xlowend, yhighend, ylowend,
	cam_p->dist_lin_radius);
}


inline void
distortion_wrapper (cam_t * cam_p, double officialx, double officialy,
		    double *distortedx, double *distortedy)
{
  distortion_func_t *distfunc;
  double r;
  double xnew, ynew, xadd, yadd;

  // We clip x/y to within the useful radius,
  // calculate the distortion result for that point,
  // and then linearly extend beyond that result.

  xnew = officialx;
  ynew = officialy;
  xadd = 0.;
  yadd = 0.;

  r = hypot (officialx, officialy);
  if (r > cam_p->dist_lin_radius)
    {
      double scale = cam_p->dist_lin_radius / r;
      xnew = officialx * scale;
      ynew = officialy * scale;
      xadd = officialx - xnew;
      yadd = officialy - ynew;
    }

  //MDBG ("distwrap clipped input to %f %f +linear %f %f",
  //      xnew, ynew, xadd, yadd);

  distfunc = camdistortion[cam_p->camdistortion].func;
  distfunc (xnew, ynew, distortedx, distortedy, cam_p->distortionparam);

  *distortedx += xadd;
  *distortedx += yadd;
}


inline double
distance_3d (double x1, double y1, double z1, double x2, double y2, double z2)
{
  return sqrt ((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2) +
	       (z1 - z2) * (z1 - z2));
}

// double camposrot[6] = camera position+rotation x/y/z/a/b/g
// double objpos[3] = object position x/y/z
// Normally, camposrot will be cam_p->posrot, except during scene calib
// where many different pos+rot will be guessed.
static inline void
calc_campix (cam_t * cam_p, double *camposrot,
	     double *objpos, double *campixx, double *campixy)
{
  double rotx, roty, rotz;
  double officialx, officialy;
  camtype_func_t *camfunc;

  FITDBG ("  position obj x %f y %f z %f", objpos[0], objpos[1], objpos[2]);

  FITDBG ("  position cam %s x %f y %f z %f", cam_p->name,
	  camposrot[0], camposrot[1], camposrot[2]);

  // now object moves to objx-camx, objy-camy, objz-camz
  FITDBG ("  translated objx %f y %f z %f",
	  objpos[0] - camposrot[0],
	  objpos[1] - camposrot[1], objpos[2] - camposrot[2]);

  FITDBG ("  rotation cam %s a %f b %f g %f", cam_p->name,
	  camposrot[3], camposrot[4], camposrot[5]);

  rotate_3d (objpos[0] - camposrot[0],
	     objpos[1] - camposrot[1],
	     objpos[2] - camposrot[2],
	     camposrot[3], camposrot[4], camposrot[5], &rotx, &roty, &rotz);

  FITDBG ("  rotated objx %f y %f z %f", rotx, roty, rotz);

  camfunc = camtype[cam_p->camtype].func;

  FITDBG ("  camera typeparam %f %f %f", cam_p->typeparam[0],
	  cam_p->typeparam[1], cam_p->typeparam[2]);

  camfunc (rotx, roty, rotz, &officialx, &officialy, cam_p->typeparam);

  FITDBG ("  camera official x %f y %f", officialx, officialy);

  FITDBG ("  camera distortionparam %f %f %f",
	  cam_p->distortionparam[0],
	  cam_p->distortionparam[1], cam_p->distortionparam[2]);

  distortion_wrapper (cam_p, officialx, officialy, campixx, campixy);

  FITDBG ("  camera distorted x %f y %f", *campixx, *campixy);
}
