// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <stdio.h>
#include <mosquitto.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <jansson.h>

#include "mqttpub.h"
#include "error.h"


/* Generic formatted mqtt publish.
   struct mosquitto *mosq,	mosquitto connection
   char *topicpart1,		topic substrings which will get concatenated
   ...char *topicpartN,		  with '/' in between.
   NULL,			end of topic substrings
   char *format,		printf format string
   ...args			printf arguments

   Note: having vprintf for topic too would be very convenient,
   but is actually impossible. vprintf() explicitly leaves va_arg
   in undefined state -- see man va_arg, man vprintf.
   And neither a NULL in-between or an argumentcount in front will help,
   since skipping arguments will fail as it is "unknowable" which argument
   size is being examined/skipped (short/double/pointer etc.).

   Why va_args gets to be undefined: vprintf may/will make a private copy
   using va_copy() and iterate over that instead; and/or may stop
   iterating as soon as max string length has been reached; 1$ 2$ argument
   reordering; etc.
*/
void
mqttprintf (struct mosquitto *mosq, ...)
{
  va_list args;
  char topic[TOPICLEN + 5];
  int topiclen;
  char *topicpart;
  char *format;
  char msg[MSGLEN + 5];
  int res;

  topiclen = 0;
  va_start (args, mosq);
  while ((topicpart = va_arg (args, char *)) != NULL)
    {
      int partlen = strlen (topicpart);
      if (topiclen + 1 + partlen > TOPICLEN)
	{
	  STDERR ("too long topic, increase TOPICLEN");
	  exit (1);
	}
      if (topiclen > 0)
	{
	  topic[topiclen] = '/';
	  topiclen++;
	}
      strncpy (topic + topiclen, topicpart, partlen + 1);
      // partlen + 1 to copy \0 too, else gcc complains
      topiclen += partlen;
    }
  topic[topiclen] = '\0';

  format = va_arg (args, char *);

  res = vsnprintf (msg, MSGLEN, format, args);
  /* Note that "args" is UNDEFINED now, see above. */

  va_end (args);

  if (res < 0 || res >= MSGLEN)
    {
      STDERR ("too long message, increase MSGLEN");
      exit (1);
    }

  mosquitto_publish (mosq, NULL, topic, strlen (msg), msg, 0, 0);
}


/* Generic mqtt json publish.
   struct mosquitto *mosq,	mosquitto connection
   char *topicpart1,		topic substrings which will get concatenated
   ...char *topicpartN,		  with '/' in between.
   NULL,			end of topic substrings
   json_t *json

   Use like:
   mqttpubjson(mosq, "a", "b", "c", NULL, json_pack("i", 42));

   Note: use "O" for objects that you want to keep afterwards.

   json_pack result will be decref'd == gets stolen.
   json NULL (or error) is fine, won't publish anything.
*/
void
mqttpubjson (struct mosquitto *mosq, ...)
{
  va_list args;
  char topic[TOPICLEN + 5];
  int topiclen;
  char *topicpart;
  json_t *jroot;
  char *jstring;

  topiclen = 0;
  va_start (args, mosq);
  while ((topicpart = va_arg (args, char *)) != NULL)
    {
      int partlen = strlen (topicpart);
      if (topiclen + 1 + partlen > TOPICLEN)
	{
	  STDERR ("too long topic, increase TOPICLEN");
	  exit (1);
	}
      if (topiclen > 0)
	{
	  topic[topiclen] = '/';
	  topiclen++;
	}
      strncpy (topic + topiclen, topicpart, partlen + 1);
      // partlen + 1 to copy \0 too, else gcc complains
      topiclen += partlen;
    }
  topic[topiclen] = '\0';

  jroot = va_arg (args, json_t *);
  va_end (args);

  jstring = json_dumps (jroot, JSON_INDENT (2));
  json_decref (jroot);

  if (jstring == NULL)
    return;

  mosquitto_publish (mosq, NULL, topic, strlen (jstring), jstring, 0, 0);

  free (jstring);
}
